package org.ow2.odis.admin;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.options.AddActivateFileNode;
import org.ow2.odis.admin.options.AddConfigFile;
import org.ow2.odis.admin.options.AddPlugins;
import org.ow2.odis.admin.options.ChangeOdisCfg;
import org.ow2.odis.admin.options.DisplayConfig;
import org.ow2.odis.admin.options.DisplayHelp;
import org.ow2.odis.admin.options.GenActivateFile;
import org.ow2.odis.admin.options.NoAction;
import org.ow2.odis.admin.options.NoStopOnError2;
import org.ow2.odis.admin.options.SetProxy;
import org.ow2.odis.admin.options.SetTrustAnchors;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.exception.OdisDomainException;
import org.ow2.odis.exception.OdisException;
import org.ow2.odis.exception.StateException;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.lifeCycle.domain.StateDomainInitialized;
import org.ow2.odis.lifeCycle.domain.StateDomainLoaded;
import org.ow2.odis.lifeCycle.domain.StateDomainStarted;
import org.ow2.odis.lifeCycle.domain.StateDomainStopped;
import org.ow2.odis.lifeCycle.domain.StateDomainUnloaded;
import org.ow2.odis.node.NodeManager;

/**
 * This class provides the launch functionality for an ODIS deployment.<br>
 * Available options are : <li><b>-add <I>file</I></b> adds <I>file</I>'s contents to the configuration (option may occur more than once)
 * <li><B>-act <I>file</I></B> restricts activation to the nodes named in <I>file</I> (one name per line in <I>file</I>, option may occur
 * more than once) <li><B>-noStopOnError</B> requires to process all configuration elements (by default, upon configuration error,
 * subsequent elements are not processed) <li><B>-gen <I>file</I></B> generates a default activation file <li><B>-display</B> displays the
 * effective configuration (XML form) <li><B>-serv <I>idServer</I></B> launches a collocated Joram server with id <I>idServer</I> <li>
 * <B>-h</B> displays help
 */
public final class Launch implements LaunchMBean {
    /**
     *
     */
    public static String ODIS_CFG = "odis.xml";

    /** Logger for this class. */
    private static final Logger LOGGER = Monolog.initialize().getLogger(
            Launch.class.getName());

    /**
     * The Constant EMPTY_CONF. Name of a resource that is an ODIS configuration file containing no element.
     */
    private static final String EMPTY_CONF = "org/ow2/odis/admin/emptyConfig.xml";

    /** Sax builder to construct the JDom document. */
    private SAXBuilder builder = null;

    /** The current document representing the ODIS deployment. */
    static public Document doc = null;

    /** The responsible for XML validation (currently unused). */
    private IValidateXML validateXML = null;

    /** Denotes that the Joram server is launched within the same JVM. */
    public static boolean joramCollocated = false;

    /** Denotes that the ODIS Domain is started. */
    private boolean isStarted = false;

    /**
     * If false, the Domain will only log errors found in configuration files and throw no exception.<br>
     * In the default behavior, when <code>-noStopOnError</code> option is <u>not</u> used, exception are raised, which may or may not lead
     * to a full stop. Best effort is the policy in all cases.
     */
    public boolean stopOnError = true;

    /** Unique instance of Launch class (singleton pattern). */
    private static Launch instance = new Launch();

    /**
     * Accessor of unique instance. Implements singleton pattern.
     * @return the Launch singleton
     */
    static public synchronized Launch getInstance() {
        if (instance == null) {
            instance = new Launch();
        }
        return instance;
    }

    /**
     *
     */
    static public List lsActions = Collections
            .synchronizedList(new ArrayList());

    /**
     * Constructor.
     */
    private Launch() {
        Launch.lsActions = Collections.synchronizedList(new ArrayList());
        JMX.getInstance().registerMBean("Launch", this);
        PluginsRegistry.getInstance();
        builder = new SAXBuilder();
        try {
            URL url = NodeManager.class.getResource("/" + EMPTY_CONF);
            // by Arnaud : prevent the following exception in OSGi
            // java.net.MalformedURLException: unknown protocol: bundle
            // doc = builder.build(url);
            InputStream is = url.openStream();
            doc = builder.build(is);
            // \Arnaud
        } catch (JDOMException e) {
            LOGGER.log(BasicLevel.ERROR, "error in contructor", e);
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "error in contructor", e);
        }
    }

    /**
     * Uses the active document to instantiate the deployment.
     * @throws OdisException the ODIS Exception
     */
    public void start() throws OdisException {
        if (isStarted) {
            LOGGER.log(BasicLevel.FATAL, "unable to stop started config ");
        } else {
            isStarted = true;
            Domain domainMgr = Domain.getInstance();
            domainMgr.attribute.setStopOnError(stopOnError);
            if (validateXML != null) {
                try {
                    validateXML.validateDocument(doc);
                } catch (Exception e) {
                    LOGGER.log(BasicLevel.ERROR, "XSD ERROR", e);
                    if (stopOnError) {
                        throw new OdisException(e);
                    }
                }
            }
            domainMgr.getDomainState().setLifeState(domainMgr,
                    new StateDomainLoaded(), doc.getRootElement());
            domainMgr.getDomainState().setLifeState(domainMgr,
                    new StateDomainInitialized());
            domainMgr.getDomainState().setLifeState(domainMgr,
                    new StateDomainStarted());
            synchronized (this) {
                try {
                    wait();
                } catch (InterruptedException ex) {
                    ex.printStackTrace();
                }
            }
        }
    }

    /**
     * Stops the deployment.
     * @throws StateException if the life cycled is violated
     */
    public void stop() throws StateException {
        int count = 0;
        try {
            while (!isStarted && count < 60) {
                count++;
                Thread.sleep(1000);
            }
        } catch (InterruptedException e) {
            LOGGER.log(BasicLevel.FATAL,
                    "unexpected InterruptException during stop");
            throw new RuntimeException(
                    "unexpected InterruptException during stop");
        }
        if (count < 60) {
            synchronized (this) {
                notifyAll();
            }
            Domain domain = Domain.getInstance();
            try {
                domain.getDomainState().setLifeState(domain,
                        new StateDomainStopped());
                domain.getDomainState().setLifeState(domain,
                        new StateDomainUnloaded());
            } catch (OdisDomainException e) {
                throw new RuntimeException(e.getMessage());
            }
            ConnectionManager connectionMgr = ConnectionManager.getInstance();
            NodeManager nodeMgr = NodeManager.getInstance();
            nodeMgr.unloadComponent();
            connectionMgr.unloadComponent();
            isStarted = false;
            JMX.getInstance().unregisterMBean("Launch", this);
            PluginsRegistry.getInstance().removePlugins();
            JMX.destroy();
            instance = null;
        } else {
            LOGGER.log(BasicLevel.FATAL, "unable to stop unstarted config");
        }
    }

    /**
     * Launch main function.
     * @param args
     * @args all command line parameters allowed by this class. (see class definition)
     * @throws OdisException if something gets wrong
     */
    public static void main(String[] args) throws OdisException {
        Launch launch = Launch.getInstance();
        launch.init(args);
    }

    /**
     * Launch main function.
     * @param args
     * @args all command line parameters allowed by this class. (see class definition)
     * @throws OdisException if something gets wrong
     */
    public synchronized void init(String[] args) throws OdisException {
        IAction changeOdisCfg = new ChangeOdisCfg();
        Launch.lsActions.add(changeOdisCfg);
        Launch.lsActions.add(new DisplayHelp());
        Launch.lsActions.add(new AddConfigFile());
        Launch.lsActions.add(new NoAction());
        Launch.lsActions.add(new GenActivateFile());
        Launch.lsActions.add(new NoStopOnError2());
        Launch.lsActions.add(new AddPlugins());
        Launch.lsActions.add(new SetProxy());
        Launch.lsActions.add(new SetTrustAnchors());
        // As AddActivateFileNode will desactivate some declared nodes
        // It must be set after AddConfigFile
        Launch.lsActions.add(new AddActivateFileNode());
        Launch.lsActions.add(new DisplayConfig());
        // -------------------------------------------------------
        // verification que le fichier de config par defaut est toujours
        // odis.xml
        if (args.length > 0) {
            changeOdisCfg.actions(0, args);
        }
        // -------------------------------------------------------
        // chargement des plugins donnees en param de lancement
        PluginsRegistry.getInstance().loadPlugin(ODIS_CFG);
        // -------------------------------------------------------
        // traitement des differentes options mise au lancement
        try {
            Iterator itAction;
            int index = 0;
            while (index < args.length) {
                itAction = lsActions.iterator();
                int oldIndex = index;
                while (itAction.hasNext() && index < args.length) {
                    IAction action = (IAction) itAction.next();
                    index = action.actions(index, args);
                }
                if (oldIndex == index) {
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER.log(BasicLevel.WARN, "unknown param :"
                                + args[index]);
                    }
                    index++;
                }
            }
            // -------------------------------------------------------
            // traitement post declaration
            itAction = lsActions.iterator();
            while (itAction.hasNext()) {
                ((IAction) itAction.next()).postDeclare();
            }
            XMLOutputter outp = new XMLOutputter();
            ByteArrayOutputStream stream = new ByteArrayOutputStream();
            try {
                outp.output(doc, stream);
                String chaine = new String(stream.toByteArray());
                Domain.getInstance().initAndStartByStringLater(chaine);
            } catch (IOException ex) {
                ex.printStackTrace();
            }
            // -------------------------------------------------------
            // lancement de l'infrastructure
            // launch.start();
            this.start();
        } catch (OdisException e) {
            LOGGER.log(BasicLevel.FATAL,
                    "Launch will stop due to OdisException :", e);
            try {
                Domain.getInstance().unloadComponent();
            } catch (OdisException e1) {
                e1.printStackTrace();
            }
            doc = null;
            JMX.getInstance().unregisterMBean("Launch", instance);
            instance = null;
            JMX.destroy();
            throw e;
        } catch (RuntimeException e) {
            LOGGER.log(BasicLevel.FATAL, "RuntimeException occured :", e);
            new DisplayHelp().postDeclare();
            // <JDO "stop on unhandled exception"/>
            throw e;
        }
    }

    /**
     * Sets the validateXML.
     * @param pValidateXML the new validate XML
     */
    public void setValidateXML(IValidateXML pValidateXML) {
        this.validateXML = pValidateXML;
    }

    /**
     * Gets the validateXML.
     * @return the validateXML
     */
    public IValidateXML getValidateXML() {
        return validateXML;
    }
}
