package org.ow2.odis.admin;

import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.ConnectionFactory;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.exception.OdisFactoryException;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.services.IFactoryService;

/**
 * The Class PluginsRegistry.<br>
 * Provide the ability to add new optional fonctionality to the Odis domain.<br>
 * <i>note: implement singleton pattern</i>
 * @author olivier Beyler FT R&D
 */
public final class PluginsRegistry implements PluginsRegistryMBean {
    /**
     * The sequence of plugin identity. Each plugin need to have an unique identity.
     */
    private static int seqIdentity = 0;

    private static final String CST_PLUGINS = "PLUGINS";

    private static final String CST_PLUGIN = "PLUGIN";

    private static final String CST_NAME = "NAME";

    private static final String CST_CLASS = "CLASS";

    private static final String CST_VALUE = "value";

    private static final Logger LOGGER = Monolog.initialize().getLogger(
            PluginsRegistry.class.getName());

    /** The plugins. */
    private List plugins = new ArrayList();

    /** The instance. */
    private static final PluginsRegistryMBean instance = new PluginsRegistry();

    /**
     * The Class PluginEntry.
     */
    final class PluginEntry {
        /** The identity. */
        int identity;

        /** The name. */
        String name = "";

        /** The classe. */
        IPlugin classe = null;

        /**
         * constructor.
         * @param identity the identity
         * @param name the name
         * @param classe the classe
         */
        public PluginEntry(int identity, String name, IPlugin classe) {
            super();
            this.identity = identity;
            this.name = name;
            this.classe = classe;
        }
    }

    /**
     * Instantiates a new plugins registry.
     */
    private PluginsRegistry() {
        JMX.getInstance().registerMBean("PluginsRegistry", this);
    }

    /**
     * Gets the instance.
     * @return the instance
     */
    public static PluginsRegistryMBean getInstance() {
        return instance;
    }

    /**
     * Gets the list plugins.
     * @return the list plugins
     * @see com.francetelecom.odis.component.admin.PluginsRegistryMBean#listOfPlugins()
     */
    public String listOfPlugins() {
        StringBuffer result = new StringBuffer("");
        Iterator itPlugins = plugins.iterator();
        while (itPlugins.hasNext()) {
            PluginEntry entry = (PluginEntry) itPlugins.next();
            if (entry.classe != null) {
                result.append("[");
                result.append(entry.identity);
                result.append("] ");
                result.append(entry.name);
                result.append(System.getProperty("line.separator"));
            }
        }
        return result.toString();
    }

    /**
     * Adds the plugin.
     * @param classe the classe
     * @return the int
     * @throws InstantiationException the instantiation exception
     * @see com.francetelecom.odis.component.admin.PluginsRegistryMBean#addPlugin(java.lang.String)
     */
    public int addPlugin(String classe) throws InstantiationException {
        try {
            int result = -1;
            IFactoryService factory = Domain.getInstance().getFactoryInstance();
            IPlugin plugin = (IPlugin) factory.getInstanceOf(classe, "1");
            if (plugin != null) {
                result = seqIdentity++;
                plugins.add(new PluginEntry(result, plugin.getName(), plugin));
                plugin.bind();
            } else {
                throw new InstantiationException("class unknown `" + classe
                        + "'");
            }
            return result;
        } catch (OdisFactoryException ex) {
            LOGGER.log(BasicLevel.ERROR, "unable to addPlugin", ex);
            throw new InstantiationException();
        }
    }

    /**
     * Removes the plugin.
     * @param identity the identity
     * @return true, if remove plugin
     * @see com.francetelecom.odis.component.admin.PluginsRegistryMBean#removePlugin(int)
     */
    public boolean removePlugin(final int identity) {
        boolean result = false;
        Iterator itPlugins = plugins.iterator();
        while (itPlugins.hasNext()) {
            PluginEntry entry = (PluginEntry) itPlugins.next();
            if (entry.identity == identity && entry.classe != null) {
                entry.classe.unbind();
                entry.classe = null;
                result = true;
            }
        }
        return result;
    }

    /**
     * Removes all the plugin.
     * @see com.francetelecom.odis.component.admin.PluginsRegistryMBean#removePlugin(int)
     */
    public void removePlugins() {
        while (!plugins.isEmpty()) {
            PluginEntry entry = (PluginEntry) plugins.remove(0);
            if (entry.classe != null) {
                entry.classe.unbind();
            }
        }
    }

    /**
     * Gets the plugin description.
     * @param identity the identity
     * @return the plugin description
     * @see com.francetelecom.odis.component.admin.PluginsRegistryMBean#getPluginDescription(int)
     */
    public String getPluginDescription(final int identity) {
        String result = "no description provided";
        Iterator itPlugins = plugins.iterator();
        while (itPlugins.hasNext()) {
            PluginEntry entry = (PluginEntry) itPlugins.next();
            if (entry.identity == identity && entry.classe != null
                    && entry.classe.getDescription() != null) {
                result = entry.classe.getDescription();
            }
        }
        return result;
    }

    /**
     * @param filename
     */
    public void loadPlugin(String filename) {
        try {
            URL url = ConnectionFactory.class.getResource("/" + filename);
            if (url == null) {
                url = new URL(filename);
            }
            if (url != null) {
                Document doc = new SAXBuilder().build(url.openStream());
                Element root = doc.getRootElement();
                Element cf = root.getChild(CST_PLUGINS);
                Iterator it = cf.getChildren(CST_PLUGIN).iterator();
                while (it.hasNext()) {
                    Element element = (Element) it.next();
                    String key = element.getAttributeValue(CST_NAME);
                    Element classe = element.getChild(CST_CLASS);
                    if (classe == null && LOGGER.isLoggable(BasicLevel.WARN)) {
                        StringBuffer warn;
                        warn = new StringBuffer(
                                "Bad configuration into file \"");
                        // warn.append(filename);
                        warn.append(url);
                        warn.append("\" for plugin \"");
                        warn.append(key);
                        warn.append("\" <CLASS value=\"...\"/> is missing.");
                        LOGGER.log(BasicLevel.WARN, warn.toString());
                    } else {
                        addPlugin(classe.getAttributeValue(CST_VALUE));
                    }
                }
            } else {
                StringBuffer warn = new StringBuffer("No file \"");
                warn.append(filename);
                warn.append("\" ");
                LOGGER.log(BasicLevel.WARN, warn.toString());
            }
        } catch (JDOMException e) {
            LOGGER.log(BasicLevel.ERROR, "Init of plugins failed", e);
        } catch (Exception e) {
            // IOException or Null pointer exception (URL = null)
            // have the same behavior
            StringBuffer warn = new StringBuffer("No or bad file \"");
            warn.append(filename);
            warn.append("\" ");
            LOGGER.log(BasicLevel.WARN, warn.toString(), e);
        }
    }
}
