package org.ow2.odis.bufferIn;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.engine.IEngineIn;
import org.ow2.odis.engine.IEngineProceed;
import org.ow2.odis.policy.AbstractPolicy;

/**
 * The Class SynchronizeManager.
 */
public class BufferInManager {

    /** The Constant EOL. */
    private static final String EOL = "\n";
    /** The size byte. */
    private transient long sizeByte = 0;
    /** The number object. */
    private transient long numberObject = 0;
    /** The proceed. */
    protected transient IEngineProceed proceed = null;
    /** The engine in. */
    protected transient IEngineIn engineIn = null;
    /** List of synchronized Object. */
    protected transient List lsSynchObject = null;
    /** Last synchronized Object. */
    protected transient BufferListObject lastSynchObject = null;
    /** The factory. */
    protected transient BufferListObjectFactory factory = null;
    /** Constant. */
    static protected final int READY = 0;
    /** Constant. */
    static protected final int NOT_READY = 1;
    /** The state. */
    private transient int state = READY;
    /** provide log for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            BufferInManager.class.getName());

    /**
     * Constructor.
     * @param pProceed the proceed
     * @param pEngineIn the engine in
     */
    public BufferInManager(IEngineProceed pProceed, IEngineIn pEngineIn) {
        super();
        lsSynchObject = Collections.synchronizedList(new ArrayList(pEngineIn.getNumberOfObjectMax()));
        this.proceed = pProceed;
        this.engineIn = pEngineIn;
    }

    /**
     * initialize the Synchronize object factory.
     */
    public void init() {
        factory = new BufferListObjectFactory();
    }

    /**
     * use by send to freeze the thread send while the manager is not able to synchronize object.
     * @throws InterruptedException the interrupted exception
     */
    public void pause() throws InterruptedException {
        if (state == NOT_READY) {
            synchronized (this) {
                if (checkStateAgain()) {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        StringBuffer msg = new StringBuffer("checkStateAgain");
                        msg.append(" returned true immediately ");
                        msg.append("- NOT pausing ");
                        msg.append(engineIn);
                        LOGGER.log(BasicLevel.INFO, msg.toString());
                    }
                    state = READY;
                } else {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        StringBuffer msg = new StringBuffer("checkStateAgain");
                        msg.append("pausing BufferInManager (NOT READY) for ");
                        msg.append(engineIn);
                        LOGGER.log(BasicLevel.INFO, msg.toString());
                    }
                    wait();
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO, "wait() exited for " + engineIn + ((state == READY) ? " - READY"
                                : " - NOT READY"));
                    }
                }
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                StringBuffer msg = new StringBuffer("checkStateAgain");
                msg.append(" state found READY immediately ");
                msg.append("- NOT pausing ");
                msg.append(engineIn);
                LOGGER.log(BasicLevel.INFO, msg.toString());
            }
        }
    }

    /**
     * check if putting object condition into synchronized buffer are satisfied.
     * @param length the length
     * @param number the number
     * @return boolean value
     */
    private boolean checkState(final int length, final int number) {
        final long nObject = getNumberObject();
        boolean result = true;
        if (engineIn.getNumberOfObjectMax() > 0 && engineIn.getNumberOfObjectMax() <= nObject + number) {
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                final StringBuffer errMsg;
                errMsg = new StringBuffer("Can't synchronize data :");
                errMsg.append(" too many objects in the list");
                errMsg.append(EOL);
                errMsg.append("\tlist size     = ");
                errMsg.append(nObject);
                errMsg.append(EOL);
                errMsg.append("\ttrying to add = ");
                errMsg.append(number);
                errMsg.append(EOL);
                errMsg.append("\tmax size      = ");
                errMsg.append(engineIn.getNumberOfObjectMax());
                LOGGER.log(BasicLevel.WARN, errMsg.toString());
            }
            state = NOT_READY;
            result = false;
        }
        if (result && engineIn.getSizeByteMax() > 0 && engineIn.getSizeByteMax() <= sizeByte + length) {
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                final StringBuffer errMsg;
                errMsg = new StringBuffer(
                        "Can't synchronize data : list global capacity exceeded ");
                errMsg.append(EOL);
                errMsg.append("\tlist size (bytes) =");
                errMsg.append(sizeByte);
                errMsg.append(EOL);
                errMsg.append("\ttrying to add     =");
                errMsg.append(length);
                errMsg.append(EOL);
                errMsg.append("\tlist capacity     =");
                errMsg.append(engineIn.getSizeByteMax());
                LOGGER.log(BasicLevel.WARN, errMsg.toString());
            }
            state = NOT_READY;
            result = false;
        }
        return result;
    }

    /**
     * check if the restarting condition of putting object into synchronized buffer are satisfied.
     * @return boolean value
     */
    private boolean checkStateAgain() {
        final boolean result = ((engineIn.getSizeByteMax() > sizeByte * 2 || engineIn.getSizeByteMax() == 0) && (engineIn.getNumberOfObjectMax()) > getNumberObject() * 2 || engineIn.getNumberOfObjectMax() == 0);
        return result;
    }

    /**
     * verify if the manager is able to synchronize data ask to the factory a synchronizeObject add it to the list.
     * @param msg the message
     * @param objects the objects
     * @param linkId the link id
     * @param context the context
     * @return true if succeed false in other case
     */
    public boolean synchronizeData(final byte[] msg, final List objects,
            final String linkId, Context context) {
        boolean result = checkState(msg.length, objects.size());
        if (result) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Add " + objects.size() + " objects to the buffer in.");
            }
            numberObject += objects.size();
            sizeByte += msg.length;
            // create a synchronized object with the objects in
            final BufferListObject ob = factory.getObject(linkId, msg, objects,
                    context);
            // add the list of objects to the list of synchronized objects
            lsSynchObject.add(ob);
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Buffer in size: " + lsSynchObject.size());
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "checkState() returned FALSE for " + engineIn);
            }
        }
        // signal the engine proceed for new object to proceed
        proceed.activeProceed();
        return result;
    }

    /**
     * Gets the number object.
     * @return the exact number of object present into all the SynchronizedObject contain into the manager list
     */
    public long getNumberObject() {
        return numberObject;
    }

    /**
     * Re synchronize data.
     * @param manager the manager
     * @return true, if successful
     */
    synchronized public boolean resynchronizeData(final BufferInManager manager) {
        while (!manager.lsSynchObject.isEmpty()) {
            lsSynchObject.add(manager.lsSynchObject.remove(0));
        }
        return true;
    }

    /**
     * Re synchronize data.
     * @param policy the policy
     * @return true, if successful
     */
    public boolean resynchronizeData(final AbstractPolicy policy) {
        return true;
    }

    /**
     * Gets the next object.
     * @return the oldest synchonizedObject contain into the list
     */
    public Object getNextObject() {
        Object result = null;
        Object objectUnitaire;
        if (!lsSynchObject.isEmpty()) {
            lastSynchObject = (BufferListObject) lsSynchObject.get(0);
            objectUnitaire = lastSynchObject.getObject();
            if (objectUnitaire == null && lastSynchObject.objects.isEmpty()) {
                final boolean remove = lsSynchObject.remove(lastSynchObject);
                if (!remove) {
                    LOGGER.log(BasicLevel.ERROR,
                            "try to remove unknown object");
                }
                lastSynchObject.delete();
                objectUnitaire = getNextObject();
            }
            result = objectUnitaire;
        }
        return result;
    }

    /**
     * Gets the current context.
     * @return the current context
     */
    public Context getCurrentContext() {
        Context result = null;
        if (lastSynchObject != null) {
            result = lastSynchObject.getContext();
        }
        return result;
    }

    /**
     * Release.
     * @param object the object
     * @return true if succeed false in other case
     */
    public boolean release(final Object object) {
        if (!lsSynchObject.isEmpty()) {
            lastSynchObject = (BufferListObject) lsSynchObject.get(0);
            lastSynchObject.releaseObject(object);
            numberObject--;
            if (lastSynchObject.getRemaindingObjects() == 0) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "remove empty synchronizeObject from the list");
                }
                sizeByte -= lastSynchObject.size;
                lsSynchObject.remove(lastSynchObject);
                lastSynchObject.delete();
            }
            if (!lsSynchObject.isEmpty()) {
                proceed.activeProceed();
            }
        }
        // <JDO "add 1 test : avoid checking state when unnecessary"/>
        if ((state == NOT_READY) && checkStateAgain()) {
            // le buffer out est suffisament vide pour accepter a nouveau de
            // recevoir des nouveaux messages
            synchronized (this) {
                state = READY;
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "notifying BufferInManager (READY) for " +
                            engineIn);
                }
                notifyAll();
            }
        }
        return true;
    }

    /**
     * Gets the size byte.
     * @return the size byte
     */
    public long getSizeByte() {
        return sizeByte;
    }
}
