package org.ow2.odis.bufferIn;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Collections;
import java.util.List;

import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.decoder.IDecoder;
import org.ow2.odis.engine.IEngineIn;
import org.ow2.odis.engine.IEngineProceed;
import org.ow2.odis.policy.AbstractPolicy;
import org.ow2.odis.port.PortIn;
import org.ow2.odis.util.Filter;

/**
 * The Class PersistenceBufferInManager.
 */
public class PersistenceBufferInManager extends BufferInManager {
    /** The path. */
    String path = "";

    /**
     * Instantiates a new persistence BufferIn manager.
     * @param proceed the proceed
     * @param engineIn the engine in
     * @param pPath the path
     */
    public PersistenceBufferInManager(IEngineProceed proceed,
            IEngineIn engineIn, String pPath) {
        super(proceed, engineIn);
        this.path = pPath;
    }

    /**
     * @see com.francetelecom.odis.component.bufferin.BufferInManager#init()
     */
    public void init() {
        factory = new PersistenceBufferListObjectFactory(path);
    }

    /**
     * Restores data from IN link persistence files.
     * @param policy the policy
     * @return true if success, otherwise false
     * @see com.francetelecom.odis.component.bufferin.BufferInManager#resynchronizeData(com.francetelecom.odis.component.policy.AbstractPolicy)
     */
    public boolean resynchronizeData(AbstractPolicy policy) {
        // check if the repository exists
        File repertoire = new File(path + File.separator + "IN");
        if (!repertoire.isDirectory()) {
            return false;
        }
        // detect all potential links
        File[] list = repertoire.listFiles();
        for (int i = 0; i < list.length; i++) {
            // check if its a potential link (sub-directory)
            if (list[i].isFile()) {
                // if it's a file it's not a potential link
                continue;
            }
            String linkName = list[i].getName();
            // create a filter to restrict file extensions
            FilenameFilter filter = new Filter(".sav");
            FilenameFilter filtertemp = new Filter(".temp");
            File[] lsSaveFile = list[i].listFiles(filtertemp);
            // get the associated link
            for (int index = 0; index < lsSaveFile.length; index++) {
                lsSaveFile[index].delete();
            }
            lsSaveFile = list[i].listFiles(filter);
            if (lsSaveFile.length <= 0) {
                // there is no file for us -> nothing to do
                continue;
            }
            // get the associated link
            PortIn portIn = (PortIn) policy.getLinkById(linkName);
            if (portIn == null) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    LOGGER.log(BasicLevel.WARN,
                            "Potential save file found on link " + linkName
                                    + " but this link is unknown");
                }
                continue;
            }
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN, "About to restore "
                        + lsSaveFile.length + " message(s) from IN link "
                        + linkName + " of " + proceed.getNodeName());
            }
            for (int index = 0; index < lsSaveFile.length; index++) {
                // get data from the file
                String fileName = lsSaveFile[index].getName();
                String filePath = lsSaveFile[index].getPath();
                int nbrDeMessageStillPendingIntoMessage = 0;
                int nbrOctets = 0;
                byte[] msg = null;
                int nbrOctetsRead = 0;
                try {
                    RandomAccessFile fichier = new RandomAccessFile(
                            lsSaveFile[index], "r");
                    nbrDeMessageStillPendingIntoMessage = fichier.readInt();
                    nbrOctets = fichier.readInt();
                    msg = new byte[nbrOctets];
                    fichier.read(msg);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Restored from "
                                + new String(msg));
                    }
                    nbrOctets = (int) (fichier.length() - fichier
                            .getFilePointer());
                    msg = new byte[nbrOctets];
                    nbrOctetsRead = fichier.read(msg);
                    fichier.close();
                } catch (FileNotFoundException e) {
                    LOGGER
                            .log(
                                    BasicLevel.ERROR,
                                    "Persistence file "
                                            + fileName
                                            + " has been unexpectedly deleted during restoration");
                    continue;
                } catch (IOException e) {
                    LOGGER.log(BasicLevel.ERROR,
                            "IO Error while restoring file " + fileName, e);
                    continue;
                }
                if (nbrOctets != nbrOctetsRead) {
                    LOGGER.log(BasicLevel.ERROR,
                            "Couldn't read all bytes into msg for file "
                                    + fileName);
                }
                IDecoder decoder = portIn.getDecoder();
                if (decoder == null) {
                    LOGGER.log(BasicLevel.ERROR,
                            "No decoder is defined for IN link " + linkName);
                    continue;
                }
                List objects = null;
                try {
                    objects = decoder.decode(msg);
                } catch (RuntimeException e1) {
                    LOGGER.log(BasicLevel.ERROR, "Decode Msg failed ", e1);
                    if (msg == null) {
                        LOGGER.log(BasicLevel.ERROR, "message was null");
                    } else {
                        LOGGER.log(BasicLevel.ERROR, "message size :"
                                + msg.length);
                        StringBuffer buffer = new StringBuffer("message =");
                        for (int u = 0; u < msg.length; u++) {
                            Byte b = new Byte(msg[u]);
                            buffer.append(b);
                        }
                        LOGGER.log(BasicLevel.ERROR, buffer.toString());
                    }
                }
                // remove some objects from the list
                while (objects.size() > nbrDeMessageStillPendingIntoMessage) {
                    objects.remove(0);
                }
                if (objects.size() > 0) {
                    // recreate a persistent object with the data read
                    PersistenceBufferListObject po = null;
                    // TODO : utilisation du contexte !!!
                    po = new PersistenceBufferListObject(filePath, portIn
                            .getId(), msg, objects, null, false);
                    // TODO : utilisation du contexte !!!
                    lsSynchObject.add(po);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Buffer in size: "
                                + lsSynchObject.size());
                    }
                } else {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Remove empty .sav file ");
                    }
                    lsSaveFile[index].delete();
                }
                // signal the engine proceed for new object to proceed
            }
        }
        Collections.sort(lsSynchObject);
        if (!lsSynchObject.isEmpty()) {
            proceed.activeProceed();
        }
        return true;
    }
}
