package org.ow2.odis.bufferIn;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.List;

import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.connection.context.Context;

/**
 * The aim of this class is to allow message to be save and restore from a file. <br>
 * The file structure is designed like this : <table border="1" cellspacing="0" cellpadding="0">
 * <tr>
 * <td>int</td>
 * <td>number of object still available into this message</td>
 * </tr>
 * <tr>
 * <td>int</td>
 * <td>represent length of identity</td>
 * </tr>
 * <tr>
 * <td>byte[]</td>
 * <td>represent identity of this message</td>
 * </tr>
 * <tr>
 * <td>byte[]</td>
 * <td>represent body of the message</td>
 * </tr>
 * </table> <br>
 * As the number of object is the first entry of the file, when updating the file <br>
 * only the start of the file is impacted by this update.
 */
public class PersistenceBufferListObject extends BufferListObject implements
        Comparable {
    /** use for serialization. */
    private static final long serialVersionUID = 1L;

    /** payload of objects. */
    private final transient byte[] msg;

    /** represent the name of the file used to persist message. */
    private final transient String name;

    /**
     * Constructor.
     * @param pName name of the file
     * @param pId unique identity of this message
     * @param pMsg byte array is a representation of the message
     * @param pObjects list of object represented by pMsg
     * @param save ask for saving or not his message
     * @param context the context
     */
    public PersistenceBufferListObject(final String pName, final String pId,
            final byte[] pMsg, final List pObjects, final Context context,
            final boolean save) {
        super(pId, pMsg, pObjects, context);
        name = pName;
        msg = pMsg;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                    "create a PersistenceSynchronizedObject");
        }
        if (save) {
            save();
        }
    }

    /**
     * save the synchronize object.
     * @return true if succeed false in other case
     */
    public final boolean save() {
        boolean result = true;
        File file;
        RandomAccessFile fichier;
        StringBuffer fileName = new StringBuffer(name);
        fileName.append(".temp");
        byte[] array;
        try {
            file = new File(fileName.toString());
            file.createNewFile();
            fichier = new RandomAccessFile(fileName.toString(), "rw");
            final int nbRObject = getRemaindingObjects();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                final StringBuffer debugMsg;
                debugMsg = new StringBuffer(
                        "save a PersistenceSynchronizedObject ");
                debugMsg.append(name);
                debugMsg.append(" with ");
                debugMsg.append(nbRObject);
                LOGGER.log(BasicLevel.DEBUG, debugMsg.toString());
            }
            fichier.writeInt(nbRObject);
            array = idObject.getBytes();
            fichier.writeInt(array.length);
            fichier.write(array);
            fichier.write(msg);
            fichier.close();
            file.renameTo(new File(name));
        } catch (FileNotFoundException e) {
            LOGGER.log(BasicLevel.ERROR, "enable to find file", e);
            result = false;
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "io exception occured on " + name, e);
            result = false;
        }
        return result;
    }

    /**
     * update the persistence support.
     * @return true if succeed false in other case
     */
    public final boolean update() {
        boolean result = true;
        RandomAccessFile fichier;
        try {
            fichier = new RandomAccessFile(name, "rw");
            final int nbRObject = getRemaindingObjects();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                final StringBuffer debug;
                debug = new StringBuffer(
                        "update a PersistenceSynchronizedObject ");
                debug.append(name);
                debug.append(" with ");
                debug.append(nbRObject);
                LOGGER.log(BasicLevel.DEBUG, debug.toString());
            }
            fichier.writeInt(nbRObject);
            fichier.close();
        } catch (FileNotFoundException e) {
            LOGGER.log(BasicLevel.ERROR, "enable to find file", e);
            result = false;
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "io exception occured", e);
            result = false;
        }
        return result;
    }

    /**
     * destroy persistence support.
     * @return true if succeed false in other case
     */
    public final boolean delete() {
        File file;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                    "delete a PersistenceSynchronizedObject " + name);
        }
        file = new File(name);
        if (!file.exists() && LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                  "PB try to delete a file witch doesn't exist");
            
        }
        file.delete();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                    "PersistenceSynchronizedObject is deleted" + name);
        }
        return super.delete();
    }

    /**
     * Compare to.
     * @param o2 the o2
     * @return int value
     */
    public int compareTo(Object o2) {
        return this.name.compareTo(((PersistenceBufferListObject) o2).name);
    }
}
