package org.ow2.odis.bufferOut;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * <code>BufferObject</code> is the encoded version of an object produced by the proceed. It contains a byte array and a transaction id.
 * When a proceed creates a list of objects, all <code>BufferObject</code> instances have the same transaction id.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class BufferObject {
    /**
     * Identity of serializable class.
     */
    private static final long serialVersionUID = 1L;

    /**
     * Logger of this class.
     */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            BufferObject.class.getName());

    /**
     * Byte array holding the encoded object data.
     */
    private byte[] msg = null;

    /**
     * Id of the transaction the message belongs to.
     */
    private long idTransaction = 0;

    /**
     * Constructor.
     * @param idTransaction id of the transaction the message belongs to
     * @param pMsg payload (the encoded object data)
     */
    public BufferObject(long idTransaction, byte[] pMsg) {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Creating BufferObject");
        }
        this.idTransaction = idTransaction;
        this.msg = pMsg;
    }

    /**
     * Constructor.
     */
    protected BufferObject() {
        super();
    }

    /**
     * Provides size information.
     * @return the estimated size of this BufferObject
     */
    public int getSize() {
        return msg.length + 8;
    }

    /**
     * destroy synchronizedObject.
     * @return true if success.
     */
    public boolean release() {
        idTransaction = -1;
        msg = null;
        return true;
    }

    /**
     * Getter.
     * @return the id of the transaction this BufferObject belongs to
     */
    public long getIdTransaction() {
        return idTransaction;
    }

    /**
     * Sets the transaction id.
     * @param idTransaction the id of the transaction this BufferObject belongs to
     */
    public void setIdTransaction(final long idTransaction) {
        this.idTransaction = idTransaction;
    }

    /**
     * Gets the encoded object data.
     * @return a byte array containing this <code>BufferObject</code> data
     */
    public byte[] getMsg() {
        return msg;
    }

    /**
     * Sets the encoded object data.
     * @param msg a byte array containing this <code>BufferObject</code> data
     */
    public void setMsg(final byte[] msg) {
        this.msg = msg;
    }
}
