package org.ow2.odis.bufferOut;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.util.Filter;

/**
 * <code>BufferObjectFactory</code> creates either a <code>BufferObject</code> or a <code>PersistenceBufferObject</code>.<br>
 * In case of a <code>PersistenceBufferObject</code> each object is assigned an id. This id will be used to order objects when the system
 * restores them. As this id has to be unique, the initial value is set to the current System timestamp and will then increase for each new
 * object.
 * @author Olivier Beyler FranceTelecom R&D Grenoble
 */
public class BufferObjectFactory {
    /** Logger for this class */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            BufferObjectFactory.class.getName());

    /**
     * Unique id of a message.<br>
     * It's initialized with the system time value and incremented with each new bufferObject creation.
     */
    static long idMessage = System.currentTimeMillis();

    /** The associated BufferOut */
    final private BufferOut bufferOut;

    /**
     * Constructor.
     * @param bufferOut the BufferOut associated with this factory
     */
    public BufferObjectFactory(BufferOut bufferOut) {
        super();
        this.bufferOut = bufferOut;
        if (bufferOut.getAttribut().isPersistant()) {
            File repository = new File(bufferOut.path);
            repository.mkdirs();
        }
    }

    /**
     * Creates a persistent or non-persistent object depending on persistence settings.
     * @param idTransaction reference to transaction id
     * @param msg the payload of the object
     * @return the BufferObject created with all parameters
     */
    public BufferObject createObject(final long idTransaction, final byte[] msg) {
        BufferObject bufferObject = null;
        if (bufferOut.getAttribut().isPersistant()) {
            bufferObject = new PersistenceBufferObject(idTransaction, msg,
                    bufferOut.path, idMessage++);
        } else {
            bufferObject = new BufferObject(idTransaction, msg);
        }
        return bufferObject;
    }

    /**
     * Restores the BufferObject to the associated BufferOut
     */
    public synchronized void restoreBufferOut() {
        // check if the repository exist
        final File repertoire = new File(bufferOut.path + File.separator);
        if (!repertoire.isDirectory()) {
            if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                LOGGER.log(BasicLevel.ERROR, "Cannot restore " + bufferOut
                        + " - path is not a directory : " + repertoire);
            }
            return;
        }
        FilenameFilter filter = new Filter(".temp");
        File[] lsSaveFile = repertoire.listFiles(filter);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Removing " + lsSaveFile.length
                    + " temporary persistence files");
        }
        for (int index = 0; index < lsSaveFile.length; index++) {
            lsSaveFile[index].delete();
        }
        filter = new Filter(".sav");
        lsSaveFile = repertoire.listFiles(filter);
        if (lsSaveFile.length <= 0) {
            // there is no file for us -> nothing to do
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "No persistence file found in "
                        + bufferOut.path + " - Restore complete");
            }
        } else {
            // get the associated link
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN, "Restoring " + lsSaveFile.length
                        + " message(s) for OUT buffer " + bufferOut.getId());
            }
            String fileName;
            PersistenceBufferObject persistenceBO;
            List lsPersistentBufferObjects = new ArrayList();
            for (int index = 0; index < lsSaveFile.length; index++) {
                // get information from file
                fileName = lsSaveFile[index].getName();
                persistenceBO = new PersistenceBufferObject();
                if (!persistenceBO.read(lsSaveFile[index])) {
                    continue;
                }
                // checks that the object transaction id is not higher than the
                // current BufferOut transaction id.
                if (persistenceBO.getIdTransaction() > bufferOut
                        .getIdTransaction()) {
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER.log(BasicLevel.WARN, "File " + fileName
                                + " ignored : incompatible idTransaction ("
                                + persistenceBO.getIdTransaction() + " > "
                                + bufferOut.getIdTransaction() + ")");
                    }
                    lsSaveFile[index].delete();
                    continue;
                }
                //
                lsPersistentBufferObjects.add(persistenceBO);
            }
            Collections.sort(lsPersistentBufferObjects);
            while (!lsPersistentBufferObjects.isEmpty()) {
                bufferOut
                        .putData((PersistenceBufferObject) lsPersistentBufferObjects
                                .remove(0));
            }
        }
        // look for context sub-directories
        filter = new Filter(".ctx");
        lsSaveFile = repertoire.listFiles(filter);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Found " + lsSaveFile.length
                    + " context subdirectories");
        }
        for (int index = 0; index < lsSaveFile.length; index++) {
            if (lsSaveFile[index].isDirectory()) {
                String[] identities = lsSaveFile[index].getName().split("!");
                if (identities.length != 2) {
                    LOGGER.log(BasicLevel.ERROR,
                            "Context persistence directory "
                                    + lsSaveFile[index].getName()
                                    + " ignored : malformed name");
                } else {
                    // ask for sub-buffer creation (including associated tasks)
                    Context ctx = new Context(identities[0], identities[1]
                            .substring(0, identities[1].indexOf('.')));
                    ((ContextualBufferOut) bufferOut)
                            .createContextBufferOut(ctx);
                }
            } else {
                LOGGER.log(BasicLevel.ERROR, "Context persistence file "
                        + lsSaveFile[index].getName()
                        + " ignored : not a directory");
            }
        }
    }
}
