package org.ow2.odis.bufferOut;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.model.BufferOutAttribute;
import org.ow2.odis.model.Const;
import org.ow2.odis.policy.IPolicyElement;
import org.ow2.odis.routing.Destination;

/**
 * A <code>BufferOut</code> contains a list of <code>BufferObject</code> or <code>PersistantBufferObject</code>.<br>
 * A BufferOut is identified either by object type or by routing information.<br>
 * All OUT buffers may be added data using one of the putData() methods. This operation is checked against the buffer's capacity-related
 * properties. If this check fails, the buffer reports to be not ready and put operations are paused.<br>
 * All OUT buffers may be read using the getObjects() method. This operation delivers data according to the buffer's threshold-related
 * properties. Objects must be explicitly released by the reader : this triggers persistence cleanup (if persistence is used) and capacity
 * re-checking for a paused buffer.<br>
 * The buffer will ask the EngineOut to read its data according to its threshold and timer parameters.<br>
 * Subtypes of BufferOut may implement additional functionalities and restrict some usages but should otherwise adhere to the above
 * description.
 * @author Beyler Olivier FranceTelecom R&D Grenoble 2005
 */
public class BufferOut implements IPolicyElement {
    
    /** The End Of Line constant. */
    static final String EOL = System.getProperty("line.separator");

    /** Logger for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            BufferOut.class.getName());

    /** Singleton timer for all BufferOut operations. */
    private static Timer bufferOutTimer = null;

    /**
     * Accessor for the bufferOutTimer.
     * @return the timer
     */
    protected synchronized static Timer getTimer() {
        if (bufferOutTimer == null) {
            bufferOutTimer = new Timer("BufferOutTimer");
        }
        return bufferOutTimer;
    }

    /** contain all settings of bufferOut. */
    protected transient BufferOutAttribute attribute = null;

    /** reference of the engine Out to be able to notify it that sending buffer need to be activate. */
    protected transient EngineOut engineOut = null;

    /** list of synchronized object. */
    private transient List lsSynchronizedObject = Collections
            .synchronizedList(new ArrayList());

    /** factory which create BufferObject. */
    private transient BufferObjectFactory factory = null;

    //
    // /** timer use to auto activate sending after timeout. */
    // private transient Timer timer = null;
    /** current id transaction. */
    private transient long idTransaction = -1;

    /** path where factory can find save file. */
    protected transient String path = "";

    /** to schedule task maxElapsTime. */
    private ActiveTimer activeTimer = null;

    /** to schedule task cyclicTime. */
    private CyclicTimer cyclicTimer = null;

    /**
     * Default constructor (for subclasses).
     */
    protected BufferOut() {
        super();
    }

    /**
     * Constructor.
     * @param pAttribut the attribute describing the BufferOut
     * @param pEngineOut reference on EngineOut to be able to notify it when the buffer need to be processing
     */
    public BufferOut(final BufferOutAttribute pAttribut,
            final EngineOut pEngineOut) {
        super();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer;
            buffer = new StringBuffer("Creating a new BufferOut for ");
            buffer.append(pEngineOut.getNode().getName());
            buffer.append(" ");
            buffer.append(pAttribut.getId());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
        this.attribute = pAttribut;
        pAttribut.setBufferOut(this);
        this.engineOut = pEngineOut;
    }

    /**
     * String representation of BufferOut.
     * @return the string
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer("BufferOut_");
        buffer.append(attribute.getTypeObject());
        buffer.append("_");
        buffer.append(attribute.getRouting());
        return buffer.toString();
    }

    /**
     * Stores a message in this buffer.<br>
     * As BufferOut is non contextual, context information is just ignored. (Contextual subclasses will override.)
     * @param pIdTransaction the message's transaction id
     * @param ctx the Context of the recipient - ignored
     * @param msg the message as a byte array
     * @return true, if put data
     */
    public boolean putData(final long pIdTransaction, final Context ctx,
            final byte[] msg) {
        if (ctx != null && LOGGER.isLoggable(BasicLevel.WARN)) {
            LOGGER.log(BasicLevel.WARN, "Context data (" + ctx
                    + ") will be ignored by non contextual buffer " + this);
        }
        return putData(pIdTransaction, msg);
    }

    /**
     * Stores a message in this buffer.<br>
     * First checks whether adding data is currently allowed.<br>
     * If it is, asks the factory for a synchronized Object and stores it in memory.
     * @param pIdTransaction Transaction id
     * @param msg the message as a byte array
     * @return true if success, otherwise false
     */
    public boolean putData(final long pIdTransaction, final byte[] msg) {
        boolean result = true;
        schedulerTask();
        if (msg != null) {
            if (attribute.getState() == Const.NOT_READY) {
                result = false;
            }
            if (result && !checkAutorizePutting(msg.length)) {
                attribute.setState(Const.NOT_READY);
                result = false;
            }
            if (result) {
                final BufferObject bufferObject = factory.createObject(
                        pIdTransaction, msg);
                putData(bufferObject);
            }
        }
        return result;
    }

    /**
     * Adds data to this buffer.<br>
     * This is a memory only operation (persistence has been done when creating the BufferObject).
     * @param bufferObject the Object which has to be sent
     */
    public void putData(final BufferObject bufferObject) {
        attribute.incSizeByte(bufferObject.getSize());
        lsSynchronizedObject.add(bufferObject);
        attribute.sizeCurrentTransaction++;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            final StringBuffer buffer;
            buffer = new StringBuffer("After putData(), ");
            buffer.append(this);
            buffer.append(" holds ");
            buffer.append(lsSynchronizedObject.size());
            buffer.append(" BufferObject\n  sizeCurrentTransaction is ");
            buffer.append(attribute.sizeCurrentTransaction);
            buffer.append(" with ");
            buffer.append(getPendingMessage());
            buffer.append(" pending messages\n  attribute declares ");
            buffer.append(attribute.getSizeByte());
            buffer.append(" in ");
            buffer.append(attribute.getNbMsg());
            LOGGER.log(BasicLevel.DEBUG, buffer.toString());
        }
    }

    /**
     * Checks the bufferOut capacity to allow putting of a new message.<br>
     * The aim is to evaluate if after adding a new message containing <code>length</code> bytes the buffer is over his limit or not.
     * @param length length of the new message
     * @return true if putting is allowed
     */
    private boolean checkAutorizePutting(final int length) {
        boolean result = true;
        // checks the number of objects in the buffer
        if (attribute.getSizeObjectMax() != 0
                && attribute.getSizeObjectMax() <= attribute.getNbMsg()) {
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                final StringBuffer warn;
                warn = new StringBuffer(
                        "Can't synchronize data : too many object in the list");
                warn.append("\n\tactual list size =");
                warn.append(lsSynchronizedObject.size());
                warn.append("\n\tlimit is set to  =");
                warn.append(attribute.getSizeObjectMax());
                warn.append("\n\tnb pending msg   =");
                warn.append(getPendingMessage());
                LOGGER.log(BasicLevel.WARN, warn);
            }
            synchronized (this) {
                attribute.setState(Const.NOT_READY);
            }
            engineOut.activeSend();
            result = false;
        } // checks the number of bytes in the buffer
        else {
            if (attribute.getSizeByteMax() != 0
                    && attribute.getSizeByteMax() <= attribute.getSizeByte()
                            + length) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    final StringBuffer warn;
                    warn = new StringBuffer(
                            "Can't synchronize data : too many bytes in the list");
                    warn.append("\n\tactual list size in bytes =");
                    warn.append(attribute.getSizeByte());
                    warn.append("\n\ttry to add data (bytes)   =");
                    warn.append(length);
                    warn.append("\n\tlimit is set to           =");
                    warn.append(attribute.getSizeByteMax());
                    LOGGER.log(BasicLevel.WARN, warn);
                }
                synchronized (this) {
                    attribute.setState(Const.NOT_READY);
                }
                // try to force sending
                engineOut.activeSend();
                result = false;
            }
        }
        return result;
    }

    /**
     * Gets an object if it has a valid transaction id.
     * @param index of the object in the list
     * @return object value only if its idTransaction is less than the current idTransaction, otherwise null
     */
    private BufferObject getObject(final int index) {
        BufferObject result = null;
        if (lsSynchronizedObject.size() > index) {
            BufferObject bufferObject = (BufferObject) lsSynchronizedObject
                    .get(index);
            if (bufferObject.getIdTransaction() != idTransaction) {
                result = bufferObject;
            }
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                StringBuffer buffer = new StringBuffer("getObject returns ");
                buffer.append(result);
                LOGGER.log(BasicLevel.DEBUG, buffer.toString());
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                StringBuffer buffer = new StringBuffer(
                        "getObject returns null because it holds ");
                buffer.append(lsSynchronizedObject.size());
                buffer.append(" objects and getObject() was called for index ")
                        .append(index);
                LOGGER.log(BasicLevel.DEBUG, buffer.toString());
            }
        }
        return result;
    }

    /**
     * Gets the number of pending messages.
     * @return the number of messages which are not contained in the current uncommitted transaction
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getPendingMessage()
     */
    public int getPendingMessage() {
        return (attribute.getNbMsg() - attribute.sizeCurrentTransaction);
    }

    /**
     * Gets the objects.
     * @return a list which contains the bufferObjects, but the list is limited by the thresholdNumberOfObject or
     * signalThresholdNumberOfObject.
     */
    public List getObjects() {
        final List result = new ArrayList();
        int size = attribute.getThresholdNumberOfObject();
        if (size <= 0) {
            size = attribute.getSignalThresholdNumberOfObject();
        }
        // <JDO "+3 lines : avoid useless calls to getObject(int)"/>
        if (size > lsSynchronizedObject.size()) {
            size = lsSynchronizedObject.size();
        }
        BufferObject bufferObject;
        for (int i = 0; i < size; i++) {
            bufferObject = getObject(i);
            if (bufferObject != null) {
                result.add(bufferObject);
            }
        }
        return result;
    }

    /**
     * Gets all objects.
     * @return a list containing all bufferObjects, ignoring thresholdNumberOfObject, signalThresholdNumberOfObject and transaction ids.
     */
    public List getAllObjects() {
        return new ArrayList(lsSynchronizedObject);
    }

    /**
     * Releases the object and if appropriate unblock a thread waiting on pause.
     * @param objects the objects
     * @return true if succeed false in other case
     */
    public final boolean release(final List objects) {
        return release(objects, true);
    }

    /**
     * Releases the object and if appropriate unblock a thread waiting on pause.
     * @param objects the objects
     * @param checkPuttingAgain denotes the state check for unblocking threads should be done
     * @return true if succeed false in other case
     */
    public final boolean release(final List objects, boolean checkPuttingAgain) {
        final Iterator itObject = objects.iterator();
        boolean result = true;
        BufferObject bufferObject;
        boolean isRemove;
        while (itObject.hasNext()) {
            bufferObject = (BufferObject) itObject.next();
            isRemove = lsSynchronizedObject.remove(bufferObject);
            if (isRemove) {
                attribute.incSizeByte(-bufferObject.getSize());
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "Releasing an object of size "
                            + bufferObject.getSize());
                    LOGGER.log(BasicLevel.DEBUG, "Attribute now declares "
                            + attribute.getSizeByte() + " bytes in "
                            + attribute.getNbMsg() + " messages");
                }
            } else {
                LOGGER.log(BasicLevel.ERROR,
                        "try to release one or more unknown object(s)");
            }
            result = result && isRemove;
            bufferObject.release();
        }
        if (checkPuttingAgain && checkAutorizePuttingAgain()) {
            // the BufferOut is empty enough to accept new messages
            synchronized (this) {
                attribute.setState(Const.READY);
                notifyAll();
            }
        }
        if (!result) {
            LOGGER.log(BasicLevel.ERROR,
                    "Tried to release one or more unknown object(s)");
        }
        return result;
    }

    /**
     * Gets the priority.
     * @return the priority defined in the attribute by the configuration
     */
    public int getPriority() {
        return attribute.getPriority();
    }

    /**
     * The pause is activated when one of thresholds has been reached.<br>
     * The pause is de-activated only if the checkAutorisePuttingAgain returns true.
     * @throws java.lang.InterruptedException in case of stopped request
     * @throws InterruptedException the interrupted exception
     */
    public void pause() throws InterruptedException {
        synchronized (this) {
            if (attribute.getState() == Const.NOT_READY) {
                if (checkAutorizePuttingAgain()) {
                    attribute.setState(Const.READY);
                } else {
                    engineOut.activeSend();
                    wait();
                }
            }
        }
    }

    /**
     * Manages BufferOut hysteresis.<br>
     * When a BufferOut has been in overflow, it accepts new data once it is less than 50% full in terms of number of messages and size in
     * bytes.
     * @return true if both conditions are satisfied
     */
    private boolean checkAutorizePuttingAgain() {
        final boolean respectNbrObject = attribute.getSizeObjectMax() <= 0
                || attribute.getSizeObjectMax() > (attribute.getNbMsg() * 2);
        final boolean respectSizeOctet = attribute.getSizeByteMax() <= 0
                || attribute.getSizeByteMax() > (attribute.getSizeByte() * 2);
        return (respectNbrObject && respectSizeOctet);
    }

    // -------------------------
    // inner class
    // -------------------------
    /**
     * Definition of ActiveTimer class <br>
     * This class is responsible for waking up the engine out.
     */
    class ActiveTimer extends TimerTask {
        
        /**
         * Constructor.
         */
        public ActiveTimer() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "create ActiveSend");
            }
        }

        /**
         * When the task wakes up, it calls this function.
         */
        public void run() {
            if (attribute.getMaxElapsTime() > 0) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "ActiveTimer engineOut.activeSend().");
                }
                engineOut.activeSend();
            }
        }
    }

    /**
     * Definition of CyclicTimer class <br>
     * This class is responsible for waking up the engine out.
     */
    class CyclicTimer extends TimerTask {
        
        /**
         * Constructor.
         */
        public CyclicTimer() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "create CyclicTimer");
            }
        }

        /**
         * When the task wakes up, it calls this function.
         */
        public void run() {
            if (attribute.getCyclicTime() > 0) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "CyclicTimer engineOut.activeSend().");
                }
                engineOut.activeSend();
            }
        }
    }

    /**
     * Sets the current id transaction.
     * @param idTransaction id of transaction
     */
    public final void setCurrentIdTransaction(final long idTransaction) {
        // if idTransaction == 0 the restoration mechanism will re-affect 0 if
        // no file is present
        // this is nominal behaviour and thus not logged
        if ((this.idTransaction == idTransaction && idTransaction > 0)
                && (LOGGER.isLoggable(BasicLevel.WARN))) {
            LOGGER.log(BasicLevel.WARN,
                    "Current IdTransaction is the same as the new :"
                            + idTransaction + "for " + getId());
        }
        // change the current Id for the transaction
        this.idTransaction = idTransaction;
        // reinitialize the number of objects present in the transaction
        attribute.sizeCurrentTransaction = 0;
        // if the threshold is reached, ask to the EngineOut to wake up
        if (lsSynchronizedObject != null
                && attribute.getNbMsg() >= attribute
                        .getSignalThresholdNumberOfObject()) {
            engineOut.activeSend();
        }
    }

    /**
     * Getter of type Object.
     * @return String
     */
    public String getTypeObject() {
        return attribute.getTypeObject();
    }

    /**
     * The BufferOut id is represented by routing name (if defined) or typeObject (default).
     * @return the id of this BufferOut
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getId()
     */
    public String getId() {
        return attribute.getId();
    }

    /**
     * Gets the destination associated to this BufferOut, which is its id.
     * @return the destination
     */
    public Destination getDestination() {
        return new Destination(getId());
    }

    /**
     * Gets the size of the buffer in number of objects.
     * @return the number of object contain by the BufferOut
     */
    protected int getSizeObject() {
        return lsSynchronizedObject.size();
    }

    // ---------------------------------------------
    // implement interface ILifeCycle
    // ---------------------------------------------
    /**
     * Attribute data is already set and the initComponent function uses it to initialize the component itself.
     */
    public void initComponent() {
        // -----------------------------------
        // determine the location needed by the factory to restore the messages
        // -----------------------------------
        path = Domain.getInstance().getPath();
        path = path + File.separator + engineOut.getNode().getName()
                + File.separator + "OUT";
        path = path + File.separator + getId();
        // -----------------------------------
        // create the collection
        // -----------------------------------
        int value = (attribute.getSizeObjectMax() < 0) ? 5000 : attribute
                .getSizeObjectMax();
        lsSynchronizedObject = Collections
                .synchronizedList(new ArrayList(value));
        // -----------------------------------
        // create the factory
        // -----------------------------------
        factory = new BufferObjectFactory(this);
        if (attribute.isPersistant()) {
            factory.restoreBufferOut();
        }
    }

    /**
     * Gets the Buffer into operational state.<br>
     * If the buffer is persistent, it will restore messages. If a timeout is set, a timer is created.
     * @return 0
     */
    public int launchComponent() {
        // -----------------------------------
        // sets the timer
        // -----------------------------------
        schedulerTask();
        // -------------------------------------
        // checks whether sending is already needed
        // if the threshold is reached ask to the engine out to wake up
        if (lsSynchronizedObject.size() >= attribute
                .getSignalThresholdNumberOfObject()) {
            engineOut.activeSend();
        }
        return 0;
    }

    /**
     * Schedules task's cyclicTime.
     */
    private void schedulerTask() {
        if (attribute.getCyclicTime() > 0 && cyclicTimer == null) {
            synchronized (this) {
                /*
                 * <JDO "use singleton timer for all buffers out"/> // recreate a new timer if need if (timer == null) { timer = new
                 * Timer(); if (LOGGER.isLoggable(BasicLevel.DEBUG)) { LOGGER.log(BasicLevel.DEBUG, "new timer = " + timer); } } //
                 * schedule a new task cyclicTimer = new CyclicTimer(); timer.schedule(cyclicTimer, 0, attribut.getCyclicTime()); if
                 * (LOGGER.isLoggable(BasicLevel.DEBUG)) { LOGGER.log(BasicLevel.DEBUG, "schedule new cyclicTimer = " + cyclicTimer); }
                 */
                // schedule a new task
                cyclicTimer = new CyclicTimer();
                getTimer().schedule(cyclicTimer, 0, attribute.getCyclicTime());
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "schedule new cyclicTimer = "
                            + cyclicTimer);
                }
            }
        }
        // to schedule task maxElapsTime
        if (activeTimer != null) {
            // cancel a scheduled task
            boolean res = activeTimer.cancel();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "activeTimer canceled = " + res);
            }
        }
        if (attribute.getMaxElapsTime() > 0) {
            synchronized (this) {
                /*
                 * <JDO "use singleton Timer for all buffers out"/> // recreate a new timer if needed if (timer == null) { timer = new
                 * Timer(); if (LOGGER.isLoggable(BasicLevel.DEBUG)) { LOGGER.log(BasicLevel.DEBUG, "new timer = " + timer); } } //
                 * schedule a new task activeTimer = new ActiveTimer(); timer.schedule(activeTimer, attribut.getMaxElapsTime()); if
                 * (LOGGER.isLoggable(BasicLevel.DEBUG)) { LOGGER.log(BasicLevel.DEBUG, "schedule new activeTimer = " + activeTimer); }
                 */
                // schedule a new task
                activeTimer = new ActiveTimer();
                getTimer().schedule(activeTimer, attribute.getMaxElapsTime());
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "schedule new activeTimer = "
                            + activeTimer);
                }
            }
        }
    }

    /**
     * Gets the attribute associated with this buffer.
     * @return the attribut
     */
    public BufferOutAttribute getAttribut() {
        return attribute;
    }

    /**
     * Gets the transaction id.
     * @return the id transaction
     */
    public long getIdTransaction() {
        return idTransaction;
    }

    /**
     * Denotes the fact that this buffer is able to handle contexts.
     * @return always false, although context-able subclasses will override
     */
    public boolean isContextual() {
        return false;
    }
}
