package org.ow2.odis.bufferOut;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextManager;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.exception.OdisContextException;
import org.ow2.odis.exception.OdisNoBufferException;
import org.ow2.odis.lifeCycle.Cause;
import org.ow2.odis.model.BufferOutAttribute;
import org.ow2.odis.node.Node;
import org.ow2.odis.routing.Destination;

/**
 * <code>BufferOutManager</code> is in charge of BufferOut creation and management.<br>
 * Only one <code>BufferOut</code> will be created by <code>BufferOutAttribut</code>. When a persistent <code>BufferOutManager</code>
 * starts it restores the current transaction id.
 * @author Olivier Beyler FT R&D Grenoble 2005
 */
public class BufferOutManager {
    /** Logger for this class. */
    private static final Logger LOGGER = Monolog.initialize().getLogger(
            BufferOutManager.class.getName());

    /** Map of all BufferOut indexed by their id. */
    private Map mapIdBufferOut = new HashMap();

    /** Reference of the associated engineOut. */
    private final EngineOut engineOut;

    /** Id of the current transaction. */
    private long idCurrentTransaction = 0;

    /** Denotes whether the managed buffer are persistent. */
    private boolean isPersistant = false;

    /**
     * <code>true</code> until the BufferOutManager has checked whether it is persistent, and done restoreIdTransaction() if needed.
     */
    private boolean isPersistenceCheckNeeded = true;

    /** Starting path of the persistence files */
    private String path = "";

    /** Constant : file extension for transaction id persistence */
    private static final String ID_EXTENTION = ".id";

    /** All BufferOut watchers for this manager. */
    private List bufferOutWatchers = new ArrayList();

    /**
     * The Constructor.
     * @param pEngineOut reference to the engineOut which will read the buffers
     */
    public BufferOutManager(EngineOut pEngineOut) {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create BufferOutManagerFactory");
        }
        this.engineOut = pEngineOut;
    }

    /**
     * Creates a BufferOut.
     * @param attribute attribute describing the BufferOut
     */
    public void createBufferOut(BufferOutAttribute attribute) {
        BufferOut bufferOut = null;
        bufferOut = (BufferOut) mapIdBufferOut.get(attribute.getId());
        if (bufferOut == null) {
            // Use double lock protection
            synchronized (this) {
                if (bufferOut == null) {
                    if (attribute.isContextual()) {
                        bufferOut = new ContextualBufferOut(attribute, engineOut);
                        ContextManager.getInstance().addContextWatcher(
                                (ContextualBufferOut) bufferOut);
                    } else {
                        bufferOut = new BufferOut(attribute, engineOut);
                    }
                    // <JDO reason="only setPersistant() restores the
                    // transaction id"/>
                    // isPersistant = bufferOut.getAttribut().isPersistant();
                    if (isPersistenceCheckNeeded) {
                        isPersistenceCheckNeeded = false;
                        if (engineOut.getNode().attribute.isPersistant()) {
                            setPersistant(true);
                        } else if (engineOut.getNode().attribute
                                .getStrPersistant() != null) {
                            setPersistant(engineOut.getNode().attribute
                                    .getStrPersistant().equalsIgnoreCase("YES"));
                        }
                    }
                    if (attribute.getStrPersistant() == null) {
                        attribute.setPersistant(isPersistant);
                    }
                    // NOTE : This propagation to the Buffer itself should be
                    // done within a consistent lifecycle !!!
                    // </JDO>
                    bufferOut.setCurrentIdTransaction(idCurrentTransaction);
                    mapIdBufferOut.put(attribute.getId(), bufferOut);
                    bufferOut.initComponent();
                }
            }
        }
    }

    /**
     * Creates and initializes a context-specific sub-buffer for a ContextualBufferOut.<br>
     * If a peer is available for that context, the new buffer is registered in the policy instance. If no peer is available yet,
     * registration will take place later.
     * @param attribute the parent ContextualBufferOut's attribute
     * @param ctx the wanted context
     * @return the created buffer
     */
    public ContextBufferOut createContextBufferOut(BufferOutAttribute attribute,
            Context ctx) {
        if (attribute.isContextual()) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO,
                        "creating ContextBuffer for contextual buffer "
                                + attribute.getId() + " on " + ctx);
            }
            ContextBufferOut cbo = new ContextBufferOut(attribute, engineOut,
                    ctx);
            cbo.initComponent();
            engineOut.getPolicyInstance().addPolicyElement(cbo);
            return cbo;
        } else {
            LOGGER.log(BasicLevel.ERROR,
                    "ContextBuffer creation asked for non contextual buffer "
                            + attribute.getId() + " on " + ctx);
            return null; // THIS SHOULD NOT HAPPEN !!!!!!!
        }
    }

    /**
     * Returns the BufferOut associated with the passed name.<br>
     * @param name identity of a buffer
     * @return the BufferOut associated with the name or null if none found
     */
    public BufferOut getBufferOut(String name) {
        BufferOut bufferOut = null;
        bufferOut = (BufferOut) mapIdBufferOut.get(name);
        if ((bufferOut == null) && LOGGER.isLoggable(BasicLevel.WARN)) {
            LOGGER.log(BasicLevel.WARN,
                    "Trying to access undeclared BufferOut or link " + name);
        }
        return bufferOut;
    }

    /**
     * Returns the BufferOut associated with the passed Destination.<br>
     * Note : the Context in the Destination is ignored.
     * @param dest the dest
     * @return the BufferOut associated with the destination or null if none found
     */
    public BufferOut getBufferOut(Destination dest) {
        BufferOut bufferOut = null;
        bufferOut = (BufferOut) mapIdBufferOut.get(dest.getName());
        if ((bufferOut == null) && LOGGER.isLoggable(BasicLevel.WARN)) {
            LOGGER.log(BasicLevel.WARN,
                    "Trying to access undeclared BufferOut or link "
                            + dest.getName());
        }
        return bufferOut;
    }

    /**
     * Tries to find a buffer for the specified destination, then asks it to take care of the data.<br>
     * If the buffer refuses the data (typically in case of overflow), this method will wait for the buffer availability and signal the
     * Node for it to take appropriate measures.
     * @param pIdTransaction the message's transaction id
     * @param dest the destination for the message
     * @param msg the message as a byte array
     * @throws OdisNoBufferException if no BufferOut is found for the provided Destination
     * @throws OdisContextException if a non contextual destination is provided for a contextual buffer
     * @throws InterruptedException if interrupted during the a buffer's pause()
     */
    public void putData(final long pIdTransaction, final Destination dest,
            final byte[] msg) throws OdisNoBufferException,
            OdisContextException, InterruptedException {
        BufferOut bo = getBufferOut(dest);
        if (bo == null) {
            throw new OdisNoBufferException(
                    "No buffer found for destination : " + dest);
        }
        if (dest.getContext() == null && bo.isContextual()) {
            throw new OdisContextException("Non contextual destination " + dest
                    + " provided for contextual buffer " + bo);
        } else {
            // a non contextual buffer will simply ignore the context
            // information
            if (!bo.putData(pIdTransaction, dest.getContext(), msg)) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer errMsg = new StringBuffer(bo.toString());
                    errMsg.append(" refuses to take care of object for ");
                    errMsg.append(getNode());
                    LOGGER.log(BasicLevel.WARN, errMsg.toString());
                }
                Cause cause = new Cause(Cause.BUFFER_OUT_OVERFLOW, bo.getId());
                getNode().addCause(cause);
                bo.pause();
                while (!bo.putData(pIdTransaction, dest.getContext(), msg)) {
                    bo.pause();
                }
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer errMsg = new StringBuffer(bo.toString());
                    errMsg.append(" accepts messages again : end of error.");
                    LOGGER.log(BasicLevel.WARN, errMsg.toString());
                }
                getNode().causeEnds(cause);
            }
        }
    }

    /**
     * Provides a list of all managed buffers, with names and key parameters.
     * @return a String array with all BufferOuts' data
     */
    public String[] getBufferOutList() {
        Collection c = mapIdBufferOut.values();
        Iterator it = c.iterator();
        StringBuffer buffer = null;
        BufferOut bo = null;
        String[] result = new String[c.size()];
        int i = 0;
        while (it.hasNext()) {
            buffer = new StringBuffer(100);
            bo = (BufferOut) it.next();
            buffer.append(bo.getId());
            buffer.append(" ( : ");
            buffer.append(bo.getAttribut().getSizeObjectMax());
            buffer.append(" : ");
            buffer.append(bo.getAttribut().getSizeByteMax());
            buffer.append(" : ");
            buffer.append(bo.getSizeObject());
            buffer.append(" : ");
            buffer.append(bo.getAttribut().getSizeByte());
            buffer.append(" ) ");
            result[i] = buffer.toString();
            i++;
        }
        return result;
    }

    /**
     * Getter.
     * @return id of the current transaction
     */
    public long getIdCurrentTransaction() {
        return idCurrentTransaction;
    }

    /**
     * Sets new current transaction id for each BufferOut.<br>
     * Saves it if persistence is used and purge old id persistence file.
     * @param idCurrentTransaction new current transaction id
     */
    synchronized public void setIdCurrentTransaction(long idCurrentTransaction) {
        long oldIdTransaction = this.idCurrentTransaction;
        this.idCurrentTransaction = idCurrentTransaction;
        if (isPersistant) {
            try {
                File f = new File(path + oldIdTransaction + ID_EXTENTION);
                if (f.exists()) {
                    f.renameTo(new File(path + idCurrentTransaction
                            + ID_EXTENTION));
                } else {
                    f = new File(path + idCurrentTransaction + ID_EXTENTION);
                    f.createNewFile();
                }
            } catch (FileNotFoundException e) {
                LOGGER.log(BasicLevel.ERROR, "enable to find file", e);
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR, "io exception occured on " + path,
                        e);
            }
        }
        Iterator it = mapIdBufferOut.values().iterator();
        while (it.hasNext()) {
            BufferOut bufferOut = (BufferOut) it.next();
            bufferOut.setCurrentIdTransaction(idCurrentTransaction);
        }
    }

    /**
     * Getter.
     * @return true if persistence is used
     */
    public boolean isPersistant() {
        return isPersistant;
    }

    /**
     * Setter of persistence.<br>
     * Manages the file used to store the persistent transaction id.
     * @param isPersistant defines if the associated BufferOut has to be persistent or not
     */
    public void setPersistant(boolean isPersistant) {
        this.isPersistant = isPersistant;
        if (isPersistant) {
            StringBuffer startOfPath = new StringBuffer(Domain.getInstance()
                    .getPath());
            startOfPath.append(File.separator);
            startOfPath.append(engineOut.getNode().getName());
            startOfPath.append(File.separator);
            path = startOfPath.toString();
            File file = new File(path);
            file.mkdirs();
            restoreIdTransaction();
        }
        // <JDO reason="shouldn't be done there">
        // // forward the persistence setting to the BufferOuts
        // final Iterator it = mapIdBufferOut.values().iterator();
        // while (it.hasNext()) {
        // BufferOut bufferOut = (BufferOut) it.next();
        // bufferOut.getAttribut().setPersistant(isPersistant);
        // }
    }

    /**
     * Restores from persistence file the highest transaction id known for this node (and deletes the others).
     */
    public void restoreIdTransaction() {
        // get the list of transaction id persistence files
        FilenameFilter filtertemp = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.endsWith(ID_EXTENTION);
            }
        };
        File repertoire = new File(path);
        File[] lsIdFile = repertoire.listFiles(filtertemp);
        // find the higher transaction id
        long maxId = 0;
        String fileName = null;
        String[] name = null;
        for (int i = 0; i < lsIdFile.length; i++) {
            fileName = lsIdFile[i].getName();
            name = fileName.split("\\D");
            if (name.length == 0 || name[0] == null) {
                LOGGER
                        .log(BasicLevel.ERROR,
                                "Error during retrieve id transaction file "
                                        + fileName);
                return;
            }
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Retrieving transaction id file "
                        + fileName);
            }
            long id = 0;
            try {
                id = Long.parseLong(name[0]);
            } catch (NumberFormatException e) {
                LOGGER.log(BasicLevel.ERROR,
                        "Bad name for id transaction file :" + fileName, e);
            }
            if (id > maxId) {
                maxId = id;
            }
        }
        long id;
        // cleanup
        for (int i = 0; i < lsIdFile.length; i++) {
            name = lsIdFile[i].getName().split("\\D");
            id = Long.parseLong(name[0]);
            if (id != maxId) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "Retrieving an older transaction id file with id : "
                                    + id);
                }
                lsIdFile[i].delete();
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "The older transaction file with id " + id
                                    + " has been deleted.");
                }
            }
        }
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Setting transaction id to " + maxId
                    + " for proceed's node " + engineOut.getNode().getName());
        }
        getNode().getEngineProceed().setCurrentIdTransaction(maxId);
        setIdCurrentTransaction(maxId);
    }

    // --------------------------------------------------------------------------
    // implement interface ILifeCycle
    // --------------------------------------------------------------------------
    /**
     * Launch the component.
     * @return 0
     */
    public int launchComponent() {
        final BufferOut[] bufferouts = (BufferOut[]) mapIdBufferOut.values()
                .toArray(new BufferOut[mapIdBufferOut.size()]);
        for (int i = 0; i < bufferouts.length; i++) {
            // <JDO "removed 1 line to prevent double restoration"/>
            // bufferouts[i].initComponent();
            bufferouts[i].launchComponent();
        }
        return 0;
    }

    /**
     * Allows access to the Node this manager serves.
     * @return the Node
     */
    public Node getNode() {
        return engineOut.getNode();
    }

    /**
     * Adds a BufferOut watcher.
     * @param bow
     */
    public void addBufferOutWatcher(IBufferOutWatcher bow) {
        bufferOutWatchers.add(bow);
    }

    /**
     * Asks for buffer removal allowance to all registered watchers.<br>
     * Note that all watchers are called even if a previous one has already refused the operation.
     * @param context the involved Context
     * @param contents the contents of the involved ContextBufferOut
     * @param reason the String value of the policy which triggered the removal
     * @return true if all watchers allowed the operation
     */
    public boolean isContextBufferOutRemovalAllowed(Context context,
            List contents, String reason) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                    "Notifying ContextBufferOut removal for : " + context
                            + " - Policy : " + reason);
        }
        boolean allow = true;
        for (Iterator it = bufferOutWatchers.iterator(); it.hasNext();) {
            allow &= ((IBufferOutWatcher) it.next())
                    .isContextBufferOutRemovalAllowed(context, contents, reason);
        }
        return allow;
    }
}
