package org.ow2.odis.bufferOut;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.TimerTask;

import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextManager;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.model.BufferOutAttribute;
import org.ow2.odis.routing.Destination;

/**
 * A <code>ContextBufferOut</code> is one of the context-specific buffers under a parent <code>ContextualBufferOut</code>.<br>
 * Writes occur only through the parent, but for reading operation this buffer behaves as a standard policy element.<br>
 * Note : the buffer attribute is shared with the parent for global capacity control, hence some methods are overridden when context
 * specific information is needed.<br>
 * A drop() method is provided to dispose of obsolete context-specific buffers. Cleanup policies are parameter-based. Some cleanup
 * operations are conducted by the ContextBufferOut itself, and notified to the parent.
 * @author ojoh6423
 */
public class ContextBufferOut extends BufferOut {
    /** The contextual multi-buffer this context-specific buffer belongs to. */
    ContextualBufferOut parent = null;

    /** A reference to this buffer for the inner DropTimer. */
    ContextBufferOut thisBuffer = this;

    /** The Context this buffer is used for. */
    Context context = null;

    /** Timestamp of the last significant event for the dropping policy. */
    long dropTimestamp = 0;

    /**
     * String identifier for this buffer.<br>
     * Replaces the id held in the attribute for standard buffers (as the attribute is here shared). The ContextBufferOut id is represented
     * by routing name and context elements.
     */
    String id = null;

    /**
     * Constructor.
     * @param pAttribute the buffer attribute to be used (should be the parent's attribute),
     * @param pEngineOut a reference to the EngineOut that should be activated when appropriate,
     * @param ctx the Context this buffer is specific to.
     */
    public ContextBufferOut(final BufferOutAttribute pAttribute,
            final EngineOut pEngineOut, Context ctx) {
        super();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(
                    "Creating a new ContextBufferOut for ");
            buffer.append(pEngineOut.getNode().getName());
            buffer.append(" ctx:");
            buffer.append(ctx);
            buffer.append(" ");
            buffer.append(pAttribute.getId());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
        attribute = pAttribute;
        engineOut = pEngineOut;
        parent = (ContextualBufferOut) pAttribute.getBufferOut();
        context = ctx;
        id = attribute.getId() + File.separator + ctx.getIdentity() + "!"
                + ctx.getSubIdentity() + ".ctx";
        if (ContextManager.getInstance().getSessionForContext(ctx) == null) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                StringBuffer buffer = new StringBuffer(
                        "No known session for context : ");
                buffer.append(ctx);
                LOGGER.log(BasicLevel.INFO, buffer.toString());
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                StringBuffer buffer = new StringBuffer("Context ");
                buffer.append(ctx);
                buffer.append(" held by session ");
                buffer.append(ContextManager.getInstance()
                        .getSessionForContext(ctx));
                LOGGER.log(BasicLevel.INFO, buffer.toString());
            }
        }
    }

    /**
     * String representation of ContextualBufferOut.
     * @return the string representation
     */
    public String toString() {
        return "Context" + super.toString() + context;
    }

    /**
     * The ContextBufferOut id is represented by routing name and context elements.
     * @return the id of this ContextBufferOut
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getId()
     */
    public String getId() {
        return id;
    }

    /**
     * Gets the destination associated with this BufferOut, which is its parent's id associated with the context.
     * @return the destination
     */
    public Destination getDestination() {
        return new Destination(attribute.getId(), context);
    }

    /**
     * Gets the number of pending messages.<br>
     * Overrides the method to take only this context buffer messages into account.
     * @return the number of messages which are not contained in the current uncommitted transaction
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getPendingMessage()
     */
    public int getPendingMessage() {
        return (getSizeObject() - attribute.sizeCurrentTransaction);
    }

    /**
     * Gets the transaction id.<br>
     * A ContextBufferOut shares its transaction id with its parent ContextualBufferOut
     * @return the id transaction
     */
    public long getIdTransaction() {
        return parent.getIdTransaction();
    }

    /**
     * Gets the objects.<br>
     * Simply adds an IO operation notification (for context-specific buffer clearing policies) to the standard getObjects().
     * @return a list which contains the bufferObjects, but the list is limited by the thresholdNumberOfObject or
     * signalThresholdNumberOfObject;
     */
    public List getObjects() {
        notifyIOAction();
        return super.getObjects();
    }

    /**
     * Gets rid of that buffer.<br>
     * <li>remaining messages are lost (released, and thus counted out of the parent ContextualBufferOut), <li>for a persistent buffer, the
     * context-specific persistence directory is deleted.
     */
    public void drop() {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Dropping : " + this);
        }
        boolean clean = false;
        while (!clean) {
            List allObjects = getAllObjects();
            clean = release(allObjects);
        }
        // for (Iterator it = getObjects().iterator(); it.hasNext();) {
        // BufferObject o = (BufferObject) it.next();
        // attribut.incSizeByte(-o.getSize());
        // }
        // remove the associated subdirectory
        if (attribute.isPersistant()) {
            final File repertoire = new File(path + File.separator);
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Persistence directory is "
                        + repertoire);
            }
            if (!repertoire.isDirectory()) {
                if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                    LOGGER.log(BasicLevel.ERROR, "Path error for " + toString()
                            + " - cannot delete " + repertoire);
                }
                return;
            }
            if (repertoire.delete()) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Directory removed : "
                            + repertoire);
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                    LOGGER.log(BasicLevel.ERROR, "Failed to delete "
                            + repertoire);
                }
            }
        }
    }

    /**
     * Schedules a drop() on this buffer.<br>
     * Uses a DropTimer task on the global BufferOut timer if the delay is greater than zero.
     * @param onlyIfEmpty the only if empty
     * @param delay the delay
     */
    public void scheduleDrop(boolean onlyIfEmpty, long delay) {
        if (delay <= 0) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Drop scheduled with 0 delay for "
                        + this);
            }
            if (getSizeObject() == 0 || !onlyIfEmpty) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "Processing immediate drop.");
                }
                if (checkDrop()) {
                    drop();
                    parent.bufferRemoved(thisBuffer);
                } else {
                    if (attribute.getClearDelay() <= 0) {
                        if (LOGGER.isLoggable(BasicLevel.WARN)) {
                            LOGGER
                                    .log(BasicLevel.WARN,
                                            "Drop rejected by a watcher, delay is 0 - won't reset timer");
                        }
                    } else {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER
                                    .log(BasicLevel.DEBUG,
                                            "Drop rejected by a watcher - resetting timer");
                        }
                        scheduleDrop(onlyIfEmpty, attribute.getClearDelay());
                    }
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "Buffer not empty : won't drop.");
                }
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "Scheduling drop with parameters : " + onlyIfEmpty
                                + "/" + delay);
            }
            dropTimestamp = System.currentTimeMillis();
            DropTimer task = new DropTimer(onlyIfEmpty, delay);
            getTimer().schedule(task, delay);
        }
    }

    /**
     * When conditions for clearing this buffer are reached, checks whether the BufferOutManager allows it.
     * @return true if operation should be conducted
     */
    private boolean checkDrop() {
        List contents = new ArrayList();
        for (Iterator it = getAllObjects().iterator(); it.hasNext();) {
            contents.add(((BufferObject) it.next()).getMsg());
        }
        boolean allow = parent.bufferOutManager
                .isContextBufferOutRemovalAllowed(context, contents, attribute
                        .getStrClear());
        return allow;
    }

    /**
     * Changes the context.<br>
     * This is not a plain setter and should be used only when an implicit Context is replaced by an explicit one (i.e. this buffer's
     * Context should be the one used to index this buffer in the ContextualBufferOut map).<br>
     * The changes include : <li>changing context itself <li>recomputing id, <li>recomputing path, <li>renaming the persistence directory
     * if the buffer is persistent.<br>
     * Note that if persistence-related operations fail, the buffer will NOT report a failure to the caller : it will actually operate with
     * the new context value, although persistence may not be assured.
     * @param newContext the new context
     * @return always true, except if the current context was already explicit
     */
    public boolean changeContext(Context newContext) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Context change required from "
                    + context + " to " + newContext);
        }
        if (!context.isImplicit()) {
            if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                LOGGER.log(BasicLevel.ERROR, "Cannot change explicit context "
                        + context);
            }
            return false;
        }
        context = newContext;
        id = attribute.getId() + File.separator + context.getIdentity() + "!"
                + context.getSubIdentity() + ".ctx";
        if (attribute.isPersistant()) {
            File repertoire = new File(path + File.separator);
            if (repertoire.isDirectory()) {
                String newPath = Domain.getInstance().getPath()
                        + File.separator + engineOut.getNode().getName()
                        + File.separator + "OUT" + File.separator + getId();
                File newRepertoire = new File(newPath + File.separator);
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER
                            .log(BasicLevel.DEBUG,
                                    "Persistent context change - new path : "
                                            + newPath);
                }
                if (repertoire.renameTo(newRepertoire)) {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO, "Directory " + repertoire
                                + " renamed as " + newRepertoire);
                    }
                    path = newPath;
                } else {
                    if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                        LOGGER.log(BasicLevel.ERROR, "Rename error for "
                                + repertoire
                                + " - persistence may not be assured.");
                    }
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                    LOGGER
                            .log(
                                    BasicLevel.ERROR,
                                    "Path error for "
                                            + toString()
                                            + " - cannot rename - persistence may not be assured.");
                }
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "Non persistent context change - new id : " + id);
            }
        }
        return true;
    }

    /**
     * Takes appropriate actions concerning clearing policy when the peer associated with this buffer's context returns.<br>
     * If the clearing policy is based on peer presence : <li>update dropTimestamp to prevent previously scheduled deletion.
     */
    public void notifyPeerReturn() {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Notifying peer return for context : "
                    + context);
        }
        if (attribute.isClearWhenPeerLost()) {
            dropTimestamp = System.currentTimeMillis();
        } else {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, " --- ignored");
            }
        }
    }

    /**
     * Takes appropriate actions when a read or write is done on this buffer.<br>
     * If the clearing policy is based on usage : <li>schedule a drop task if none is already running, <li>update dropTimestamp to prevent
     * previously scheduled deletion.
     */
    public void notifyIOAction() {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Notifying IO action for context : "
                    + context);
        }
        if (attribute.isClearWhenUnused()) {
            long previousTimestamp = dropTimestamp;
            dropTimestamp = System.currentTimeMillis();
            if (previousTimestamp == 0) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            " --- rescheduling 'unused' drop");
                }
                scheduleDrop(attribute.isClearWhenEmpty(), attribute
                        .getClearDelay());
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, " --- ignored");
            }
        }
    }

    // -------------------------
    // inner class
    // -------------------------
    /**
     * The DropTimer task checks for buffer drop conditions and drops it if necessary.<br>
     * This timer task is set at buffer creation for UNUSED and UNUSED_AND_EMPTY strategies or when the peer is lost for PEER_LOST or
     * PEER_LOST_AND_EMPTY strategies. It may reschedule another DropTimer task.
     */
    class DropTimer extends TimerTask {
        /**
         * Denotes the drop() should happen only if the buffer is empty when the timer expires.
         */
        boolean onlyIfEmpty;

        /** The delay. */
        long delay;

        /**
         * Constructor.
         * @param onlyIfEmpty the only if empty
         * @param delay the delay
         */
        public DropTimer(boolean onlyIfEmpty, long delay) {
            super();
            this.onlyIfEmpty = onlyIfEmpty;
            this.delay = delay;
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "DropTimer task " + this
                        + " created for " + thisBuffer);
            }
        }

        /**
         * Drops the buffer if the optional EMPTY condition is met and the timer has not been reset.
         */
        public void run() {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "DropTimer task executing for "
                        + this);
            }
            if (getSizeObject() == 0 || !onlyIfEmpty) {
                long elapsedSinceTimestamp = System.currentTimeMillis()
                        - dropTimestamp;
                if (elapsedSinceTimestamp >= delay) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG,
                                "DropTimer : actually dropping");
                    }
                    if (checkDrop()) {
                        drop();
                        parent.bufferRemoved(thisBuffer);
                    } else {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER
                                    .log(BasicLevel.DEBUG,
                                            "DropTimer : drop rejected by a watcher - resetting timer");
                        }
                        scheduleDrop(onlyIfEmpty, attribute.getClearDelay());
                    }
                } else {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG,
                                "DropTimer : Timer was reset - won't drop.");
                    }
                    // re-schedule if UNUSED mode
                    // (if PEER_LOST mode, being here means the peer has come
                    // back alive : the task will trigger again on next loss)
                    if (attribute.isClearWhenUnused()) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER
                                    .log(
                                            BasicLevel.DEBUG,
                                            "Rescheduling 'unused' drop with parameters : "
                                                    + onlyIfEmpty
                                                    + "/"
                                                    + (attribute.getClearDelay() - elapsedSinceTimestamp));
                        }
                        scheduleDrop(onlyIfEmpty, attribute.getClearDelay()
                                - elapsedSinceTimestamp);
                    }
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER
                            .log(BasicLevel.DEBUG,
                                    "DropTimer : ContextBufferOut not empty - won't drop, nor reschedule.");
                }
                // no reschedule : buffer won't become empty if it's not used
                // again or the peer is found again
                dropTimestamp = 0;
            }
        }
    }
}
