package org.ow2.odis.bufferOut;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.IContextWatcher;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.model.BufferOutAttribute;

/**
 * A ContextualBufferOut offers a unique BufferOut facade for a collection of context specific buffers.<br>
 * Context specific buffers (ContextBufferOut) are created on-the-fly when a message addressing a new Context arrives. All those buffers
 * share the same BufferOutAttribut, so that capacity checking is done globally. Each has its own timer tasks, however.<br>
 * The ContextualBufferOut alone is seen by the EngineProceed, but each individual ContextBufferOut is seen as a separate policy element
 * from the EngineOut side, so as to prevent a session failure for one peer from freezing the whole system.<br>
 * The ContextualBufferOut registers in the ContextManager as a ContextWatcher and takes care of making sub-buffers available or not for
 * the policy instance, depending on the availability of a peer for their context.<br>
 * Context buffers are always cleared as soon as the peer is lost if the context is implicit (but UNUSED conditions may apply too) ; for
 * explicit contexts, the behavior depends on the value specified in the configuration file for the &lt;CLEAR_WHEN/&gt; tag : <li>PEER_LOST
 * : the buffer is cleared after a delay when the peer is lost (provided it didn't connect again), even if it still contains data ; the
 * delay is specified in the tag, in seconds, and defaults to 0. (Actually, Context declaration and deletion, notified by the
 * ContextManager, are used to detect peer presence.) <li>PEER_LOST_AND_EMPTY : same as "PEER_LOST" but the buffer must be empty <li>UNUSED
 * : it is cleared if no read or write operation happens over the specified delay (defaults to 300), <li>UNUSED_AND_EMPTY : same as above
 * but the buffer must be empty over the delay
 * @author ojoh6423
 */
public class ContextualBufferOut extends BufferOut implements IContextWatcher {
    
    /** Logger for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            ContextualBufferOut.class.getName());

    /** A map of context-specific buffers, indexed by context. */
    private Map mapContextBuffers = new HashMap();

    /** The BufferOutManager instance responsible for this buffer. */
    protected BufferOutManager bufferOutManager = null;

    /**
     * Constructor.
     * @param pAttribut the attribute describing the ContextualBufferOut
     * @param pEngineOut the EngineOut the buffer belongs to (for notification purposes)
     */
    public ContextualBufferOut(final BufferOutAttribute pAttribut,
            final EngineOut pEngineOut) {
        super(pAttribut, pEngineOut);
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO,
                    "The above buffer was a ContextualBufferOut");
        }
        bufferOutManager = pEngineOut.getBufferOutManager();
    }

    /**
     * Initializes the component.<br>
     * After standard BufferOut initialization, remove the empty sub-buffers.
     */
    public void initComponent() {
        super.initComponent();
        Collection cbs = mapContextBuffers.values();
        for (Iterator i = cbs.iterator(); i.hasNext();) {
            ContextBufferOut cbo = (ContextBufferOut) i.next();
            if (cbo.getSizeObject() == 0) {
                i.remove();
                cbo.drop();
            }
        }
    }

    /**
     * String representation of ContextualBufferOut.
     * @return the string representation
     */
    public String toString() {
        return "Contextual" + super.toString();
    }

    /**
     * Stores a message in the buffer.<br>
     * This method first locates the ContextBufferOut for the provided Context, or asks for its creation if it doesn't exist. Then the
     * storage is delegated to that buffer.
     * @param context the context
     * @param pIdTransaction the id transaction
     * @param msg the msg
     * @return true, if put data
     */
    public boolean putData(final long pIdTransaction, Context context,
            final byte[] msg) {
        ContextBufferOut subBuffer = null;
        if (mapContextBuffers.containsKey(context)) {
            subBuffer = (ContextBufferOut) mapContextBuffers.get(context);
        } else {
            subBuffer = bufferOutManager.createContextBufferOut(getAttribut(),
                    context);
            mapContextBuffers.put(context, subBuffer);
        }
        if (subBuffer == null) {
            LOGGER.log(BasicLevel.FATAL, "Buffer vanished for context : "
                    + context);
            return false; // WHAT SHOULD WE REALLY DO ?????
        } else {
            subBuffer.notifyIOAction();
            return subBuffer.putData(pIdTransaction, msg);
        }
    }

    /**
     * This should not be done directly on this kind of buffer.
     * @param pIdTransaction Transaction id
     * @param msg the message as a byte array
     * @return true, if put data
     * @throws RuntimeException always
     */
    public boolean putData(final long pIdTransaction, final byte[] msg) {
        StringBuffer sb = new StringBuffer(
                "INTERNAL : Method putData(long, byte[]) was called on ");
        sb.append(toString()).append(" for transaction #").append(
                pIdTransaction);
        LOGGER.log(BasicLevel.FATAL, sb.toString());
        throw new RuntimeException(sb.toString());
    }

    /**
     * This should not be done directly on this kind of buffer.
     * @param bufferObject the Object which has to be sent
     * @throws RuntimeException always
     */
    public final void putData(final BufferObject bufferObject) {
        StringBuffer sb = new StringBuffer(
                "INTERNAL : Method putData(BufferObject) was called on ");
        sb.append(toString()).append(" for BufferObject in transaction #")
                .append(bufferObject.getIdTransaction());
        LOGGER.log(BasicLevel.FATAL, sb.toString());
        throw new RuntimeException(sb.toString());
    }

    /**
     * Creates a context-specific sub-buffer for this buffer.
     * @param ctx the wanted context
     * @return the created buffer
     */
    public ContextBufferOut createContextBufferOut(Context ctx) {
        ContextBufferOut cbo = (ContextBufferOut) mapContextBuffers.get(ctx);
        if (cbo == null) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "Creating a new ContextBufferOut for " + ctx);
            }
            cbo = bufferOutManager.createContextBufferOut(getAttribut(), ctx);
            mapContextBuffers.put(ctx, cbo);
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Reusing buffer " + cbo
                        + "for context " + ctx);
            }
        }
        return cbo;
    }

    /**
     * Denotes the fact that this buffer is able to handle contexts.
     * @return always true
     */
    public boolean isContextual() {
        return true;
    }

    /**
     * Takes appropriate measures when a Context is lost (the associated communication channel has become unavailable).<br>
     * In all cases, the ContextBuffer associated with the Context is removed from the policy elements list.<br>
     * The buffer is actually removed if it is empty (it will be re-created when a new message wants to get out) or if the Context was
     * implicit (there is no way to consistently identify the same peer after reconnection). Richer strategies will be provided in a future
     * release.
     * @param ctx the ctx
     */
    public void contextLost(Context ctx) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Context lost : " + ctx);
        }
        ContextBufferOut cbo = (ContextBufferOut) mapContextBuffers.get(ctx);
        if (cbo != null) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Asking PolicyInstance to remove "
                        + cbo + " for lost context " + ctx);
            }
            engineOut.getPolicyInstance().removePolicyElement(cbo);
            if (attribute.isClearWhenPeerLost()) {
                if (ctx.isImplicit()) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG,
                                "Asking for immediate drop of " + cbo);
                    }
                    cbo.scheduleDrop(false, 0);
                } else {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Scheduling drop of "
                                + cbo);
                    }
                    cbo.scheduleDrop(attribute.isClearWhenEmpty(), attribute
                            .getClearDelay());
                }
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER
                        .log(BasicLevel.DEBUG,
                                "No buffer found for lost context");
            }
        }
    }

    /**
     * Takes appropriate measures when a new Context is associated with a valid session.<br>
     * If a ContextBuffer is already associated with the Context, it is simply added to the policy elements list.<br>
     * If no such buffer exists, nothing is done : the buffer will be created when a message addresses the Context and added to the policy
     * at that time.
     * @param ctx the ctx
     */
    public void contextSet(Context ctx) {
        ContextBufferOut cbo = (ContextBufferOut) mapContextBuffers.get(ctx);
        if (cbo != null) {
            cbo.notifyPeerReturn();
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Asking PolicyInstance to add "
                        + cbo + " for context " + ctx);
            }
            engineOut.getPolicyInstance().addPolicyElement(cbo);
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "No buffer found for new context "
                        + ctx);
            }
        }
    }

    /**
     * Takes appropriate measures when a Context is replaced by another.<br>
     * Processing has no options (will have in future releases) : <li>If a ContextBuffer is already associated with the new context, it is
     * added to the policy elements list and if a buffer was associated with the old context, that buffer is dropped (even if not empty).
     * <li>If no such buffer exists but a buffer was associated with the old context, a new buffer is associated with the new context and
     * added to the policy elements list, and all objects in the old buffer are inserted in the new one (we cannot simply reuse the buffer
     * because the persistence path would be wrong). <li>In other cases, nothing is done : the needed buffer will be created when a message
     * addresses the Context and added to the policy at that time.
     * @param oldContext the old context
     * @param newContext the new context
     */
    public void contextReplaced(Context oldContext, Context newContext) {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Context " + oldContext
                    + " has been replaced by " + newContext);
        }
        ContextBufferOut ocbo = (ContextBufferOut) mapContextBuffers
                .get(oldContext);
        ContextBufferOut ncbo = (ContextBufferOut) mapContextBuffers
                .get(newContext);
        if (ncbo != null) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Found an existing buffer for "
                        + newContext + " : " + ncbo);
            }
            ncbo.notifyPeerReturn();
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Asking PolicyInstance to add "
                        + ncbo + " for context " + newContext);
            }
            engineOut.getPolicyInstance().addPolicyElement(ncbo);
            if (ocbo != null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "Asking PolicyInstance to remove " + ocbo
                                    + " for context " + oldContext);
                }
                engineOut.getPolicyInstance().removePolicyElement(ocbo);
                if (oldContext.isImplicit()) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER
                                .log(BasicLevel.DEBUG,
                                        "Old buffer was implicit : scheduling immediate drop");
                    }
                    ocbo.scheduleDrop(false, 0);
                } else {
                    // <JDO "triggered by the timer task"/>
                    // mapContextBuffers.remove(oldContext);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER
                                .log(BasicLevel.DEBUG,
                                        "Old buffer was explicit : checking drop conditions");
                    }
                    if (attribute.isClearWhenPeerLost()) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER
                                    .log(BasicLevel.DEBUG,
                                            " --- scheduling drop");
                        }
                        ocbo.scheduleDrop(attribute.isClearWhenEmpty(), attribute
                                .getClearDelay());
                    }
                }
            }
        } else {
            if (ocbo != null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "No preexisting buffer found. Reusing buffer "
                                    + ocbo + " for context " + newContext);
                }
                engineOut.getPolicyInstance().addPolicyElement(ocbo);
                ocbo.changeContext(newContext);
                mapContextBuffers.put(newContext, ocbo);
                mapContextBuffers.remove(oldContext);
            }
        }
    }

    /**
     * Takes appropriate measures when a ContextBuffer is dropped.<br>
     * Removes the buffer from the map.<br>
     * This method is called by the ContextBuffer when it drops, which usually happens asynchronously due to the timer mechanism.
     * @param cbo the cbo
     */
    protected void bufferRemoved(ContextBufferOut cbo) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Child buffer removed : " + cbo);
        }
        Object context = cbo.getDestination().getContext();
        if (mapContextBuffers.containsKey(context)) {
            mapContextBuffers.remove(context);
        } else {
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN,
                        "Buffer not found in map for context : " + context);
            }
        }
    }
}
