package org.ow2.odis.bufferOut;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * <code>PersistenceBufferObject</code> plays the same role as <code>BufferObject</code> and adds a persistence functionality.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class PersistenceBufferObject extends BufferObject implements Comparable {
    /**
     * Id of serializable class.
     */
    private static final long serialVersionUID = 1L;

    /**
     * Name or full path of the file holding the persistent object.
     */
    transient private String name = "";

    /**
     * Id of the object. Used to order the objects during restoration.
     */
    private long idMessage = -2;

    /**
     * Constructor.
     * @param idTransaction the transaction this object belongs to
     * @param pMsg byte array representing this object
     * @param path location where the file version of this object is saved for persistence
     * @param idMessage identity of this object
     */
    public PersistenceBufferObject(long idTransaction, byte[] pMsg,
            String path, long idMessage) {
        super(idTransaction, pMsg);
        this.idMessage = idMessage;
        name = path + File.separator + idMessage + ".sav";
        try {
            File file = new File(name + ".temp");
            file.createNewFile();
            RandomAccessFile fichier = new RandomAccessFile(name + ".temp",
                    "rw");
            fichier.writeLong(idTransaction);
            fichier.writeLong(idMessage);
            fichier.write(pMsg);
            fichier.close();
            file.renameTo(new File(name));
        } catch (FileNotFoundException e) {
            LOGGER.log(BasicLevel.ERROR, "Unable to open persistence file "
                    + name, e);
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "IO error on file " + name, e);
        }
    }

    /**
     * Constructor.
     */
    PersistenceBufferObject() {
        super();
    }

    /**
     * Reconstruct the object by reading data from the file.
     * @param file the file containing the previously saved data
     * @return true if success
     */
    boolean read(final File file) {
        long nbrOctets = 0;
        long nbrOctetsRead = 0;
        boolean result = true;
        try {
            name = file.getAbsolutePath();
            final RandomAccessFile fichier = new RandomAccessFile(file, "r");
            setIdTransaction(fichier.readLong());
            idMessage = fichier.readLong();
            nbrOctets = (long) (fichier.length() - fichier.getFilePointer());
            final byte[] msg = new byte[(int) nbrOctets];
            nbrOctetsRead = fichier.read(msg);
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "PBO-Restored " + name + " : id "
                        + idMessage + " - length " + nbrOctetsRead);
            }
            setMsg(msg);
            fichier.close();
        } catch (FileNotFoundException e) {
            LOGGER.log(BasicLevel.ERROR, "File " + file.getName()
                    + " not found for restoration. Object lost.");
            result = false;
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "IO error during restoration on file "
                    + file.getName(), e);
            result = false;
        }
        if (nbrOctets != nbrOctetsRead) {
            LOGGER.log(BasicLevel.ERROR,
                    "Couldn't read all bytes into the message for "
                            + file.getName());
            result = false;
        }
        return result;
    }

    /**
     * Deletes the persistent object.
     * @return true if success
     */
    public boolean release() {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG,
                    "Deleting a PersistenceSynchronizedObject " + name);
        }
        final File file = new File(name);
        if (!file.exists() && LOGGER.isLoggable(BasicLevel.WARN)) {
            LOGGER.log(BasicLevel.WARN,
                    "Problem trying to delete non-existent persistence file "
                            + name);
        }
        file.delete();
        return super.release();
    }

    /**
     * Compares two PersistentBufferObject.<br>
     * This is used during restoration to sort objects.
     * @param object the object this object is to be compared to
     * @return 0 if this has the same id as the argument ; -1 if this has a lower id ; +1 if this has a higher id
     */
    public int compareTo(Object object) {
        final PersistenceBufferObject pbo = (PersistenceBufferObject) object;
        return new Long(idMessage).compareTo(new Long(pbo.idMessage));
    }

    /**
     * Gets the id message.
     * @return id of the message
     */
    public long getIdMessage() {
        return idMessage;
    }

    /**
     * Sets the id message.
     * @param idMessage id of the message
     */
    public void setIdMessage(final long idMessage) {
        this.idMessage = idMessage;
    }
}
