package org.ow2.odis.connection;

import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.Launch;
import org.ow2.odis.connection.composite.CompositeFactory;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.connection.memory.MemoryFactory;
import org.ow2.odis.connection.socket.SocketConnectionFactory;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.services.IFactoryService;

/**
 * Create connection attribute and the connection In, Out, Contextual . It delegate all the instance creation to specialized factories.
 * Factories needed are created during init and referenced into a hash map. Factories can be also declared into a configuration file.
 * @version 1.0
 * @author Beyler Olivier FranceTelecom R&D
 */
public class ConnectionFactory {
    /** The Constant CONNECTION_SOCKET. */
    public static final String CONNECTION_SOCKET = "SOCKET";

    /** The Constant CONNECTION_MEMORY. */
    public static final String CONNECTION_MEMORY = "MEMORY";

    /** The Constant CONNECTION_COMPOSITE. */
    public static final String CONNECTION_COMPOSITE = "COMPOSITE";

    /** The Constant CONNECTION_CFG. */
    public static final String CST_CONNECTION = "CONNECTION";

    /**
     *
     */
    public static final String CST_CONNECTIONS_FACTORY = "CONNECTIONS_FACTORY";

    /**
     *
     */
    public static final String CST_NAME = "name";

    /**
     *
     */
    public static final String CST_VALUE = "value";

    /**
     *
     */
    public static final String CST_CLASS = "CLASS";

    /** The instance. */
    private static ConnectionFactory instance = null;

    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            ConnectionFactory.class.getName());

    /**
     * implement singleton pattern.
     * @return the unique instance
     */
    public static ConnectionFactory getInstance() {
        if (instance == null) {
            instance = new ConnectionFactory();
            instance.init();
        }
        return instance;
    }

    private final Map mapModeConnectionFactory = new HashMap();

    /**
     * Instantiates a new connection factory.
     */
    private ConnectionFactory() {
        super();
    }

    /**
     * init the Connection factory with default connection provided by Odis and allow user to add a new connection factory. Default
     * connection is a connection witch not need additional librairies. Others connections witch need additional librairies can be added by
     * using a connection.xml file
     */
    void init() {
        // initialize the map of factory with default connection
        addFactory(CONNECTION_MEMORY, MemoryFactory.getInstance());
        addFactory(CONNECTION_SOCKET, SocketConnectionFactory.getInstance());
        addFactory(CONNECTION_COMPOSITE, CompositeFactory.getInstance());
        // look inside configuration file if new connection is available.
        try {
            URL url = ConnectionFactory.class
                    .getResource("/" + Launch.ODIS_CFG);
            if (url == null) {
                url = new URL(Launch.ODIS_CFG);
            }
            Document doc = new SAXBuilder().build(url.openStream());
            Element root = doc.getRootElement();
            Element cf = root.getChild(CST_CONNECTIONS_FACTORY);
            Iterator it = cf.getChildren(CST_CONNECTION).iterator();
            while (it.hasNext()) {
                Element element = (Element) it.next();
                String key = element.getAttributeValue(CST_NAME);
                Element classe = element.getChild(CST_CLASS);
                if (classe == null && LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer warn;
                    warn = new StringBuffer("Bad configuration into file \"");
                    warn.append(Launch.ODIS_CFG);
                    warn.append("\" for connection \"");
                    warn.append(key);
                    warn.append("\" <CLASS value=\"...\"/> is missing.");
                    LOGGER.log(BasicLevel.WARN, warn.toString());
                } else {
                    addFactory(key, classe.getAttributeValue(CST_VALUE));
                }
            }
        } catch (JDOMException e) {
            LOGGER
                    .log(BasicLevel.ERROR, "Init of connection Factory failed",
                            e);
        } catch (Exception e) {
            // IOException or Null pointer exception (URL = null)
            // have the same behavior
            StringBuffer warn = new StringBuffer("No or bad file \"");
            warn.append(Launch.ODIS_CFG);
            warn.append("\" present into classpath.");
            LOGGER.log(BasicLevel.WARN, warn.toString());
        }
    }

    /**
     * addFactory give the opportunity to add new connection factory to the runtime.
     * @param key name of the new connection
     * @param factory name of the factory class
     */
    public void addFactory(final String key, final String factory) {
        IConnectionFactory cf;
        IFactoryService loader = Domain.getInstance().getFactoryInstance();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer info;
            info = new StringBuffer("Set a new Connection Factory for ");
            info.append(key);
            info.append(" -> ");
            info.append(factory);
            LOGGER.log(BasicLevel.INFO, info);
        }
        try {
            cf = (IConnectionFactory) loader.getInstanceOf(factory, "1");
            addFactory(key, cf);
        } catch (Exception e) {
            StringBuffer err = new StringBuffer(
                    "Set a new Connection Factory for ");
            err.append(key);
            err.append(" failed raison is : ");
            LOGGER.log(BasicLevel.ERROR, err.toString(), e);
        }
    }

    /**
     * addFactory give the opportunity for plugin or OSGI service to add new connection factory to the runtime.
     * @param key name of the new connection
     * @param factory instance of the factory class
     */
    public void addFactory(final String key, IConnectionFactory factory) {
        IConnectionFactory old = (IConnectionFactory) mapModeConnectionFactory
                .get(key);
        if (factory == null) {
            StringBuffer err = new StringBuffer(
                    "Unable to set a new Connection Factory for ");
            err.append(key);
            LOGGER.log(BasicLevel.ERROR, err.toString());
        } else {
            if (old != null) {
                StringBuffer buffer = new StringBuffer("remove old factory ->");
                buffer.append(old.getClass().getName());
                buffer.append(" set new factory ");
                buffer.append(factory.getClass().getName());
            }
        }
        mapModeConnectionFactory.put(key, factory);
    }

    /**
     * @param key
     * @param factory
     */
    public void removeFactory(final String key, IConnectionFactory factory) {
        IConnectionFactory old = (IConnectionFactory) mapModeConnectionFactory
                .get(key);
        if (factory.equals(old)) {
            mapModeConnectionFactory.remove(key);
        }
    }

    /**
     * create a new connection In or delegate the creation of this new connection to a dedicated factory.
     * @param attribute the attribute
     * @return the connection In just created
     */
    public IConnectionIn newConnectionIn(final AbstractConnectionAttribute attribute) {
        final String mode;
        IConnectionIn result = null;
        final IConnectionFactory factory;
        mode = attribute.getTypeConnection();
        factory = (IConnectionFactory) mapModeConnectionFactory.get(mode);
        if (factory == null) {
            StringBuffer err = new StringBuffer("Unknown connection type \"");
            err.append(mode);
            err.append("\" => can not create it ");
            LOGGER.log(BasicLevel.ERROR, err);
        } else {
            result = factory.newConnectionIn(attribute);
            if (result == null) {
                StringBuffer err = new StringBuffer("Connection In type \"");
                err.append(mode);
                err.append("\" is not available ");
                LOGGER.log(BasicLevel.ERROR, err);
            }
        }
        return result;
    }

    /**
     * Creates a new OUT connection or delegates the creation of this new connection to a dedicated factory.
     * @param attribute the attribute holding the connection parameters
     * @return the connection just created
     */
    public IConnectionOut newConnectionOut(final AbstractConnectionAttribute attribute) {
        IConnectionOut result = null;
        final String mode;
        final IConnectionFactory factory;
        mode = attribute.getTypeConnection();
        factory = (IConnectionFactory) mapModeConnectionFactory.get(mode);
        if (factory == null) {
            StringBuffer err = new StringBuffer("Unknown connection type \"");
            err.append(mode);
            err.append("\" => can not create it ");
            LOGGER.log(BasicLevel.ERROR, err);
        } else {
            result = factory.newConnectionOut(attribute);
            if (result == null) {
                StringBuffer err = new StringBuffer("Connection Out type \"");
                err.append(mode);
                err.append("\" is not available");
                LOGGER.log(BasicLevel.ERROR, err);
            }
        }
        return result;
    }

    /**
     * Creates a new OUT connection or delegates the creation of this new connection to a dedicated factory.
     * @param attribute the attribute holding the connection parameters
     * @return the connection just created
     */
    public IContextualConnectionOut newContextualConnectionOut(
            final AbstractConnectionAttribute attribute) {
        String mode = attribute.getTypeConnection();
        IContextualConnectionOut result = null;
        IConnectionFactory factory = (IConnectionFactory) mapModeConnectionFactory
                .get(mode);
        if (factory == null) {
            StringBuffer err = new StringBuffer("Unknown connection type \"");
            err.append(mode);
            err.append("\" => can not create it ");
            LOGGER.log(BasicLevel.ERROR, err);
        } else {
            result = factory.newContextualConnectionOut(attribute);
            if (result == null) {
                StringBuffer err = new StringBuffer("Connection type \"");
                err.append(mode);
                err.append("\" is not Context-able => can't create it");
                LOGGER.log(BasicLevel.ERROR, err);
            }
        }
        return result;
    }

    /**
     * Creates a new connection attribute.
     * @param mode denotes the type of attribute
     * @return a connection attribute
     */
    public AbstractConnectionAttribute newAttribute(final String mode) {
        AbstractConnectionAttribute result = null;
        IConnectionFactory factory = (IConnectionFactory) mapModeConnectionFactory
                .get(mode);
        if (null != factory) {
            result = factory.newAttribute(Domain.getInstance().attribute);
        } else {
            StringBuffer err = new StringBuffer("Unknown connection type \"");
            err.append(mode);
            err.append("\"");
            err.append("connection type known are : ");
            Iterator it = mapModeConnectionFactory.keySet().iterator();
            while (it.hasNext()) {
                err.append(it.next());
                err.append(" ");
            }
            LOGGER.log(BasicLevel.ERROR, err.toString());
        }
        return result;
    }
}
