package org.ow2.odis.connection;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.jdom.Element;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.port.PortIn;

/**
 * A ConnectionManager manages all connections for an ODIS Domain.<br>
 * Implements singleton pattern.<br>
 * Note : the ConnectionManager actually only manages the connection <u>attributes</u> and helps stopping properly the connections when
 * unloading the Domain.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class ConnectionManager implements ConnectionManagerMBean {
    /** Logger for this class. */
    final static Logger LOGGER = Monolog.initialize().getLogger(
            ConnectionManager.class.getName());

    /**
     * map defining a relation between name of a connection an the connection IN.
     */
    private static Map mapCnxIN = Collections.synchronizedMap(new HashMap());

    /**
     * map defining a relation between name of a connection an the connection OUT.
     */
    private static Map mapCnxOUT = Collections.synchronizedMap(new HashMap());

    /**
     * Map listing existing IN connections by node name and link name.<br>
     * Key : node name ; Value : a submap with Key : link name ; value : the connection.
     */
    public static Map mapNodeCnxIN = Collections.synchronizedMap(new HashMap());

    /**
     * Map listing existing OUT connections by node name and link name.<br>
     * Key : node name ; Value : a submap with Key : link name ; value : the connection. public for JUnit tests
     */
    public static Map mapNodeCnxOUT = Collections
            .synchronizedMap(new HashMap());

    /** Map of all connection attributes, indexed by name. */
    private static Map mapCnxAttribut = Collections
            .synchronizedMap(new HashMap());

    /** The singleton. */
    private static ConnectionManager instance = null;

    /**
     * Gets the singleton.
     * @return the ConnectionManager for the running Domain
     */
    public static synchronized ConnectionManager getInstance() {
        if (instance == null) {
            instance = new ConnectionManager();
        }
        return instance;
    }

    /**
     * Default constructor.
     */
    private ConnectionManager() {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create ConnectionManager");
        }
        JMX.getInstance().registerMBean("Connection Manager", this);
    }

    /**
     * Returns an IN connection for the specified node and attribute parameters.<br>
     * First tries to find an existing IN connection for that node and link name (ignoring other attribute values), then creates one if
     * none has been found.
     * @param nodeName name of the node needing the connection
     * @param cnxAttribute all parameters
     * @return the found or created connection
     */
    public IConnectionIn getConnectionIn(String nodeName,
            AbstractConnectionAttribute cnxAttribute) {
        Map m = (Map) mapNodeCnxIN.get(nodeName);
        if (m == null) {
            m = new HashMap();
            mapNodeCnxIN.put(nodeName, m);
        }
        IConnectionIn ci = (IConnectionIn) m.get(cnxAttribute.getName());
        if (ci == null) {
            ci = ConnectionFactory.getInstance().newConnectionIn(cnxAttribute);
            if (ci != null) {
                m.put(cnxAttribute.getName(), ci);
            }
        }
        return ci;
    }

    /**
     * Returns an OUT connection for the specified node and attribute parameters.<br>
     * First tries to find an existing OUT connection for that node and link name (ignoring other attribute values), then creates one if
     * none has been found.
     * @param nodeName name of the node needing the connection
     * @param cnxAttribute all parameters
     * @return the found or created connection
     */
    public IConnectionOut getConnectionOut(String nodeName,
            AbstractConnectionAttribute cnxAttribute) {
        Map m = (Map) mapNodeCnxOUT.get(nodeName);
        if (m == null) {
            m = new HashMap();
            mapNodeCnxOUT.put(nodeName, m);
        }
        IConnectionOut co = (IConnectionOut) m.get(cnxAttribute.getName());
        if (co == null) {
            co = ConnectionFactory.getInstance().newConnectionOut(cnxAttribute);
            if (co != null) {
                m.put(cnxAttribute.getName(), co);
            }
        }
        return co;
    }

    /**
     * Returns a context-capable OUT connection for the specified node and attribute parameters.<br>
     * First tries to find an existing IN connection for that node and link name, that allows two-way communication (ignoring other
     * attribute values); then, if none has been found, tries to find an existing context-capable OUT connection; creates it if still none
     * found.
     * @param nodeName name of the node needing the connection
     * @param cnxAttribute all parameters
     * @return the found or created connection
     */
    public IContextualConnectionOut getContextualConnectionOut(String nodeName,
            AbstractConnectionAttribute cnxAttribute) {
        Map m = (Map) mapNodeCnxIN.get(nodeName);
        IContextualConnectionOut cio = null;
        if (m != null) {
            try {
                // Check the IN connection we found can be used as two-way cnx
                // by casting it to IContextualConnectionOut
                cio = (IContextualConnectionOut) m.get(cnxAttribute.getName());
                if (cio != null)
                    cio.usedAsTwoWay();
            } catch (Exception e) {
                LOGGER.log(BasicLevel.ERROR,
                        "Found a non two-way capable IN connection for link : "
                                + cnxAttribute.getName());
                return null;
            }
        }
        if (cio == null) {
            m = (Map) mapNodeCnxOUT.get(nodeName);
            if (m == null) {
                m = new HashMap();
                mapNodeCnxOUT.put(nodeName, m);
            }
            try {
                cio = (IContextualConnectionOut) m.get(cnxAttribute.getName());
            } catch (Exception e) {
                LOGGER.log(BasicLevel.ERROR,
                        "Found a non two-way capable OUT connection for link : "
                                + cnxAttribute.getName());
                return null;
            }
            if (cio == null) {
                cio = ConnectionFactory.getInstance()
                        .newContextualConnectionOut(cnxAttribute);
                if (cio != null) {
                    m.put(cnxAttribute.getName(), cio);
                }
            }
        } else {
            m.put(cnxAttribute.getName(), cio);
        }
        return cio;
    }


    /**
     * Adds a new IN connection, associated with a PortIn.
     * @param connection the connection
     * @param portIn the associated PortIn
     */
    public void addConnectionIn(final IConnectionIn connection,
            final PortIn portIn) {
        mapCnxIN.put(portIn, connection);
    }

  
    /**
     * Suspends the connection manager. <br>
     * Unused, except in integration tests.
     */
    public void suspendComponent() {
    }

    /**
     * Stops the connection manager. <br>
     * Unused, except in integration tests.
     */
    public void stopComponent() {
    }


    /**
     * Unloads the connection manager, breaks all dependencies with other components, clears the statics.<br>
     * <b>This removes access to the singleton</b>, ie : if a getInstance() is called, a new ConnectionManager will be created.<br>
     * @return always 0 for success (cannot fail)
     */
    public int unloadComponent() {
        final IConnectionIn[] connectionins = (IConnectionIn[]) mapCnxIN.values()
                .toArray(new IConnectionIn[mapCnxIN.size()]);
        for (int i = 0; i < connectionins.length; i++) {
            connectionins[i].unloadComponent();
        }
        final IConnectionOut[] connectionOuts = (IConnectionOut[]) mapCnxOUT
                .values().toArray(new IConnectionOut[mapCnxOUT.size()]);
        for (int i = 0; i < connectionOuts.length; i++) {
            connectionOuts[i].unloadComponent();
        }
        mapCnxIN.clear();
        mapCnxOUT.clear();
        mapNodeCnxIN.clear();
        mapNodeCnxOUT.clear();
        instance = null;

        JMX.getInstance().unregisterMBean("Connection Manager", this);
        return 0;
    }

    /**
     * Gets a connection attribute by its name.
     * @param name the name of the wanted attribute
     * @return the attribute or null if the name was not found
     */
    public static AbstractConnectionAttribute getCnxAttribute(String name) {
        return (AbstractConnectionAttribute) mapCnxAttribut.get(name);
    }

    /**
     * Gets a list of all declared connections as a String array.
     * @return an array holding a String in the format <b>[<i>type</i>] <i>name</i></b> for each connection
     */
    public String[] getListConnection() {
        final Collection col = mapCnxAttribut.values();
        final String[] result = new String[col.size()];
        final Iterator itAttribut = col.iterator();
        int index = 0;
        AbstractConnectionAttribute attribute;
        while (itAttribut.hasNext()) {
            attribute = (AbstractConnectionAttribute) itAttribut.next();
            result[index] = "[" + attribute.getTypeConnection() + "] "
                    + attribute.getName();
            index++;
        }
        return result;
    }

    /**
     * Creates a connection attribute by parsing the configuration XML.
     * @param declaration the configuration XML as a JDOM Element
     * @return the created attribute
     */
    public AbstractConnectionAttribute createCnxAttribut(Element declaration) {
        String type = declaration.getName();
        // -------------------------------------
        // define the connection attribute
        AbstractConnectionAttribute cnxAttribut = ConnectionFactory.getInstance()
                .newAttribute(type);
        if (cnxAttribut != null) {
            cnxAttribut.setter(declaration);
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "New Connection Attribute created : "
                        + cnxAttribut);
            }
        }
        return cnxAttribut;
    }

    /**
     * When the domain already exist, this function give the opportunity to add new connection.
     * @param lsCnxAttribut Collection
     */
    public void add(Collection lsCnxAttribut) {
        AbstractConnectionAttribute cnxAttribut;
        Iterator itCnxAttribut = lsCnxAttribut.iterator();
        while (itCnxAttribut.hasNext()) {
            cnxAttribut = (AbstractConnectionAttribute) itCnxAttribut.next();
            mapCnxAttribut.put(cnxAttribut.getName(), cnxAttribut);
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "New CnxAttribut mapped : "
                        + cnxAttribut);
            }
        }
    }

    /**
     * Removes all references to connections used by the specified Node.<br>
     * This method should be called when a Node is unloaded, and only at that time.
     * @param nodeName name of the unloaded Node
     */
    public static void unregister(String nodeName) {
        mapNodeCnxIN.remove(nodeName);
        mapNodeCnxOUT.remove(nodeName);
    }
}
