package org.ow2.odis.connection.composite;

// ~--- non-JDK imports --------------------------------------------------------
import java.util.Observable;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.ConnectionFactory;
import org.ow2.odis.connection.IConnectionOut;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionInitialized;
import org.ow2.odis.lifeCycle.connection.StateConnectionLoaded;
import org.ow2.odis.lifeCycle.connection.StateConnectionStarted;
import org.ow2.odis.lifeCycle.connection.StateConnectionStopped;
import org.ow2.odis.lifeCycle.connection.StateConnectionUnloaded;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.CompositeConnectionAttribute;

/**
 * @author olivier beyler - orange-labs
 */
public class CompositeOut implements IContextualConnectionOut {
    /**
     * Logger for this class.
     */
    private static final Logger LOGGER = Monolog.initialize().getLogger(
            CompositeOut.class.getName());

    /** Attribute describing this composite */
    private CompositeConnectionAttribute attribute = null;

    /** State of this connection. By default it is Loaded */
    private AbstractConnectionState cnxState = new StateConnectionLoaded();

    /** Denotes we run as contextual */
    private boolean contextual = false;

    /**
     * Index of the currently used underlying connection in the connections array
     */
    private int current = 0;

    /** The underlying connections */
    private IConnectionOut[] connections;

    /** Validity of underlying connections */
    private boolean[] valid;

    /**
     * Constructor for a non-contextual composite.
     * @param pAttribut
     */
    public CompositeOut(AbstractConnectionAttribute pAttribut) {
        this(pAttribut, false);
    }

    /**
     * Constructor.
     * @param pAttribute
     * @param contextual denotes a contextual use
     */
    public CompositeOut(AbstractConnectionAttribute pAttribute, boolean contextual) {
        this.contextual = contextual;
        attribute = (CompositeConnectionAttribute) pAttribute;
        attribute.addObserver(this);
    }

    public boolean send(byte[] msg) {
        boolean sent = false;
        int i;
        for (i = 0; !sent && (i < connections.length); i++) {
            if (valid[i]) {
                sent = connections[i].send(msg);
            }
        }
        if (sent) {
            current = i;
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN,
                        "Composite : message sent on connection #" + current);
            }
        } else {
            LOGGER
                    .log(BasicLevel.ERROR,
                            "Composite : message could not be sent on any underlying connection");
        }
        return sent;
    }

    public AbstractConnectionAttribute getAttribute() {
        return attribute;
    }

    public AbstractConnectionState getConnectionState() {
        return cnxState;
    }

    public void initComponent() throws OdisConnectionException {
        // Check our attribute
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisConnectionException(e);
        }
        // Now create underlying connections
        AbstractConnectionAttribute[] members = attribute.getMembers();
        if (contextual) {
            connections = new IContextualConnectionOut[members.length];
        } else {
            connections = new IConnectionOut[members.length];
        }
        valid = new boolean[members.length];
        boolean none = true;
        for (int i = 0; i < members.length; i++) {
            ConnectionFactory cf = ConnectionFactory.getInstance();
            if (contextual) {
                connections[i] = cf.newContextualConnectionOut(members[i]);
            } else {
                connections[i] = cf.newConnectionOut(members[i]);
            }
            valid[i] = (connections[i] != null);
            if (valid[i]) {
                none = false;
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "Composite : underlying connection valid for priority "
                                    + i + " : " + connections[i]);
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    LOGGER.log(BasicLevel.WARN,
                            "Composite : underlying connection invalid for priority "
                                    + i);
                }
            }
        }
        if (none) {
            LOGGER.log(BasicLevel.ERROR,
                    "Composite : no valid underlying connection");
        }
        // Now try to init underlying connections
        none = true;
        int i;
        for (i = 0; i < connections.length; i++) {
            if (valid[i]) {
                try {
                    connections[i].getConnectionState().setLifeState(connections[i],
                            new StateConnectionInitialized());
                    none = false;
                } catch (OdisConnectionException e) {
                    valid[i] = false;
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER.log(BasicLevel.WARN,
                                "Composite : invalidating connection #" + i
                                        + " could not init because :", e);
                    }
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Composite : connection #" + i
                            + " was invalid - didn't try to init");
                }
            }
        }
        if (none) {
            throw new OdisConnectionException(
                    "Composite init failed : no underlying connection could init");
        }
    }

    public void launchComponent() throws OdisConnectionException {
        boolean none = true;
        int i;
        for (i = 0; i < connections.length; i++) {
            if (valid[i]) {
                try {
                    connections[i].getConnectionState().setLifeState(connections[i],
                            new StateConnectionStarted());
                    none = false;
                } catch (OdisConnectionException e) {
                    valid[i] = false;
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER.log(BasicLevel.WARN,
                                "Composite : invalidating connection #" + i
                                        + " could not launch because :", e);
                    }
                }
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Composite : connection #" + i
                            + " was invalid - didn't try to launch");
                }
            }
        }
        if (none) {
            throw new OdisConnectionException(
                    "Composite launch failed : no underlying connection could launch");
        }
    }

    public void setConnectionState(AbstractConnectionState state) {
        cnxState = state;
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
    }

    public void stopComponent() {
        int i;
        for (i = 0; i < connections.length; i++) {
            // We try to stop even invalid connections, to release everything we
            // can
            try {
                connections[i].getConnectionState().setLifeState(connections[i],
                        new StateConnectionStopped());
            } catch (Exception e) {
                int level = (valid[i]) ? BasicLevel.WARN : BasicLevel.INFO;
                if (LOGGER.isLoggable(level)) {
                    LOGGER.log(level, "Composite : valid connection #" + i
                            + " failed to stop properly because :", e);
                }
            }
        }
    }

    public void unloadComponent() {
        attribute.deleteObserver(this);
        int i;
        for (i = 0; i < connections.length; i++) {
            // We try to unload even invalid connections, to release everything
            // we can
            try {
                connections[i].getConnectionState().setLifeState(connections[i],
                        new StateConnectionUnloaded());
            } catch (Exception e) {
                int level = (valid[i]) ? BasicLevel.WARN : BasicLevel.INFO;
                if (LOGGER.isLoggable(level)) {
                    LOGGER.log(level, "Composite : valid connection #" + i
                            + " failed to unload properly because :", e);
                }
            }
        }
    }

    public void update(Observable o, Object arg) {
    }

    public boolean send(ContextMessage msg) {
        if (!contextual) {
            return false;
        }
        boolean sent = false;
        int i;
        for (i = 0; !sent && (i < connections.length); i++) {
            if (valid[i]) {
                sent = ((IContextualConnectionOut) connections[i]).send(msg);
            }
        }
        if (sent) {
            current = i;
            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN,
                        "Composite : message sent on connection #" + current);
            }
        } else {
            LOGGER
                    .log(BasicLevel.ERROR,
                            "Composite : message could not be sent on any underlying connection");
        }
        return sent;
    }

    public void usedAsTwoWay() {
        // actually, this should not happen as we do not provide a CompositeIn
        LOGGER.log(BasicLevel.FATAL,
                "INTERNAL : CompositeOut.usedAsTwoWay() was called !");
        int i;
        for (i = 0; i < connections.length; i++) {
            if (valid[i]) {
                ((IContextualConnectionOut) connections[i]).usedAsTwoWay();
            }
        }
    }
}
