package org.ow2.odis.connection.context;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.Logger;

/**
 * An ODIS Context identifies a peer.<br>
 * In a non-contextual situation, it usually identifies a source link. In a contextual situation, it identifies a specific peer by identity
 * and sub-identity.<br>
 * A contextual connection should give each session an implicit context (typically IP address and port for a socket-based connection).
 * Business methods (protocol or proceed) may later replace it with an explicit context.<br>
 * Setters are purposefully not provided for identity data as changing them directly would likely change the hash value for the context,
 * which is not wanted, because contexts are typically used to index hash maps.
 */
public class Context {
    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            Context.class.getName());

    /** The identity. */
    private String identity;

    /** The sub-identity. */
    private String subIdentity;

    /** The sub-identity. */
    private String detail = "";

    /** Denotes the implicit nature of the context. */
    private boolean implicit = false;

    /**
     * Gets the identity.
     * @return the identity
     */
    public String getIdentity() {
        return identity;
    }

    /**
     * Gets the sub-identity.
     * @return the sub-identity
     */
    public String getSubIdentity() {
        return subIdentity;
    }

    /**
     * Gets the detail.
     * @return the detail
     */
    public String getDetail() {
        return detail;
    }

    /**
     * Sets the detail.
     * @param detail the detail
     */
    public void setDetail(String detail) {
        this.detail = detail;
    }

    /**
     * Instantiates a new explicit context.
     * @param pIdentity the identity
     * @param pSubIdentity the sub-identity
     */
    public Context(String pIdentity, String pSubIdentity) {
        this.identity = pIdentity;
        this.subIdentity = pSubIdentity;
    }

    /**
     * Instantiates a new explicit context.
     * @param pIdentity the identity
     * @param pSubIdentity the sub-identity
     * @param detail the detail
     */
    public Context(String pIdentity, String pSubIdentity, String detail) {
        this(pIdentity, pSubIdentity);
        this.detail = detail;
    }

    /**
     * Instantiates a new context.
     * @param pIdentity the identity
     * @param pSubIdentity the sub-identity
     * @param detail
     * @param implicit true if the context is implicit (business methods should not create implicit contexts)
     */
    public Context(String pIdentity, String pSubIdentity, String detail,
            boolean implicit) {
        this(pIdentity, pSubIdentity, detail);
        this.implicit = implicit;
    }

    /**
     * Instantiates a new context.
     * @param pIdentity the identity
     * @param pSubIdentity the sub-identity
     * @param implicit true if the context is implicit (business methods should not create implicit contexts)
     */
    public Context(String pIdentity, String pSubIdentity, boolean implicit) {
        this(pIdentity, pSubIdentity);
        this.implicit = implicit;
    }

    /**
     * Provides a String representation of this context.
     * @return a string formed with identity and sub-identity separated by '!'
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer(identity);
        buffer.append("!");
        buffer.append(subIdentity);
        buffer.append("!");
        buffer.append(detail);
        return buffer.toString();
    }

    /**
     * Indicates whether some other object is "equal to" this one.<br>
     * An Object is equal to a Context if : <li>it is a Context <li>it has the same identity and sub-identity
     * @return true if the objects are equal
     */
    public boolean equals(Object obj) {
        if (obj != null) {
            Context copy = null;
            try {
                copy = (Context) obj;
            } catch (ClassCastException e) {
                return false;
            }
            // compare identity
            boolean identityEquals = false;
            if (copy.identity != null) {
                identityEquals = copy.identity.equals(identity);
            } else {
                identityEquals = (identity == null);
            }
            // compare subIdentity
            boolean subIdentityEquals = false;
            if (copy.subIdentity != null) {
                subIdentityEquals = copy.subIdentity.equals(subIdentity);
            } else {
                subIdentityEquals = (subIdentity == null);
            }
            return (identityEquals && subIdentityEquals);
        } else {
            return false;
        }
    }

    /**
     * Returns a hash code value for the object.<br>
     * The hash code for the Context is the hash code for its String representation.
     */
    public int hashCode() {
        return toString().hashCode();
    }

    /**
     * Tells whether the Context is implicit or not.
     * @return true if the Context is implicit.
     */
    public boolean isImplicit() {
        return implicit;
    }
}
