package org.ow2.odis.connection.context;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisContextException;

/**
 * This manager tracks the associations between contexts and sessions.<BR>
 * A context is an ODIS Context, but we make no assumptions on sessions : a session is just an Object.<BR>
 * An ODIS Connection interested in context handling should provide a default value for the context associated with a session before
 * calling PortIn.onMessage() for the first time. That Context should be tagged as implicit unless it allows to consistently identify the
 * same peer after a reconnection. That context may be updated later by external (business) methods. The possibilities are : <li>in an
 * IProtocolCodec (NIO Connection) : the session is available, so use getContextForSession() and updateContextForSession() <li>in an
 * IProceed : use ProceedManager.getCurrentReceiveContext() then updateContext()<br>
 * The ContextManager provides methods for retrieving the session associated with a context (or vice-versa) and changing the associations.
 * It notifies the changes to all registered ContextWatchers.<br>
 * Open issues : should we have a separate map for each Connection/Node ? Do we allow a client to use distinct connections in the same
 * Domain with the same Context ? Or should we have one ContextManager per Connection ?
 * @author ojoh6423
 */
public class ContextManager {
    /** Logger for this class. */
    final static Logger LOGGER = Monolog.initialize().getLogger(
            ContextManager.class.getName());

    /**
     * The map of sessions indexed by contexts.
     */
    private Map mapSessionByContext = new HashMap();

    /**
     * Reverse map of contexts indexed by sessions.<br>
     * Open issue : does the cost of searching for a session through the context indexed map justify the reverse map ?
     */
    private Map mapContextBySession = new HashMap();

    /** All context watchers for this manager. */
    private List contextWatchers = new ArrayList();

    /** Singleton pattern. */
    private static ContextManager instance = null;

    /**
     * Gets the unique instance of ContextManager.<br>
     * Applies singleton pattern.
     * @return the ContextManager instance
     */
    public static synchronized ContextManager getInstance() {
        if (instance == null) {
            instance = new ContextManager();
        }
        return instance;
    }

    /**
     * Default constructor.
     */
    private ContextManager() {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Creating ContextManager");
        }
    }

    /**
     * Adds a context/session association.
     * @param context the context
     * @param session the session
     * @throws OdisContextException if the Context is already associated with a different session or the session with a different Context.
     */
    public void addSessionContext(Context context, Object session)
            throws OdisContextException {
        addSessionContext(context, session, false);
    }

    /**
     * Adds a context/session association.<br>
     * In override mode, if the context already refers to a session, that association is dropped. If the session already refers to another
     * context, that association is dropped and the watchers are notified that the orphaned context is lost.
     * @param context the context
     * @param session the session
     * @param override true if existing associations should be replaced
     * @throws OdisContextException if override is false and the Context is already associated with a different session or the session with
     * a different Context.
     */
    public synchronized void addSessionContext(Context context, Object session,
            boolean override) throws OdisContextException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Adding session " + session
                    + " for context " + context);
        }
        Object oldSession = mapSessionByContext.get(context);
        if (oldSession != null) {
            if (oldSession.equals(session)) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    StringBuffer buf = new StringBuffer("Context ");
                    buf.append(context).append(
                            " is already mapped with same session ").append(
                            session);
                    LOGGER.log(BasicLevel.INFO, buf);
                }
                return;
            } else {
                StringBuffer buf = new StringBuffer("Context ");
                buf.append(context).append(
                        " is already mapped with another session ").append(
                        session);
                if (override) {
                    LOGGER.log(BasicLevel.WARN, buf.toString());
                    mapContextBySession.remove(oldSession);
                } else {
                    throw new OdisContextException(buf.toString());
                }
            }
        }
        Context oldContext = (Context) mapContextBySession.get(session);
        if (oldContext != null) {
            StringBuffer buf = new StringBuffer("Session ");
            buf.append(session).append(
                    " is already mapped with another context ").append(
                    oldContext);
            if (override) {
                LOGGER.log(BasicLevel.WARN, buf.toString());
                notifyContextChange(oldContext, true);
                mapSessionByContext.remove(oldContext);
            } else {
                throw new OdisContextException(buf.toString());
            }
        }
        associate(context, session);
        if (oldSession == null) {
            notifyContextChange(context, false);
        }
    }

    /**
     * Updates an existing context.<br>
     * This should be done only to replace the default context set by Odis by a business-significant context.
     * @param oldContext the existing context
     * @param newContext the new context
     * @throws OdisContextException if the old context is not mapped yet or the new context is.
     */
    public synchronized void updateContext(Context oldContext,
            Context newContext) throws OdisContextException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Changing context " + oldContext
                    + " into " + newContext);
        }
        if (mapSessionByContext.containsKey(oldContext)) {
            if (oldContext.equals(newContext)) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Request to update context "
                            + oldContext
                            + " with an equal context : nothing done.");
                }
            } else {
                if (mapSessionByContext.containsKey(newContext)) {
                    StringBuffer buf = new StringBuffer("Context ");
                    buf.append(newContext).append(
                            " is already mapped with session ").append(
                            mapSessionByContext.get(newContext)).append(
                            ". Cannot use it to update context ").append(
                            oldContext);
                    throw new OdisContextException(buf.toString());
                } else {
                    associate(newContext, mapSessionByContext.get(oldContext));
                    mapSessionByContext.remove(oldContext);
                    notifyContextChange(oldContext, newContext);
                }
            }
        } else {
            StringBuffer buf = new StringBuffer("Context ");
            buf.append(oldContext)
                    .append(" is not mapped yet : cannot update.");
            throw new OdisContextException(buf.toString());
        }
    }

    /**
     * Updates the context for a session.<br>
     * This should be done only to replace the default context set by Odis by a business-significant context.
     * @param context the new context
     * @param session the session
     * @throws OdisContextException if the session is not mapped yet or the context is.
     */
    public synchronized void updateContextForSession(Context context,
            Object session) throws OdisContextException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Updating context (" + context
                    + ") for session " + session);
        }
        Context oldContext = (Context) mapContextBySession.get(session);
        if (oldContext != null) {
            if (oldContext.equals(context)) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Session " + session
                            + " already mapped with a context equal to "
                            + context + " : nothing done.");
                }
            } else {
                mapSessionByContext.remove(oldContext);
                associate(context, session);
                notifyContextChange(oldContext, context);
            }
        } else {
            StringBuffer buf = new StringBuffer("Session ");
            buf.append(session).append(" is not mapped yet.");
            throw new OdisContextException(buf.toString());
        }
    }

    /**
     * Actually updates the maps.<br>
     * For internal use by other methods when all checking is done.
     * @param context the context
     * @param session the session
     */
    private void associate(Context context, Object session) {
        mapSessionByContext.put(context, session);
        mapContextBySession.put(session, context);
    }

    /**
     * Getter.
     * @param session a session
     * @return the associated context or null if session is unknown
     */
    public Context getContextForSession(Object session) {
        return (Context) mapContextBySession.get(session);
    }

    /**
     * Getter.
     * @param context a context
     * @return the associated session or null if session is unknown
     */
    public Object getSessionForContext(Context context) {
        return mapSessionByContext.get(context);
    }

    /**
     * Getter.
     * @param session a session
     * @return true if the session is currently mapped with a context
     */
    public boolean isSessionMapped(Object session) {
        return mapContextBySession.containsKey(session);
    }

    /**
     * Getter.
     * @param context a context
     * @return true if the context is currently mapped with a session
     */
    public boolean isContextMapped(Context context) {
        return mapSessionByContext.containsKey(context);
    }

    /**
     * Removes a context/session association from the maps.<br>
     * Notifies the watchers of the context loss.
     * @param context the context identifying the association.
     * @throws OdisContextException if the context is unknown
     */
    public void removeSessionForContext(Context context)
            throws OdisContextException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Removing session for context "
                    + context);
        }
        if (mapSessionByContext.containsKey(context)) {
            notifyContextChange(context, true);
            mapContextBySession.remove(mapSessionByContext.get(context));
            mapSessionByContext.remove(context);
        } else {
            throw new OdisContextException("Context " + context
                    + " is unknown. Can't remove associated session.");
        }
    }

    /**
     * Removes a context/session association from the map.<br>
     * Notifies the watchers of the context loss.
     * @param session the session identifying the association.
     * @throws OdisContextException if the session is unknown
     */
    public void removeSession(Object session) throws OdisContextException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Removing session " + session);
        }
        if (mapContextBySession.containsKey(session)) {
            notifyContextChange((Context) mapContextBySession.get(session),
                    true);
            mapSessionByContext.remove(mapContextBySession.get(session));
            mapContextBySession.remove(session);
        } else {
            throw new OdisContextException("Session " + session
                    + " is unknown. Can't remove.");
        }
    }

    /**
     * Adds a ContextWatcher to the list.
     * @param cw the new ContextWatcher
     */
    public void addContextWatcher(IContextWatcher cw) {
        contextWatchers.add(cw);
    }

    /**
     * Notifies a context change to all registered watchers.
     * @param ctx the involved Context
     * @param lost true if the Context has disappeared, false if it was created
     */
    public void notifyContextChange(Context ctx, boolean lost) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Notifying change : " + ctx
                    + (lost ? " lost" : " set"));
        }
        for (Iterator it = contextWatchers.iterator(); it.hasNext();) {
            if (lost) {
                ((IContextWatcher) it.next()).contextLost(ctx);
            } else {
                ((IContextWatcher) it.next()).contextSet(ctx);
            }
        }
    }

    /**
     * Notifies a context change (used when an implicit Context is replaced by an Explicit one).
     * @param octx the previous Context
     * @param nctx the new Context
     */
    public void notifyContextChange(Context octx, Context nctx) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Notifying change : " + octx
                    + " becomes " + nctx);
        }
        for (Iterator it = contextWatchers.iterator(); it.hasNext();) {
            ((IContextWatcher) it.next()).contextReplaced(octx, nctx);
        }
    }
}
