package org.ow2.odis.connection.context;
/**
 * This interface specifies a Context directory API.<br>
 * A context directory is use to get the proper technical context related to a given business contexts for a given communication medium. In
 * other words, it may be queried for IP address, mail address, SMS number, etc. based on a business ID.<br>
 * The context directory may either refer to a fully external directory or be a memory copy, in which case refresh policy is not specified.
 * Optionally, the context directory may support update at runtime.
 * @author ojoh6423
 */
public interface IContextDirectory {
    /**
     * Initializes the directory based on repository location data.<br>
     * Initialization may connect or prepare to connect to the external repository, or load all data, as appropriate.
     * @param repositoryLocation an implementation specific array of String parameters allowing to locate the repository
     * @return true if successful
     */
    boolean init(String[] repositoryLocation);

    /**
     * Refreshes data using the repository, if appropriate.<br>
     * This may mean reloading all data, updating a cache or doing nothing, according to the implementation.<br>
     * If a lengthy execution is likely, processing should be done in a separate thread.<br>
     * Implementations should allow queries to the ContextDirectory during the refreshing process.
     * @return true if successful (or if there's nothing to do)
     */
    boolean refresh();

    /**
     * Provides a Context that may be used by a link of type linkType to address the recipient identified by businessContext.
     * @param businessContext business ID of the recipient
     * @param linkType ID of the communication medium
     * @return technical Context holding adequate routing data for addressing that recipient on that type of link, or null if none found
     */
    Context getAddress(Context businessContext, String linkType);

    /**
     * Adds (or replaces) the technical Context associated with specified recipient on specified link type.<br>
     * An read-only implementation will always return false.
     * @param businessContext business ID of the recipient
     * @param linkType ID of the communication medium
     * @param technicalContext technical Context holding adequate routing data for addressing that recipient on that type of link
     * @return true if update was successful, false if it failed or was not allowed
     */
    boolean addAddress(Context businessContext, String linkType,
            Context technicalContext);
}
