package org.ow2.odis.connection.memory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.model.MemoryConnectionAttribute;

/**
 * Manage a list of object into memory.
 * <p>
 * the main idea of this connection is to provide an easy solution to link two nodes using a memory buffer. To avoid a crash due to a
 * memory overload two limits are set. The default bytes array size is limited to 100000 witch can be dispached on 1000 objects When the
 * limit is reached the send function is blocked until the buffer respect at least the half of these limit. <br>
 * <img src="doc-files/hysteresis.png"/> <br>
 * Of course user can change this limit but at his own responsibility. The limit can be inactivated if their values are negative.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class Memory {
    /** Information which define the memory parameters. */
    public final MemoryConnectionAttribute attribute;

    /** The Constant READY. */
    static protected final int READY = 0;

    /** The Constant NOT_READY. */
    static protected final int NOT_READY = 1;

    /** The Constant EOL. */
    static final String EOL = System.getProperty("line.separator");

    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            Memory.class.getName());

    /** list of the message of the memory. */
    transient List lsMsg = Collections.synchronizedList(new ArrayList(1000));

    /** reference on a port MemoryIn. */
    protected MemoryIn memoryIn = null;

    /** reference on a port MemoryOut. */
    protected MemoryOut memoryOut = null;

    /**
     * Constructor.
     * @param attribute definition of this Memory
     */
    protected Memory(MemoryConnectionAttribute attribute) {
        super();
        this.attribute = attribute;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create " + Memory.class.getName());
        }
    }

    /**
     * setter.
     * @param memoryIn associated memoryIn
     */
    public void setMemoryIn(MemoryIn memoryIn) {
        this.memoryIn = memoryIn;
        if (!lsMsg.isEmpty()) {
            activeMemoryIn();
        }
        check();
    }

    /**
     * setter.
     * @param memoryOut associated memoryOut
     */
    public void setMemoryOut(MemoryOut memoryOut) {
        this.memoryOut = memoryOut;
        check();
    }

    /**
     * Check if there is still a MemoryIn or MemoryOut Connected. <br>
     * if not the memory will be deleted and all message list will be cleared
     */
    public void check() {
        if (memoryOut == null && memoryIn == null) {
            MemoryFactory.getInstance().removeMemory(this);
            lsMsg.clear();
        }
    }

    /**
     * active the MemoryIn that the list of message is not empty.
     */
    protected void activeMemoryIn() {
        if (memoryIn != null) {
            memoryIn.onMessage();
        }
    }

    /**
     * getter.
     * @return attribute
     */
    public MemoryConnectionAttribute getAttribute() {
        return attribute;
    }

    /**
     * check if the level of byte and nbr object less than half the max authorized.
     * @return boolean value
     */
    protected boolean checkStateAgain() {
        boolean result = true;
        // teste la taille en octet
        if (attribute.getSizeByteMax() > 0) {
            result = ((attribute.getSizeByte() * 2) < attribute.getSizeByteMax());
        }
        // teste la taille en nombre d'objet
        if (attribute.getSizeObjectMax() > 0) {
            result = result
                    && (lsMsg.size() * 2 < (attribute.getSizeObjectMax()));
        }
        if (result) {
            attribute.setState(Memory.READY);
        }
        return result;
    }

    /**
     * assure l'histeresis de la memory.
     * @param length the length
     * @return true if ok
     */
    protected boolean checkState(final int length) {
        boolean result = true;
        final int nbMsgPresent = attribute.getNumberMessage();
        if (attribute.getSizeObjectMax() > 0
                && attribute.getSizeObjectMax() < nbMsgPresent + 1) {
            if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                final StringBuffer errMsg = new StringBuffer("MEMORY [");
                errMsg.append(getAttribute().getName());
                errMsg.append("] reject sendingMessage ");
                errMsg.append(" due to too many message present :");
                errMsg.append(nbMsgPresent);
                errMsg.append(EOL);
                errMsg.append("\t->the max object size has been set to  ");
                errMsg.append(attribute.getSizeObjectMax());
                errMsg.append(EOL);
                LOGGER.log(BasicLevel.ERROR, errMsg.toString());
            }
            attribute.setState(NOT_READY);
            result = false;
        } else {
            if (attribute.getSizeByteMax() > 0
                    && (attribute.getSizeByteMax()) < length
                            + attribute.getSizeByte()) {
                if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                    final StringBuffer errMsg = new StringBuffer("MEMORY [");
                    errMsg.append(getAttribute().getName());
                    errMsg
                            .append("] reject sendingMessage due to too many data present :");
                    errMsg.append(EOL);
                    errMsg.append("\t->max data authorized      :");
                    errMsg.append(attribute.getSizeByteMax());
                    errMsg.append(EOL);
                    errMsg.append("\t->try to add               :");
                    errMsg.append(length);
                    errMsg.append(" bytes");
                    errMsg.append(EOL);
                    errMsg.append("\t->memory was already contain :");
                    errMsg.append(attribute.getSizeByte());
                    errMsg.append(" (on ");
                    errMsg.append(nbMsgPresent);
                    errMsg.append(" message(s))");
                    errMsg.append(EOL);
                    LOGGER.log(BasicLevel.ERROR, errMsg.toString());
                }
                attribute.setState(NOT_READY);
                result = false;
            }
        }
        return result;
    }
}
