package org.ow2.odis.connection.memory;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.ow2.odis.connection.AbstractConnectionfactory;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.IConnectionOut;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.model.AbstractAttribute;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.MemoryConnectionAttribute;

/**
 * implement Factory (for memory) and singleton design pattern . It control that a memory with the same attribute value is created only once
 * time. 
 * @author Beyler Olivier FranceTelecom R&D
 */
public class MemoryFactory extends AbstractConnectionfactory {
    /**
     * singleton pattern.
     */
    private static MemoryFactory instance = new MemoryFactory();

    /**
     * map between attribute and memory
     */
    private final transient Map mapMemory;

    /**
     * map between attribute and memoryIn
     */
    private final transient Map mapMemoryIn;

    /**
     * map between attribute and memoryOut
     */
    private final transient Map mapMemoryOut;

    /**
     * constructor.
     */
    private MemoryFactory() {
        super();
        mapMemory = new HashMap();
        mapMemoryIn = Collections.synchronizedMap(new HashMap());
        mapMemoryOut = Collections.synchronizedMap(new HashMap());
    }

    /**
     * getter. (implement Singleton pattern)
     * @return unique instance of MemoryFactory
     */
    public static MemoryFactory getInstance() {
        return instance;
    }

    /**
     * return a unique memory for a attribute. If the memory doesn't exist it will create it.
     * @param attribute
     * @return Memory
     */
    public Memory getMemory(final MemoryConnectionAttribute attribute) {
        Memory result = (Memory) mapMemory.get(attribute);
        if (result == null) {
            synchronized (this) {
                result = (Memory) mapMemory.get(attribute);
                if (result == null) {
                    result = new Memory(attribute);
                    mapMemory.put(attribute, result);
                }
            }
        }
        return result;
    }

    /**
     * return a unique memory in for an attribute. If the memory doesn't exist it will create it.
     * @param attribut
     * @return memory
     */
    public IConnectionIn newConnectionIn(final AbstractConnectionAttribute attribut) {
        MemoryIn result = (MemoryIn) mapMemoryIn.get(attribut);
        if (result == null) {
            synchronized (this) {
                result = (MemoryIn) mapMemoryIn.get(attribut);
                if (result == null) {
                    result = new MemoryIn((MemoryConnectionAttribute) attribut);
                    mapMemoryIn.put(attribut, result);
                }
            }
        }
        return result;
    }

    /**
     * return a unique memory in for a attribute. If the memory doesn't exist it will create it.
     * @param attribute the attribute
     * @return memory
     */
    public IConnectionOut newConnectionOut(final AbstractConnectionAttribute attribute) {
        MemoryOut result = (MemoryOut) mapMemoryOut.get(attribute);
        if (result == null) {
            synchronized (this) {
                result = (MemoryOut) mapMemoryOut.get(attribute);
                if (result == null) {
                    result = new MemoryOut((MemoryConnectionAttribute) attribute);
                    mapMemoryOut.put(attribute, result);
                }
            }
        }
        return result;
    }

    /**
     * return a unique memory in for a attribute. If the memory doesn't exist it will create it.
     * @param attribute the attribute
     * @return memory
     */
    public IContextualConnectionOut newContextualConnectionOut(
            AbstractConnectionAttribute attribute) {
        return (IContextualConnectionOut) newConnectionOut(attribute);
    }

    /**
     * remove MemoryIn reference from the map of the factory.
     * @param memory the memory port witch need to be unloaded
     */
    public void removeMemoryIn(MemoryIn memory) {
        mapMemoryIn.remove(memory.getAttribute());
    }

    /**
     * remove MemoryOut reference from the map of the factory.
     * @param memory the Memory port witch need to be unloaded
     */
    public void removeMemoryOut(MemoryOut memory) {
        mapMemoryOut.remove(memory.getAttribute());
    }

    /**
     * remove Memory reference from the map of the factory.
     * @param memory the Memory port witch need to be unloaded
     */
    public void removeMemory(Memory memory) {
        mapMemory.remove(memory.getAttribute());
    }

    public AbstractConnectionAttribute newAttribute(final AbstractAttribute attribute) {
        return new MemoryConnectionAttribute(attribute);
    }

    /**
     * @return
     */
    protected String getTypeName() {
        return ("Memory");
    }
}
