package org.ow2.odis.connection.memory;

import java.util.Observable;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionLoaded;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.MemoryConnectionAttribute;
import org.ow2.odis.port.PortIn;

/**
 * Manage a list of object into memory.
 * <p>
 * the main idea of this connection is to provide an easy solution to link two nodes using a memory buffer. To avoid a crash due to a
 * memory overload two limits are set. The default bytes array size is limited to 100000 witch can be dispach on 1000 objects When the
 * limit is reached the send function is blocked until the buffer respect at least the half of these limit. <br>
 * <img src="doc-files/hysteresis.png"/> <br>
 * Of course user can change this limit but at his own responsability. The limit can be disactivated if their values are negative.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class MemoryIn implements IConnectionIn {
    /** State of this connection. By default it is Loaded */
    private AbstractConnectionState cnxState = new StateConnectionLoaded();

    /** reference on Port In. It will be notified that a new Message comes. */
    transient PortIn portIn = null;

    /**
     * as each system can use different line separator, the EOL is dependent of the system.
     */
    static final String EOL = System.getProperty("line.separator");

    /** attribute witch give the information of memory configuration. */
    public final MemoryConnectionAttribute attribute;

    /** The Constant READY. */
    protected static final int READY = 0;

    /** The Constant NOT_READY. */
    protected static final int NOT_READY = 1;

    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            MemoryIn.class.getName());

    /** reference memory instance witch contain objects for the memoryIn. */
    private final Memory memory;

    /** if true activation have been request while the portIn wasn't set. */
    boolean isMessagePresent = false;

    /**
     * Constructor.
     * @param pAttribut the attribute
     */
    protected MemoryIn(MemoryConnectionAttribute pAttribut) {
        super();
        this.attribute = pAttribut;
        memory = MemoryFactory.getInstance().getMemory(pAttribut);
        memory.setMemoryIn(this);
        pAttribut.addObserver(this);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create " + MemoryIn.class.getName());
        }
    }

    /**
     * @return
     */
    public int getPendingMessage() {
        return memory.lsMsg.size();
    }

    /**
     * Ackwnolege the message.
     * @param valid the valid
     */
    public void ackMessage(final boolean valid) {
        if (valid) {
            final Byte[] bytes = (Byte[]) memory.lsMsg.remove(0);
            attribute.incSizeByte(-bytes.length);
            attribute.incNumberMessage(-1);
        }
        // verifie si on est en dessous des criteres de relance
        memory.checkStateAgain();
        if (!memory.lsMsg.isEmpty()) {
            onMessage();
        }
    }

    /**
     * Gets the context message.
     * @return byteMessage
     */
    public ContextMessage getContextMessage() {
        byte[] msg = null;
        if (!memory.lsMsg.isEmpty()) {
            final Byte[] bytes = (Byte[]) memory.lsMsg.get(0);
            msg = new byte[bytes.length];
            for (int i = 0; i < bytes.length; i++) {
                msg[i] = bytes[i].byteValue();
            }
        }
        return new ContextMessage(msg, new Context(attribute.getName(), ""));
    }

    /**
     * Gets the attribute.
     * @return CnxAttribut
     */
    public AbstractConnectionAttribute getAttribute() {
        return attribute;
    }

    /**
     * check if the port In is already set if not set it.
     * @param port the port
     * @return the portIn associate to this memory
     */
    public IConnectionIn setPortIn(final PortIn port) {
        if (this.portIn != null) {
            LOGGER.log(BasicLevel.ERROR,
                    "MEMORY is not able to have more than one consumer : old port :"
                            + this.portIn.getId()
                            + " is unsubcribe  the new consumer is port :"
                            + port.getId());
        }
        this.portIn = port;
        if (isMessagePresent) {
            onMessage();
        }
        return this;
    }

    /**
     * Sets the cnx state.
     * @param state the state
     */
    public void setConnectionState(AbstractConnectionState state) {
        this.cnxState = state;
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * Gets the cnx state.
     * @return the cnx state
     */
    public AbstractConnectionState getConnectionState() {
        return cnxState;
    }

    /**
     * Update.
     * @param object the object
     * @param arg the arg
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(final Observable object, final Object arg) {
    }

    /**
     * Inits the component.
     * @throws OdisConnectionException the odis cnx exception
     */
    public void initComponent() throws OdisConnectionException {
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisConnectionException(e);
        }
    }

    /**
     * Launch component.
     */
    public void launchComponent() {
    }

    /**
     * Stop component.
     */
    public void stopComponent() {
    }

    /**
     * On message.
     */
    public void onMessage() {
        if (portIn != null) {
            portIn.onMessage();
            isMessagePresent = false;
        } else {
            isMessagePresent = true;
        }
    }

    /**
     */
    public void unloadComponent() {
        attribute.deleteObserver(this);
        memory.setMemoryIn(null);
        MemoryFactory.getInstance().removeMemoryIn(this);
    }
}
