package org.ow2.odis.connection.memory;

import java.util.Observable;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.IConnectionOut;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionLoaded;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.MemoryConnectionAttribute;

/**
 * Manage a list of object into memory.
 * <p>
 * the main idea of this connection is to provide an easy solution to link two nodes using a memory buffer. To avoid a crash due to a
 * memory overload two limits are set. The default bytes array size is limited to 100000 witch can be dispach on 1000 objects When the
 * limit is reached the send function is blocked until the buffer respect at least the half of these limit. <br>
 * <img src="doc-files/hysteresis.png"/> <br>
 * Of course user can change this limit but at his own responsibility. The limit can be disactivated if their values are negative.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class MemoryOut implements IConnectionOut, IContextualConnectionOut {
    /** State of this connection. By default it is Loaded */
    private AbstractConnectionState cnxState = new StateConnectionLoaded();

    /** attribute witch give the information of memory configuration. */
    public final MemoryConnectionAttribute attribute;

    /** The Constant READY. */
    protected static final int READY = 0;

    /** The Constant NOT_READY. */
    protected static final int NOT_READY = 1;

    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            MemoryOut.class.getName());

    /** reference memory instance witch contain objects for the memoryOut. */
    private final Memory memory;

    /**
     * Instantiates a new memory out.
     * @param attribute the attribute
     */
    protected MemoryOut(MemoryConnectionAttribute attribute) {
        super();
        this.attribute = attribute;
        memory = MemoryFactory.getInstance().getMemory(attribute);
        memory.setMemoryOut(this);
        ;
        attribute.addObserver(this);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create " + MemoryOut.class.getName());
        }
    }

    /**
     * Gets the attribute.
     * @return CnxAttribut
     */
    public AbstractConnectionAttribute getAttribute() {
        return attribute;
    }

    /**
     * Store the byte array into the memory. Update attribute value :
     * <ul>
     * <li>numberMessage <li>sizeByte
     * </ul>
     * <br>
     * and notify the port that a new message has arrived
     * @param pMsg bytes array witch have to be sent
     * @return true if succeed false in other case
     */
    public boolean send(final byte[] pMsg) {
        boolean result = true;
        if (attribute.getState() == Memory.NOT_READY) {
            if (!memory.checkStateAgain()) {
                return false;
            }
        }
        if (attribute.getState() == Memory.READY) {
            if (!memory.checkState(pMsg.length)) {
                return false;
            }
        }
        attribute.incSizeByte(pMsg.length);
        Byte[] bytes = new Byte[pMsg.length];
        for (int i = 0; i < pMsg.length; i++) {
            bytes[i] = new Byte(pMsg[i]);
        }
        memory.lsMsg.add(bytes);
        attribute.incNumberMessage(1);
        memory.activeMemoryIn();
        return result;
    }

    /**
     * change the state of the memoryOut.
     * @param state AbstractCnxState
     */
    public void setConnectionState(AbstractConnectionState state) {
        this.cnxState = state;
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * getter.
     * @return AbstractCnxState
     */
    public AbstractConnectionState getConnectionState() {
        return cnxState;
    }

    /**
     * Update.
     * @param object the object
     * @param arg the arg
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(final Observable object, final Object arg) {
        memory.checkState(0);
    }

    /**
     * Inits the component.
     * @throws OdisConnectionException the odis connection exception

     */
    public void initComponent() throws OdisConnectionException {
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisConnectionException(e);
        }
    }

    /**
     * Launch component.
     */
    public void launchComponent() {
    }

    /**
     * Stop component.
     */
    public void stopComponent() {
    }

    /**
     * Unload component.
     */
    public void unloadComponent() {
        attribute.deleteObserver(this);
        memory.setMemoryOut(null);
        MemoryFactory.getInstance().removeMemoryOut(this);
    }

    public boolean send(ContextMessage pMsg) {
        return send(pMsg.getPayload());
    }

    public void usedAsTwoWay() {
    }
}
