package org.ow2.odis.connection.socket;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Observable;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionLoaded;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.Const;
import org.ow2.odis.model.SocketConnectionAttribute;
import org.ow2.odis.port.PortIn;

/**
 * base class of TCPClientConnectionIn and TCPServerConnectionIn.
 * @author Beyler Olivier FranceTelecom R&D
 */
public abstract class AbstractTCPConnectionIn implements IConnectionIn {
    /** The list of context messages. */
    private List lsContextMessages = Collections
            .synchronizedList(new ArrayList(5000));

    /** The connection socket. */
    protected AbstractTCPConnectionIn connectionSocket = null;

    /** define where the connection get the message. */
    protected PortIn portIn = null;

    /** attribute with contain all properties of the connection. */
    public final SocketConnectionAttribute attribut;

    /** connection state. */
    private AbstractConnectionState cnxState = new StateConnectionLoaded();

    /**
     * Constructor.
     * @param attribute the attribute
     */
    AbstractTCPConnectionIn(SocketConnectionAttribute attribute) {
        super();
        this.attribut = attribute;
        attribute.addObserver(this);
        connectionSocket = this;
    }

    /** logger of this class. */
    static final Logger loggerCnxIn = Monolog.initialize().getLogger(
            AbstractTCPConnectionIn.class.getName());

    /**
     * getter.
     * @return IProtocolIn
     */
    IProtocolIn getProtocolIn() {
        IProtocolIn result = null;
        try {
            result = attribut.createProtocolInInstance();
        } catch (OdisAttributeException e) {
            loggerCnxIn
                    .log(
                            BasicLevel.FATAL,
                            "protocol should have been ejected during resolve component!",
                            e);
        }
        return result;
    }

    /**
     * Wait until be able to add new message and then add it.
     * @param bytes array
     * @param context the context
     * @throws InterruptedException on a user request an Interrupted Exception can occurred to break the pause
     */
    public void addMessage(byte[] bytes, Context context)
            throws InterruptedException {
        if (!checkState(bytes.length)) {
            pause();
        }
        attribut.setState(Const.READY);
        attribut.incNumberMessage(1);
        attribut.incSizeByte(bytes.length);
        final ContextMessage contextMessage = new ContextMessage(bytes, context);
        lsContextMessages.add(contextMessage);
        if (portIn != null) {
            portIn.onMessage();
        }
    }

    /**
     * use by send to freeze the thread send while the.
     * @throws InterruptedException the interrupted exception
     */
    private void pause() throws InterruptedException {
        if (attribut.getState() == Const.NOT_READY) {
            synchronized (this) {
                if (!checkStateAgain()) {
                    if (loggerCnxIn.isLoggable(BasicLevel.INFO)) {
                        loggerCnxIn.log(BasicLevel.INFO,
                                "Buffer full : entering wait() for " + this);
                    }
                    wait();
                    if (loggerCnxIn.isLoggable(BasicLevel.INFO)) {
                        loggerCnxIn.log(BasicLevel.INFO,
                                "Buffer available again : wait() exited for "
                                        + this);
                    }
                }
            }
        }
    }

    /**
     * check if all condition are respected if a new message with length byte is added.
     * @param length int size of message added
     * @return boolean
     */
    private boolean checkState(int length) {
        int nbMsgPresent = lsContextMessages.size();
        if (attribut.getSizeObjectMax() > 0
                && attribut.getSizeObjectMax() < nbMsgPresent) {
            if (loggerCnxIn.isLoggable(BasicLevel.WARN)) {
                StringBuffer errMsg = new StringBuffer("SOCKET [");
                errMsg.append(attribut.getName());
                errMsg
                        .append("] refusing to addMessage() : too many messages present (");
                errMsg.append(nbMsgPresent);
                errMsg
                        .append(")\n\twhile the socket buffer size has been set to : ");
                errMsg.append(attribut.getSizeObjectMax());
                loggerCnxIn.log(BasicLevel.WARN, errMsg.toString());
            }
            attribut.setState(Const.NOT_READY);
            return false;
        }
        if (attribut.getSizeByteMax() > 0
                && (attribut.getSizeByteMax()) < length
                        + attribut.getSizeByte()) {
            if (loggerCnxIn.isLoggable(BasicLevel.WARN)) {
                StringBuffer errMsg = new StringBuffer("SOCKET [");
                errMsg.append(attribut.getName());
                errMsg
                        .append("] refusing to addMessage() : socket buffer full");
                errMsg.append("\n\t->capacity (bytes) : ");
                errMsg.append(attribut.getSizeByteMax());
                errMsg.append("\n\t->current size     : ");
                errMsg.append(attribut.getSizeByte());
                errMsg.append(" (on ");
                errMsg.append(nbMsgPresent);
                errMsg.append(" message(s))");
                errMsg.append("\n\t->trying to add    : ");
                errMsg.append(length);
                errMsg.append(" bytes");
                loggerCnxIn.log(BasicLevel.WARN, errMsg.toString());
            }
            attribut.setState(Const.NOT_READY);
            return false;
        }
        return true;
    }

    /**
     * Check state again.
     * @return true if only half buffer is at least free
     */
    private boolean checkStateAgain() {
        boolean result;
        result = (attribut.getSizeByteMax() < attribut.getSizeByte() * 2);
        result = result
                && (attribut.getSizeObjectMax()) < lsContextMessages.size() * 2;
        return result;
    }

    /**
     * Gets the pending message.
     * @return the pending message
     */
    public int getPendingMessage() {
        return lsContextMessages.size();
    }

    /**
     * Ack message.
     * @param valid the valid
     */
    public void ackMessage(boolean valid) {
        if (valid) {
            final ContextMessage contextMessage = (ContextMessage) lsContextMessages
                    .remove(0);
            attribut.incSizeByte(-contextMessage.getPayload().length);
            attribut.incNumberMessage(-1);
        }
        synchronized (this) {
            if (loggerCnxIn.isLoggable(BasicLevel.INFO)) {
                loggerCnxIn.log(BasicLevel.INFO,
                        "Buffer available again : notifying " + this);
            }
            notifyAll();
        }
    }

    /**
     * Gets the context message.
     * @return the context message
     */
    public ContextMessage getContextMessage() {
        ContextMessage result = null;
        if (lsContextMessages.isEmpty() != true) {
            result = (ContextMessage) lsContextMessages.get(0);
        }
        return result;
    }

    /**
     * Sets the port in.
     * @param portIn the port in
     * @return the i connection in
     */
    public IConnectionIn setPortIn(PortIn portIn) {
        this.portIn = portIn;
        return this;
    }

    /**
     * Inits the component.
     * @throws OdisConnectionException the odis cnx exception
     */
    public void initComponent() throws OdisConnectionException {
        try {
            attribut.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisConnectionException(e);
        }
    }

    /**
     * Unload component.
     */
    public void unloadComponent() {
        attribut.deleteObserver(this);
    }

    /**
     * Sets the cnx state.
     * @param state the state
     */
    public void setConnectionState(AbstractConnectionState state) {
        this.cnxState = state;
        if (loggerCnxIn.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            loggerCnxIn.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * Gets the cnx state.
     * @return the cnx state
     */
    public AbstractConnectionState getConnectionState() {
        return cnxState;
    }

    /**
     * Update.
     * @param o the o
     * @param arg the arg
     * @see java.util.Observer#update(java.util.Observable, java.lang.Object)
     */
    public void update(Observable o, Object arg) {
    }

    /**
     * Gets the attribut.
     * @return the attribut
     */
    public AbstractConnectionAttribute getAttribute() {
        return attribut;
    }
}
