package org.ow2.odis.connection.socket;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The aim of this class is to send message on a socket by using a protocoleOut. <br>
 * The send is intependent of the fact that the socket is created by the node (the link is viewed as a client) or by an other node (the
 * link is viewed as a server)
 */
public class MessageSender {
    /**
     * will be used for writing a message on this socket.
     */
    private Socket socket = null;

    /**
     * the protocol instance witch will be used for writing a message on this socket.
     */
    private IProtocolOut protocol = null;

    /**
     * logger of the class
     */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            MessageSender.class.getName());

    /**
     * Until the sending message as no problem is set to true and in case of error the value is false.
     */
    public boolean stopSending = false;

    /**
     * Constructor.
     * @param pSocket socket used to send message
     */
    MessageSender(Socket pSocket) {
        super();
        socket = pSocket;
    }

    /**
     * send the msg by using write protocol function.
     * @param msg byte array
     * @return true if succeed false in other case
     */
    public boolean send(byte[] msg) {
        if (socket == null || protocol == null) {
            return false;
        }
        try {
            try {
                InputStream streamIn = socket.getInputStream();
                OutputStream streamOut = socket.getOutputStream();
                if (protocol.isStreamStillOpen()) {
                    if (protocol.write(streamIn, streamOut, msg)) {
                        return true;
                    }
                } else {
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER
                                .log(BasicLevel.WARN,
                                        "Stream is not still open ");
                    }
                }
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR, "IOException on read protocol", e);
            }
            try {
                socket.close();
                socket = null;
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR, "IOException during closing connection",
                        e);
                socket = null;
            }
        } catch (Exception e) {
            LOGGER.log(BasicLevel.ERROR, "Listen exception :", e);
        }
        LOGGER.log(BasicLevel.ERROR, "END LISTENING ");
        stopSending = true;
        return false;
    }

    /**
     * getter.
     * @return the protocol class witch will be used for writing a message on this socket.
     */
    public IProtocolOut getProtocol() {
        return protocol;
    }

    /**
     * setter.
     * @param protocol class witch will be used for writing a message on this socket.
     */
    public void setProtocol(IProtocolOut protocol) {
        this.protocol = protocol;
    }
}
