package org.ow2.odis.connection.socket;

import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.model.SocketConnectionAttribute;

/**
 * This module provide the ability to accept TCP connection and instantiate the client witch manage this connection.
 * @author O.Beyler
 */
public class TCPClientConnectionIn extends AbstractTCPConnectionIn {
    /** The end. */
    boolean end = false;

    /** The Constant LOGGER. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            TCPClientConnectionIn.class.getName());

    /** The socket. */
    private Socket socket = null;

    /** The tcp server. */
    private Thread tcpServer;

    /** The running. */
    boolean running = true;

    /**
     * Instantiates a new TCP client connection in.
     * @param attribute the attribute
     */
    public TCPClientConnectionIn(SocketConnectionAttribute attribute) {
        super(attribute);
        connectionSocket = this;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Create TCPClientConnection");
        }
    }

    /**
     * Create a socket a message listener to look for message on this socket. <br>
     * If the socket is broken it will recreate a new one.
     * @author Beyler Olivier FranceTelecom R&D
     */
    class TCPKeepConnection implements Runnable {
        /**
         * Instantiates a new TCP keep connection.
         */
        public TCPKeepConnection() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Create TCPServer");
            }
        }

        /**
         * Run.
         * @see java.lang.Runnable#run()
         */
        public void run() {
            try {
                while (running) {
                    try {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG,
                                    "Try to get a socket on "
                                            + attribut.getAddress() + ":"
                                            + attribut.getPort());
                        }
                        socket = new Socket(attribut.getAddress(), attribut
                                .getPort());
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG, "new socket");
                        }
                        MessageListener listener = new MessageListener(
                                connectionSocket);
                        listener.setProtocol(getProtocol());
                        listener.setSocket(socket);
                        listener.start();
                        while (listener.stopListening == false) {
                            Thread.sleep(5000);
                        }
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG,
                                    "the client socket is closed");
                        }
                    } catch (java.net.ConnectException e) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG,
                                    "Retry to get a socket"
                                            + attribut.getAddress() + ":"
                                            + attribut.getPort());
                        }
                        Thread.sleep(5000);
                    }
                }
            } catch (InterruptedException ex) {
                running = false;
            } catch (UnknownHostException e) {
                LOGGER.log(BasicLevel.ERROR, "create socket failed", e);
            } catch (IOException e) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    final StringBuffer msgDbg;
                    msgDbg = new StringBuffer("create socket failed on");
                    msgDbg.append(attribut.getAddress());
                    msgDbg.append(":");
                    msgDbg.append(attribut.getPort());
                    msgDbg.append(" exception :");
                    msgDbg.append(e.getMessage());
                    LOGGER.log(BasicLevel.WARN, msgDbg.toString());
                }
            }
        }

        /**
         * Gets the protocol.
         * @return IProtocolIn
         */
        IProtocolIn getProtocol() {
            IProtocolIn result = null;
            try {
                Class classe = Class.forName(attribut.getProtocolIn());
                result = (IProtocolIn) classe.newInstance();
                result.setSocketCnxAttribut(attribut);
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "create Protocol "
                            + attribut.getProtocolIn());
                }
            } catch (ClassNotFoundException e) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "protocol class "
                            + attribut.getProtocolIn() + " is unknown", e);
                }
            } catch (InstantiationException e) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "protocol class "
                            + attribut.getProtocolIn()
                            + " unable to instantiate", e);
                }
            } catch (IllegalAccessException e) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "protocol class "
                            + attribut.getProtocolIn() + " illegal access", e);
                }
            }
            return result;
        }
    }

    /**
     * Sets the running. *
     * @param running the new running
     */
    public void setRunning(boolean running) {
        this.running = running;
    }

    /**
     * Launch component.
     * @throws OdisConnectionException the odis cnx exception
     */
    public void launchComponent() throws OdisConnectionException {
        tcpServer = new Thread(new TCPKeepConnection());
        tcpServer.start();
    }

    /**
     * Stop component.
     */
    public void stopComponent() {
        running = false;
        tcpServer.interrupt();
        try {
            if (socket != null) {
                socket.close();
            }
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "ERROR :: stopComponent "
                    + e.getMessage());
        }
    }
}
