package org.ow2.odis.connection.socket;

import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.model.SocketConnectionAttribute;

/**
 * This module provide the ability to accept TCP connection and instantiate the client witch manage this connection.
 * @author O.Beyler
 */
public class TCPClientConnectionOut extends AbstractTCPConnectionOut {
    /** logger of the class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            TCPClientConnectionOut.class.getName());

    /** socket used by this client for the connection. */
    private Socket socket = null;

    /** Thread witch will survey the connection regularly. */
    private Thread tcpServer;

    /**
     * while the connection is active the process is running to check regularly the connection.
     */
    private boolean running = true;

    /** time (ms) before checking socket connection. */
    private static final int CHECK_CNX_DELAY = 1000;

    /** time before try to restore connection. */
    private static final int DELAY_BEFORE_RESTORE_CNX = 5000;

    /** The client. */
    MessageSender client = null;

    /**
     * Instantiates a new TCP client connection out.
     * @param attribute the attribute
     */
    public TCPClientConnectionOut(SocketConnectionAttribute attribute) {
        super(attribute);
        connectionSocket = this;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Create "
                    + TCPClientConnectionOut.class.getName());
        }
    }

    /**
     * iner classe.Witch survey the client connection and recreate it if need.
     * @author Beyler Olivier FranceTelecom R&D
     */
    class TCPKeepConnection implements Runnable {
        /**
         * Instantiates a new TCP keep connection.
         */
        public TCPKeepConnection() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "Create TCPServer");
            }
        }

        /**
         * implements Runnable interface.
         */
        public void run() {
            while (running) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    final StringBuffer msgDbg;
                    msgDbg = new StringBuffer("Try to get a socket on ");
                    msgDbg.append(attribut.getAddress());
                    msgDbg.append(":");
                    msgDbg.append(attribut.getPort());
                    LOGGER.log(BasicLevel.DEBUG, msgDbg.toString());
                }
                try {
                    socket = new Socket(attribut.getAddress(), attribut
                            .getPort());
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "new socket");
                    }
                    client = new MessageSender(socket);
                    client.setProtocol(getProtocol());
                    while (running && !client.stopSending) {
                        try {
                            // validate the client every CHECK_CNX_DELAY ms
                            // tand que le client n'est pas arrete on boucle
                            // TODO a transformer en wait...
                            Thread.sleep(CHECK_CNX_DELAY);
                        } catch (InterruptedException ex) {
                            running = false;
                        }
                    }
                } catch (UnknownHostException e) {
                    LOGGER.log(BasicLevel.ERROR, "create socket failed", e);
                } catch (IOException e) {
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        final StringBuffer msgDbg;
                        msgDbg = new StringBuffer("create socket failed on");
                        msgDbg.append(attribut.getAddress());
                        msgDbg.append(":");
                        msgDbg.append(attribut.getPort());
                        msgDbg.append(" exception :");
                        msgDbg.append(e.getMessage());
                        LOGGER.log(BasicLevel.WARN, msgDbg.toString());
                    }
                }
                if (running) {
                    try {
                        // on attend 5s avant de chercher a recrer une socket
                        Thread.sleep(DELAY_BEFORE_RESTORE_CNX);
                    } catch (InterruptedException e) {
                        running = false;
                    }
                }
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "the client socket is closed");
                }
            }
        }

        /**
         * getter of protocol.
         * @return the protocol used by the client
         */
        final IProtocolOut getProtocol() {
            IProtocolOut result = null;
            try {
                Class classe = Class.forName(attribut.getProtocolOut());
                result = (IProtocolOut) classe.newInstance();
                result.setSocketCnxAttribut(attribut);
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "protocol is created");
                }
            } catch (ClassNotFoundException e) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer log = new StringBuffer("protocol class ");
                    log.append(attribut.getProtocolIn());
                    log.append(" is unknown");
                    LOGGER.log(BasicLevel.WARN, log.toString(), e);
                }
            } catch (InstantiationException e) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer log = new StringBuffer("protocol class ");
                    log.append(attribut.getProtocolIn());
                    log.append(" unable to instantiate");
                    LOGGER.log(BasicLevel.WARN, log.toString(), e);
                }
            } catch (IllegalAccessException e) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer log = new StringBuffer("protocol class ");
                    log.append(attribut.getProtocolIn());
                    log.append(" illegal access");
                    LOGGER.log(BasicLevel.WARN, log.toString(), e);
                }
            }
            return result;
        }
    }

    /**
     * setter for running.
     * @param pRunning boolean
     */
    public final void setRunning(final boolean pRunning) {
        this.running = pRunning;
    }

    /**
     * try to send message to the client.
     * @param pMsg byte[] message
     * @return true if succed false in other case
     */
    public final boolean send(byte[] pMsg) {
        boolean result = false;
        if (client != null) {
            result = client.send(pMsg);
        }
        return result;
    }

    /**
     * start the survey thread (tcpServer).
     * @throws OdisConnectionException the odis cnx exception
     */
    public void launchComponent() throws OdisConnectionException {
        tcpServer = new Thread(new TCPKeepConnection(), "TCPClientConnection_TCP");
        tcpServer.start();
    }

    /**
     * Stop component.
     */
    public void stopComponent() {
        running = false;
        if (tcpServer != null) {
            tcpServer.interrupt();
        }
        try {
            socket.close();
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "ERROR :: stopComponent "
                    + e.getMessage());
        }
    }
}
