package org.ow2.odis.connection.socket;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.model.SocketConnectionAttribute;

/**
 * This module provide the ability to accept TCP connection and instantiate the client witch manage this connection.
 * @author O.Beyler
 */
public class TCPServerConnectionOut extends AbstractTCPConnectionOut {
    /** The end. */
    boolean end = false;

    /** logger of the class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            TCPServerConnectionOut.class.getName());

    /** The tcp server. */
    private TCPServer tcpServer;

    /** The running. */
    boolean running = true;

    /** list of connected clients. */
    List lsSenders = new ArrayList();

    /**
     * Instantiates a new TCP server connection out.
     * @param attribute the attribute
     */
    public TCPServerConnectionOut(SocketConnectionAttribute attribute) {
        super(attribute);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Create "
                    + TCPServerConnectionOut.class.getName());
        }
    }

    /**
     * inner class to implement the socket server.
     */
    class TCPServer extends Thread {
        /** The socket server. */
        private ServerSocket socketServer;

        /**
         * constructor.
         */
        public TCPServer() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "Create TCPServer (TCPServerConnectionOut");
            }
        }

        /**
         * create a socket server on specified port.
         */
        public void init() {
            try {
                socketServer = new ServerSocket(attribut.getPort());
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "create ServerSocket on port "
                            + attribut.getPort());
                }
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR,
                        "Unable to create ServerSocket on port "
                                + attribut.getPort(), e);
            }
        };

        /**
         * implement run thread, while the server is running it will accepte the new connection.
         */
        public void run() {
            try {
                while (running) {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER
                                .log(BasicLevel.INFO,
                                        "server wait new connection");
                    }
                    Socket socket = socketServer.accept();
                    socket.setSoTimeout(attribut.getTimeOut());
                    attribut.incNumberConnection(1);
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO,
                                "server accept new connection");
                    }
                    MessageSender sender = new MessageSender(socket);
                    try {
                        sender
                                .setProtocol(attribut
                                        .createProtocolOutInstance());
                    } catch (OdisAttributeException e) {
                        throw new RuntimeException(e);
                    }
                    lsSenders.add(sender);
                }
            } catch (RuntimeException e) {
                LOGGER.log(BasicLevel.ERROR, "RuntimeException", e);
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR, "IOException", e);
            }
            try {
                socketServer.close();
            } catch (IOException e) {
                LOGGER.log(BasicLevel.WARN, "Exception on socketServer close",
                        e);
            }
        }
    }

    /**
     * Start.
     */
    public synchronized void start() {
    }

    /**
     * Sets the running. *
     * @param running the new running
     */
    public void setRunning(boolean running) {
        this.running = running;
    }

    /**
     * launch the server.
     * @throws OdisConnectionException the odis cnx exception
     */
    public void launchComponent() throws OdisConnectionException {
        tcpServer = new TCPServer();
        tcpServer.init();
        tcpServer.start();
    }

    /**
     * send the message to the connected client.
     * @param pMsg message byte array
     * @return true if at least the number of client has received the message has reached the minClient attribute value.
     */
    public boolean send(byte[] pMsg) {
        List clientDeconnected = new ArrayList();
        Iterator itClient = lsSenders.iterator();
        boolean resultSending = true;
        boolean result;
        if (lsSenders.size() >= attribut.getMinClient()) {
            while (itClient.hasNext()) {
                MessageSender client = (MessageSender) itClient.next();
                if (!client.stopSending) {
                    result = client.send(pMsg);
                    if (!result) {
                        clientDeconnected.add(client);
                    }
                }
            }
            Iterator itClientDeconnected = clientDeconnected.iterator();
            while (itClientDeconnected.hasNext()) {
                lsSenders.remove(itClientDeconnected.next());
                attribut.incNumberConnection(-1);
            }
        }
        resultSending = lsSenders.size() >= attribut.getMinClient();
        return resultSending;
    }

    /**
     * stop component.
     */
    public void stopComponent() {
        running = false;
        tcpServer.interrupt();
        try {
            ServerSocket server = tcpServer.socketServer;
            if (server != null) {
                server.close();
            }
        } catch (IOException e) {
            LOGGER.log(BasicLevel.DEBUG, "Closing socket failed...");
        }
    }
}
