package org.ow2.odis.domain;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.Launch;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisDomainException;
import org.ow2.odis.exception.OdisNodeException;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.lifeCycle.connection.IConnectionLifeCycle;
import org.ow2.odis.lifeCycle.domain.AbstractDomainState;
import org.ow2.odis.lifeCycle.domain.IDomainLifeCycle;
import org.ow2.odis.lifeCycle.domain.StateDomainInitial;
import org.ow2.odis.lifeCycle.node.AbstractNodeState;
import org.ow2.odis.lifeCycle.node.StateNodeInitialized;
import org.ow2.odis.lifeCycle.node.StateNodeStarted;
import org.ow2.odis.lifeCycle.node.StateNodeStopped;
import org.ow2.odis.lifeCycle.node.StateNodeSuspended;
import org.ow2.odis.lifeCycle.node.StateNodeUnloaded;
import org.ow2.odis.loader.GenericFactory;
import org.ow2.odis.model.AbstractAttribute;
import org.ow2.odis.model.AbstractConnectionAttribute;
import org.ow2.odis.model.Const;
import org.ow2.odis.model.DomainAttribute;
import org.ow2.odis.model.NodeAttribute;
import org.ow2.odis.node.Node;
import org.ow2.odis.node.NodeManager;
import org.ow2.odis.services.IFactoryService;
import org.ow2.odis.util.TraceException;
import org.xml.sax.InputSource;

/**
 * The aim of a domain manager is to create and manage the domain.<br>
 * There is only one domain per JVM. Copyright France Telecom 2005, All Rights Reserved.
 * @author Beyler Olivier FranceTelecom R&D
 */
public final class Domain implements IDomainLifeCycle, DomainMBean {

    /** The Constant VERSION. */
    private static final String VERSION = "META-INF/version.info";
    /** The Constant CONNECTIONS. Name of an XML tag in ODIS 3.1 format. */
    private static final String CONNECTIONS = "CONNECTIONS";
    /** The Constant NODE. Name of an XML tag in ODIS 3.1 format. */
    public static final String NODE = "NODE";
    /** The Constant NODES. Name of an XML tag in ODIS 3.1 format. */
    public static final String NODES = "NODES";
    /** The Constant DOMAIN. Name of an XML tag in ODIS 3.1 format. */
    public static final String DOMAIN = "DOMAIN";
    /** The Constant OLDDOMAIN. Name of an XML tag in ODIS 2.* format. */
    public static final String OLDDOMAIN = "deploiement";
    /** The Constant NAME. Name of an XML attribute in ODIS 3.1 format. */
    private static final String NAME = "name";
    /** Logger for the class. */
    private static final Logger LOGGER = Monolog.initialize().getLogger(Domain.class.getName());
    /** Logger for the version. */
    private static final Logger LOGGER_VERSION = Monolog.initialize().getLogger("VERSION");
    /**
     * Unique instance of Domain<br>
     * Implements <i>singleton</i> pattern.
     */
    private static Domain instance = null;
    /**
     * State of the domain.<br>
     * Follows the Domain life-cycle.
     * @see com.francetelecom.odis.component.lifeCycle.domain
     */
    private AbstractDomainState domainState;
    /** This attribute contains all properties of the domain. */
    public final DomainAttribute attribute;
    /** The factory instance. */
    private IFactoryService factoryInstance = new GenericFactory();
    /**
     * The list of elements to initialize later.<br>
     * This is used when the Domain is asked to add nodes/connections and is not ready to initialize them at once.
     * @see method {@link #initAndStartByStringLater(String)}
     */
    private List lsInitLater = new ArrayList();

    /**
     * Gets the single instance of Domain.
     * @return single instance of Domain
     */
    public static Domain getInstance() {
        if (instance == null) {
            synchronized (LOGGER) {
                if (instance == null) {
                    instance = new Domain();
                    String info = instance.getInfoVersion();
                    if (info != null) {
                        String[] infos = info.split("\n");
                        for (int i = 0; i < infos.length; i++) {
                            LOGGER_VERSION.log(BasicLevel.DEBUG, infos[i]);
                        }
                    }
                }
            }
        }
        return instance;
    }

    /**
     * Hides the default constructor.
     */
    private Domain() {
        domainState = new StateDomainInitial();
        attribute = new DomainAttribute(null);
        JMX.getInstance().registerMBean("Domain", this);
    }

    /**
     * Defines administration elements which will be added to the document :
     * <li> AdminNode node definition
     * <li> Admin_In link definition
     * <li> Admin_Out link definition <br>
     * The added elements definitions are restored from 3 files.<br>
     * Note : <i>this is not used in ODIS 3.1</i>
     * @param doc the document the elements are to be added to
     * @since odis framework 2.4
     */
    private static void addAdminElements(Document doc) {
        Element root = doc.getRootElement();
        List nodes = root.getChildren(NODE);
        Iterator itNodes = nodes.iterator();
        // recherche de la presence du Node Admin
        Element adminNodeElt = null;
        while (itNodes.hasNext()) {
            try {
                Element node = (Element) itNodes.next();
                if ("AdminNode".equals(node.getAttribute(NAME))) {
                    adminNodeElt = node;
                }
            } catch (ClassCastException e) {
                // nothing
            }
        }
        if (adminNodeElt == null) {
            // le noeud d'administration n'a pas ete trouve
            addConfigFile("com/francetelecom/odis/component/domain/AdminNode.xml", doc);
        }
        // recherche de la presence des liens Admin_In et Admin_Out
        List links = root.getChildren(CONNECTIONS);
        Iterator itLinks = links.iterator();
        // recherche de la presence des liens de Node Admin
        Element adminLinkInElt = null;
        Element adminLinkOutElt = null;
        while (itLinks.hasNext()) {
            try {
                Element node = (Element) itLinks.next();
                if ("Admin_In".equals(node.getAttribute(NAME))) {
                    adminLinkInElt = node;
                }
                if ("Admin_Out".equals(node.getAttribute(NAME))) {
                    adminLinkOutElt = node;
                }
            } catch (ClassCastException e) {
            }
        }
        if (adminLinkInElt == null) {
            // le lien entrant d'administration n'a pas ete trouve
            addConfigFile("com/francetelecom/odis/component/domain/AdminLinkIn.xml", doc);
        }
        if (adminLinkOutElt == null) {
            // le lien sortir d'administration n'a pas ete trouve
            addConfigFile("com/francetelecom/odis/component/domain/AdminLinkOut.xml", doc);
        }
    }

    /**
     * Opens a resource file, which is an ODIS XML document, and merges its contents with the document doc.
     * @param fileName name of a file holding ODIS elements definitions
     * @param doc the document the file's elements are to be added to
     */
    public static void addConfigFile(final String fileName, Document doc) {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Adding a new Config File :" + fileName);
        }
        URL url = NodeManager.class.getResource("/" + fileName);
        if (url == null) {
            LOGGER.log(BasicLevel.FATAL, "Unable to find resource /" + fileName);
            return;
        }
        addConfigUrl(url, doc);
    }

    /**
     * Opens an URL, which references an ODIS XML document, and merges its contents with the document doc.
     * @param url URL of a file holding ODIS elements definitions
     * @param doc the document the file's elements are to be added to
     */
    public static void addConfigUrl(URL url, Document doc) {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Adding a new Config File by URL :" + url);
        }
        SAXBuilder builder = new SAXBuilder();
        if (builder == null) {
            LOGGER.log(BasicLevel.FATAL, "Unable to create SAXBuilder");
            return;
        }
        if (url == null) {
            LOGGER.log(BasicLevel.FATAL, "Unable to find resource /" + url);
            return;
        }
        try {
            if (doc == null) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "Reading resource to initialize domain's document");
                }
                InputStream is = url.openStream();
                doc = builder.build(is);
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO,
                            "Reading resource to update domain's document");
                }
                final Document temp = builder.build(url.openStream());
                final Element root = doc.getRootElement();
                Element nodes = root.getChild(NODES);
                if (nodes == null) {
                    nodes = new Element(NODES);
                    root.addContent(nodes);
                }
                Element connections = root.getChild(CONNECTIONS);
                if (connections == null) {
                    connections = new Element(CONNECTIONS);
                    root.addContent(connections);
                }
                final List lsDeploymentElt = new ArrayList();
                final List lsNodeElt = new ArrayList();
                final List lsConnexionElt = new ArrayList();
                // retrieves existing elements from domain's document
                extractInformationsFromDocument(doc, lsDeploymentElt, lsNodeElt, lsConnexionElt);
                // retrieves elements from new document
                extractInformationsFromDocument(temp, lsDeploymentElt, lsNodeElt, lsConnexionElt);
                // merges all elements into root element
                root.addContent(lsDeploymentElt);

                nodes.addContent(lsNodeElt);
                connections.addContent(lsConnexionElt);
            }
        } catch (JDOMException e) {
            LOGGER.log(BasicLevel.ERROR,
                    "JDMOException while adding configuration resource " + url, e);
            e.printStackTrace();
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR,
                    "IOException while adding configuration resource " + url, e);
            e.printStackTrace();
        }
    }

    /**
     * Extracts ODIS elements from a Jdom document and adds them into the passed lists according to their type.
     * @param docu the source document
     * @param lsDeploymentElt list of deployment elements
     * @param lsNodeElt list of node elements
     * @param lsConnexionElt list of connection elements
     */
    public static void extractInformationsFromDocument(final Document docu, final List lsDeploymentElt, final List lsNodeElt,
            final List lsConnexionElt) {
        Element root = docu.getRootElement();
        List ls = root.getContent();
        Object[] obs = ls.toArray();
        // for all elements in source document
        for (int i = 0; i < obs.length; i++) {
            Object ob = obs[i];
            if (ob instanceof org.jdom.Comment) {
                // we remove the comments as they become usually meaningless in a merged document
                org.jdom.Comment comment = (org.jdom.Comment) ob;
                comment.detach();
            } else if (ob instanceof Element) {
                Element elt = (Element) ob;
                if (DOMAIN.equals(elt.getName())) {
                    elt.detach();
                    if (lsDeploymentElt.size() > 0) {
                        LOGGER.log(BasicLevel.ERROR, "Can't have more than one deployment information !!!");
                        LOGGER.log(BasicLevel.ERROR, "new information will be ignored");
                    } else {
                        lsDeploymentElt.add(elt);
                    }
                }
                if (OLDDOMAIN.equals(elt.getName())) {
                    LOGGER.log(BasicLevel.WARN, "ODIS 3.1 : Please use <DOMAIN> instead of <deploiement> - adding element anyway.");
                    elt.detach();
                    if (lsDeploymentElt.size() > 0) {
                        LOGGER.log(BasicLevel.ERROR, "Can't have more then one deployment information !!!");
                    } else {
                        lsDeploymentElt.add(elt);
                    }
                }
                if (NODES.equals(elt.getName())) {
                    List nodes = elt.getChildren();
                    while (!nodes.isEmpty()) {
                        Object e = nodes.remove(0);
                        try {
                            ((Element) e).detach();
                            LOGGER.log(BasicLevel.DEBUG, "found node"+e.toString());
                            lsNodeElt.add(e);
                        } catch (ClassCastException exp) {
                        }
                    }
                }
                if (CONNECTIONS.equals(elt.getName())) {
                    List connections = elt.getChildren();
                    while (!connections.isEmpty()) {
                        Object e = connections.remove(0);
                        try {
                            ((Element) e).detach();
                            LOGGER.log(BasicLevel.DEBUG, "found connection"+e.toString());
                            lsConnexionElt.add(e);
                        } catch (ClassCastException exp) {
                        }
                    }
                }
            }
        }
    }

    /**
     * Gets the name.
     * @return the name of the domain
     */
    public String getName() {
        return attribute.getName();
    }

    /**
     * Gets the path.
     * @return the path used for the persistence tree
     */
    public String getPath() {
        return attribute.getPath();
    }

    /**
     * Gets the domain state.
     * @return the domain state
     * @see com.francetelecom.odis.component.lifeCycle.domain.IDomainLifeCycle#getDomainState()
     */
    public AbstractDomainState getDomainState() {
        return domainState;
    }

    /**
     * Sets the domain state.
     * @param state the state
     * @see com.francetelecom.odis.component.lifeCycle.domain.IDomainLifeCycle#setDomainState(com.francetelecom.odis.component.lifeCycle.domain.AbstractDomainState)
     */
    public void setDomainState(final AbstractDomainState state) {
        domainState = state;
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * Initializes the domain by creating and initializing all nodes.
     * @throws OdisDomainException if a node refuses to initialize
     * @see com.francetelecom.odis.component.lifeCycle.domain.IDomainLifeCycle#initComponent()
     */
    public void initComponent() throws OdisDomainException {
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e) {
            throw new OdisDomainException(e);
        }
        List lsExceptions = new ArrayList();
        NodeAttribute nodeAttribut;
        Node node;
        NodeAttribute[] tabNodeAttributs = (NodeAttribute[]) attribute.lsNodeAttributs.toArray(new NodeAttribute[attribute.lsNodeAttributs.size()]);
        for (int i = 0; i < tabNodeAttributs.length; i++) {
            nodeAttribut = (NodeAttribute) tabNodeAttributs[i];
            node = new Node(nodeAttribut);
            try {
                node.getNodeState().setLifeState(node, new StateNodeInitialized());
                NodeManager.register(node);
            } catch (OdisNodeException e) {
                try {
                    NodeManager.register(node);
                } catch (Exception ex) {
                    LOGGER.log(BasicLevel.ERROR, "register failed", ex);
                }
                lsExceptions.add(e.getMessage());
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            if (attribute.isStopOnError()) {
                throw new OdisDomainException(errors);
            }
            LOGGER.log(BasicLevel.ERROR, "fail during node initialize : " + errors);
        }
    }

    /**
     * Launches the domain by starting all nodes.
     * @throws OdisDomainException if a node refuses to start
     * @see com.francetelecom.odis.component.lifeCycle.domain.IDomainLifeCycle#launchComponent()
     */
    public void launchComponent() throws OdisDomainException {
        List lsExceptions = new ArrayList();
        NodeAttribute[] tabNodeAttributs = (NodeAttribute[]) attribute.lsNodeAttributs.toArray(new NodeAttribute[attribute.lsNodeAttributs.size()]);
        for (int i = 0; i < tabNodeAttributs.length; i++) {
            NodeAttribute nodeAttribut = (NodeAttribute) tabNodeAttributs[i];
            try {
                AbstractNodeState nodeState = nodeAttribut.getNode().getNodeState();
                nodeState.setLifeState(nodeAttribut.getNode(), new StateNodeStarted());
            } catch (OdisNodeException e) {
                lsExceptions.add(e.getMessage());
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            if (attribute.isStopOnError()) {
                throw new OdisDomainException(errors);
            }
            LOGGER.log(BasicLevel.ERROR, "fail during node initialize : " + lsExceptions.toString());
        }
    }

    /**
     * Create the domain attribut.
     * @param element org.jdom.Element
     */
    public void createDomainAttribut(final Element element) {
        final Element root = element;
        Element domainDeclaration = root.getChild(DOMAIN);
        if (domainDeclaration == null) {
            domainDeclaration = root.getChild(OLDDOMAIN);
            if (domainDeclaration != null) {
                StringBuffer buffer = new StringBuffer("Use ");
                buffer.append(DOMAIN);
                buffer.append(" instead of ");
                buffer.append(OLDDOMAIN);
                LOGGER.log(BasicLevel.WARN, buffer.toString());
            }
        }
        attribute.setter(domainDeclaration);
    }

    /**
     * Creates the attribut.
     * @param element org.jdom.Element
     * @throws OdisDomainException the odis domain exception
     */
    public void createAttribut(final Element element) throws OdisDomainException {
        // ------------------------
        // create domain attribut
        createDomainAttribut(element);
      }

    /**
     * create nodes attributs.
     * @param element org.jdom.Element
     * @return a collection of NodeAttribut
     */
    private Collection createNodesAttributs(final Element element) {
        NodeManager nodeManager = NodeManager.getInstance();
        Element base = element.getChild(NODES);
        List result = new ArrayList();
        if (base == null) {
            LOGGER.log(BasicLevel.INFO, "<NODES></NODES> not defined ");
        } else {
            List lsNodesDeclaration = base.getChildren(NODE);
            result = new ArrayList(lsNodesDeclaration.size());
            Iterator itNodeDeclaration = lsNodesDeclaration.iterator();
            Element nodeDeclaration = null;
            NodeAttribute nodeAtribut = null;
            while (itNodeDeclaration.hasNext()) {
                nodeDeclaration = (Element) (itNodeDeclaration.next());
                nodeAtribut = nodeManager.createNodeAttribut(nodeDeclaration);
                result.add(nodeAtribut);
            }
        }
        return result;
    }

    /**
     * create cnx attribut.
     * @param element the element
     * @return a collection of CnxAttribut
     */
    private Collection createCnxAttributs(final Element element) {
        ConnectionManager connexionManager = ConnectionManager.getInstance();
        Element base = element.getChild(CONNECTIONS);
        List result = new ArrayList();
        if (base == null) {
            LOGGER.log(BasicLevel.INFO, "<CONNECTIONS></CONNECTIONS> not defined ");
        } else {
            List lsCnxDeclaration = base.getChildren();
            result = new ArrayList(lsCnxDeclaration.size());
            Iterator itCnxDeclaration = lsCnxDeclaration.iterator();
            Element cnxDeclaration = null;
            AbstractConnectionAttribute cnxAttribut = null;
            while (itCnxDeclaration.hasNext()) {
                cnxDeclaration = (Element) itCnxDeclaration.next();
                cnxAttribut = connexionManager.createCnxAttribut(cnxDeclaration);
                if (cnxAttribut == null) {
                    LOGGER.log(BasicLevel.ERROR, "unable to create attribute");
                } else {
                    result.add(cnxAttribut);
                }
            }
        }
        return result;
    }

    /**
     * Adds configuration elements to the Domain if it is started.<br>
     * If it is not, all actions are postponed until it starts.
     * @param config the config
     * @see #initAndStartByString(String, boolean)
     * @see #invokeLater()
     */
    public void initAndStartByStringLater(final String config) {
        if (getDomainState().getState() != IDomainLifeCycle.STATE_STARTED) {
            lsInitLater.add(config);
        } else {
            initAndStartByString(config, true);
        }
    }

    /*
     * (non-Javadoc)
     * @see com.francetelecom.odis.component.domain.DomainMBean#initByString(java.lang.String)
     */
    public void initByString(final String config) {
        initAndStartByString(config, false);
    }

    /**
     * Called after the Domain has started to execute any pending configuration actions requested while the Domain was not ready.
     */
    public void invokeLater() {
        Iterator it = lsInitLater.iterator();
        while (it.hasNext()) {
            String init = (String) it.next();
            initAndStartByString(init, true);
        }
        lsInitLater.clear();
    }

    /**
     * Adds configuration elements to the Domain based on a String parameter, and optionally starts them.
     * @param config configuration string (ODISXMLSchema3_1.xsd format)
     * @param start if true the nodes created will be started
     * @throws RuntimeException if called when the Domain is not started
     */
    public synchronized void initAndStartByString(final String config, final boolean start) {
        if (getDomainState().getState() != IDomainLifeCycle.STATE_STARTED) {
            throw new RuntimeException("InitByString called before domain reach Started State");
        }
        try {
            InputSource inputSource;
            SAXBuilder builder;
            Document doc;
            builder = new SAXBuilder();
            inputSource = new InputSource(new StringReader(config));
            doc = builder.build(inputSource);
            Element element = doc.getRootElement();
            Collection lsNodeAttributs = createNodesAttributs(element);
            Collection lsCnxAttributs = createCnxAttributs(element);
            List lsExceptions = new ArrayList();
            // check duplicates and conflicts
            if (lsNodeAttributs != null && !lsNodeAttributs.isEmpty()) {
                try {
                    lsNodeAttributs = checkDuplicate(attribute.lsNodeAttributs, lsNodeAttributs);
                    lsNodeAttributs = checkInvalidNode(attribute.lsNodeAttributs, lsNodeAttributs);
                } catch (OdisDomainException e) {
                    lsExceptions.add(e);
                }
            }
            if (lsCnxAttributs != null && !lsCnxAttributs.isEmpty()) {
                try {
                    lsCnxAttributs = checkDuplicate(attribute.lsCnxAttributs, lsCnxAttributs);
                    lsCnxAttributs = checkInvalidConnection(attribute.lsCnxAttributs, lsCnxAttributs);
                } catch (OdisDomainException e) {
                    lsExceptions.add(e);
                }
            }
            if (!lsExceptions.isEmpty()) {
                String errors = TraceException.formatListException(lsExceptions);
                if (Launch.getInstance().stopOnError) {
                    throw new RuntimeException(errors);
                }
            }
            attribute.lsNodeAttributs.addAll(lsNodeAttributs);
            attribute.lsCnxAttributs.addAll(lsCnxAttributs);
            // connections
            if (lsCnxAttributs != null && !lsCnxAttributs.isEmpty()) {
                ConnectionManager.getInstance().add(lsCnxAttributs);
            }
            // nodes
            if (lsNodeAttributs != null && !lsNodeAttributs.isEmpty()) {
                Iterator itNodeAttribut = lsNodeAttributs.iterator();
                while (itNodeAttribut.hasNext()) {
                    NodeAttribute nodeAttribut = (NodeAttribute) itNodeAttribut.next();
                    Node node = new Node(nodeAttribut);
                    try {
                        NodeManager.register(node);
                        node.getNodeState().setLifeState(node, new StateNodeInitialized());
                        if (start) {
                            node.getNodeState().setLifeState(node, new StateNodeStarted());
                        }
                    } catch (OdisNodeException e) {
                        lsExceptions.add(e);
                        StringBuffer buff = new StringBuffer("Exception on Node [");
                        buff.append(node).append("]");
                        buff.append(node).append(e.getMessage());
                        LOGGER.log(BasicLevel.ERROR, buff.toString());
                        LOGGER.log(BasicLevel.DEBUG, "stacktrace :", e);
                    }
                }
                if (!lsExceptions.isEmpty()) {
                    String errors;
                    errors = TraceException.formatListException(lsExceptions);
                    if (attribute.isStopOnError()) {
                    	instance = null;
                        throw new RuntimeException(errors);
                    }
                    LOGGER.log(BasicLevel.ERROR, "errors during node initialization : see previous messages");
                }
            }
        } catch (JDOMException e) {
            LOGGER.log(BasicLevel.ERROR, "JDMOException initByString Domain", e);
            throw new RuntimeException(e.getMessage());
        } catch (IOException e) {
            LOGGER.log(BasicLevel.ERROR, "IOException initByString Domain", e);
        } catch (RuntimeException e) {
            LOGGER.log(BasicLevel.ERROR, "RunTimeException initByString Domain" + e.getMessage());
            throw e;
        }
    }

    /**
     * Checks a collection for full duplicates and removes them.
     * @param lsAttributs the collection to check
     * @return Collection a new collection with duplicates pruned out
     */
    public Collection checkDuplicate(Collection lsAttributs) {
        Object localAttribut;
        Collection lsResults = new ArrayList(lsAttributs.size());
        List lsToRemoves = new ArrayList(lsAttributs.size());
        Iterator iterator = lsAttributs.iterator();
        while (iterator.hasNext()) {
            localAttribut = iterator.next();
            if (!lsResults.contains(localAttribut)) {
                lsResults.add(localAttribut);
            } else {
                lsToRemoves.add(localAttribut);
            }
        }
        if (!lsToRemoves.isEmpty()) {
            StringBuffer buffer = new StringBuffer("Duplicate Information found");
            buffer.append(Const.EOL);
            buffer.append(" only one added to domain ");
            buffer.append(Const.EOL);
            buffer.append(" it concern this : ");
            buffer.append(Const.EOL);
            Iterator it = lsToRemoves.iterator();
            while (it.hasNext()) {
                AbstractAttribute att = (AbstractAttribute) it.next();
                try {
                    ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
                    att.writeXML(byteStream);
                    buffer.append(byteStream.toString());
                } catch (IOException e) {
                    LOGGER.log(BasicLevel.ERROR, "Unexpected exception", e);
                }
                buffer.append(Const.EOL);
            }
            LOGGER.log(BasicLevel.WARN, buffer.toString());
        }
        return lsResults;
    }

    /**
     * Checks a collection of attributes for full duplicates against a collection of running attributes
     * @param lsRunningAttributs list of running attributes
     * @param toCheckAttributs list of attributes to check
     * @return Collection a new collecion with duplicates pruned out
     */
    public Collection checkDuplicate(Collection lsRunningAttributs, Collection toCheckAttributs) {
        Object localAttribut;
        Collection lsAttributs = checkDuplicate(toCheckAttributs);
        Collection lsResults = new ArrayList(lsAttributs.size());
        Iterator iterator = lsAttributs.iterator();
        while (iterator.hasNext()) {
            localAttribut = iterator.next();
            if (!lsRunningAttributs.contains(localAttribut)) {
                lsResults.add(localAttribut);
            } else {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    StringBuffer buffer;
                    buffer = new StringBuffer("Duplicate (already running, not added) :: ");
                    buffer.append(localAttribut.toString());
                    LOGGER.log(BasicLevel.DEBUG, buffer.toString());
                }
            }
        }
        return lsResults;
    }

    /**
     * Checks invalid connections : two connections cannot have the same name and be different. If best effort behaviour is required :
     * <li>two conflicting new connections are both ignored
     * <li>a new connection conflicting with a running one is ignored
     * <li>all running connections are always kept
     * @param lsRunningAttributs list of running connection attributes
     * @param toCheckAttributs list of new connexion attributes to check
     * @return collection of valid new connexion attributes
     * @throws OdisDomainException if any conflict exists and isStopOnError()
     */
    public Collection checkInvalidConnection(Collection lsRunningAttributs, Collection toCheckAttributs) throws OdisDomainException {
    	AbstractConnectionAttribute cnxAttribut;
        HashMap map = new HashMap();
        List dupNames = new ArrayList();
        Iterator it = toCheckAttributs.iterator();
        // Look for duplicate names within new attributes
        while (it.hasNext()) {
            cnxAttribut = (AbstractConnectionAttribute) it.next();
            if (map.containsKey(cnxAttribut.getName())) {
                dupNames.add(cnxAttribut.getName());
                LOGGER.log(BasicLevel.ERROR, "Conflicting new connection declarations for name " + cnxAttribut.getName());
            } else {
                map.put(cnxAttribut.getName(), cnxAttribut);
            }
        }
        // Stop or prune duplicates
        if (!dupNames.isEmpty()) {
            if (attribute.isStopOnError()) {
                throw new OdisDomainException("Conflicting new connection declarations for names " + dupNames.toString());
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Removing conflicting new connection declarations from configuration");
                }
                it = dupNames.iterator();
                while (it.hasNext()) {
                    map.remove(it.next());
                }
            }
        }
        // Check the new connections against running ones
        it = lsRunningAttributs.iterator();
        dupNames.clear();
        while (it.hasNext()) {
            cnxAttribut = (AbstractConnectionAttribute) it.next();
            if (map.containsKey(cnxAttribut.getName())) {
                LOGGER.log(BasicLevel.ERROR, "Connection " + cnxAttribut.getName() + " already running. Removing conflicting new declaration");
                dupNames.add(cnxAttribut.getName());
                map.remove(cnxAttribut.getName());
            }
        }
        if (!dupNames.isEmpty() &&attribute.isStopOnError()) {
            throw new OdisDomainException("Some new connection declarations conflict with running ones :: " + dupNames.toString());            
        }
        return map.values();
    }

    /**
     * Checks invalid nodes : two nodes cannot have the same name and be different. If best effort behavior is required :
     * <li>two conflicting new nodes are both ignored
     * <li>a new node conflicting with a running one is ignored
     * <li>all running nodes are always kept
     * @param lsRunningAttributs list of running node attributes
     * @param toCheckAttributs list of new node attributes to check
     * @return collection of valid new node attributes
     * @throws OdisDomainException if any conflict exists and isStopOnError()
     */
    public Collection checkInvalidNode(Collection lsRunningAttributs, Collection toCheckAttributs) throws OdisDomainException {
        NodeAttribute nodeAttribut;
        HashMap map = new HashMap();
        List dupNames = new ArrayList();
        Iterator it = toCheckAttributs.iterator();
        // Look for duplicate names within new attributes
        while (it.hasNext()) {
            nodeAttribut = (NodeAttribute) it.next();
            if (map.containsKey(nodeAttribut.getName())) {
                dupNames.add(nodeAttribut.getName());
                LOGGER.log(BasicLevel.ERROR, "Conflicting new node declarations for name " + nodeAttribut.getName());
            } else {
                map.put(nodeAttribut.getName(), nodeAttribut);
            }
        }
        // Stop or prune duplicates
        if (!dupNames.isEmpty()) {
            if (attribute.isStopOnError()) {
                throw new OdisDomainException("Conflicting new node declarations for names " + dupNames.toString());
            } else {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "Removing conflicting new node declarations from configuration");
                }
                it = dupNames.iterator();
                while (it.hasNext()) {
                    map.remove(it.next());
                }
            }
        }
        // Check the new nodes against running ones
        it = lsRunningAttributs.iterator();
        dupNames.clear();
        while (it.hasNext()) {
            nodeAttribut = (NodeAttribute) it.next();
            if (map.containsKey(nodeAttribut.getName())) {
                LOGGER.log(BasicLevel.ERROR, "Node " + nodeAttribut.getName() + " already running. Removing conflicting new declaration");
                dupNames.add(nodeAttribut.getName());
                map.remove(nodeAttribut.getName());
            }
        }
        if (!dupNames.isEmpty() && attribute.isStopOnError()) {
            throw new OdisDomainException("Some new node declarations conflict with running ones :: " + dupNames.toString());
        }
        return map.values();
    }

    /**
     * Stops the Domain.<br>
     * This is done by stopping all nodes (suspending them first if they were started).
     * @throws OdisDomainException if a node fails to stop properly
     */
    public void stopComponent() throws OdisDomainException {
        List lsExceptions = new ArrayList();
        NodeAttribute nodeAttribut;
        AbstractNodeState nodeState;
        NodeAttribute[] tabNodeAttributs = (NodeAttribute[]) attribute.lsNodeAttributs.toArray(new NodeAttribute[attribute.lsNodeAttributs.size()]);
        for (int i = 0; i < tabNodeAttributs.length; i++) {
            nodeAttribut = (NodeAttribute) tabNodeAttributs[i];
            nodeState = nodeAttribut.getNode().getNodeState();
            try {
                if (nodeState.getState() == IConnectionLifeCycle.STATE_STARTED) {
                    nodeState.setLifeState(nodeAttribut.getNode(), new StateNodeSuspended());
                    // update nodeState
                    nodeState = nodeAttribut.getNode().getNodeState();
                }
                nodeState.setLifeState(nodeAttribut.getNode(), new StateNodeStopped());
            } catch (OdisNodeException e) {
                e.printStackTrace();
                lsExceptions.add(e);
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisDomainException(errors);
        }
    }

    /**
     * Unloads the Domain.<br>
     * This is done by unloading all nodes.
     * @throws OdisDomainException if a node fails to unload properly
     * @see com.francetelecom.odis.component.lifeCycle.domain.IDomainLifeCycle#unloadComponent()
     */
    public void unloadComponent() throws OdisDomainException {
        
        List lsExceptions = new ArrayList();
        Object[] tabNodeAttributs = attribute.lsNodeAttributs.toArray();
        AbstractNodeState nodeState;
        for (int i = 0; i < tabNodeAttributs.length; i++) {
            Node n = ((NodeAttribute) tabNodeAttributs[i]).getNode();
            if (n != null) {
                nodeState = n.getNodeState();
                try {
                    nodeState.setLifeState(((NodeAttribute) tabNodeAttributs[i]).getNode(), new StateNodeUnloaded());
                } catch (OdisNodeException e) {
                    lsExceptions.add(e);
                }
            }
        }
        JMX.getInstance().unregisterMBean("Domain", this);
        instance = null;
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisDomainException(errors);
        }
    }

    /**
     * Exports the XML description of the domain.
     * @return string
     * @throws IOException exception
     */
    public String exportXML() throws IOException {
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        attribute.writeXML(byteStream);
        return byteStream.toString();
    }

    /**
     * Exports the full XML description of the domain.
     * @return string
     * @throws IOException exception
     */
    public String exportFullXML() throws IOException {
        ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
        attribute.writeXML(byteStream, true);
        return byteStream.toString();
    }

    /**
     * Gets the info version.
     * @return the info version
     * @see com.francetelecom.odis.component.domain.DomainMBean#getInfoVersion()
     */
    public String getInfoVersion() {
        URL url = Domain.class.getResource("/META-INF/maven/org.ow2.odis/odis-core/pom.properties");
        Properties properties = new Properties();
        try {
            if (url != null) {
                properties.load(url.openStream());
            }
        } catch (IOException e) {
        }
        BufferedReader br = null;
        StringBuffer buffer = new StringBuffer("#===============================================\n");
        buffer.append("#        ");
        buffer.append(properties.getProperty("artifactId", "not defined"));
        buffer.append("  Version ");
        buffer.append(properties.getProperty("version", "not defined"));
        buffer.append("\n");
        url = Domain.class.getResource("/" + VERSION);
        if (url == null) {
            LOGGER.log(BasicLevel.ERROR, "url null unable to find file version");
            buffer.append("unable to find file version");
        } else {
            try {
                br = new BufferedReader(new InputStreamReader(url.openStream()));
                String ligne;
                while ((ligne = br.readLine()) != null) {
                    buffer.append(ligne);
                    buffer.append("\n");
                }
            } catch (FileNotFoundException e) {
                LOGGER.log(BasicLevel.ERROR, "unable to find file version");
            } catch (IOException e) {
                LOGGER.log(BasicLevel.ERROR, "unable to read file version");
            } finally {
                try {
                    if (br != null) {
                        br.close();
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
        return buffer.toString();
    }

    /**
     * Replaces the factory by the specified one.
     * @param factory the factory
     */
    public void setFactoryInstance(IFactoryService factory) {
        if (factory == null) {
            LOGGER.log(BasicLevel.ERROR, "Could not set factory to null");
        }
        factoryInstance = factory;
    }

    /**
     * Gets the factory instance.
     * @return The FactoryInstance of the domain.
     */
    public IFactoryService getFactoryInstance() {
        if (factoryInstance == null) {
            factoryInstance = new GenericFactory();
        }
        return factoryInstance;
    }

    /**
     * Resets the factory to the default value : GenericFactory.
     */
    public void resetFactoryInstance() {
        factoryInstance = new GenericFactory();
        LOGGER.log(BasicLevel.DEBUG, "Factory has been reset to GenericFactory");
    }
}
