package org.ow2.odis.encoder;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.exception.OdisEncoderException;
import org.ow2.odis.exception.OdisFactoryException;

/**
 * Manage Encoder creation. As the encoding function is sequential, inside a node only one encoder by type is created. 2 encoders cannot
 * have the same type even their class'name is different.
 * @author olivier Beyler
 */
public class EncoderFactory {
    /**
     * log of this class.
     */
    final static Logger LOGGER = Monolog.initialize().getLogger(
            EncoderFactory.class.getName());

    /**
     * contain the association between the class name of the encoder and the encoder type. example : "com.francetelecom.TensionEncoder"
     * ->Tension's encoder type this map is only use during encoder's registration. to avoid differents encoders with the same type.
     */
    private final transient Map mapEncoderDeclaration;

    /**
     * contain the association between the type of the encoder. Type is known by using the getType() function of this encoder. example
     * :"Tension"-> Tension's Encoder.
     */
    private final transient Map mapEncoder;

    /**
     * constructor.
     */
    public EncoderFactory() {
        super();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Create EncoderFactory ");
        }
        mapEncoder = new HashMap();
        mapEncoderDeclaration = new HashMap();
    }

    /**
     * Register the encoder with the type of object decoded by this decoder.<br>
     * Register two encoder with the same type is refused. <br>
     * Register a encoder with an unknown class is refused.
     * @param name name of this decoder
     * @param version
     * @throws OdisEncoderException
     */
    public synchronized void registerEncoder(final String name,
            final String version) throws OdisEncoderException {
        IEncoder encoder = null;
        try {
            encoder = (IEncoder) Domain.getInstance().getFactoryInstance()
                    .getInstanceOf(name, version);
            final String typeEncoder = encoder.getType();
            final String lastDeclaration = (String) mapEncoderDeclaration
                    .get(typeEncoder);
            if (lastDeclaration == null) {
                mapEncoderDeclaration.put(typeEncoder, name);
                mapEncoder.put(typeEncoder, encoder);
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, "encoder " + name
                            + " is registred");
                }
            } else if (lastDeclaration.equals(name)) {
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "encoder " + name
                            + " is already registred");
                }
            } else {
                StringBuffer buff;
                buff = new StringBuffer(
                        "Try to register two differents encoder with same type :");
                buff.append("\nencoder :");
                buff.append(lastDeclaration);
                buff.append("\nencoder :");
                buff.append(name);
                buff.append("\nhave the same type :");
                buff.append(typeEncoder);
                LOGGER.log(BasicLevel.ERROR, buff.toString());
                throw new OdisEncoderException(buff.toString());
            }
        } catch (ClassCastException e) {
            // <JDO>This to allow Launch to crash properly</JDO>
            throw new OdisEncoderException(e);
        } catch (OdisFactoryException e) {
            throw new OdisEncoderException(e);
        }
    }

    /**
     * @param name is the type's encoder wanted
     * @return associate encoder. <br>
     * note : if the encoder is not found return <i>null</i> and a log error is created.
     */
    public IEncoder getEncoder(final String name) {
        IEncoder encoder = (IEncoder) mapEncoder.get(name);
        if (encoder == null) {
            encoder = (IEncoder) mapEncoder.get(Object.class.getName());
            if (encoder == null) {
                LOGGER.log(BasicLevel.ERROR,
                        "Try to use unregister encoder type :" + name);
            }
        }
        return encoder;
    }
}
