package org.ow2.odis.engine;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.bufferIn.BufferInManager;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.decoder.IDecoder;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.lifeCycle.Behavior;
import org.ow2.odis.lifeCycle.engine.IEngineLifeCycle;
import org.ow2.odis.lifeCycle.engine.StateEngineStopped;
import org.ow2.odis.model.EngineInAttribute;
import org.ow2.odis.node.Node;
import org.ow2.odis.port.PortFactory;
import org.ow2.odis.port.PortIn;
import org.ow2.odis.util.Semaphore;
import org.ow2.odis.util.TraceException;

/**
 * The Class EngineIn.
 */
public final class EngineIn extends AbstractEngine implements IEngineIn {
    /** provide log service. */
    final static Logger LOGGER = Monolog.initialize().getLogger(
            EngineIn.class.getName());

    /**
     * this variable is set to true if the semaphore is set during a suspend time in this case when the process is restarted the semaphore
     * is set.
     */
    private transient boolean wakeUpDuringSuspendState = false;

    /** The number of error while decoding. */
    private transient long nbrErrorDecoding = 0;

    /** The behavior. */
    final public Behavior behavior;

    /** The port factory. */
    final public PortFactory portFactory;

    /** The attribute. */
    private EngineInAttribute attribut;

    /** The semaphore. */
    private final transient Semaphore sem = new Semaphore();

    /**
     * Instantiates a new engine in.
     * @param pNode the node
     * @param attribute the attribute
     */
    public EngineIn(Node pNode, EngineInAttribute attribute) {
        super("EngineIn_" + pNode.getName(), pNode);
        this.attribut = attribute;
        this.attribut.setEngineIn(this);
        behavior = new Behavior(this);
        portFactory = new PortFactory(this);
    }

    /**
     * use name to identify the node.
     * @return EngineIn_<i>node name</i>
     */
    public String toString() {
        return (node == null) ? "EngineIn_XX" : "EngineIn_" + node.getName();
    }

    /**
     * <i>note : Thread start function override</i> <br>
     * set the right SynchronizeManagerIn then start thread.
     */
    public synchronized void start() {
        final BufferInManager manager = node.getSynchronizeManagerIn();
        // try to restore all saved messages witch have not been released
        manager.resynchronizeData(attribut.getPolicyInstance());
        // ------------------
        super.start();
    }

    /**
     * Active read.
     */
    public void activeRead() {
        sem.release();
    }

    /**
     * <i>note : Thread run function override</i> <br>
     * infinite loop processing news messages from link.
     */
    public void run() {
        ContextMessage contextMessage = null;
        byte[] msg = null;
        boucleRun: while (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
            try {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, node.getName()
                            + " waiting new message from link ");
                }
                sem.waiting();
                if (engineState.getState() == IEngineLifeCycle.STATE_SUSPENDED) {
                    // deroutage des reveils effectu?s pendant l'etat suspendu
                    // mais prise en compte quand meme pour que quand l'etat est
                    // remis a start on sache
                    // qu'un reveil force doit etre fait
                    wakeUpDuringSuspendState = true;
                } else if (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO, node.getName()
                                + " has new message");
                    }
                    PortIn portIn = (PortIn) attribut.getPolicyInstance()
                            .getPolicyElement();
                    // while the policy component is able to elect a portIn
                    // to extract a message from.
                    while (portIn != null
                            && engineState.getState() == IEngineLifeCycle.STATE_STARTED
                            && wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG,
                                    "get Message from link " + portIn.getId());
                        }
                        contextMessage = portIn.getContextMessage();
                        msg = contextMessage.getPayload();
                        if (msg == null) {
                            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                                LOGGER.log(BasicLevel.WARN,
                                        "link has null msg or empty msg");
                            }
                        } else if (msg.length == 0) {
                            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                                LOGGER.log(BasicLevel.WARN,
                                        "link has empty msg");
                            }
                        } else {
                            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                                LOGGER.log(BasicLevel.DEBUG,
                                        "decode Message from link");
                            }
                            final IDecoder decoder = portIn.getDecoder();
                            List objects = null;
                            try {
                                objects = decoder.decode(msg);
                            } catch (RuntimeException e1) {
                                LOGGER.log(BasicLevel.ERROR,
                                        "Decode Msg failed ", e1);
                                nbrErrorDecoding++;
                                if (msg == null) {
                                    LOGGER.log(BasicLevel.ERROR,
                                            "message was null");
                                } else {
                                    LOGGER.log(BasicLevel.ERROR,
                                            "message size :" + msg.length);
                                    final StringBuffer buffer = new StringBuffer(
                                            "message =");
                                    Byte character;
                                    for (int i = 0; i < msg.length; i++) {
                                        character = new Byte(msg[i]);
                                        buffer.append(character);
                                    }
                                    LOGGER.log(BasicLevel.ERROR, buffer
                                            .toString());
                                }
                            }
                            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                                LOGGER.log(BasicLevel.DEBUG, "ask to "
                                        + node.getName()
                                        + "'s manager to take care of message");
                            }
                            try {
                                if (objects != null && objects.size() > 0) {
                                    BufferInManager sm = node
                                            .getSynchronizeManagerIn();
                                    final Context context = contextMessage
                                            .getContext();
                                    if (!sm.synchronizeData(msg, objects,
                                            portIn.getId(), context)) {
                                        sm.pause();
                                        while (!sm.synchronizeData(msg,
                                                objects, portIn.getId(),
                                                context)
                                                && wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                                            sm.pause();
                                        }
                                    }
                                }
                            } catch (InterruptedException e) {
                                break boucleRun;
                            }
                        }
                        if (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                                LOGGER.log(BasicLevel.DEBUG, node.getName()
                                        + "'s manager ack message");
                            }
                            portIn.ackMessage(true);
                        }
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG, node.getName()
                                    + " new message from link");
                        }
                        portIn = (PortIn) attribut.getPolicyInstance()
                                .getPolicyElement();
                    }
                }
            } catch (InterruptedException e) {
                // the end of this engine has been requested
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "User interrupt engine In of "
                            + node.getName());
                }
            }
        }
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Engine In of " + node.getName()
                    + " is stopped");
        }
        engineState = new StateEngineStopped();
    }


    /**
     * Gets the node name.
     * @return node name
     */
    public String getNodeName() {
        return getNode().getName();
    }

    /**
     * acknowledge the message coming from the link.
     * @param linkId the link id
     */
    public void ackMessage(final String linkId) {
        final PortIn link = (PortIn) attribut.getPolicyInstance().getLinkById(
                linkId);
        link.ackMessage(true);
    }

    /**
     * initialize the EngineIn.
     * @throws OdisEngineException the odis engine exception
     */
    public void initComponent() throws OdisEngineException {
        List lsExceptions = new ArrayList();
        try {
            attribut.checkProperties();
        } catch (OdisAttributeException e1) {
            throw new OdisEngineException(e1);
        }
        List lsElement;
        lsElement = attribut.getPolicyInstance().getLsPolicyElements();
        final Iterator itElement = lsElement.iterator();
        PortIn portIn;
        while (itElement.hasNext()) {
            portIn = (PortIn) itElement.next();
            // rejette les messages en cours de traitements dans les portIn
            try {
                portIn.initComponent();
            } catch (OdisConnectionException e) {
                lsExceptions.add(e);
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            OdisEngineException exception = new OdisEngineException(errors);
            throw exception;
        }
    }

    /**
     * set the state of the engine to suspend witch allow him to be notified that an event happends but not to react on it.
     */
    public void suspendComponent() {
        if ((engineState.getState() == IEngineLifeCycle.STATE_SUSPENDED)  &&
            LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Engine In of " + node.getName()
                    + " is already suspended");
        }
        wakeUpDuringSuspendState = false;
    }

    /**
     * set the wanted state to stopped and wait until the component is not stopped.
     */
    public void stopComponent() {
        int result = 0;
        wantedState = new StateEngineStopped();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Stopping for Node " + getNodeName());
        }
        if (Thread.currentThread() == this) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO,
                        "Stopping from current Thread for Node "
                                + getNodeName());
            }
            // je suis demandeur du stopped alors que je suis le thread en cours
            // donc je m'arrete directe
            engineState = new StateEngineStopped();
        } else {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO,
                        "Stopping from another Thread for Node "
                                + getNodeName());
            }
            while (this.isAlive()) {
                try {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Interrupting Engine IN.");
                    }
                    this.interrupt();
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG,
                                "Engine IN interrupted. Waiting for join().");
                    }
                    this.join(1000L);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "Joined.");
                    }
                } catch (InterruptedException e) {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG,
                                "Interrupted while trying to interrupt/join");
                    }
                    continue;
                }
            }
        }
        final List lsElement = attribut.getPolicyInstance()
                .getLsPolicyElements();
        final Iterator itElement = lsElement.iterator();
        PortIn portIn;
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Stopping the PortIns");
        }
        while (itElement.hasNext()) {
            portIn = (PortIn) itElement.next();
            // rejette les messages en cours de traitements dans les portIn
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "stop engineIn Link"
                        + getNodeName());
            }
            portIn.stopComponent();
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "stop engineIn Link"
                        + getNodeName() + " done " + result);
            }
        }
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Stop engineIn done " + getNodeName());
        }
    }

    /**
     * Unloads all dependencies.
     * @throws OdisEngineException if a dependent element won't unload
     */
    public void unloadComponent() throws OdisEngineException {
        final List lsElement = attribut.getPolicyInstance()
                .getLsPolicyElements();
        final Iterator itElement = lsElement.iterator();
        PortIn portIn;
        while (itElement.hasNext()) {
            portIn = (PortIn) itElement.next();
            // decharge les portIn
            portIn.unloadComponent();
        }
    }

    /**
     * Launches the component. If the component was suspended it restarts.
     * @throws OdisEngineException if a dependent element couldn't be launched
     */
    public void launchComponent() throws OdisEngineException {
        List lsExceptions = new ArrayList();
        final List lsElement = attribut.getPolicyInstance()
                .getLsPolicyElements();
        final Iterator itElement = lsElement.iterator();
        while (itElement.hasNext()) {
            PortIn portIn = (PortIn) itElement.next();
            // rejette les messages en cours de traitements dans les portIn
            try {
                portIn.launchComponent();
            } catch (OdisConnectionException e) {
                lsExceptions.add(e.getMessage());
            }
        }
        start();
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            OdisEngineException exception = new OdisEngineException(errors);
            throw exception;
        }
    }

    /**
     * Re-launches the component. If the component was suspended it restarts.
     */
    public void reLaunchComponent() {
        if (wakeUpDuringSuspendState) {
            sem.release();
        }
    }

    // --------------------------------------------------------------------------
    // --------------
    // Getter and Setter for each private attribute
    // --------------------------------------------------------------------------
    // --------------
    /**
     * @return int
     * @see com.francetelecom.odis.component.engine.IEngineIn#getNumberOfObjectMax()
     */
    public int getNumberOfObjectMax() {
        return attribut.getNumberOfObjectMax();
    }

    /**
     * @return long
     * @see com.francetelecom.odis.component.engine.IEngineIn#getSizeByteMax()
     */
    public long getSizeByteMax() {
        return attribut.getSizeByteMax();
    }

    /**
     * Gets the nbr error decoding.
     * @return number of error during decoding.
     */
    public long getNbrErrorDecoding() {
        return nbrErrorDecoding;
    }

    /**
     * Gets the engine in.
     * @return the engine in
     */
    public EngineIn getEngineIn() {
        return this;
    }

    /**
     * Gets the behavior.
     * @return the behavior
     */
    public Behavior getBehavior() {
        return behavior;
    }
}
