package org.ow2.odis.engine;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.bufferOut.BufferObject;
import org.ow2.odis.bufferOut.BufferOut;
import org.ow2.odis.bufferOut.BufferOutManager;
import org.ow2.odis.bufferOut.IBufferOutWatcher;
import org.ow2.odis.connection.context.ContextManager;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.lifeCycle.Behavior;
import org.ow2.odis.lifeCycle.Cause;
import org.ow2.odis.lifeCycle.engine.IEngineLifeCycle;
import org.ow2.odis.lifeCycle.engine.StateEngineStopped;
import org.ow2.odis.model.EngineOutAttribute;
import org.ow2.odis.model.ToAttribute;
import org.ow2.odis.node.Node;
import org.ow2.odis.policy.AbstractPolicy;
import org.ow2.odis.port.PortFactory;
import org.ow2.odis.port.PortOut;
import org.ow2.odis.routing.Routing;
import org.ow2.odis.util.CodingDecoding;
import org.ow2.odis.util.Semaphore;
import org.ow2.odis.util.TraceException;

/**
 * The EngineOut is the third engine of an ODIS Node, responsible for sending data.<br>
 * It runs as a Thread.<br>
 * In STARTED state, the EngineOut asks its policy instance for a non-empty buffer, retrieves the data in the buffer and passes them to the
 * adequate OUT ports, as defined by the Routing instance and the available destination information.
 */
public final class EngineOut extends AbstractEngine {
    /** Logger for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            EngineOut.class.getName());

    /** The buffer out manager instance for this engine. */
    private BufferOutManager bufferOutManager = null;

    /** The routing instance for this engine. */
    private final Routing routing;

    /** The behavior instance for this engine. */
    private Behavior behavior = null;

    /** The semaphore used to pause the run() loop. */
    private Semaphore sem = new Semaphore();

    /** The attribute holding the engine's properties. */
    private final EngineOutAttribute attribut;

    /** The port factory instance for this engine, used to get OUT ports. */
    private final PortFactory portFactory;

    /** The delay (in ms) between retries when sending data fails. */
    private static final int SLEEP_DELAY_BEFORE_RETRY = 1000;

    /**
     * If the engine is waken up while in SUSPENDED state, this attribute is set to true.<br>
     * When the engine reaches STARTED state again, it wakes up immediately if this attribute is true.
     */
    private boolean wakeUpDuringSuspendedState;

    /** This attribute is true while the engine is retrying to send data. */
    private boolean isOnError = false;

    /** How many times this engine went into a retrying cycle to send data. */
    private long nbErrors = 0;

    /**
     * Instantiates a new EngineOut.
     * @param pNode the node this engine belongs
     * @param pAttribut the properties attribute
     */
    public EngineOut(Node pNode, EngineOutAttribute pAttribut) {
        super("EngineOut_" + pNode.getName(), pNode);
        this.attribut = pAttribut;
        pAttribut.setEngineOut(this);
        behavior = new Behavior(this);
        bufferOutManager = new BufferOutManager(this);
        portFactory = new PortFactory(this);
        routing = new Routing();
    }

    /**
     * Returns a String representation of this engine, based on the associated Node's name.
     * @return the string representation
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer("EngineOut_");
        if (node == null) {
            buffer.append("XX");
        } else {
            buffer.append(node.getName());
        }
        return buffer.toString();
    }

    /**
     * Activates sending operations.<br>
     * This should be called as a request to send what data is available. It wakes up the engine.
     */
    public void activeSend() {
        sem.release();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, node.getName() + " active send");
        }
    }

    /**
     * Infinite loop processing new messages available in the buffers.<br>
     * <i>Overrides java.lang.Thread.run()</i>
     */
    public void run() {
        try {
            while (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, node.getName()
                            + " waiting new message from link ");
                }
                // wait on the semaphore for new data to send
                try {
                    sem.waiting();
                } catch (InterruptedException e) {
                    // typically when this engine has been requested to stop
                    break;
                }
                if (engineState.getState() == IEngineLifeCycle.STATE_SUSPENDED) {
                    // the engine was waken up while in SUSPENDED state : no
                    // processing will take place put this has to be remembered,
                    // so
                    // as to wake the engine as soon as it reaches STARTED state
                    // again.
                    wakeUpDuringSuspendedState = true;
                } else if (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                    // the engine is NOT stopping, this is the regular case
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, node.getName()
                                + " has new message to send");
                    }
                    // ask the policy to choose a non empty BufferOut
                    BufferOut bufferOut = (BufferOut) attribut
                            .getPolicyInstance().getPolicyElement();
                    if (bufferOut == null
                            && LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, node.getName()
                                + " policy has nothing to send");
                    }
                    // while the policy is able to select a BufferOut to extract
                    // a message from
                    // AND the engine isn't stopping
                    while (bufferOut != null
                            && wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                        // retrieve objects from the buffer
                        // (the number of objects in the list depends on the
                        // buffer's properties)
                        List lsObjects = bufferOut.getObjects();
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG, "get "
                                    + lsObjects.size() + " objects from "
                                    + bufferOut.toString());
                        }
                        // make a message with the objects
                        byte[] msg = createMessage(lsObjects);
                        // First try to reach the contextual destination if any
                        if (routing.sendContextualMessage(bufferOut
                                .getDestination(), msg)) {
                            // No failure : try to reach the standard
                            // destinations
                            Collection failures = null;
                            if (routing.sendMessage(bufferOut.getDestination(),
                                    msg, failures)) {
                                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                                    LOGGER.log(BasicLevel.INFO,
                                            "SEND MESSAGE OK -> "
                                                    + bufferOut.toString());
                                }
                                bufferOut.release(lsObjects);
                            } else {
                                // there are failures on non contextual
                                // connections : we'll retry
                                nbErrors++;
                                isOnError = true;
                                Cause cause = new Cause(Cause.LINK_ERROR,
                                        "unable to send message");
                                node.getEngineProceed().getBehavior().addCause(
                                        cause);
                                node.getEngineIn().getBehavior()
                                        .addCause(cause);
                                if (LOGGER.isLoggable(BasicLevel.ERROR)) {
                                    StringBuffer errMsg = new StringBuffer(
                                            "Message cannot be sent on ");
                                    errMsg.append(bufferOut.getId());
                                    errMsg
                                            .append(" -> engineIn and proceed will be suspended for node ");
                                    errMsg.append(node.getName());
                                    LOGGER.log(BasicLevel.ERROR, errMsg
                                            .toString());
                                }
                                int countRetry = 1;
                                while (!routing.sendMessage(bufferOut
                                        .getDestination(), msg, failures)
                                        && wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                                    Thread.sleep(SLEEP_DELAY_BEFORE_RETRY);
                                    countRetry++;
                                }
                                if (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED
                                        && LOGGER.isLoggable(BasicLevel.ERROR)) {
                                    StringBuffer errMsg = new StringBuffer(
                                            "End of error. Messages have been sent for node ");
                                    errMsg.append(node.getName());
                                    errMsg.append(EOL);
                                    errMsg
                                            .append("\t-> number of retries done before success : ");
                                    errMsg.append(countRetry);
                                    LOGGER.log(BasicLevel.ERROR, errMsg
                                            .toString());
                                }
                                cause.setEnd();
                                node.getEngineProceed().getBehavior()
                                        .checkBehaviours();
                                node.getEngineIn().behavior.checkBehaviours();
                                // <JDO reason=
                                // "prevent duplication on retry (seen on SOCKETIN)"
                                // >
                                // Don't release if retry cycle aborted due to
                                // stop request
                                if (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                                    bufferOut.release(lsObjects);
                                }
                                // </JDO>
                            }
                        } else {
                            // Recipient unavailable : log the fact and exclude
                            // the buffer from policy
                            ContextManager.getInstance().notifyContextChange(
                                    bufferOut.getDestination().getContext(),
                                    true);
                            if (LOGGER.isLoggable(BasicLevel.WARN)) {
                                StringBuffer errMsg = new StringBuffer(
                                        toString());
                                errMsg
                                        .append(" could not send a message contextually on ");
                                errMsg.append(bufferOut.getDestination());
                                errMsg
                                        .append(" -> the message will be sent when the recipient is available");
                                LOGGER.log(BasicLevel.ERROR, errMsg.toString());
                            }
                        }
                        isOnError = false;
                        bufferOut = (BufferOut) attribut.getPolicyInstance()
                                .getPolicyElement();
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            if (bufferOut == null) {
                                LOGGER.log(BasicLevel.DEBUG,
                                        "there is no more msg to send");
                            } else {
                                LOGGER.log(BasicLevel.DEBUG, "the buffer ["
                                        + bufferOut.getId() + "] has still "
                                        + bufferOut.getPendingMessage()
                                        + " pending msg");
                            }
                        }
                    }
                }
            }
        } catch (InterruptedException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Engine Out of " + node.getName()
                        + " is interrupted");
            }
        }
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Engine Out of " + node.getName()
                    + " is stopped");
        }
    }

    /**
     * Creates a message holding the passed objects.
     * @param lsObject a list of objects
     * @return a byte array holding the objects
     */
    private byte[] createMessage(List lsObject) {
        byte[][] vector = new byte[lsObject.size()][];
        // build a vector with the serialized objects (byte[]) to send
        for (int i = 0; i < lsObject.size(); i++) {
            BufferObject bo = (BufferObject) lsObject.get(i);
            if (bo != null) {
                vector[i] = bo.getMsg();
            } else {
                vector[i] = null;
            }
        }
        return CodingDecoding.getLinearByte(vector);
    }

    /**
     * Gets the node name.
     * @return node name
     */
    public String getNodeName() {
        return getNode().getName();
    }

    /**
     * Inits the component.
     * @throws OdisEngineException the odis engine exception
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#initComponent()
     */
    public void initComponent() throws OdisEngineException {
        List lsExceptions = new ArrayList();
        try {
            attribut.checkProperties();
        } catch (OdisAttributeException e1) {
            throw new OdisEngineException(e1);
        }
        // <JDO/> THIS HAS BEEN DONE WITHIN THE checkProperties() ABOVE
        // ======================================================
        // // parcours des bufferOuts pour les initialiser
        // Collection colBufferAttributs;
        // colBufferAttributs = attribut.mapIdBufferOutAttribut.values();
        // final Iterator itBufferOutAttribut = colBufferAttributs.iterator();
        // BufferOutAttribut bufferOutAttribut;
        // while (itBufferOutAttribut.hasNext()) {
        // bufferOutAttribut = (BufferOutAttribut) itBufferOutAttribut.next();
        // bufferOutAttribut.getBufferOut().initComponent();
        // }
        // parcours des To pour les initialiser
        Collection colToAttributs;
        colToAttributs = attribut.getLsTo();
        final Iterator itTo = colToAttributs.iterator();
        ToAttribute toAttribut;
        while (itTo.hasNext()) {
            toAttribut = (ToAttribute) itTo.next();
            PortOut portOut = toAttribut.getPortOut();
            try {
                portOut.initComponent();
            } catch (OdisConnectionException e) {
                lsExceptions.add(e.getMessage());
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            OdisEngineException exception = new OdisEngineException(errors);
            throw exception;
        }
    }

    /**
     * Suspend component.
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#suspendComponent()
     */
    public void suspendComponent() {
        wakeUpDuringSuspendedState = false;
    }

    /**
     * Stop component.
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#stopComponent()
     */
    public void stopComponent() {
        wantedState = new StateEngineStopped();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "stop engineOut of" + getNodeName());
        }
        if (Thread.currentThread() == this) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER
                        .log(BasicLevel.INFO,
                                "stop engineOut from current Thread of"
                                        + getNodeName());
            }
            // the current Thread asked to stop, so we just have to stop
            engineState = new StateEngineStopped();
        } else {
            while (this.isAlive()) {
                try {
                    this.interrupt();
                    this.join(1000L);
                } catch (InterruptedException e) {
                    continue;
                }
            }
        }
        Collection colToAttributs;
        colToAttributs = attribut.getLsTo();
        final Iterator itTo = colToAttributs.iterator();
        ToAttribute toAttribut;
        while (itTo.hasNext()) {
            toAttribut = (ToAttribute) itTo.next();
            PortOut portOut = toAttribut.getPortOut();
            if (portOut != null) {
                portOut.stopComponent();
            }
        }
    }

    /**
     * Launch component.
     * @throws OdisEngineException the odis engine exception
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#launchComponent()
     */
    public void launchComponent() throws OdisEngineException {
        List lsExceptions = new ArrayList();
        bufferOutManager.launchComponent();
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Engine Out of " + node.getName()
                    + " is starting");
        }
        Collection list = routing.getListPortOut();
        Iterator it = list.iterator();
        while (it.hasNext()) {
            PortOut route = (PortOut) it.next();
            try {
                route.launchComponent();
            } catch (OdisConnectionException e) {
                lsExceptions.add(e.getMessage());
            }
        }
        start();
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisEngineException(errors);
        }
    }

    /**
     * Re launch component.
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#reLaunchComponent()
     */
    public void reLaunchComponent() {
        if (wakeUpDuringSuspendedState) {
            sem.release();
        }
    }

    /**
     * Unload component.
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#unloadComponent()
     */
    public void unloadComponent() {
        bufferOutManager = null;
        Collection colToAttributs;
        colToAttributs = attribut.getLsTo();
        final Iterator itTo = colToAttributs.iterator();
        ToAttribute toAttribut;
        while (itTo.hasNext()) {
            toAttribut = (ToAttribute) itTo.next();
            PortOut portOut = toAttribut.getPortOut();
            if (portOut != null) {
                portOut.unloadComponent();
            }
        }
    }

    /**
     * Gets the buffer out manager.
     * @return the buffer out manager
     */
    public BufferOutManager getBufferOutManager() {
        return bufferOutManager;
    }

    /**
     * Sets the buffer out manager.
     * @param pBufferOutManager the new buffer out manager
     */
    public void setBufferOutManager(BufferOutManager pBufferOutManager) {
        this.bufferOutManager = pBufferOutManager;
    }

    /**
     * Gets the engine out.
     * @return the engine out
     */
    public EngineOut getEngineOut() {
        return this;
    }

    /**
     * Gets the routing.
     * @return the routing
     */
    public Routing getRouting() {
        return routing;
    }

    /**
     * Checks if is on error.
     * @return true, if is on error
     */
    public boolean isOnError() {
        return isOnError;
    }

    /**
     * Gets the behavior.
     * @return the behavior
     */
    public Behavior getBehavior() {
        return behavior;
    }

    /**
     * Gets the nbr error.
     * @return the nbr error
     */
    public long getNbrError() {
        return nbErrors;
    }

    /**
     * Gets the port factory.
     * @return the port factory
     */
    public PortFactory getPortFactory() {
        return portFactory;
    }

    /**
     * Gets the policy instance.
     * @return the policy instance
     */
    public AbstractPolicy getPolicyInstance() {
        return attribut.getPolicyInstance();
    }

    /**
     * Adds a BufferOut watcher.
     * @param bow
     */
    public void addBufferOutWatcher(IBufferOutWatcher bow) {
        bufferOutManager.addBufferOutWatcher(bow);
    }
}
