package org.ow2.odis.engine;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.bufferIn.BufferInManager;
import org.ow2.odis.bufferOut.BufferOutManager;
import org.ow2.odis.encoder.EncoderFactory;
import org.ow2.odis.encoder.IEncoder;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisContextException;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.exception.OdisException;
import org.ow2.odis.exception.OdisNoBufferException;
import org.ow2.odis.lifeCycle.Behavior;
import org.ow2.odis.lifeCycle.engine.IEngineLifeCycle;
import org.ow2.odis.lifeCycle.engine.StateEngineStopped;
import org.ow2.odis.model.EngineProceedAttribute;
import org.ow2.odis.node.Node;
import org.ow2.odis.proceed.ProceedManager;
import org.ow2.odis.routing.Destination;
import org.ow2.odis.routing.IRouting;
import org.ow2.odis.routing.RoutingObject;
import org.ow2.odis.stat.StatProvider;
import org.ow2.odis.util.Semaphore;

/**
 * Engine of a node which provides process functionality.
 * @author Beyler Olivier
 * @version 1.0
 * @since JDK 1.4
 */
public class EngineProceed extends AbstractEngine implements IEngineLifeCycle,
        IEngineProceed {
    /** Logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            EngineProceed.class.getName());

    /** The Constant MAX_TRANSACTION_ID. */
    private static final long MAX_TRANSACTION_ID = Long.MAX_VALUE;

    /** The Constant MIN_TRANSACTION_ID. */
    private static final long MIN_TRANSACTION_ID = 1L;

    /** The semaphore. */
    private Semaphore sem = new Semaphore();

    /** The run once. */
    private boolean runOnce = false;

    /** define the node transaction id. */
    private long transactionId = MIN_TRANSACTION_ID;

    /** The timer. */
    private Timer timer = null;

    /** The encoder factory. */
    private EncoderFactory encoderFactory = null;

    /**
     * this variable is set to true. If the semaphore is set during a suspend time in this case when the process is restarted the semaphore
     * is set because we know that process have received a notification without proceed it
     */
    private boolean wakeUpDuringSuspendState = false;

    /** The behavior. */
    private Behavior behavior = null;

    /** The Constant WAITING_PROCEED. */
    static final String WAITING_PROCEED = "waiting proceed";

    /** The Constant RUNNING_PROCEED. */
    static final String RUNNING_PROCEED = "running proceed";

    /** The is routing process. */
    private boolean isRoutingProcess = false;

    /** The attribute. */
    private EngineProceedAttribute attribute = null;

    /** The active proceed. */
    private ActiveProceed activeProceed = null;

    /**
     * Constructor.
     * @param pNode node of this engine proceed. keep a reference to the node
     * @param pAttribute the attribute
     */
    public EngineProceed(Node pNode, EngineProceedAttribute pAttribute) {
        super("EngineProceed_" + pNode.getName(), pNode);
        this.attribute = pAttribute;
        pAttribute.setEngineProceed(this);
        behavior = new Behavior(this);
        encoderFactory = new EncoderFactory();
    }

    /**
     * Uses name to identify the engine node.
     * @return the string
     */
    public String toString() {
        if (node == null) {
            return "EngineProceed_XX";
        } else {
            return "EngineProceed_" + node.getName();
        }
    }

    /**
     * Activates the proceed.
     * @see com.francetelecom.odis.component.engine.IEngineProceed#activeProceed()
     */
    public void activeProceed() {
        LOGGER.log(BasicLevel.DEBUG, "activeProceed");
        sem.release();
    }

    /**
     * Sends data to the appropriate BufferOut.
     * @param ls the list of objects to be sent
     * @param bufferOutManager the buffer out manager
     * @param isRouting denotes usage of routing functionality
     * @throws InterruptedException if interrupted
     */
    private void toBufferOut(List ls, BufferOutManager bufferOutManager,
            boolean isRouting) throws InterruptedException {
        if (ls != null) {
            Iterator it = ls.iterator();
            Collection destinations = null;
            IEncoder encoder = null;
            byte[] msg = null;
            while (it.hasNext()) {
                try {
                    Object ob = it.next();
                    msg = null;
                    if (ob == null) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            StringBuffer buffer = new StringBuffer(
                                    "Proceed of ");
                            buffer.append(getNodeName());
                            buffer
                                    .append(" has return some null object into his result list");
                            LOGGER.log(BasicLevel.DEBUG, buffer.toString());
                        }
                    } else {
                        String className = findClassName(ob, isRouting);
                        destinations = findDestinations(ob, isRouting,
                                className);
                        encoder = getEncoder(className);
                        if (encoder == null) {
                            throw new OdisException("Encoder == null");
                        }
                        msg = encodeMsg(ob, encoder, isRouting);
                        if (msg == null && LOGGER.isLoggable(BasicLevel.WARN)) {
                            StringBuffer errMsg = new StringBuffer(toString());
                            errMsg.append("Encoding ");
                            errMsg.append(ob);
                            errMsg.append(" produce a null byte array.");
                            LOGGER.log(BasicLevel.WARN, errMsg.toString());
                        }
                        putMessageIntoBuffersOut(msg, destinations,
                                bufferOutManager);
                    }
                } catch (OdisException e) {
                    StringBuffer errMsg = new StringBuffer(
                            "This exception occured for engine Proceed");
                    errMsg.append(getNodeName());
                    errMsg.append(": ");
                    errMsg.append(org.ow2.odis.model.Const.EOL);
                    LOGGER.log(BasicLevel.ERROR, errMsg.toString(), e);
                }
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                StringBuffer buffer = new StringBuffer("Proceed of ");
                buffer.append(getNodeName());
                buffer.append(" has return null object");
                LOGGER.log(BasicLevel.DEBUG, buffer.toString());
            }
        }
    }

    /**
     * For all the destinations, put message into destination's buffer out.
     * @param msg
     * @param destinations
     * @param bufferOutManager
     * @throws InterruptedException
     */
    void putMessageIntoBuffersOut(byte[] msg, Collection destinations,
            BufferOutManager bufferOutManager) throws InterruptedException {
        // start a transaction for the object proceeded
        long idTransaction = getCurrentIdTransaction();
        Iterator itDestination = destinations.iterator();
        while (itDestination.hasNext()) {
            Destination destination = (Destination) itDestination.next();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                StringBuffer errMsg = new StringBuffer(toString());
                errMsg.append("send msg to buffer ");
                errMsg.append(destination);
                LOGGER.log(BasicLevel.DEBUG, errMsg.toString());
            }
            try {
                bufferOutManager.putData(idTransaction, destination, msg);
            } catch (OdisNoBufferException e) {
                if (LOGGER.isLoggable(BasicLevel.WARN)) {
                    StringBuffer errMsg = new StringBuffer(toString());
                    errMsg.append(" found no buffer for destination ");
                    errMsg.append(destination);
                    LOGGER.log(BasicLevel.WARN, errMsg.toString());
                }
            } catch (OdisContextException e) {
                LOGGER.log(BasicLevel.FATAL, "Non contextual destination "
                        + destination.getName()
                        + " points to a contextual buffer (" + e.getMessage()
                        + ")");
            }
        }
    }

    /**
     * encodeMsg : in case of routing object, the encoded message is the payload. In the other case it's the object
     * @param ob
     * @param isRouting
     * @return object
     */
    static byte[] encodeMsg(Object ob, IEncoder encoder, boolean isRouting)
            throws OdisException {
        byte[] msg = null;
        Object payload = findPayload(ob, isRouting);
        if (encoder != null && payload != null) {
            msg = encoder.encode(payload);
        }
        return msg;
    }

    /**
     * findPayload extract object's payload. The payload can be either the object or can be contained into a routing object.
     * @param ob
     * @param isRouting
     * @return object
     */
    static Object findPayload(Object ob, boolean isRouting)
            throws OdisException {
        Object payload = null;
        if (isRouting && ob != null) {
            LOGGER.log(BasicLevel.DEBUG, "Do routing process");
            try {
                RoutingObject rob = (RoutingObject) ob;
                if (rob.getPayload() != null) {
                    payload = rob.getPayload();
                } else {
                    LOGGER.log(BasicLevel.WARN,
                            Const.ROUTING_OBJECT_WITH_EMPTY_PAYLOAD);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        StringBuffer debMsg = new StringBuffer();
                        debMsg.append("Routing Object :");
                        debMsg.append(rob.toString());
                        LOGGER.log(BasicLevel.DEBUG, debMsg);
                    }
                }
            } catch (ClassCastException e) {
                throw new OdisException(Const.INVALID_ROUTING_OBJECT, e);
            }
        } else {
            LOGGER.log(BasicLevel.DEBUG, "Do no routing process");
            payload = ob;
        }
        return payload;
    }

    /**
     * findClassName : return the class name of the payload in case of a routing object or the object class name in the other case.
     * @param ob
     * @param isRouting
     * @return String
     */
    static String findClassName(Object ob, boolean isRouting)
            throws OdisException {
        String className = null;
        if (ob != null) {
            if (isRouting) {
                Object payload = findPayload(ob, isRouting);
                if (payload != null) {
                    className = payload.getClass().getName();
                }
            } else {
                className = ob.getClass().getName();
            }
        }
        return className;
    }

    /**
     * findDestinations : construct a destination list. The destination list is directly extract from the routing object informations or
     * deduce from classname in case of no routing.
     * @param resultOfProceed : the object produce by the proceed.
     * @param isRouting : is the proceed implement IRouting ?
     * @return Collection
     */
    static Collection findDestinations(Object resultOfProceed,
            boolean isRouting, String className) throws OdisException {
        Collection destinations = null;
        if (isRouting) {
            LOGGER.log(BasicLevel.DEBUG, "Do routing process");
            if (resultOfProceed != null) {
                // extract object's payload and get the wanted destinations
                try {
                    destinations = ((RoutingObject) resultOfProceed)
                            .getLsDestinations();
                } catch (ClassCastException e) {
                    throw new OdisException(Const.INVALID_ROUTING_OBJECT, e);
                }
            } else {
                LOGGER.log(BasicLevel.DEBUG, Const.ERR_PARAM_IS_NULL
                        + "resultOfProceed");
            }
        } else {
            LOGGER.log(BasicLevel.DEBUG, "do no routing process");
            // the objects class determines the destination
            if (className == null) {
                throw new OdisException(Const.ERR_PARAM_IS_NULL + "className");
            }
            if (resultOfProceed != null) {
                destinations = new ArrayList(1);
                destinations.add(new Destination(className));
            }
        }
        return destinations;
    }

    /**
     * Main loop for Proceed invocation.
     */
    public void run() {
        IProceed process = attribute.getProceedInstance();
        isRoutingProcess = (process instanceof IRouting);
        BufferInManager bufferIn = node.getSynchronizeManagerIn();
        BufferOutManager bufferOutManager = node.getEngineOut()
                .getBufferOutManager();
        List ls = null;
        long start;
        long delay;
        while (wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
            try {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, WAITING_PROCEED);
                }
                sem.waiting();
                if (engineState.getState() == IEngineLifeCycle.STATE_SUSPENDED) {
                    wakeUpDuringSuspendState = true;
                } else {
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, RUNNING_PROCEED);
                    }
                    Object source = bufferIn.getNextObject();
                    // if auto-activation is on
                    if (runOnce) {
                        runOnce = false;
                        boolean error = false;
                        // // get the proceed
                        // process = attribut.getProceedInstance();
                        process
                                .notifyFWEvent(new FWEvent(
                                        FWEvent.ENGINE_PROCEED,
                                        FWEvent.BEFORE_PROCEED));
                        try {
                            attribute.incInvocation();
                            start = StatProvider.getNanoTime();
                            // call the proceed
                            ls = process.proceed(source);
                            delay = StatProvider.getNanoTime() - start;
                            StatProvider.getStatEngine().probe(attribute, delay);
                        } catch (RuntimeException e) {
                            LOGGER.log(BasicLevel.ERROR,
                                    "Node exception during proceed on node"
                                            + node.getName(), e);
                            error = true;
                            attribute.incNbrProceedError();
                            while (retryInCaseOfException(source) && error) {
                                error = false;
                                try {
                                    attribute.incInvocation();
                                    start = StatProvider.getNanoTime();
                                    ls = attribute.getProceedInstance().proceed(
                                            source);
                                    delay = StatProvider.getNanoTime() - start;
                                    StatProvider.getStatEngine().probe(
                                            attribute, delay);
                                } catch (RuntimeException e2) {
                                    LOGGER.log(BasicLevel.ERROR,
                                            "Node exception during proceed on node"
                                                    + node.getName(), e);
                                    error = true;
                                }
                            }
                        }
                        process.notifyFWEvent(new FWEvent(
                                FWEvent.ENGINE_PROCEED, FWEvent.AFTER_PROCEED));
                        if (ls == null && LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            StringBuffer buffer = new StringBuffer(
                                    "Proceed of ");
                            buffer.append(getNodeName());
                            buffer.append(" has return null object");
                            LOGGER.log(BasicLevel.DEBUG, buffer.toString());
                        }
                        if (!error && ls != null) {
                            // put data to BufferOut
                            toBufferOut(ls, bufferOutManager, isRoutingProcess);
                            // end the transaction for the proceeded object
                            bufferOutManager
                                    .setIdCurrentTransaction(nextIdTransaction());
                        }
                        continue;
                    }
                    boolean error;
                    // proceed object coming from buffer in
                    while (source != null
                            && engineState.getState() == IEngineLifeCycle.STATE_STARTED
                            && wantedState.getState() != IEngineLifeCycle.STATE_STOPPED) {
                        error = false;
                        // get the proceed
                        process = attribute.getProceedInstance();
                        process
                                .notifyFWEvent(new FWEvent(
                                        FWEvent.ENGINE_PROCEED,
                                        FWEvent.BEFORE_PROCEED));
                        try {
                            attribute.incInvocation();
                            start = StatProvider.getNanoTime();
                            // call the proceed
                            ls = process.proceed(source);
                            delay = StatProvider.getNanoTime() - start;
                            StatProvider.getStatEngine().probe(attribute, delay);
                        } catch (RuntimeException e) {
                            LOGGER.log(BasicLevel.ERROR,
                                    "Node exception during proceed on node"
                                            + node.getName(), e);
                            error = true;
                            attribute.incNbrProceedError();
                        }
                        if (!error) {
                            // put data to BufferOut
                            if (ls != null) {
                                toBufferOut(ls, bufferOutManager,
                                        isRoutingProcess);
                            } else {
                                if (ls == null
                                        && LOGGER.isLoggable(BasicLevel.DEBUG)) {
                                    StringBuffer buffer = new StringBuffer(
                                            "Proceed of ");
                                    buffer.append(getNodeName());
                                    buffer.append(" has return null object");
                                    LOGGER.log(BasicLevel.DEBUG, buffer
                                            .toString());
                                }
                            }
                            // release the proceeded object
                            bufferIn.release(source);
                            // end the current object proceed transaction
                            bufferOutManager
                                    .setIdCurrentTransaction(nextIdTransaction());
                            // get the next object to proceed
                            source = bufferIn.getNextObject();
                        } else {
                            // An error occurred during proceed.
                            // If retryInCaseOfException refuses to retry, the
                            // next object is selected ; otherwise the same
                            // object is kept
                            if (retryInCaseOfException(source) == false) {
                                bufferIn.release(source);
                                source = bufferIn.getNextObject();
                            }
                        }
                        process = attribute.getProceedInstance();
                        process.notifyFWEvent(new FWEvent(
                                FWEvent.ENGINE_PROCEED, FWEvent.AFTER_PROCEED));
                    }
                }
            } catch (InterruptedException e) {
                // the end of this engine has been requested
                if (LOGGER.isLoggable(BasicLevel.INFO)) {
                    LOGGER.log(BasicLevel.INFO, "User interrupt proceed");
                }
            }
        } // end main "while" loop
    }

    /**
     * Action when an exception has occurred during the business process.
     * @param object the object to be re-processed
     * @return true if the policy wants to retry this object
     */
    private boolean retryInCaseOfException(Object object) {
        boolean result = false;
        if (attribute.getOnException() != null) {
            // ask the retryPolicy what to do
            try {
                if (attribute.getOnException().retry(object)) {
                    if (LOGGER.isLoggable(BasicLevel.INFO)) {
                        LOGGER.log(BasicLevel.INFO,
                                "Retry policy has decided to retry object "
                                        + object.toString());
                    }
                    result = true;
                } else {
                    if (LOGGER.isLoggable(BasicLevel.WARN)) {
                        LOGGER.log(BasicLevel.WARN,
                                "Retry policy has decided to evict object "
                                        + object.toString());
                    }
                    result = false;
                }
            } catch (RuntimeException e) {
                LOGGER.log(BasicLevel.ERROR,
                        "Retry policy has generated an exeption on "
                                + object.toString());
                LOGGER.log(BasicLevel.ERROR, "=> object automatically evicted");
                result = false;
            }
        } else {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "No retry policy is defined => object "
                                + object.toString() + " automatically evicted");
            }
        }
        return result;
    }

    /**
     * Selects the appropriate encoder.
     * @param name the name of the wanted encoder
     * @return the IEncoder associated with this name
     */
    IEncoder getEncoder(String name) {
        IEncoder encoder = null;
        encoder = encoderFactory.getEncoder(name);
        if (encoder == null) {
            StringBuffer errMsg = new StringBuffer(
                    "Unable to find encoder for ");
            errMsg.append(name);
            errMsg.append(" of ");
            errMsg.append(getNodeName());
            errMsg.append("'s node.");
            LOGGER.log(BasicLevel.ERROR, errMsg.toString());
        } else if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            StringBuffer buffer = new StringBuffer("use encoder ");
            buffer.append(encoder.getClass().getName());
            buffer.append(name);
            LOGGER.log(BasicLevel.DEBUG, buffer.toString());
        }
        return encoder;
    }

    /**
     * Increments transaction id.
     * @return the next transaction id.
     */
    private long nextIdTransaction() {
        transactionId++;
        if (transactionId > MAX_TRANSACTION_ID) {
            transactionId = MIN_TRANSACTION_ID;
        }
        return transactionId;
    }

    /**
     * Sets the current transaction id.
     * @param pTransactionId the new transaction id
     */
    public void setCurrentIdTransaction(long pTransactionId) {
        this.transactionId = pTransactionId;
    }

    /**
     * Gets the current id transaction.
     * @return the current transaction id.
     */
    private long getCurrentIdTransaction() {
        return transactionId;
    }

    // ------------------------------------------------
    // Inner class
    // ------------------------------------------------
    /**
     * Definition of ActiveProceed class.<br>
     * This class handles proceed auto-activation by setting a semaphore after defined interval (milliseconds)
     * @author Olivier Beyler
     */
    class ActiveProceed extends TimerTask {
        /** constant. */
        static final String CREATE_ACTIVE_PROCEED = "create ActiveProceed";

        /** constant. */
        static final String LAUNCH_TIMER = "launch timer";

        /**
         * Instantiates a new active proceed.
         */
        public ActiveProceed() {
            super();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, CREATE_ACTIVE_PROCEED);
            }
        }

        /**
         * Run.
         * @see java.util.TimerTask#run()
         */
        public void run() {
            if (attribute != null && attribute.getInterval() != 0) {
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG, LAUNCH_TIMER);
                }
                runOnce = true;
                if (sem != null) {
                    sem.release();
                }
            }
            if (attribute != null && attribute.getInterval() == 0) {
                timer.cancel();
            }
        }
    }

    /**
     * If the state is STATE_INITIALIZED the thread engine hasn't started yet so it start the thread if the other case the main think is to
     * do is to manage the interval.
     */
    public void launchComponent() {
        start();
        if (wakeUpDuringSuspendState) {
            sem.release();
        }
    }

    /**
     * Re launch component.
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#reLaunchComponent()
     */
    public void reLaunchComponent() {
        if (wakeUpDuringSuspendState) {
            sem.release();
        }
        IProceed process = attribute.getProceedInstance();
        if (process != null) {
            process.notifyFWEvent(new FWEvent(FWEvent.ENGINE_PROCEED,
                    FWEvent.RELAUNCH));
        }
    }

    /**
     * Inits the component.
     * @throws OdisEngineException the odis engine exception
     * @see com.francetelecom.odis.component.lifeCycle.engine.IEngineLifeCycle#initComponent()
     */
    public void initComponent() throws OdisEngineException {
        try {
            attribute.checkProperties();
        } catch (OdisAttributeException e1) {
            throw new OdisEngineException(e1);
        }
        boolean init = false;
        StringBuffer cause = null;
        try {
            init = attribute.getProceedInstance().init(attribute.getProperties());
            if (!init) {
                cause = new StringBuffer(" has returned false");
            }
        } catch (RuntimeException e) {
            cause = new StringBuffer(" has throwed a ");
            cause.append(e.getClass().getSimpleName());
            cause.append(EOL);
            cause.append("see stacktrace :");
            cause.append(EOL);
            StackTraceElement[] stack = e.getStackTrace();
            String thisClass = this.getClass().getName();
            for (int i = 0; i < stack.length; i++) {
                String a = stack[i].getClassName();
                if (a.equals(thisClass)) {
                    break;
                }
                cause.append("  ");
                cause.append(stack[i].toString());
                cause.append(EOL);
            }
        }
        if (!init) {
            StringBuffer buff = new StringBuffer(
                    "Init of the engine Proceed of node <");
            buff.append(node.getName());
            buff.append("> has failed");
            buff.append(EOL);
            buff.append("cause :");
            buff.append(attribute.getProceedClassName());
            buff.append(".init(String)");
            buff.append(cause.toString());
            throw new OdisEngineException(buff.toString());
        }
        schedulerTask();
    }

    /**
     * Scheduler task.
     */
    private void schedulerTask() {
        // to schedule task activeProceed
        if (activeProceed != null) {
            // cancel a scheduled task
            boolean res = activeProceed.cancel();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "activeProceed canceled = " + res);
            }
        }
        if (attribute.getInterval() > 0) {
            synchronized (this) {
                // recreate a new timer if need
                if (timer == null) {
                    timer = new Timer("TIMER_" + node.getName());
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "new timer = " + timer);
                    }
                }
                // schedule a new task
                activeProceed = new ActiveProceed();
                timer.schedule(activeProceed, 0, attribute.getInterval());
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "schedule new activeProceed = " + activeProceed);
                }
            }
        }
    }

    /**
     * suspend the engine.
     */
    public void suspendComponent() {
        wakeUpDuringSuspendState = false;
        IProceed process = attribute.getProceedInstance();
        if (process != null) {
            process.notifyFWEvent(new FWEvent(FWEvent.ENGINE_PROCEED,
                    FWEvent.SUSPEND));
        }
    }

    /**
     * Stops the engine.
     */
    public void stopComponent() {
        wantedState = new StateEngineStopped();
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "stop proceed of" + getNodeName());
        }
        // stop the timer
        attribute.setInterval(0);
        try {
            if (timer != null) {
                timer.cancel();
            }
        } catch (RuntimeException e1) {
            e1.printStackTrace();
        }
        if (Thread.currentThread() == this) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO,
                        "Stopping proceed from current Thread of "
                                + getNodeName());
            }
            // I'm the one asking for stopped state and I'm in the current
            // thread
            // => I can simply set the state
            engineState = new StateEngineStopped();
        } else {
            while (this.isAlive()) {
                try {
                    this.interrupt();
                    this.join(1000L);
                } catch (InterruptedException e) {
                    continue;
                }
            }
        }
        attribute.getProceedInstance().close();
        ProceedManager.getInstance().unregisterProceed(
                attribute.getProceedInstance());
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Stop proceed complete for "
                    + getNodeName());
        }
    }

    /**
     * Unloads the engine.
     */
    public void unloadComponent() {
        attribute = null;
        encoderFactory = null;
        timer = null;
    }

    // -------------------------------------------------------
    // Getter (and Setter) for each private attribute
    // -------------------------------------------------------
    /**
     * Getter for a Node.
     * @return Node witch contains this engineProceed.
     * @see com.francetelecom.odis.component.engine.AbstractEngine#getNode()
     */
    public Node getNode() {
        return node;
    }

    /**
     * Gets the attribute.
     * @return the engineProceedAttribut
     */
    public EngineProceedAttribute getAttribute() {
        return attribute;
    }

    /**
     * Gets the node name.
     * @return node name
     * @see com.francetelecom.odis.component.engine.IEngineProceed#getNodeName()
     */
    public String getNodeName() {
        return getNode().getName();
    }

    /**
     * Gets the encoder factory.
     * @return the encoder factory
     */
    public EncoderFactory getEncoderFactory() {
        return encoderFactory;
    }

    /**
     * Gets the behavior.
     * @return the behavior
     */
    public Behavior getBehavior() {
        return behavior;
    }

    /**
     * Checks if is run once.
     * @return true, if is run once
     */
    public boolean isRunOnce() {
        return runOnce;
    }

    /**
     * Sets the run once.
     * @param pRunOnce the new run once
     */
    public void setRunOnce(boolean pRunOnce) {
        this.runOnce = pRunOnce;
    }
}
