package org.ow2.odis.jmx;

import java.lang.reflect.InvocationTargetException;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * JMX allow user to use JMX administration. Implement Singleton design pattern. All this class is make by using introspection method to be
 * JDK 1.4 compliant and give the opportunity of JDK 1.5 users to use JMX.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class JMX {
    /**
     * logger of this class.
     */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            JMX.class.getName());

    /**
     * implement singleton design pattern.
     */
    static private JMX instance = new JMX();

    /**
     * Constant.
     */
    static final private String JMX_DESACTIVATED = "JMX is desactived ";

    /**
     * Constant.
     */
    static final private String CAUSE = "cause ";

    /**
     * true =No error has occured during JMX setup. if true the JMX instance/methode can be use.
     */
    static private boolean useJMX = true;

    /**
     * represent a constructor for ObjectName class (JDK 1.5)
     */
    private java.lang.reflect.Constructor constructor = null;

    /**
     * Register method object.
     */
    private java.lang.reflect.Method methodeRegister = null;

    /**
     * Unregister method object.
     */
    private java.lang.reflect.Method methodeUnregister = null;

    /**
     * represent an PlateformMBeanServer in case of JDK >1.4.
     */
    private Object plateformMBeanServer = null;

    /**
     * implement Singleton pattern.
     * @return unique instance of JMX class
     */
    public static JMX getInstance() {
        return instance;
    }

    /**
     * destroy JMX singleton. In fact the JMX is only replace by a new one. Oldest is unreferenced.
     */
    public static void destroy() {
        instance = new JMX();
    }

    /**
     * Private Constructor.
     */
    private JMX() {
        super();
        try {
            Class factory = Class
                    .forName("java.lang.management.ManagementFactory");
            // $ANALYSIS-IGNORE,codereview.java.rules.reflection.
            // RuleReflectionGetMethod
            java.lang.reflect.Method methode = factory.getMethod(
                    "getPlatformMBeanServer", (Class[]) null);
            plateformMBeanServer = methode.invoke(factory, (Object[]) null);
            Class[] paramRegister = new Class[2];
            paramRegister[0] = Object.class;
            paramRegister[1] = Class.forName("javax.management.ObjectName");
            Class[] typeParamConstructor = new Class[1];
            typeParamConstructor[0] = String.class;
            constructor = paramRegister[1].getConstructor(typeParamConstructor);
            // $ANALYSIS-IGNORE,codereview.java.rules.reflection.
            // RuleReflectionGetMethod
            methodeRegister = plateformMBeanServer.getClass().getMethod(
                    "registerMBean", paramRegister);
            Class[] paramUnregister = new Class[1];
            paramUnregister[0] = Class.forName("javax.management.ObjectName");
            // $ANALYSIS-IGNORE,codereview.java.rules.reflection.
            // RuleReflectionGetMethod
            methodeUnregister = plateformMBeanServer.getClass().getMethod(
                    "unregisterMBean", paramUnregister);
        } catch (RuntimeException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, JMX_DESACTIVATED);
            }
            useJMX = false;
        } catch (ClassNotFoundException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, JMX_DESACTIVATED);
            }
            useJMX = false;
        } catch (NoSuchMethodException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, JMX_DESACTIVATED);
            }
            useJMX = false;
        } catch (IllegalAccessException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, JMX_DESACTIVATED);
            }
            useJMX = false;
        } catch (InvocationTargetException e) {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, JMX_DESACTIVATED);
            }
            useJMX = false;
        }
    }

    /**
     * register the specified MBean with Name identity
     * @param name String represent MBean identity
     * @param mbean instance of Mbean
     */
    public void registerMBean(String name, Object mbean) {
        if (useJMX) {
            try {
                Object[] paramConstructor = new String[1];
                String chaine = mbean.getClass().getName() + ":type=" + name;
                paramConstructor[0] = chaine;
                Object objectName = constructor.newInstance(paramConstructor);
                Object[] registerParams = new Object[2];
                registerParams[0] = mbean;
                registerParams[1] = objectName;
                methodeRegister.invoke(plateformMBeanServer, registerParams);
            } catch (RuntimeException e) {
                LOGGER
                        .log(BasicLevel.ERROR, JMX_DESACTIVATED + CAUSE + name,
                                e);
                useJMX = false;
            } catch (InstantiationException e) {
                LOGGER
                        .log(BasicLevel.ERROR, JMX_DESACTIVATED + CAUSE + name,
                                e);
                useJMX = false;
            } catch (IllegalAccessException e) {
                LOGGER
                        .log(BasicLevel.ERROR, JMX_DESACTIVATED + CAUSE + name,
                                e);
                useJMX = false;
            } catch (InvocationTargetException e) {
                if (e.getCause().getClass().getName().endsWith(
                        "InstanceAlreadyExistsException")) {
                    useJMX = true;
                } else {
                    LOGGER.log(BasicLevel.ERROR, JMX_DESACTIVATED + CAUSE
                            + name, e);
                    useJMX = false;
                }
            }
        }
    }

    /**
     * Unregister the specified MBean with Name identity
     * @param name String represent MBean identity
     * @param mbean instance of Mbean
     */
    public void unregisterMBean(String name, Object mbean) {
        if (useJMX) {
            try {
                Object[] paramConstructor = new String[1];
                String chaine = mbean.getClass().getName() + ":type=" + name;
                paramConstructor[0] = chaine;
                Object objectName = constructor.newInstance(paramConstructor);
                Object[] unregisterParams = new Object[1];
                unregisterParams[0] = objectName;
                methodeUnregister
                        .invoke(plateformMBeanServer, unregisterParams);
            } catch (Exception e) {
                LOGGER.log(BasicLevel.WARN, "unable to unregister " + name, e);
            }
        }
    }
}
