package org.ow2.odis.lifeCycle;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.lifeCycle.engine.AbstractEngineState;
import org.ow2.odis.lifeCycle.engine.IEngineLifeCycle;
import org.ow2.odis.lifeCycle.engine.StateEngineStarted;
import org.ow2.odis.lifeCycle.engine.StateEngineSuspended;

/**
 * aim : in charge of collecting Cause witch can influence on component behavior. It can decide himself to suspend or restart component to
 * do it It collect all cause and do a distinction between cause request of an user and cause by the system an "human" cause cannot be
 * remove by the system. <br>
 * behavior is only active if the state of component is suspended or started
 * @author Beyler Olivier FranceTelecom R&D
 */
public final class Behavior {
    /**
     * class log.
     */
    private static final Logger LOGGER = Monolog.initialize().getLogger(
            Behavior.class.getName());

    /**
     * list of automatic cause.
     */
    List lsCause = null;

    /**
     * define hight level of user cause.
     */
    Cause userCause = null;

    /**
     * object who has delegate his behavior to this class.
     */
    IEngineLifeCycle object = null;

    /**
     * constructor.
     * @param object object who delegates his behavior.
     */
    public Behavior(final IEngineLifeCycle object) {
        super();
        lsCause = Collections.synchronizedList(new ArrayList());
        this.object = object;
    }

    /**
     * add a cause to the list if the cause is set by a user action only the most important is kept.
     * @param cause new cause
     */
    public synchronized void addCause(final Cause cause) {
        if (cause.getValue() == Cause.USER_REQUEST) {
            Cause lastUserCause = getUserCause();
            if (lastUserCause == null) {
                userCause = cause;
            } else if (lastUserCause.getUserLevelRequest() > cause
                    .getUserLevelRequest()) {
                userCause = cause;
            }
        } else {
            lsCause.add(cause);
        }
        checkBehaviours();
    }

    /**
     * remove the cause set by an user if the cause has the same or is less important than the usercause stored.
     * @param cause cause who need to be removed.
     */
    public synchronized void removeUserCause(final Cause cause) {
        if ((userCause != null)
                && (cause.getValue() == Cause.USER_REQUEST)
                && (cause.getUserLevelRequest() <= userCause
                        .getUserLevelRequest())) {
            userCause = null;
        }
        checkBehaviours();
    }

    /**
     * check the present of cause set by an user action.
     * @return a hight level Cause generated by an user
     */
    private Cause getUserCause() {
        return userCause;
    }

    /**
     * Do an iteration on each cause to see if some cause have to be remove.
     */
    private synchronized void purgeCause() {
        List lsRemove = new ArrayList();
        Iterator itCause = lsCause.iterator();
        // recupere la liste des causes qui sont terminees
        while (itCause.hasNext()) {
            Cause cause = (Cause) itCause.next();
            if (cause.getEnd() != -1) {
                lsRemove.add(cause);
            }
        }
        // Enleve les causes terminees de la liste des causes a prendre en
        // compte
        while (!lsRemove.isEmpty()) {
            lsCause.remove(lsRemove.remove(0));
        }
        if (userCause != null && userCause.getEnd() != -1) {
            userCause = null;
        }
    }

    /**
     * check if the object need to change is life state.
     */
    public void checkBehaviours() {
        purgeCause();
        AbstractEngineState state;
        int engineState = object.getEngineState().getState();
        try {
            if ((engineState == IEngineLifeCycle.STATE_SUSPENDED)
                    || (engineState == IEngineLifeCycle.STATE_STARTED)) {
                if (lsCause.isEmpty() && userCause == null) {
                    state = object.getEngineState();
                    state.setLifeState(object, new StateEngineStarted());
                } else {
                    state = object.getEngineState();
                    state.setLifeState(object, new StateEngineSuspended());
                }
            }
        } catch (OdisEngineException e) {
            LOGGER.log(BasicLevel.ERROR, "Unexpected exception !", e);
        }
    }
}
