package org.ow2.odis.lifeCycle.engine;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.model.AbstractAttribute;

/**
 * Base class for all engine states. <i>Follows State design pattern.</i><br>
 * <img src="doc-files/EngineLifecycle.png"/>
 * @author Beyler Olivier FranceTelecom R&D
 */
public abstract class AbstractEngineState {
    /** Logger for this class. */
    final static Logger LOGGER = Monolog.initialize().getLogger(
            AbstractEngineState.class.getName());

    /**
     * An int value identifying this state.<br>
     * This should be set by the constructor of each state.
     */
    protected int state;

    /**
     * Gets the identifier value for this state.
     * @return the state
     */
    public final int getState() {
        return state;
    }

    /**
     * Gets the state name.
     * @return a String representing the state in human language
     */
    public final String getStateName() {
        return IEngineLifeCycle.STATE_NAMES[getState()];
    }

    /**
     * Transition method allowing a change of life state.<br>
     * This method should check whether the transition to the wanted state is allowed by the engine life-cycle and call the appropriate
     * action method provided by IEngineLifeCycle.
     * @param engine the engine
     * @param wantedState the wanted state
     * @throws OdisEngineException if the transition is not allowed or the actions fail
     */
    public abstract void setLifeState(IEngineLifeCycle engine,
            AbstractEngineState wantedState) throws OdisEngineException;

    /**
     * Transition method allowing a change of life state when an attribute is needed to provide parameters.<br>
     * This method should check whether the transition to the wanted state is allowed by the engine life-cycle and call the appropriate
     * action method provided by IEngineLifeCycle.<br>
     * It should not be implemented when there is no need for an attribute : an exception is raised by default.
     * @param engine the engine
     * @param wantedState the wanted state
     * @param attribute the attribute holding any needed parameters
     * @throws OdisEngineException if the attribute is irrelevant (default), the transition is not allowed or the actions fail
     */
    public void setLifeState(final IEngineLifeCycle engine,
            final AbstractEngineState wantedState,
            final AbstractAttribute attribute) throws OdisEngineException {
        invalidTreatment(engine, wantedState);
    }

    /**
     * Logs a transition failure and raises an exception.<br>
     * This method should be called whenever a transition fails.
     * @param engine the engine
     * @param wantedState the wanted state
     * @throws OdisEngineException
     */
    public void invalidTreatment(final IEngineLifeCycle engine,
            final AbstractEngineState wantedState) throws OdisEngineException {
        StringBuffer buff = new StringBuffer(engine.getClass().getName());
        buff.append(" cannot be set to ");
        buff.append(wantedState.getStateName());
        buff.append(" from ");
        buff.append(getStateName());
        LOGGER.log(BasicLevel.ERROR, buff.toString());
        throw new OdisEngineException(buff.toString());
    }
}
