package org.ow2.odis.lifeCycle.engine;

import org.ow2.odis.exception.OdisEngineException;

/**
 * The Interface IEngineLifeCycle defines the available states and transitions for an ODIS engine. <i>Follows State design pattern.</i>
 * @author Beyler Olivier FranceTelecom R&D
 */
public interface IEngineLifeCycle {
    /** The LOADED state : initial state when the engine is created. */
    int STATE_LOADED = 0;

    /**
     * The INITIALIZED state : the engine properties have been found valid ; the engine is now linked to all its partners and ready to run.
     */
    int STATE_INITIALIZED = STATE_LOADED + 1;

    /** The STARTED state : the engine runs. */
    int STATE_STARTED = STATE_INITIALIZED + 1;

    /**
     * The SUSPENDED state : the engine hasn't released any resources but no processing happens.
     */
    int STATE_SUSPENDED = STATE_STARTED + 1;

    /** The STOPPED state : processing has stopped and will not start again. */
    int STATE_STOPPED = STATE_SUSPENDED + 1;

    /** The UNLOADED state : all resources are released. */
    int STATE_UNLOADED = STATE_STOPPED + 1;

    /** The state names in human language. */
    String[] STATE_NAMES = new String[] { "LOADED", "INITIALIZED", "STARTED",
            "SUSPENDED", "STOPPED", "UNLOADED" };

    /**
     * Gets the current engine state.
     * @return the engine state
     */
    AbstractEngineState getEngineState();

    /**
     * Sets the engine state.<br>
     * This is a plain setter. It should not trigger other actions than actually changing the state value. The engine partners use the
     * transition method on the current state to require a state change :
     * <code>AbstractEngineState.setLifeState(engine, wantedState)</code>
     * @param state AbstractEngineState
     */
    void setEngineState(AbstractEngineState state);

    /**
     * Initializes the component.<br>
     * Carries out the required actions for a transition between LOADED state and INITIALIZED state. The life cycle pattern takes care of
     * checking the current state and actually changing the state ; this should not be done by this method.
     * @throws OdisEngineException if initializing the component cannot be carried out properly
     */
    void initComponent() throws OdisEngineException;

    /**
     * Launches the component.<br>
     * Carries out the required actions for a transition between INITIALIZED state and STARTED state. The life cycle pattern takes care of
     * checking the current state and actually changing the state ; this should not be done by this method.
     * @throws OdisEngineException if starting the component cannot be carried out properly
     */
    void launchComponent() throws OdisEngineException;

    /**
     * Re-launches the component.<br>
     * Carries out the required actions for a transition between SUSPENDED state and STARTED state. The life cycle pattern takes care of
     * checking the current state and actually changing the state ; this should not be done by this method.<br>
     * This method should never fail.
     */
    void reLaunchComponent();

    /**
     * Suspends the component.<br>
     * Carries out the required actions for a transition between STARTED state and SUSPENDED state. The life cycle pattern takes care of
     * checking the current state and actually changing the state ; this should not be done by this method. This method should never fail.
     */
    void suspendComponent();

    /**
     * Stops the component.<br>
     * Carries out the required actions for a transition between LOADED, INITIALIZED or SUSPENDED state and STOPPED state. The life cycle
     * pattern takes care of checking the current state and actually changing the state ; this should not be done by this method.
     * @throws OdisEngineException if stopping the component cannot be carried out properly
     */
    void stopComponent() throws OdisEngineException;

    /**
     * Unloads the component.<br>
     * Carries out the required actions for a transition between STOPPED state and UNLOADED state. The life cycle pattern takes care of
     * checking the current state and actually changing the state ; this should not be done by this method.
     * @throws OdisEngineException if unloading the component cannot be carried out properly
     */
    void unloadComponent() throws OdisEngineException;
}
