package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Observable;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.loader.GenericFactory;

/**
 * base class of Attribute.
 * @author Beyler Olivier FranceTelecom R&D
 */
public abstract class AbstractAttribute extends Observable {
    /** logger of the attribute class. */
    protected static final Logger LOGGER = Monolog.initialize().getLogger(
            AbstractAttribute.class.getName());

    /** keep a reference on the parent. */
    protected AbstractAttribute parent;

    /** version of the element class. */
    protected String classVersion = GenericFactory.DEFAULT_VERSION;

    /**
     * Constructor.
     * @param parent AbstractAttribut
     */
    public AbstractAttribute(AbstractAttribute parent) {
        this.parent = parent;
    }

    /**
     * @param out OutPutStream
     * @param value the version value
     * @param full
     * @throws IOException
     */
    protected void writeXMLVersion(OutputStream out, String value, boolean full)
            throws IOException {
        if ((value != null) &&
                (full || !value.equals(GenericFactory.DEFAULT_VERSION))) {
            out.write(Const.VERSION.getBytes());
            out.write(" =\"".getBytes());
            out.write(value.getBytes());
            out.write("\" ".getBytes());
        }
    }

    /**
     * utilities.
     * @param out OutPutStream
     * @param level int
     * @param property String
     * @param value String
     * @param def String (default value)
     * @param full boolean
     * @throws IOException if stream is broken
     */
    protected void writeXMLUtilities(OutputStream out, int level,
            String property, String value, String def, boolean full)
            throws IOException {
        if (full) {
            for (int i = 0; i < level; i++) {
                out.write(Const.bTAB);
            }
            out.write("<".getBytes());
            if (value == null) {
                out.write((property + "/>" + Const.EOL).getBytes());
            } else {
                out.write((property + " " + Const.VALUE + " = \"" + value
                        + "\"/>" + Const.EOL).getBytes());
            }
        } else {
            if (value != null && !value.equals(def)) {
                for (int i = 0; i < level; i++) {
                    out.write(Const.bTAB);
                }
                out.write("<".getBytes());
                out.write(property.getBytes());
                out.write(" ".getBytes());
                out.write(Const.VALUE.getBytes());
                out.write(" = \"".getBytes());
                out.write(value.getBytes());
                out.write("\"/>".getBytes());
                out.write(Const.EOL.getBytes());
            }
        }
    }

    /**
     * utility function to help writing xml describing a connection.
     * @param out OutputStream
     * @param level int (number of tabulation)
     * @param property String
     * @param value String
     * @param def String (default value)
     * @param full boolean (the xml can be full or without default value)
     * @throws IOException in case of fail
     */
    protected void writeXMLCNXUtilities(OutputStream out, int level,
            String property, String value, String def, boolean full)
            throws IOException {
        if (full) {
            for (int i = 0; i < level; i++) {
                out.write(Const.bTAB);
            }
            out.write("<".getBytes());
            if (value == null) {
                out.write((property + Const.VALUE + "=\"\"/>" + Const.EOL)
                        .getBytes());
            } else {
                out.write((property + " " + Const.VALUE + " = \"" + value
                        + "\"/>" + Const.EOL).getBytes());
            }
        } else {
            if (value != null && !value.equals(def)) {
                for (int i = 0; i < level; i++) {
                    out.write(Const.bTAB);
                }
                out.write("<".getBytes());
                out.write(property.getBytes());
                out.write(" ".getBytes());
                out.write(Const.VALUE.getBytes());
                out.write(" = \"".getBytes());
                out.write(value.getBytes());
                out.write("\"/>".getBytes());
                out.write(Const.EOL.getBytes());
            }
        }
    }

    /**
     * utility function to help writing xml describing a connection.
     * @param out OutputStream
     * @param level int (number of tabulation)
     * @param property String
     * @param attributName
     * @param attributValue
     * @param content
     * @throws IOException in case of fail
     */
    protected void writeXMLCNXUtilities(OutputStream out, int level,
            String property, String attributName, String attributValue,
            String content) throws IOException {
        for (int i = 0; i < level; i++) {
            out.write(Const.bTAB);
        }
        out.write("<".getBytes());
        if (attributValue == null) {
            out.write((property + attributName + "=\"\">").getBytes());
        } else {
            out
                    .write((property + " " + attributName + " = \""
                            + attributValue + "\">").getBytes());
        }
        if (content != null) {
            out.write((content).getBytes());
        }
        out.write(("</" + property + ">" + Const.EOL).getBytes());
    }

    /**
     * utility function to help writing xml describing a connection.
     * @param out OutputStream
     * @param level int (number of tabulation)
     * @param property String
     * @param endTag boolean (the xml tag is a end tag)
     * @throws IOException in case of fail
     */
    protected void writeXMLCNXUtilities(OutputStream out, int level,
            String property, boolean endTag) throws IOException {
        writeXMLCNXUtilities(out, level, property, endTag, null, null);
    }

    /**
     * utility function to help writing xml describing a connection.
     * @param out OutputStream
     * @param level int (number of tabulation)
     * @param property String
     * @param endTag boolean (the xml tag is a end tag)
     * @param attributeName
     * @param attributeValue
     * @throws IOException in case of fail
     */
    protected void writeXMLCNXUtilities(OutputStream out, int level,
            String property, boolean endTag, String attributeName,
            String attributeValue) throws IOException {
        for (int i = 0; i < level; i++) {
            out.write(Const.bTAB);
        }
        if (endTag)
            out.write("</".getBytes());
        else
            out.write("<".getBytes());
        out.write((property).getBytes());
        if (attributeName != null) {
            out.write((" " + attributeName + " = \"" + attributeValue + "\"")
                    .getBytes());
        }
        out.write((">" + Const.EOL).getBytes());
    }

    /**
     * utility function to help writing xml describing a node.
     * @param out OutputStream
     * @param level int (number of tabulation)
     * @param property String
     * @param value String
     * @param def String (default value)
     * @param full boolean (the xml can be full or without default value)
     * @throws IOException in case of fail
     */
    protected void writeXMLUtilities(OutputStream out, int level,
            String property, long value, long def, boolean full)
            throws IOException {
        if (full) {
            for (int i = 0; i < level; i++) {
                out.write(Const.bTAB);
            }
            out.write("<".getBytes());
            out
                    .write((property + " " + Const.VALUE + " = \"" + value
                            + "\"/>" + Const.EOL).getBytes());
        } else {
            if (value > def) {
                for (int i = 0; i < level; i++) {
                    out.write(Const.bTAB);
                }
                out.write("<".getBytes());
                out.write(property.getBytes());
                out.write(" ".getBytes());
                out.write(Const.VALUE.getBytes());
                out.write(" = \"".getBytes());
                out.write(("" + value).getBytes());
                out.write("\"/>".getBytes());
                out.write(Const.EOL.getBytes());
            }
        }
    }

    /**
     * write on stream the XML description of the config without optional value.
     * @param out OutputStream
     * @throws IOException if stream is broken
     */
    public void writeXML(OutputStream out) throws IOException {
        writeXML(out, false);
    }

    /**
     * write on stream the XML description of the config with or without optional value.
     * @param out OutputStream
     * @param full boolean
     * @throws IOException if stream is broken
     */
    public abstract void writeXML(OutputStream out, boolean full)
            throws IOException;

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     */
    public abstract void checkProperties() throws OdisAttributeException;

    /**
     * Merge exception to create a new BufferString with all information need to debug application.
     * @param buff StringBuffer
     * @param property String
     * @param value String
     * @param e Exception
     * @return StringBuffer new StringBuffer
     */
    protected StringBuffer concatException(StringBuffer buff,
            final String property, final String value, final Exception e) {
        StringBuffer result;
        if (buff == null) {
            result = new StringBuffer("");
        } else {
            result = buff;
            result.append(System.getProperty("line.separator"));
        }
        if (e != null) {
            result.append(e.getMessage());
        }
        result.append("property ");
        result.append(property);
        result.append(" with \"");
        result.append(value);
        result.append("\" value is invalid.");
        return result;
    }

    /**
     * Gets the parent.
     * @return AbstractAttribut
     */
    public final AbstractAttribute getParent() {
        return parent;
    }

    /**
     * Gets the P name.
     * @return String
     */
    public abstract String getPName();
}
