package org.ow2.odis.model;

import java.io.ByteArrayOutputStream;
import java.io.IOException;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * Define the attribute of a connection. This class must be overloaded.
 * @version 1.0
 * @author Beyler Olivier FranceTelecom R&D
 */
public abstract class AbstractConnectionAttribute extends AbstractAttribute {
    /**
     * Constructor.
     * @param parent AbstractAttribut
     */
    public AbstractConnectionAttribute(AbstractAttribute parent) {
        super(parent);
    }

    /** name of a connection. */
    private String name = "";

    /**
     * number of port IN or Out register to this connection. the aim of this counter is to know when the connection have to be remove from
     * the JMX MBean exposed
     */
    private int registerCount = 0;

    /**
     * getter.
     * @return int
     */
    public int getRegisterCount() {
        return registerCount;
    }

    /**
     * increase the register counter.
     */
    public void increaseRegisterCount() {
        registerCount++;
    }

    /**
     * decrease the register counter.
     */
    public void decreaseRegisterCount() {
        registerCount--;
    }

    /**
     * Gets the connection identity.
     * @return Returns the name.
     */
    public String getName() {
        return name;
    }

    /**
     * Sets the connection identity.
     * @param pName The name to set.
     */
    public void setName(String pName) {
        name = pName;
    }

    /**
     * give the type of connection.
     * @return the type of connection
     */
    public abstract String getTypeConnection();

    /**
     * Setter that uses the JDom element to construct this attribute.
     * @param declaration Element
     */
    public void setter(Element declaration) {
        name = declaration.getAttributeValue("name", "");
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, Const.LINE_SEP);
            LOGGER.log(BasicLevel.INFO, "CONNECTION  " + getTypeConnection() + ": "
                    + name);
            LOGGER.log(BasicLevel.INFO, Const.LINE_SEP);
        }
        subSetter(declaration);
    }

    /**
     * Complementary setter that uses the JDom element to complete construction of this attribute as needed by the implementing class.
     * @param declaration Element
     * @see #setter(Element)
     */
    public abstract void subSetter(Element declaration);

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also an
     * AbstractCnxAttribut and has the same name. Further checking is to be implemented by derived classes.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is named like <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        if (obj != null) {
            AbstractConnectionAttribute copy;
            try {
                copy = (AbstractConnectionAttribute) obj;
            } catch (ClassCastException e) {
                return false;
            }
            result = (getName() == null && copy.getName() == null)
                    || getName().equals(copy.getName());
        }
        return result;
    }

    /**
     * Trace on when a conflictual connection is found.
     * @param copy
     */
    protected void traceOnConfictualConnection(AbstractAttribute copy) {
        traceOnConfictualConnection(this, copy);
    }

    /**
     * @param a
     * @param b
     */
    protected static void traceOnConfictualConnection(AbstractConnectionAttribute a,
            AbstractAttribute b) {
        if (LOGGER.isLoggable(BasicLevel.WARN)) {
            try {
                StringBuffer buffer = new StringBuffer("2 declarations of ");
                buffer.append(a.getName());
                buffer.append(" connection are not equal : ");
                buffer.append(Const.EOL);
                ByteArrayOutputStream byteStream;
                byteStream = new ByteArrayOutputStream();
                a.writeXML(byteStream, true);
                buffer.append(byteStream.toString());
                byteStream = new ByteArrayOutputStream();
                b.writeXML(byteStream, true);
                buffer.append(Const.EOL);
                buffer.append(" versus ");
                buffer.append(Const.EOL);
                buffer.append(byteStream.toString());
                LOGGER.log(BasicLevel.WARN, buffer.toString());
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
    }

    /**
     * @return String
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buff = new StringBuffer("AbstractCnxAttribut(");
        buff.append(getName());
        buff.append(")@");
        buff.append(hashCode());
        return buff.toString();
    }
}
