package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.List;
import java.util.Timer;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.bufferOut.BufferOut;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.util.CheckUtilities;

/**
 * A <code>BufferOutAttribut</code> contains all informations about a <code>BufferOut</code> instance.<br>
 * All parameters must be set before the <code>BufferOut</code> is initialized.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class BufferOutAttribute extends AbstractAttribute {
    /** The string value of the &lt;CLEAR_WHEN/&gt; tag. */
    public String strClear = Const.VALUE_PEER_LOST_EMPTY;

    /**
     * Constructor.
     * @param parent AbstractAttribut
     */
    public BufferOutAttribute(AbstractAttribute parent) {
        super(parent);
    }

    /**
     * number object max present into the buffer <br>
     * 0 mean unlimited <br>
     * if the number of object present into lsSynchronizedObject is equal or greater no more object will be added.
     */
    private int numberObjectMax = 5000;

    /**
     * String witch represent number object max present into the buffer <br>
     * no check is done on this string.
     */
    private String strNumberObjectMax = "5000";

    /** type of object present into bufferOut. */
    protected String typeObject = "";

    /**
     * number of object witch can be extract from the lsSynchronizedObject to be sended <br>
     * 0 mean unlimited.
     */
    private int thresholdNumberOfObject = 0;

    /** The str threshold number of object. */
    private String strThresholdNumberOfObject = "0";

    /**
     * number of octet max present into the buffer <br>
     * 0 mean unlimited <br>
     * if the size of lsSynchronizedObject is reached no more object will be added.
     */
    private long sizeByteMax = 10000000;

    /** The str size byte max. */
    private String strSizeByteMax = "10000000";

    /** number of object before required sending buffer. */
    private int signalThresholdNumberOfObject = 1;

    /** The str signal threshold number of object. */
    private String strSignalThresholdNumberOfObject = "1";

    /** max elaps time before required sending buffer. */
    protected long maxElapsTime = 0;

    /** The str max elaps time. */
    private String strMaxElapsTime = "0";

    /** cyclic time before required sending buffer. */
    protected long cyclicTime = 0;

    /** The string cyclic time. */
    private String strCyclicTime = "0";

    /** The routing. */
    private String routing = "";

    /** The str persistant. */
    private String strPersistant = null;

    /** The is persistant. */
    private boolean isPersistant = false;

    /** The string value of the &lt;USE_CONTEXT/&gt; tag. */
    private String strContextual = null;

    /** True if the buffer is contextual. */
    private boolean isContextual = false;

    /** The priority. */
    private int priority = 1000;

    /** The str priority. */
    private String strPriority = "1000";

    /**
     * The string value of the delay (in ms) associated with the &lt;CLEAR_WHEN/&gt; tag.
     */
    private String strClearDelay = null;

    /**
     * The int value of the delay (in ms) associated with the &lt;CLEAR_WHEN/&gt; tag.
     */
    private long clearDelay = 0;

    /** The state. */
    private int state = Const.READY;

    /** The buffer out. */
    private BufferOut bufferOut = null;

    /** Number of messages in the bufferOut. */
    private int nbMsg = 0;

    /** Actual size in bytes of all messages in the bufferOut. */
    private long sizeByte = 0;

    /**
     * Gets the size byte.
     * @return the size byte
     */
    public long getSizeByte() {
        return sizeByte;
    }

    /**
     * Gets the number of messages.
     * @return the size byte
     */
    public int getNbMsg() {
        return nbMsg;
    }

    /**
     * Updates internal counters in number of messages and bytes.<br>
     * If the value is positive, one message is considered added. If it is negative, one message is considered removed. If the value is
     * zero, no message is counted.
     * @param value the size change in bytes
     */
    public void incSizeByte(long value) {
        sizeByte += value;
        if (value > 0) {
            nbMsg++;
        } else if (value < 0) {
            nbMsg--;
        }
    }

    /** Actual size of transaction present into the bufferOut. */
    public int sizeCurrentTransaction = 0;

    /** The timer. */
    protected Timer timer = null;

    /**
     * Initializes all properties (in String form) for the buffer, based on a JDOM element.
     * @param declaration JDOM base element of a bufferOut declaration
     */
    public void setter(Element declaration) {
        List lsUnusedProperties = declaration.getChildren();
        if (declaration != null) {
            // PATCH to assume 2.6 compatibility
            typeObject = declaration.getAttributeValue(Const.PROPERTY_TYPE);
            // END PATCH
            Element element;
            element = declaration.getChild(Const.PROPERTY_NUMBER_OBJECT_MAX);
            if (element != null) {
                strNumberObjectMax = element.getAttributeValue(Const.VALUE,
                        strNumberObjectMax);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_SIZE_BYTE_MAX);
            if (element != null) {
                strSizeByteMax = element.getAttributeValue(Const.VALUE,
                        strSizeByteMax);
                lsUnusedProperties.remove(element);
            }
            element = declaration
                    .getChild(Const.PROPERTY_SIGNAL_THRESHOLD_NUMBER_OBJECT);
            if (element != null) {
                strSignalThresholdNumberOfObject = element.getAttributeValue(
                        Const.VALUE, strSignalThresholdNumberOfObject);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_MAX_ELAPS_TIME);
            if (element != null) {
                strMaxElapsTime = element.getAttributeValue(Const.VALUE,
                        strMaxElapsTime);
                lsUnusedProperties.remove(element);
            }
            element = declaration
                    .getChild(Const.PROPERTY_THRESHOLD_NUMBER_OBJECT);
            if (element != null) {
                strThresholdNumberOfObject = element.getAttributeValue(
                        Const.VALUE, strThresholdNumberOfObject);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_CYCLIC_TIME);
            if (element != null) {
                strCyclicTime = element.getAttributeValue(Const.VALUE,
                        strCyclicTime);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_USE_PERSISTANCE);
            if (element != null) {
                strPersistant = element.getAttributeValue(Const.VALUE,
                        strPersistant);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_CONTEXTUAL);
            if (element != null) {
                strContextual = element.getAttributeValue(Const.VALUE,
                        strContextual);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_ROUTING);
            if (element != null) {
                routing = element.getAttributeValue(Const.VALUE, routing);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_PRIORITY);
            if (element != null) {
                strPriority = element.getAttributeValue(Const.VALUE,
                        strPriority);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_TYPE);
            if (element != null) {
                typeObject = element.getAttributeValue(Const.VALUE, "");
                lsUnusedProperties.remove(element);
            }
            // Handle CLEAR_WHEN tag only if contextual
            if ("yes".equalsIgnoreCase(strContextual)) {
                element = declaration.getChild(Const.PROPERTY_CLEAR);
                if (element != null) {
                    strClear = element.getAttributeValue(Const.VALUE, strClear)
                            .toUpperCase();
                    strClearDelay = element.getAttributeValue(Const.DELAY, "");
                    lsUnusedProperties.remove(element);
                }
            } else
                CheckUtilities
                        .checkUnusedProperties(lsUnusedProperties, LOGGER);
        }
    }

    /**
     * Gets the id.
     * @return the routing name if defined, else the data type (class name)
     */
    public String getId() {
        String result;
        if ("".equals(getRouting()) || getRouting() == null) {
            result = getTypeObject();
        } else {
            result = getRouting();
        }
        return result;
    }

    /**
     * Gets the size byte max.
     * @return the size byte max
     */
    public long getSizeByteMax() {
        return sizeByteMax;
    }

    /**
     * Gets the size object max.
     * @return the size object max
     */
    public int getSizeObjectMax() {
        return numberObjectMax;
    }

    /**
     * fix the number entry of the buffer before sending an event to the EnginOut.
     * @param pThresholdObject the threshold object
     */
    protected void setSignalThresholdObject(int pThresholdObject) {
        if ((numberObjectMax < pThresholdObject && numberObjectMax > 0) &&
                (LOGGER.isLoggable(BasicLevel.WARN))) {
            LOGGER.log(BasicLevel.WARN,"ThresholdObject "
                + pThresholdObject
                + " will never be reached with the current buffer'size limit:"
                + numberObjectMax);
        }        
        signalThresholdNumberOfObject = pThresholdObject;
        strSignalThresholdNumberOfObject = "" + pThresholdObject;
    }

    /**
     * fix the maximum number entry of the buffer. <br>
     * 0 or less means no limits
     * @param pSizeObjectMax the size object max
     */
    protected void setSizeObjectMax(final int pSizeObjectMax) {
        if ((numberObjectMax < signalThresholdNumberOfObject
                && numberObjectMax > 0) && LOGGER.isLoggable(BasicLevel.WARN)) {
                LOGGER.log(BasicLevel.WARN,"ThresholdObject "
                        + signalThresholdNumberOfObject
                        + " will never be reached with this buffer limit size :"
                        + pSizeObjectMax);
        }
        numberObjectMax = pSizeObjectMax;
        strNumberObjectMax = "" + pSizeObjectMax;
    }

    /**
     * Gets the type object.
     * @return the type object
     */
    public String getTypeObject() {
        return typeObject;
    }

    /**
     * Gets the routing.
     * @return the routing
     */
    public String getRouting() {
        return routing;
    }

    /**
     * Gets the threshold number of object.
     * @return the threshold number of object
     */
    public int getThresholdNumberOfObject() {
        return thresholdNumberOfObject;
    }

    /**
     * Gets the max elaps time.
     * @return the max elaps time
     */
    public long getMaxElapsTime() {
        return maxElapsTime;
    }

    /**
     * Gets the cyclic time.
     * @return the cyclic time
     */
    public long getCyclicTime() {
        return cyclicTime;
    }

    /**
     * Checks whether the buffer is persistent.
     * @return true if the buffer is persistent
     */
    public boolean isPersistant() {
        return isPersistant;
    }

    /**
     * Checks whether the buffer is contextual.
     * @return true if the buffer is contextual
     */
    public boolean isContextual() {
        return isContextual;
    }

    /**
     * Checks whether the buffer should be cleared on peer loss conditions.
     * @return true if PEER_LOST or PEER_LOST_AND_EMPTY policies are used
     */
    public boolean isClearWhenPeerLost() {
        return (Const.VALUE_PEER_LOST.equals(strClear) || Const.VALUE_PEER_LOST_EMPTY
                .equals(strClear));
    }

    /**
     * Checks whether the buffer should be cleared on no-use conditions.
     * @return true if UNUSED or UNUSED_AND_EMPTY policies are used
     */
    public boolean isClearWhenUnused() {
        return (Const.VALUE_UNUSED.equals(strClear) || Const.VALUE_UNUSED_EMPTY
                .equals(strClear));
    }

    /**
     * Checks whether there is an EMPTY clause on the buffer's clearing policy.
     * @return true if ***_AND_EMPTY policies are used
     */
    public boolean isClearWhenEmpty() {
        return (Const.VALUE_PEER_LOST_EMPTY.equals(strClear) || Const.VALUE_UNUSED_EMPTY
                .equals(strClear));
    }

    /**
     * Sets the persistant. *
     * @param pPersistant the new persistant
     */
    public void setPersistant(final boolean pPersistant) {
        this.isPersistant = pPersistant;
        if (pPersistant) {
            strPersistant = "YES";
        } else {
            strPersistant = "NO";
        }
    }

    /**
     * Sets the routing. *
     * @param pRouting the new routing
     */
    public void setRouting(final String pRouting) {
        this.routing = pRouting;
    }

    /**
     * Gets the signal threshold number of object.
     * @return the signal threshold number of object
     */
    public int getSignalThresholdNumberOfObject() {
        return signalThresholdNumberOfObject;
    }

    /**
     * Gets the state.
     * @return the state
     */
    public int getState() {
        return state;
    }

    /**
     * Sets the state. *
     * @param pState the new state
     */
    public void setState(final int pState) {
        this.state = pState;
    }

    /**
     * Sets the type object. *
     * @param pTypeObject the new type object
     */
    public void setTypeObject(final String pTypeObject) {
        this.typeObject = pTypeObject;
    }

    /**
     * Gets the priority.
     * @return the priority
     */
    public int getPriority() {
        return priority;
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#writeXML(java.io.OutputStream, boolean)
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        if (full
                || ((maxElapsTime > 0)
                        || (routing != null && routing.length() > 0)
                        || (signalThresholdNumberOfObject > 1)
                        || (numberObjectMax > 0) || (thresholdNumberOfObject > 1))) {
            out
                    .write((Const.TAB + Const.TAB + Const.TAB + "<BUFFER_OUT>" + Const.EOL)
                            .getBytes());
            writeXMLUtilities(out, 4, Const.PROPERTY_TYPE, typeObject, "", true);
            writeXMLUtilities(out, 4, Const.PROPERTY_MAX_ELAPS_TIME, ""
                    + strMaxElapsTime, "0", full);
            writeXMLUtilities(out, 4, Const.PROPERTY_ROUTING, routing, "",
                    false);
            writeXMLUtilities(out, 4,
                    Const.PROPERTY_SIGNAL_THRESHOLD_NUMBER_OBJECT, ""
                            + strSignalThresholdNumberOfObject, "0", full);
            writeXMLUtilities(out, 4, Const.PROPERTY_SIZE_BYTE_MAX, ""
                    + strSizeByteMax, "0", full);
            writeXMLUtilities(out, 4, Const.PROPERTY_NUMBER_OBJECT_MAX, ""
                    + strNumberObjectMax, "0", full);
            writeXMLUtilities(out, 4, Const.PROPERTY_THRESHOLD_NUMBER_OBJECT,
                    "" + strThresholdNumberOfObject, "0", full);
            writeXMLUtilities(out, 4, Const.PROPERTY_PRIORITY,
                    "" + strPriority, "0", full);
            out
                    .write((Const.TAB + Const.TAB + Const.TAB + "</BUFFER_OUT>" + Const.EOL)
                            .getBytes());
        }
    }

    /**
     * Checks all properties for format and coherence.<br>
     * This has to be done during initialization of the buffer.
     * @throws OdisAttributeException if any incorrect data is encountered
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        StringBuffer buff = null;
        if (typeObject == null || "".equals(typeObject)) {
            buff = new StringBuffer("missing type of bufferOut :");
            buff
                    .append("\n if you use a bufferOut, setting the type information is mandatory :");
            buff.append("\n example of use :");
            buff.append("\n\t<BUFFER_OUT>");
            buff.append("\n\t\t<TYPE value=\"type of buffer\"/>");
            buff.append("\n\t</BUFFER_OUT>");
            buff.append('\n');
        }
        try {
            numberObjectMax = Integer.parseInt(strNumberObjectMax);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_NUMBER_OBJECT_MAX,
                    strNumberObjectMax, e);
        }
        try {
            thresholdNumberOfObject = Integer
                    .parseInt(strThresholdNumberOfObject);
        } catch (NumberFormatException e) {
            buff = concatException(buff,
                    Const.PROPERTY_THRESHOLD_NUMBER_OBJECT,
                    strThresholdNumberOfObject, e);
        }
        try {
            sizeByteMax = Long.parseLong(strSizeByteMax);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_SIZE_BYTE_MAX,
                    strSizeByteMax, e);
        }
        try {
            signalThresholdNumberOfObject = Integer
                    .parseInt(strSignalThresholdNumberOfObject);
        } catch (NumberFormatException e) {
            buff = concatException(buff,
                    Const.PROPERTY_SIGNAL_THRESHOLD_NUMBER_OBJECT,
                    strSignalThresholdNumberOfObject, e);
        }
        try {
            maxElapsTime = Long.parseLong(strMaxElapsTime);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_MAX_ELAPS_TIME,
                    strMaxElapsTime, e);
        }
        try {
            cyclicTime = Long.parseLong(strCyclicTime);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_CYCLIC_TIME,
                    strCyclicTime, e);
        }
        // Persistant set by EngineOutAttribut.checkProperties()
        // (come from NodeAttribut persistance value).
        // Bufferout can overload persistance value.
        if (isPersistant && strPersistant != null) {
            isPersistant = strPersistant.trim().equalsIgnoreCase("YES");
        }
        if (strContextual != null) {
            isContextual = strContextual.trim().equalsIgnoreCase("YES");
        }
        try {
            priority = Integer.parseInt(strPriority);
        } catch (NumberFormatException e) {
            buff = concatException(buff, "Priority", strPriority, e);
        }
        if (isContextual) {
            if ("".equals(strClearDelay) || strClearDelay == null) {
                clearDelay = isClearWhenUnused() ? 300 : 0;
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    LOGGER.log(BasicLevel.DEBUG,
                            "Using default value for clearDelay : "
                                    + clearDelay);
                }
            } else {
                try {
                    clearDelay = Long.parseLong(strClearDelay);
                    if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                        LOGGER.log(BasicLevel.DEBUG, "ClearDelay set to "
                                + clearDelay);
                    }
                } catch (NumberFormatException e) {
                    buff = concatException(buff, "ClearDelay", strClearDelay, e);
                }
            }
        }
        if (buff != null) {
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * Gets the buffer out.
     * @return the buffer out
     */
    public BufferOut getBufferOut() {
        return bufferOut;
    }

    /**
     * Sets the buffer out. *
     * @param pBufferOut the new buffer out
     */
    public void setBufferOut(BufferOut pBufferOut) {
        this.bufferOut = pBufferOut;
    }

    /**
     * Gets the P name.
     * @return the p name
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        return null;
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also a
     * BufferOutAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        BufferOutAttribute copy = null;
        try {
            copy = (BufferOutAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.strCyclicTime.equals(this.strCyclicTime);
        result &= copy.strMaxElapsTime.equals(this.strMaxElapsTime);
        result &= copy.strNumberObjectMax.equals(this.strNumberObjectMax);
        result &= copy.strPersistant.equals(this.strPersistant);
        result &= copy.strPriority.equals(this.strPriority);
        result &= copy.strSignalThresholdNumberOfObject
                .equals(this.strSignalThresholdNumberOfObject);
        result &= copy.strSizeByteMax.equals(this.strSizeByteMax);
        result &= copy.strThresholdNumberOfObject
                .equals(this.strThresholdNumberOfObject);
        result &= copy.routing.equals(this.routing);
        result &= copy.strClear.equals(this.strClear);
        result &= copy.strClearDelay.equals(this.strClearDelay);
        return result;
    }

    /**
     * Getter
     * @return
     */
    public String getStrPersistant() {
        return strPersistant;
    }

    /**
     * Getter
     * @return
     */
    public long getClearDelay() {
        return clearDelay;
    }

    /**
     * Getter
     * @return
     */
    public String getStrClear() {
        return strClear;
    }
}
