package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.util.CheckUtilities;

/**
 * This class collects all information about a composite connection.<br>
 * Such a connection groups in fact basic connections, trying to use them in priority order.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class CompositeConnectionAttribute extends AbstractConnectionAttribute implements
        CompositeConnectionAttributeMBean {
    /** The state. */
    private int state = 0;

    /** The maximum number of messages : computed from underlying connections. */
    private int limitNumberMessage = 0;

    /** The size byte max. */
    private transient int sizeByteMax = 0;

    /** The number message. */
    protected transient int numberMessage = 0;

    /** The number of underlying connections */
    private int membersCount = 0;

    /** The underlying connections' names */
    private String[] membersNames;

    /** The underlying connections' attributes */
    private AbstractConnectionAttribute[] members;

    /**
     * Inc number message.
     * @param value the value
     */
    public void incNumberMessage(int value) {
        numberMessage += value;
    }

    /** The size byte. */
    protected transient long sizeByte = 0;

    /**
     * Inc size byte.
     * @param value the value
     */
    public void incSizeByte(long value) {
        sizeByte += value;
    }

    /**
     * Instantiates a new memory connection attribute.
     * @param parent the parent
     */
    public CompositeConnectionAttribute(AbstractAttribute parent) {
        super(parent);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create "
                    + CompositeConnectionAttribute.class.getName());
        }
    }

    /**
     * equals.
     * @param obj to be compare
     * @return boolean
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj) {
        boolean result = false;
        CompositeConnectionAttribute copy = null;
        try {
            copy = (CompositeConnectionAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = super.equals(obj);
        result &= (membersNames.length == copy.membersNames.length);
        if (result) {
            for (int i = 0; i < membersNames.length; i++) {
                if (!membersNames[i].equals(copy.membersNames[i])) {
                    result = false;
                    break;
                }
            }
        }
        return result;
    }

    /**
     * Setter.
     * @param cnxDeclaration the declaration
     */
    public void subSetter(Element cnxDeclaration) {
        List lsUnusedProperties = cnxDeclaration.getChildren();
        Object[] lsConnections = cnxDeclaration.getChildren("CONNECTION")
                .toArray();
        membersCount = lsConnections.length;
        membersNames = new String[membersCount];
        for (int i = 0; i < membersCount; i++) {
            Element elt = (Element) lsConnections[i];
            String cnxName = elt.getAttributeValue("name", "");
            String cnxPriority = elt.getAttributeValue("priority", "");
            try {
                int prio = Integer.parseInt(cnxPriority) - 1;
                if (prio >= membersCount || prio < 0) {
                    StringBuffer buff = new StringBuffer(
                            "Priority value out of bounds for composite member \"");
                    buff.append(cnxName).append("\" : ").append(cnxPriority)
                            .append(" - ignoring member");
                    LOGGER.log(BasicLevel.ERROR, buff.toString());
                } else {
                    if (membersNames[prio] == null) {
                        membersNames[prio] = cnxName;
                        lsUnusedProperties.remove(elt);
                    } else {
                        StringBuffer buff = new StringBuffer(
                                "Priority value used more than once : ");
                        buff.append(cnxPriority).append(" - members ").append(
                                cnxName).append(" (ignored) and ").append(
                                membersNames[prio]);
                        LOGGER.log(BasicLevel.ERROR, buff.toString());
                    }
                }
            } catch (NumberFormatException e) {
                StringBuffer buff = new StringBuffer(
                        "Bad priority value for composite member \"");
                buff.append(cnxName).append("\" : ").append(cnxPriority)
                        .append(" - ignoring member");
                LOGGER.log(BasicLevel.ERROR, buff.toString());
            }
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        members = new AbstractConnectionAttribute[membersCount];
        StringBuffer buff = new StringBuffer("");
        for (int i = 0; i < membersCount; i++) {
            members[i] = ConnectionManager.getCnxAttribute(membersNames[i]);
            if (members[i] == null) {
                buff.append(
                        "No underlying connection attribute found for member ")
                        .append(membersNames[i]).append(" (").append(i).append(
                                ")\n");
            }
        }
        if (!"".equals(buff.toString())) {
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * Getter
     * @return
     */
    public AbstractConnectionAttribute[] getMembers() {
        return members;
    }

    /**
     * Gets the size object max.
     * @return the max of message which can be contained by this composite (actually, that limit for the first member).
     */
    public final int getSizeObjectMax() {
        return limitNumberMessage;
    }

    /**
     * Gets the size byte max.
     * @return int
     */
    public final int getSizeByteMax() {
        return sizeByteMax;
    }

    /**
     * Gets the number message.
     * @return int
     */
    public final int getNumberMessage() {
        return numberMessage;
    }

    /**
     * Gets the size byte.
     * @return long
     */
    public final long getSizeByte() {
        return sizeByte;
    }

    /**
     * Gets the state string.
     * @return String
     */
    public final String getStateString() {
        return state == Const.NOT_READY ? Const.STR_NOT_READY : Const.STR_READY;
    }

    /**
     * Gets the state.
     * @return the state
     */
    public final int getState() {
        return state;
    }

    /**
     * Sets the state. *
     * @param state the new state
     */
    public void setState(final int state) {
        this.state = state;
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public void writeXML(final OutputStream out, boolean full)
            throws IOException {
        out
                .write((Const.TAB + "<COMPOSITE name=\"" + getName() + "\">" + Const.EOL)
                        .getBytes());
        for (int i = 0; i < membersCount; i++) {
            out.write((Const.TAB2 + "<CONNECTION name=\"" + membersNames[i]
                    + "\" priority=\"" + i + "\">" + Const.EOL).getBytes());
        }
        out.write((Const.TAB + "</COMPOSITE>" + Const.EOL).getBytes());
    }

    /**
     * Gets the type cnx.
     * @return String
     * @see com.AbstractConnectionAttribute.odis.model.AbstractCnxAttribut#getTypeCnx()
     */
    public String getTypeConnection() {
        return "COMPOSITE";
    }

    /**
     * Gets the P name.
     * @return String
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("CompositeCnx");
        return buffer.toString();
    }
}
