package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.engine.EngineIn;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.policy.AbstractPolicy;
import org.ow2.odis.policy.RoundRobinPolicy;
import org.ow2.odis.port.PortIn;
import org.ow2.odis.util.CheckUtilities;
import org.ow2.odis.util.TraceException;

/**
 * The Class EngineInAttribut.
 * @author Nicolas Tachker & Olivier Beyler
 */
public class EngineInAttribute extends AbstractAttribute {
    /** The policy. */
    private String policy = RoundRobinPolicy.class.getName();

    /** The policy instance. */
    private AbstractPolicy policyInstance;

    /** list of from attribute. */
    private final List lsFromAttributs;

    /** The engine in. */
    private EngineIn engineIn = null;

    /** The string number of object max. */
    private String strNumberOfObjectMax = Const.DEFAULT_NUMBER_OF_OBJECT;

    /** The number of object max. */
    private int numberOfObjectMax;

    /** The string size byte max. */
    private String strSizeByteMax = Const.DEFAULT_SIZE_BYTE_MAX;

    /** The size byte max. */
    private long sizeByteMax;

    /**
     * Instantiates a new engine in attribute.
     * @param parent the parent
     */
    public EngineInAttribute(AbstractAttribute parent) {
        super(parent);
        policy = RoundRobinPolicy.class.getName();
        lsFromAttributs = new ArrayList();
    }

    /**
     * setter.
     * @param declaration Element
     */
    public void setter(Element declaration) {
        List lsUnusedProperties = declaration.getChildren();
        Element element;
        if (declaration != null) {
            element = declaration.getChild(Const.PROPERTY_POLICY);
            if (element != null) {
                policy = element.getAttributeValue(Const.VALUE, policy);
                classVersion = element.getAttributeValue(Const.VERSION,
                        classVersion);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_NUMBER_OBJECT_MAX);
            if (element != null) {
                strNumberOfObjectMax = element.getAttributeValue(Const.VALUE,
                        strNumberOfObjectMax);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild(Const.PROPERTY_SIZE_BYTE_MAX);
            if (element != null) {
                strSizeByteMax = element.getAttributeValue(Const.VALUE,
                        strSizeByteMax);
                lsUnusedProperties.remove(element);
            }
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * Creates the policy instance.
     * @throws OdisAttributeException the odis attribute exception
     */
    public void createPolicyInstance() throws OdisAttributeException {
        if (policyInstance == null) {
            AbstractPolicy result = null;
            try {
                result = (AbstractPolicy) CheckUtilities.createInstance(policy,
                        classVersion, "policy", LOGGER);
            } catch (ClassCastException e) {
                StringBuffer buff = new StringBuffer("policy class \"");
                buff.append(policy);
                buff.append("\" at version \"");
                buff.append(classVersion);
                buff.append("\" is invalid");
                LOGGER.log(BasicLevel.ERROR, buff.toString());
                throw new OdisAttributeException(buff.toString(), e);
            }
            policyInstance = result;
        }
    }

    /**
     * Adds the from.
     * @param declaration the declaration
     */
    public void addFrom(Element declaration) {
        FromAttribute fromAttribut = new FromAttribute(this);
        fromAttribut.setter(declaration);
        lsFromAttributs.add(fromAttribut);
    }

    /**
     * Gets the policy.
     * @return the policy
     */
    public String getPolicy() {
        return policy;
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        // write From
        FromAttribute fromAttribut;
        Iterator itFroms = lsFromAttributs.iterator();
        while (itFroms.hasNext()) {
            fromAttribut = (FromAttribute) itFroms.next();
            fromAttribut.writeXML(out, full);
        }
        // write engineIn specific property
        out.write((Const.TAB2 + "<" + Const.ENGINE_IN + ">" + Const.EOL)
                .getBytes());
        if (full || !RoundRobinPolicy.class.getName().equals(policy)) {
            out.write((Const.TAB3 + "<" + Const.PROPERTY_POLICY + " value=\""
                    + policy + "\" ").getBytes());
            writeXMLVersion(out, classVersion, full);
            out.write(("/>" + Const.EOL).getBytes());
        }
        if (full
                || !strNumberOfObjectMax.equals(Const.DEFAULT_NUMBER_OF_OBJECT)) {
            out.write((Const.TAB3 + "<" + Const.PROPERTY_NUMBER_OBJECT_MAX
                    + " value=\"" + strNumberOfObjectMax + "\"/>" + Const.EOL)
                    .getBytes());
        }
        if (full || !strSizeByteMax.equals(Const.DEFAULT_SIZE_BYTE_MAX)) {
            out.write((Const.TAB3 + "<" + Const.PROPERTY_SIZE_BYTE_MAX
                    + " value=\"" + strSizeByteMax + "\"/>" + Const.EOL)
                    .getBytes());
        }
        out.write((Const.TAB2 + "</" + Const.ENGINE_IN + ">" + Const.EOL)
                .getBytes());
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        List lsExceptions = new ArrayList();
        StringBuffer buff = null;
        // policy
        try {
            createPolicyInstance();
        } catch (OdisAttributeException e1) {
            lsExceptions.add(e1.getMessage());
        }
        try {
            numberOfObjectMax = Integer.parseInt(strNumberOfObjectMax);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_NUMBER_OBJECT_MAX,
                    strNumberOfObjectMax, e);
        }
        try {
            sizeByteMax = Integer.parseInt(strSizeByteMax);
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_SIZE_BYTE_MAX,
                    strSizeByteMax, e);
        }
        // port<-->cnxManager
        Iterator itFrom = lsFromAttributs.iterator();
        FromAttribute fromAttribut;
        while (itFrom.hasNext()) {
            fromAttribut = (FromAttribute) itFrom.next();
            try {
                fromAttribut.checkProperties();
            } catch (OdisAttributeException e) {
                lsExceptions.add(e.getMessage());
            }
            String link = fromAttribut.getLink();
            AbstractConnectionAttribute cnxAttribut = ConnectionManager
                    .getCnxAttribute(link);
            boolean result = (cnxAttribut != null);
            if (result) {
                PortIn portIn = engineIn.portFactory.getPortIn(fromAttribut);
                IConnectionIn cnxIn = ConnectionManager.getInstance()
                        .getConnectionIn(engineIn.getNodeName(), cnxAttribut);
                cnxIn.setPortIn(portIn);
                portIn.setConnectionIn(cnxIn);
                policyInstance.addPolicyElement(portIn);
            } else {
                if (buff == null) {
                    buff = new StringBuffer("");
                }
                buff.append("Link \"");
                buff.append(link);
                buff.append("\" declared into to section FROM of node \"");
                buff.append(engineIn.getNode().getName());
                buff.append("\" is lacking in connection section");
                lsExceptions.add(buff.toString());
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisAttributeException(errors);
        }
    }

    /**
     * Sets the engine in. *
     * @param engineIn the new engine in
     */
    public void setEngineIn(EngineIn engineIn) {
        this.engineIn = engineIn;
    }

    /**
     * Gets the policy instance.
     * @return the policy instance
     */
    public AbstractPolicy getPolicyInstance() {
        if (policyInstance == null) {
            throw new RuntimeException("Bad use! policy is not already created");
        }
        return policyInstance;
    }

    /**
     * Gets the size byte max.
     * @return the size byte max
     */
    public long getSizeByteMax() {
        return sizeByteMax;
    }

    /**
     * Gets the number of object max.
     * @return the number of object max
     */
    public int getNumberOfObjectMax() {
        return numberOfObjectMax;
    }

    /**
     * Gets the P name.
     * @return the p name
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("engineIn");
        return buffer.toString();
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also an
     * EngineInAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        EngineInAttribute copy = null;
        try {
            copy = (EngineInAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.policy.equals(this.policy);
        result = copy.classVersion.equals(this.classVersion);
        result &= copy.strNumberOfObjectMax.equals(this.strNumberOfObjectMax);
        result &= copy.strSizeByteMax.equals(this.strSizeByteMax);
        result &= copy.lsFromAttributs.equals(this.lsFromAttributs);
        return result;
    }
}
