package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.bufferOut.BufferOut;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.policy.AbstractPolicy;
import org.ow2.odis.policy.RoundRobinPolicy;
import org.ow2.odis.port.PortOut;
import org.ow2.odis.util.CheckUtilities;
import org.ow2.odis.util.TraceException;

/**
 * The Class EngineOutAttribut.
 * @author Nicolas Tachker
 * @author Olivier Beyler
 */
public class EngineOutAttribute extends AbstractAttribute {
    /** The policy. */
    private String policy;

    /** The policy instance. */
    private AbstractPolicy policyInstance;

    /** The map id buffer out attribute. */
    public final Map mapIdBufferOutAttribute;

    /** The map id to ls attribute. */
    public final Map mapIdToLsAttribute;

    /** The engine out. */
    private EngineOut engineOut;

    /**
     * Instantiates a new engine out attribute.
     * @param parent the parent
     */
    public EngineOutAttribute(AbstractAttribute parent) {
        super(parent);
        mapIdBufferOutAttribute = new HashMap();
        mapIdToLsAttribute = new HashMap();
        policy = RoundRobinPolicy.class.getName();
    }

    /**
     * Sets the ter. *
     * @param declaration the new ter
     */
    public void setter(Element declaration) {
        List lsUnusedProperties = declaration.getChildren();
        Element element;
        if (declaration != null) {
            element = declaration.getChild(Const.POLICY);
            if (element != null) {
                policy = element.getAttributeValue(Const.VALUE, policy);
                classVersion = element.getAttributeValue(Const.VERSION,
                        classVersion);
            }
            lsUnusedProperties.remove(element);
            List lsElements = declaration.getChildren(Const.BUFFER_OUT);
            Iterator itElement = lsElements.iterator();
            while (itElement.hasNext()) {
                element = (Element) itElement.next();
                addBufferOut(element);
            }
            // lsChildrens.remove(lsElements); Not work so use
            Object[] tbElements = lsElements.toArray();
            for (int i = 0; i < tbElements.length; i++) {
                element = (Element) tbElements[i];
                lsUnusedProperties.remove(element);
            }
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * add a To to the List referenced by mapIdToLsAttribut.
     * @param declaration definition of a To destination
     */
    public void addTo(Element declaration) {
        ToAttribute toAttribut = new ToAttribute(this);
        toAttribut.setter(declaration);
        List lsAttribut = (List) mapIdToLsAttribute.get(toAttribut.getId());
        if (lsAttribut == null) {
            synchronized (this) {
                lsAttribut = (List) mapIdToLsAttribute.get(toAttribut.getId());
                if (lsAttribut == null) {
                    lsAttribut = new ArrayList();
                    mapIdToLsAttribute.put(toAttribut.getId(), lsAttribut);
                }
            }
        }
        lsAttribut.add(toAttribut);
    }

    /**
     * Adds the buffer out.
     * @param declaration the declaration
     */
    public void addBufferOut(Element declaration) {
        BufferOutAttribute bufferOutAttribut;
        bufferOutAttribut = new BufferOutAttribute(this);
        bufferOutAttribut.setter(declaration);
        mapIdBufferOutAttribute
                .put(bufferOutAttribut.getId(), bufferOutAttribut);
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#writeXML(java.io.OutputStream, boolean)
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        if (full || mapIdBufferOutAttribute.size() > 0) {
            out.write((Const.TAB2 + "<" + Const.ENGINE_OUT + ">" + Const.EOL)
                    .getBytes());
            Iterator itBufferOuts = mapIdBufferOutAttribute.values().iterator();
            BufferOutAttribute bufferOutAttribut;
            while (itBufferOuts.hasNext()) {
                bufferOutAttribut = (BufferOutAttribute) itBufferOuts.next();
                bufferOutAttribut.writeXML(out, full);
            }
            out.write((Const.TAB2 + "</" + Const.ENGINE_OUT + ">" + Const.EOL)
                    .getBytes());
        }
        // write To
        ToAttribute toAttribut;
        Iterator itLsTos = mapIdToLsAttribute.values().iterator();
        Iterator itTos;
        List lsTos;
        while (itLsTos.hasNext()) {
            lsTos = (List) itLsTos.next();
            itTos = lsTos.iterator();
            while (itTos.hasNext()) {
                toAttribut = (ToAttribute) itTos.next();
                toAttribut.writeXML(out, full);
            }
        }
    }

    /**
     * Creates the policy instance.
     * @throws OdisAttributeException the odis attribute exception
     */
    public void createPolicyInstance() throws OdisAttributeException {
        if (policyInstance == null) {
            AbstractPolicy result = null;
            try {
                result = (AbstractPolicy) CheckUtilities.createInstance(policy,
                        classVersion, "policy", LOGGER);
            } catch (ClassCastException e) {
                StringBuffer buff = new StringBuffer("policy class \"");
                buff.append(policy);
                buff.append("\" at version \"");
                buff.append(classVersion);
                buff.append("\" is invalid");
                LOGGER.log(BasicLevel.ERROR, buff.toString());
                throw new OdisAttributeException(buff.toString(), e);
            }
            policyInstance = result;
        }
    }

    /**
     * Gets the policy instance.
     * @return the policy instance
     */
    public AbstractPolicy getPolicyInstance() {
        if (policyInstance == null) {
            throw new RuntimeException("Bad use! policy is not already created");
        }
        return policyInstance;
    }

    /**
     * Gets the list to.
     * @return the list to
     */
    public List getLsTo() {
        List lsTo = new ArrayList();
        Iterator itLsTos = mapIdToLsAttribute.values().iterator();
        Iterator itTos;
        List lsTos;
        while (itLsTos.hasNext()) {
            lsTos = (List) itLsTos.next();
            itTos = lsTos.iterator();
            while (itTos.hasNext()) {
                lsTo.add(itTos.next());
            }
        }
        return lsTo;
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribute#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        List lsExceptions = new ArrayList();
        // policy
        try {
            createPolicyInstance();
        } catch (OdisAttributeException e1) {
            lsExceptions.add(e1.getMessage());
        }
        // checkProperties bufferOut
        Collection coBufferOutAttribut = mapIdBufferOutAttribute.values();
        Iterator itBufferOutAttribut = coBufferOutAttribut.iterator();
        BufferOutAttribute bufferOutAttribute;
        while (itBufferOutAttribut.hasNext()) {
            bufferOutAttribute = (BufferOutAttribute) itBufferOutAttribut.next();
            bufferOutAttribute.setPersistant(engineOut.getNode().attribute
                    .isPersistant());
            try {
                bufferOutAttribute.checkProperties();
            } catch (OdisAttributeException e1) {
                lsExceptions.add(e1.getMessage());
            }
        }
        // port<-->cnxManager
        // chaque port doit retrouver son bufferOut associe
        Iterator itLsTos = mapIdToLsAttribute.values().iterator();
        Iterator itTos;
        List lsTos;
        ToAttribute toAttribut;
        StringBuffer buff;
        String id;
        BufferOut bufferOut;
        while (itLsTos.hasNext()) {
            lsTos = (List) itLsTos.next();
            itTos = lsTos.iterator();
            while (itTos.hasNext()) {
                toAttribut = (ToAttribute) itTos.next();
                try {
                    toAttribut.checkProperties();
                } catch (OdisAttributeException e) {
                    lsExceptions.add(e.getMessage());
                }
                String link = toAttribut.getLink();
                AbstractConnectionAttribute cnxAttribute = ConnectionManager
                        .getCnxAttribute(link);
                boolean result = (cnxAttribute != null);
                if (result) {
                    id = toAttribut.getId();
                    bufferOutAttribute = (BufferOutAttribute) mapIdBufferOutAttribute
                            .get(id);
                    if (bufferOutAttribute == null) {
                        // No explicit configuration found : create a BufferOut
                        // with default values
                        bufferOutAttribute = new BufferOutAttribute(this);
                        bufferOutAttribute
                                .setTypeObject(toAttribut.getStrType());
                        bufferOutAttribute.setRouting(toAttribut.getRouting());
                        bufferOutAttribute
                                .setPersistant(engineOut.getNode().attribute
                                        .isPersistant());
                        mapIdBufferOutAttribute.put(id, bufferOutAttribute);
                        LOGGER.log(BasicLevel.DEBUG,
                                " create a bufferOutAttribut with default value for "
                                        + toAttribut.getId());
                    }
                    engineOut.getBufferOutManager().createBufferOut(
                            bufferOutAttribute);
                    bufferOut = engineOut.getBufferOutManager().getBufferOut(
                            bufferOutAttribute.getId());
                    PortOut portOut = engineOut.getPortFactory().getPortOut(
                            toAttribut);
                    if (bufferOutAttribute.isContextual()) {
                        portOut.setConnectionOut(ConnectionManager.getInstance()
                                .getContextualConnectionOut(
                                        engineOut.getNodeName(), cnxAttribute));
                    } else {
                        portOut.setConnectionOut(ConnectionManager.getInstance()
                                .getConnectionOut(engineOut.getNodeName(),
                                        cnxAttribute));
                        policyInstance.addPolicyElement(bufferOut);
                    }
                    engineOut.getRouting().addPortOut(id, portOut);
                } else {
                    buff = new StringBuffer("");
                    buff.append("Link \"");
                    buff.append(link);
                    buff.append("\" declared into to section To of node \"");
                    buff.append(engineOut.getNode().getName());
                    buff.append("\" is lacking in connection section");
                    LOGGER.log(BasicLevel.ERROR, buff.toString());
                    lsExceptions.add(buff.toString());
                }
            }
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisAttributeException(errors);
        }
    }

    /**
     * setter.
     * @param engineOut EngineOut
     */
    public void setEngineOut(EngineOut engineOut) {
        this.engineOut = engineOut;
    }

    /**
     * @return String
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("engineOut");
        return buffer.toString();
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also an
     * EngineOutAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        EngineOutAttribute copy = null;
        try {
            copy = (EngineOutAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.policy.equals(this.policy);
        result = copy.classVersion.equals(this.classVersion);
        result &= copy.mapIdBufferOutAttribute
                .equals(this.mapIdBufferOutAttribute);
        result &= copy.mapIdToLsAttribute.equals(this.mapIdToLsAttribute);
        return result;
    }
}
