package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.engine.EngineProceed;
import org.ow2.odis.engine.IProceed;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.exception.OdisEncoderException;
import org.ow2.odis.loader.GenericFactory;
import org.ow2.odis.proceed.IRetryPolicy;
import org.ow2.odis.proceed.ProceedManager;
import org.ow2.odis.util.CheckUtilities;

/**
 * Engine proceed attribut.
 * @author Beyler olivier
 * @version 1.0
 * @since JDK 1.4
 */
public class EngineProceedAttribute extends AbstractAttribute {
    /** The engine proceed. */
    private EngineProceed engineProceed = null;

    /** class name of the proceed. */
    private String processClassName = "";

    /** define the process use by the engine. */
    private IProceed processInstance = null;

    private String onExceptionClassName = "";

    private IRetryPolicy onException = null;

    /**
     * @return
     */
    public IRetryPolicy getOnException() {
        return onException;
    }

    /** The properties. */
    private String properties = "";

    /** number of Exception witch occured while proceed. */
    protected long nbrProceedError = 0;

    /**
     * Inc nbr proceed error.
     */
    public void incNbrProceedError() {
        nbrProceedError++;
    }

    /** interval between two proceed call (milli seconde). */
    private long interval = 0;

    /** The str interval. */
    private String strInterval = "0";

    /**
     * increase at each invocation of proceed without any link with the result witch can be null , a list or even an exception.
     */
    private long invocation = 0;

    /**
     * Gets the invocation.
     * @return the invocation
     */
    public long getInvocation() {
        return invocation;
    }

    /**
     * Inc invocation.
     */
    public void incInvocation() {
        invocation++;
    }

    /** The ls versionned encoders. */
    private final List lsVersionedEncoders;

    /**
     * default constructor.
     * @param parent AbstractAttribut
     */
    public EngineProceedAttribute(AbstractAttribute parent) {
        super(parent);
        lsVersionedEncoders = new ArrayList();
    }

    /**
     * Sets the param.
     * @param declaration the new ter
     */
    public void setter(Element declaration) {
        Element element;
        List lsUnusedProperties = declaration.getChildren();
        element = declaration.getChild(Const.PROPERTY_INTERVAL);
        if (element != null) {
            strInterval = element.getAttributeValue(Const.VALUE, strInterval);
            lsUnusedProperties.remove(element);
        }
        element = declaration.getChild(Const.PROPERTY_PROCESS);
        if (element != null) {
            processClassName = element.getAttributeValue(Const.VALUE,
                    processClassName);
            classVersion = element.getAttributeValue(Const.VERSION,
                    classVersion);
            lsUnusedProperties.remove(element);
        }
        element = declaration.getChild(Const.PROPERTY_PROPERTIES);
        if (element != null) {
            properties = element.getAttributeValue(Const.VALUE, properties);
            lsUnusedProperties.remove(element);
        }
        element = declaration.getChild(Const.PROPERTY_ON_EXCEPTION);
        if (element != null) {
            onExceptionClassName = element.getAttributeValue(Const.VALUE,
                    onExceptionClassName);
            lsUnusedProperties.remove(element);
        }
        // -----------------------
        // declare encoder
        List lsElements = declaration.getChildren(Const.PROPERTY_ENCODER);
        Iterator itElement = lsElements.iterator();
        while (itElement.hasNext()) {
            element = (Element) itElement.next();
            VersionedEncoder ve = new VersionedEncoder(element
                    .getAttributeValue(Const.VALUE), element
                    .getAttributeValue(Const.VERSION));
            lsVersionedEncoders.add(ve);
        }
        // lsChildrens.remove(lsElements); Not work so use
        Object[] tbElements = lsElements.toArray();
        for (int i = 0; i < tbElements.length; i++) {
            element = (Element) tbElements[i];
            lsUnusedProperties.remove(element);
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * Gets the proceed instance.
     * @return the proceed instance
     */
    public IProceed getProceedInstance() {
        if (processInstance == null) {
            throw new RuntimeException(
                    "Bad use! Proceed is not already created");
        }
        return processInstance;
    }

    /**
     * set the proceed class name.
     * @param type the type
     */
    public void setProceedClassName(final String type) {
        processClassName = type;
    }

    /**
     * Gets the proceed class name.
     * @return the proceed classname.
     */
    public String getProceedClassName() {
        return processClassName;
    }

    /**
     * Gets the properties.
     * @return the properties
     */
    public String getProperties() {
        return properties;
    }

    /**
     * set the properties param.
     * @param pProperties the properties
     */
    public void setProperties(String pProperties) {
        this.properties = pProperties;
    }

    /**
     * Gets the nbr proceed error.
     * @return the number of proceed error.
     */
    public long getNbrProceedError() {
        return nbrProceedError;
    }

    /**
     * Sets the nbr proceed error.
     * @param pNbrProceedError the new nbr proceed error
     */
    public void setNbrProceedError(final long pNbrProceedError) {
        this.nbrProceedError = pNbrProceedError;
    }

    /**
     * Gets the interval.
     * @return the interval
     */
    public long getInterval() {
        return interval;
    }

    /**
     * Sets the interval.
     * @param pInterval the new interval
     */
    public void setInterval(final long pInterval) {
        this.interval = pInterval;
    }

    /**
     * utility function to help writing xml describing a node.
     * @param out OutputStream
     * @param full boolean (the xml can be full or without default value)
     * @throws IOException in case of fail
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        out.write(Const.bTAB2);
        out.write(("<" + Const.ENGINE_PROCEED + ">" + Const.EOL).getBytes());
        out.write(Const.bTAB3);
        out.write(("<" + Const.PROPERTY_PROCESS + " " + Const.VALUE + "=\""
                + processClassName + "\" ").getBytes());
        writeXMLVersion(out, classVersion, full);
        out.write(("/>" + Const.EOL).getBytes());
        writeXMLUtilities(out, 3, Const.PROPERTY_INTERVAL, strInterval, "0",
                full);
        writeXMLUtilities(out, 3, Const.PROPERTY_PROPERTIES, properties, "",
                full);
        final Iterator itVersionedEnvoder = lsVersionedEncoders.iterator();
        while (itVersionedEnvoder.hasNext()) {
            VersionedEncoder ve = (VersionedEncoder) itVersionedEnvoder.next();
            writeXMLEncoder(out, 3, ve, full);
        }
        out.write(Const.bTAB2);
        out.write(("</" + Const.ENGINE_PROCEED + ">" + Const.EOL).getBytes());
    }

    /**
     * Write the XMLof Encoder
     * @param out output stream
     * @param level
     * @param value
     * @param version
     * @param full
     * @throws IOException
     */
    private void writeXMLEncoder(OutputStream out, int level,
            VersionedEncoder ve, boolean full) throws IOException {
        String value = ve.getEncoderClassName();
        if (full) {
            for (int i = 0; i < level; i++) {
                out.write(Const.bTAB);
            }
            out.write("<".getBytes());
            if (value == null) {
                out.write((Const.PROPERTY_ENCODER + "/>" + Const.EOL)
                        .getBytes());
            } else {
                out.write((Const.PROPERTY_ENCODER + " " + Const.VALUE + " = \""
                        + value + "\" ").getBytes());
                writeXMLVersion(out, classVersion, full);
                out.write(("/>").getBytes());
                out.write(Const.EOL.getBytes());
            }
        } else {
            if (value != null && !value.equals("")) {
                for (int i = 0; i < level; i++) {
                    out.write(Const.bTAB);
                }
                out.write("<".getBytes());
                out.write(Const.PROPERTY_ENCODER.getBytes());
                out.write(" ".getBytes());
                out.write(Const.VALUE.getBytes());
                out.write(" = \"".getBytes());
                out.write(value.getBytes());
                out.write("\" ".getBytes());
                writeXMLVersion(out, classVersion, full);
                out.write(("/>").getBytes());
                out.write(Const.EOL.getBytes());
            }
        }
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribute#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        StringBuffer buff = null;
        Iterator itEncoders = lsVersionedEncoders.iterator();
        while (itEncoders.hasNext()) {
            VersionedEncoder vEncoder = (VersionedEncoder) itEncoders.next();
            try {
                engineProceed.getEncoderFactory().registerEncoder(
                        vEncoder.getEncoderClassName(),
                        vEncoder.getEncoderVersion());
            } catch (OdisEncoderException e) {
                buff = concatException(buff, Const.PROPERTY_ENCODER, vEncoder
                        .toString(), e);
            }
        }
        try {
            setInterval(Long.parseLong(strInterval));
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_INTERVAL, strInterval,
                    e);
        }
        try {
            processInstance = (IProceed) CheckUtilities.createInstance(
                    getProceedClassName(), classVersion, "Engine process",
                    LOGGER);
            if (processInstance != null) {
                ProceedManager.getInstance().registerProceed(processInstance,
                        engineProceed.getNode());
            } else {
                if (!"".equals(getProceedClassName())
                        && !"".equals(classVersion)) {
                    if (buff == null) {
                        buff = new StringBuffer("");
                    } else {
                        buff.append("/n");
                    }
                    buff.append("unable to create proceed :");
                    buff.append(getProceedClassName());
                    buff.append("of verison :");
                    buff.append(classVersion);
                }
            }
        } catch (ClassCastException e) {
            if (buff == null) {
                buff = new StringBuffer("");
            } else {
                buff.append("/n");
            }
            buff.append(getProceedClassName() + " at version " + classVersion);
            buff.append("\" is invalid");
        }
        try {
            if (onExceptionClassName != null
                    && !"".equals(onExceptionClassName)) {
                onException = (IRetryPolicy) CheckUtilities.createInstance(
                        onExceptionClassName, classVersion,
                        "Exception on process", LOGGER);
                if (onException == null) {
                    buff.append("unable to create onException :");
                    buff.append(getProceedClassName());
                    buff.append("of version :");
                    buff.append(classVersion);
                }
            }
        } catch (ClassCastException e) {
            if (buff == null) {
                buff = new StringBuffer("");
            } else {
                buff.append("/n");
            }
            buff.append(getProceedClassName() + " at version " + classVersion);
            buff.append("\" is invalid");
        }
        if (buff != null) {
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * setter.
     * @param engineProceed EngineProceed
     */
    public void setEngineProceed(EngineProceed engineProceed) {
        this.engineProceed = engineProceed;
    }

    /**
     * Gets the name.
     * @return String
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("engineProceed");
        return buffer.toString();
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also an
     * EngineProceedAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        EngineProceedAttribute copy = null;
        try {
            copy = (EngineProceedAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.processClassName.equals(this.processClassName);
        result &= copy.classVersion.equals(this.classVersion);
        result &= copy.properties.equals(this.properties);
        result &= copy.strInterval.equals(this.strInterval);
        result &= copy.lsVersionedEncoders.equals(this.lsVersionedEncoders);
        return result;
    }

    /**
     * @author Arnaud Constancin _ France-Telecom R&D _ OrangeLabs <br /> 
     * this class represent an encoder class with version attribute.
     */
 
    private class VersionedEncoder {
        /** version of the element class. */
        private String encoderClassName = null;

        /** version of the element class. */
        private String encoderVersion = null;

        /**
         * Constructor.
         * @param encoderClassName string
         * @param encoderVersion string
         */
        public VersionedEncoder(String encoderClassName, String encoderVersion) {
            super();
            this.encoderClassName = encoderClassName;
            if (encoderVersion == null) {
                encoderVersion = GenericFactory.DEFAULT_VERSION;
            }
            this.encoderVersion = encoderVersion;
        }

        /**
         * Gets the encoder class name.
         * @return the encoder class name
         */
        public String getEncoderClassName() {
            return encoderClassName;
        }


        /**
         * get the encoder version
         * @return the encoder version
         * 
         */
        public String getEncoderVersion() {
            return encoderVersion;
        }

        /**
         * @return the string
         * @see java.lang.Object#toString()
         */
        public String toString() {
            return (encoderClassName + ";version=" + encoderVersion);
        }

        /**
         * @param nevc
         * @return
         */
        public boolean equals(Object obj) {
            boolean isEquals = false;
            VersionedEncoder ve = null;
            if (obj instanceof VersionedEncoder) {
                ve = (VersionedEncoder) obj;
                isEquals = versionEquals(ve.getEncoderVersion())
                        && classNameEquals(ve.getEncoderClassName());
            } else {
                // isEquals = false;
            }
            return isEquals;
        }

        /**
         * @param expectedVersion
         * @return
         */
        private boolean versionEquals(String expectedVersion) {
            if (null == encoderVersion) {
                return (null == expectedVersion);
            } else {
                // version is not null
                return encoderVersion.equals(expectedVersion);
            }
        }

        private boolean classNameEquals(String expectedClassName) {
            if (null == encoderClassName) {
                return (null == expectedClassName);
            } else {
                // className is not null
                return encoderClassName.equals(expectedClassName);
            }
        }
    }
}
