package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.decoder.IDecoder;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.util.CheckUtilities;

/**
 * The Class FromAttribut.
 * @author Nicolas Tachker
 */
public class FromAttribute extends AbstractAttribute {
    /**
     * Instantiates a new from attribut.
     * @param parent the parent
     */
    protected FromAttribute(AbstractAttribute parent) {
        super(parent);
    }

    /** The str priority. */
    private String strPriority = "1";

    /** The priority. */
    private int priority = 1;

    /** The str limit octet size. */
    private String strLimitOctetSize = "1000000";

    /** The limit octet size. */
    private long limitOctetSize = -1;

    /** The str limit msg number. */
    private String strLimitMsgNumber = "1000";

    /** The limit msg number. */
    private long limitMsgNumber = -1;

    /** The link. */
    private String link = null;

    /** The decoder. */
    private String decoder = "";

    /** The decoder instance. */
    private IDecoder decoderInstance = null;

    /**
     * Setter.
     * @param declaration
     */
    public void setter(Element declaration) {
        if (declaration != null) {
            List lsUnusedProperties = declaration.getChildren();
            link = declaration.getAttributeValue("link");
            Element element = declaration.getChild("PRIORITY");
            if (element != null) {
                strPriority = element.getAttributeValue(Const.VALUE,
                        strPriority);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild("DECODER");
            if (element != null) {
                decoder = element.getAttributeValue(Const.VALUE, decoder);
                classVersion = element.getAttributeValue(Const.VERSION,
                        classVersion);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild("LIMIT_OCTET_SIZE");
            if (element != null) {
                strLimitOctetSize = element.getAttributeValue(Const.VALUE,
                        strLimitOctetSize);
                lsUnusedProperties.remove(element);
            }
            element = declaration.getChild("limitMessageNumber");
            if (element != null) {
                strLimitMsgNumber = element.getAttributeValue(Const.VALUE,
                        strLimitMsgNumber);
                lsUnusedProperties.remove(element);
            }
            CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
        }
    }

    /**
     * Gets the decoder.
     * @return the decoder
     */
    public String getDecoder() {
        return decoder;
    }

    /**
     * Gets the link.
     * @return the link
     */
    public String getLink() {
        return link;
    }

    /**
     * Gets the limit msg number.
     * @return the limit msg number
     */
    public long getLimitMsgNumber() {
        return limitMsgNumber;
    }

    /**
     * Gets the limit octet size.
     * @return the limit octet size
     */
    public long getLimitOctetSize() {
        return limitOctetSize;
    }

    /**
     * Gets the priority.
     * @return the priority
     */
    public int getPriority() {
        return priority;
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#writeXML(java.io.OutputStream, boolean)
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        out
                .write((Const.TAB2 + "<" + Const.FROM + " link=\"" + link
                        + "\">" + Const.EOL).getBytes());
        out.write((Const.TAB3 + "<" + Const.PRIORITY + " value=\""
                + strPriority + "\"/>" + Const.EOL).getBytes());
        if (decoder != null) {
            out.write((Const.TAB3 + "<" + Const.DECODER + " " + Const.VALUE
                    + "=\"" + decoder + "\" ").getBytes());
            writeXMLVersion(out, classVersion, full);
            out.write(("/>" + Const.EOL).getBytes());
        }
        out.write((Const.TAB2 + "</" + Const.FROM + ">+Const.EOL").getBytes());
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see AbstractAttribute#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        // la partie creation de decodeur est faite par l'engineIn
        StringBuffer buff = null;
        try {
            limitOctetSize = Long.parseLong(strLimitOctetSize);
        } catch (NumberFormatException e) {
            buff = concatException(buff, "LimitOctetSize", strLimitOctetSize, e);
        }
        try {
            limitMsgNumber = Long.parseLong(strLimitMsgNumber);
        } catch (NumberFormatException e) {
            buff = concatException(buff, "LimitMsgNumber", strLimitMsgNumber, e);
        }
        try {
            priority = Integer.parseInt(strPriority);
        } catch (NumberFormatException e) {
            buff = concatException(buff, "Priority", strPriority, e);
        }
        try {
            decoderInstance = (IDecoder) CheckUtilities.createInstance(decoder,
                    classVersion, "decoder", LOGGER);
        } catch (ClassCastException e) {
            if (buff == null) {
                buff = new StringBuffer(decoder);
            } else {
                buff.append(decoder);
            }
            buff.append(" of version \"");
            buff.append(classVersion);
            buff.append("\" is not a valid implementation of ");
            buff.append(IDecoder.class.getName());
        }
        if (buff != null) {
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * Gets the decoder instance.
     * @return the decoder instance
     */
    public IDecoder getDecoderInstance() {
        return decoderInstance;
    }

    /**
     * Gets the P name.
     * @return the p name
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("from_");
        buffer.append(getLink());
        return buffer.toString();
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also a
     * FromAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        FromAttribute copy = null;
        try {
            copy = (FromAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.strLimitMsgNumber.equals(this.strLimitMsgNumber);
        result &= copy.strLimitOctetSize.equals(this.strLimitOctetSize);
        result &= copy.strPriority.equals(this.strPriority);
        result &= copy.link.equals(this.link);
        result &= copy.decoder.equals(this.decoder);
        result &= copy.classVersion.equals(this.classVersion);
        return result;
    }
}
