package org.ow2.odis.model;

import java.io.IOException;
import java.io.OutputStream;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.memory.MemoryFactory;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.util.CheckUtilities;

/**
 * This class collect all informations about a Memory connection.
 * <p>
 * @author Beyler Olivier FranceTelecom R&D
 */
public class MemoryConnectionAttribute extends AbstractConnectionAttribute implements
        MemoryConnectionAttributeMBean {
    /** The state. */
    private int state = 0;

    /**
     * Max size message present into memory be careful a message can contain more than one object By defaut the the limit is set to 10000.
     */
    private static final int DEFAULT_LIMIT_NUMBER_MESSAGE = 10000;

    /** The limit number message. */
    private int limitNumberMessage = DEFAULT_LIMIT_NUMBER_MESSAGE;

    /** The str limit number message. */
    private String strLimitNumberMessage = "10000";

    /** By default, the limit of the memory buffer size is set to 1000000 octets. */
    private static final int DEFAULT_SIZE_BYTE_MAX = 1000000;

    /** The size byte max. */
    private transient int sizeByteMax = DEFAULT_SIZE_BYTE_MAX;

    /** The str size byte max. */
    private String strSizeByteMax = "1000000";

    /** The number message. */
    protected transient int numberMessage = 0;

    /**
     * Inc number message.
     * @param value the value
     */
    public void incNumberMessage(int value) {
        numberMessage += value;
    }

    /** The size byte. */
    protected transient long sizeByte = 0;

    /**
     * Inc size byte.
     * @param value the value
     */
    public void incSizeByte(long value) {
        sizeByte += value;
    }

    /**
     * Instantiates a new memory connection attribute.
     * @param parent the parent
     */
    public MemoryConnectionAttribute(AbstractAttribute parent) {
        super(parent);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "create "
                    + MemoryConnectionAttribute.class.getName());
        }
    }

    /**
     * equals.
     * @param obj to be compare
     * @return boolean
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj) {
        boolean result = false;
        MemoryConnectionAttribute copy = null;
        try {
            copy = (MemoryConnectionAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = super.equals(obj);
        if (result) {
            result &= copy.strLimitNumberMessage
                    .equals(this.strLimitNumberMessage);
            result &= copy.strSizeByteMax.equals(this.strSizeByteMax);
            if (!result) {
                traceOnConfictualConnection(this, copy);
            }
        }
        return result;
    }

    /**
     * Clean.
     */
    public void clean() {
        IConnectionIn memoryIn = MemoryFactory.getInstance().newConnectionIn(this);
        LOGGER
                .log(BasicLevel.DEBUG, "clean "
                        + MemoryConnectionAttribute.class.getName());
        while (numberMessage > 0) {
            memoryIn.ackMessage(true);
        }
        LOGGER.log(BasicLevel.DEBUG, "create "
                + MemoryConnectionAttribute.class.getName());
    }

    /**
     * Setter.
     * @param cnxDeclaration the declaration
     */
    public void subSetter(Element cnxDeclaration) {
        List lsUnusedProperties = cnxDeclaration.getChildren();
        Element element = cnxDeclaration.getChild(Const.PROPERTY_MAX_MESSAGES);
        if (element != null) {
            strLimitNumberMessage = element.getAttributeValue(Const.VALUE,
                    strLimitNumberMessage);
            lsUnusedProperties.remove(element);
        }
        element = cnxDeclaration.getChild(Const.PROPERTY_SIZE_BYTE_MAX);
        if (element != null) {
            strSizeByteMax = element.getAttributeValue(Const.VALUE,
                    strSizeByteMax);
            lsUnusedProperties.remove(element);
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        StringBuffer buff = null;
        try {
            setSizeByteMax(Integer.parseInt(strSizeByteMax));
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_SIZE_BYTE_MAX,
                    strSizeByteMax, e);
        }
        try {
            setSizeObjectMax(Integer.parseInt(strLimitNumberMessage));
        } catch (NumberFormatException e) {
            buff = concatException(buff, Const.PROPERTY_MAX_MESSAGES,
                    strLimitNumberMessage, e);
        }
        if (buff != null) {
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * Gets the size object max.
     * @return the max of message witch can be contains by a memory.
     */
    public final int getSizeObjectMax() {
        return limitNumberMessage;
    }

    /**
     * Sets the size object max.
     * @param sizeObjectMax the size object max
     */
    public void setSizeObjectMax(final int sizeObjectMax) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "property "
                    + Const.PROPERTY_MAX_MESSAGES + " is set to "
                    + sizeObjectMax);
        }
        this.limitNumberMessage = sizeObjectMax;
        this.strLimitNumberMessage = "" + sizeObjectMax;
        setChanged();
        this.notifyObservers(this);
    }

    /**
     * Gets the size byte max.
     * @return int
     */
    public final int getSizeByteMax() {
        return sizeByteMax;
    }

    /**
     * Sets the size byte max.
     * @param sizeByteMax the size byte max
     */
    public void setSizeByteMax(final int sizeByteMax) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "property "
                    + Const.PROPERTY_SIZE_BYTE_MAX + " is set to "
                    + limitNumberMessage);
        }
        this.sizeByteMax = sizeByteMax;
        this.strSizeByteMax = "" + sizeByteMax;
        setChanged();
        this.notifyObservers(this);
    }

    /**
     * Gets the number message.
     * @return int
     */
    public final int getNumberMessage() {
        return numberMessage;
    }

    /**
     * Gets the size byte.
     * @return long
     */
    public final long getSizeByte() {
        return sizeByte;
    }

    /**
     * Gets the state string.
     * @return String
     */
    public final String getStateString() {
        return state == Const.NOT_READY ? Const.STR_NOT_READY : Const.STR_READY;
    }

    /**
     * Gets the state.
     * @return the state
     */
    public final int getState() {
        return state;
    }

    /**
     * Sets the state. *
     * @param state the new state
     */
    public void setState(final int state) {
        this.state = state;
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     */
    public void writeXML(final OutputStream out, boolean full)
            throws IOException {
        out.write((Const.TAB + "<MEMORY name=\"" + getName() + "\">" + Const.EOL)
                .getBytes());
        writeXMLCNXUtilities(out, 2, Const.PROPERTY_MAX_MESSAGES,
                strLimitNumberMessage, "" + DEFAULT_LIMIT_NUMBER_MESSAGE, full);
        writeXMLCNXUtilities(out, 2, Const.PROPERTY_SIZE_BYTE_MAX,
                strSizeByteMax, "" + DEFAULT_SIZE_BYTE_MAX, full);
        out.write((Const.TAB + "</MEMORY>" + Const.EOL).getBytes());
    }

    /**
     * Gets the type cnx.
     * @return String
     * @see com.AbstractConnectionAttribute.odis.model.AbstractCnxAttribut#getTypeCnx()
     */
    public String getTypeConnection() {
        return "MEMORY";
    }

    /**
     * Gets the P name.
     * @return String
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append("MEMORYCnx");
        return buffer.toString();
    }
}
