package org.ow2.odis.model;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Iterator;
import java.util.List;

import org.jdom.Element;
import org.objectweb.util.monolog.api.BasicLevel;
import org.ow2.odis.exception.OdisAttributeException;
import org.ow2.odis.node.Node;
import org.ow2.odis.util.CheckUtilities;

/**
 * The Class NodeAttribut.
 * @author Nicolas Tachkeni && Olivier Beyler
 */
public class NodeAttribute extends AbstractAttribute {
    /** a node has a name, this name must be unique into a domain. */
    private String name = null;

    /**
     * description of the node. This property is use only for user's convenience.
     */
    private String description = null;

    /** store the persistence attribute as string without any control on it. */
    private String strPersistent = Const.NO;

    /** use persistence inside a node. */
    private boolean persistent = false;

    /** keep a reference on engineIn attribute. */
    private final EngineInAttribute engineInAttribute;

    /** keep a reference on engineOut attribute. */
    private final EngineOutAttribute engineOutAttribute;

    /** keep a reference on engine proceed attribute. */
    private final EngineProceedAttribute engineProceedAttribute;

    /** keep a reference on real node object. */
    private Node node = null;

    /**
     * Gets the engine in attribute.
     * @return the engine in attribute
     */
    public EngineInAttribute getEngineInAttribute() {
        return engineInAttribute;
    }

    /**
     * Gets the engine out attribute.
     * @return the engine out attribute
     */
    public EngineOutAttribute getEngineOutAttribute() {
        return engineOutAttribute;
    }

    /**
     * Gets the engine proceed attribute.
     * @return the engine proceed attribute
     */
    public EngineProceedAttribute getEngineProceedAttribute() {
        return engineProceedAttribute;
    }

    /**
     * Instantiates a new node attribute.
     * @param parent the parent
     */
    public NodeAttribute(AbstractAttribute parent) {
        super(parent);
        engineInAttribute = new EngineInAttribute(this);
        engineOutAttribute = new EngineOutAttribute(this);
        engineProceedAttribute = new EngineProceedAttribute(this);
    }

    /**
     * Gets the description.
     * @return the description
     */
    public String getDescription() {
        return description;
    }

    /**
     * Gets the name.
     * @return the name
     */
    public String getName() {
        return name;
    }

    /**
     * Setter.
     * @param nodeDeclaration Element
     */
    public void setter(Element nodeDeclaration) {
        List lsUnusedProperties = nodeDeclaration.getChildren();
        // -------------------------------------
        // create node
        name = nodeDeclaration.getAttributeValue("name");
        description = nodeDeclaration.getAttributeValue(Const.DESCRIPTION, "");
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, Const.LINE_SEP);
            LOGGER.log(BasicLevel.INFO, "NODE  " + name);
            LOGGER.log(BasicLevel.INFO, Const.LINE_SEP);
        }
        Element element = nodeDeclaration.getChild(Const.CST_ISPERSISTENT);
        if (element != null) {
            strPersistent = element.getAttributeValue(Const.VALUE,
                    strPersistent);
            lsUnusedProperties.remove(element);
        }
        // -------------------------------------
        // set engineIn
        element = nodeDeclaration.getChild(Const.ENGINE_IN);
        if (element != null) {
            engineInAttribute.setter(element);
            lsUnusedProperties.remove(element);
        }
        // -------------------------------------
        // set Froms
        List lsFroms = nodeDeclaration.getChildren(Const.FROM);
        Iterator itFrom = lsFroms.iterator();
        while (itFrom.hasNext()) {
            element = (Element) itFrom.next();
            engineInAttribute.addFrom(element);
        }
        // lsChildrens.remove(lsFroms);
        Object[] tbElements = lsFroms.toArray();
        for (int i = 0; i < tbElements.length; i++) {
            element = (Element) tbElements[i];
            lsUnusedProperties.remove(element);
        }
        // -------------------------------------
        // set engineOut
        element = nodeDeclaration.getChild(Const.ENGINE_OUT);
        if (element != null) {
            engineOutAttribute.setter(element);
            lsUnusedProperties.remove(element);
        }
        // -------------------------------------
        // set Tos
        final List lsTos = nodeDeclaration.getChildren(Const.TO);
        Iterator itTo = lsTos.iterator();
        while (itTo.hasNext()) {
            element = (Element) itTo.next();
            engineOutAttribute.addTo(element);
        }
        // lsChildrens.remove(lsTos); Not work so use
        tbElements = lsTos.toArray();
        for (int i = 0; i < tbElements.length; i++) {
            element = (Element) tbElements[i];
            lsUnusedProperties.remove(element);
        }
        // -------------------------------------
        // set engineProceed
        element = nodeDeclaration.getChild(Const.ENGINE_PROCEED);
        if (element != null) {
            engineProceedAttribute.setter(element);
            lsUnusedProperties.remove(element);
        }
        CheckUtilities.checkUnusedProperties(lsUnusedProperties, LOGGER);
    }

    /**
     * Write XML.
     * @param out the out
     * @param full the full
     * @throws IOException Signals that an I/O exception has occurred.
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#writeXML(java.io.OutputStream, boolean)
     */
    public void writeXML(OutputStream out, boolean full) throws IOException {
        if (description != null) {
            out.write((Const.TAB + "<" + Const.NODE + " name=\"" + name
                    + "\" description=\"" + description + "\" >" + Const.EOL)
                    .getBytes());
        } else {
            out
                    .write((Const.TAB + "<" + Const.NODE + " name=\"" + name
                            + " >" + Const.EOL).getBytes());
        }
        engineInAttribute.writeXML(out, full);
        engineProceedAttribute.writeXML(out, full);
        engineOutAttribute.writeXML(out, full);
        out.write((Const.TAB + "</" + Const.NODE + ">" + Const.EOL).getBytes());
    }

    /**
     * Check properties.
     * @throws OdisAttributeException the odis attribute exception
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#checkProperties()
     */
    public void checkProperties() throws OdisAttributeException {
        if (strPersistent.trim().equalsIgnoreCase(Const.YES)) {
            persistent = true;
        } else if (strPersistent.trim().equalsIgnoreCase(Const.NO)) {
            persistent = false;
        } else {
            StringBuffer buff = new StringBuffer("per \"");
            buff.append(name);
            buff.append("\" contains bad property It can be only ");
            buff.append(Const.YES);
            buff.append(" or ");
            buff.append(Const.NO);
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
        if ((name.indexOf(":") > 0) || (name.indexOf(" ") > 0)) {
            StringBuffer buff = new StringBuffer("name \"");
            buff.append(name);
            buff.append("\" contains invalid character (':' or ' ').");
            LOGGER.log(BasicLevel.ERROR, buff.toString());
            throw new OdisAttributeException(buff.toString());
        }
    }

    /**
     * getter.
     * @return boolean
     */
    public boolean isPersistant() {
        return persistent;
    }

    /**
     * getter.
     * @return Node
     */
    public Node getNode() {
        return node;
    }

    /**
     * setter.
     * @param pNode Node
     */
    public void setNode(Node pNode) {
        this.node = pNode;
    }

    /**
     * To string.
     * @return the string
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buff = new StringBuffer("NodeAttribut(");
        buff.append(getName());
        buff.append(")@");
        return buff.toString();
    }

    /**
     * Compares the specified object with <i>this</i> for equality. Returns <tt>true</tt> if and only if the specified object is also a
     * NodeAttribut and their discriminating parameters are <i>equal</i>.
     * @param obj the object to be compared for equality with <i>this</i>.
     * @return <tt>true</tt> if the specified object is equal to <i>this</i>.
     */
    public boolean equals(Object obj) {
        boolean result = false;
        NodeAttribute copy = null;
        try {
            copy = (NodeAttribute) obj;
        } catch (ClassCastException e) {
            return false;
        }
        result = copy.name.equals(this.name);
        boolean temp;
        if (result) {
            temp = copy.description.equals(this.description);
            result &= temp;
            temp = copy.strPersistent.equals(this.strPersistent);
            result &= temp;
            if (!temp) {
                LOGGER.log(BasicLevel.WARN, traceBadCompare(this, copy, name));
            }
            temp = copy.engineInAttribute.equals(this.engineInAttribute);
            result &= temp;
            if (!temp) {
                LOGGER.log(BasicLevel.WARN, traceBadCompare(engineInAttribute,
                        copy.engineInAttribute, name));
            }
            temp = copy.engineOutAttribute.equals(this.engineOutAttribute);
            result &= temp;
            if (!temp) {
                LOGGER.log(BasicLevel.WARN, traceBadCompare(engineOutAttribute,
                        copy.engineOutAttribute, name));
            }
            temp = copy.engineProceedAttribute
                    .equals(this.engineProceedAttribute);
            result &= temp;
            if (!temp) {
                LOGGER.log(BasicLevel.WARN,
                        traceBadCompare(engineProceedAttribute,
                                copy.engineProceedAttribute, name));
            }
        }
        return result;
    }

    static private String declarationConfliting(String name) {
        StringBuffer errMsg = new StringBuffer(Const.EOL);
        errMsg = new StringBuffer("declaration on node ");
        errMsg.append(name);
        errMsg.append("is conficting");
        errMsg.append(Const.EOL);
        return errMsg.toString();
    }

    static private String traceBadCompare(AbstractAttribute attrib1,
            AbstractAttribute attrib2, String nodeName) {
        StringBuffer errMsg = new StringBuffer(declarationConfliting(nodeName));
        try {
            ByteArrayOutputStream byteStream;
            byteStream = new ByteArrayOutputStream();
            attrib1.writeXML(byteStream, true);
            errMsg.append(byteStream.toString());
            byteStream = new ByteArrayOutputStream();
            attrib2.writeXML(byteStream, true);
            errMsg.append("versus ");
            errMsg.append(Const.EOL);
            errMsg.append(byteStream.toString());
        } catch (IOException e) {
        }
        return errMsg.toString();
    }

    /**
     * Gets the P name.
     * @return String
     * @see com.AbstractAttribute.odis.model.AbstractAttribut#getPName()
     */
    public String getPName() {
        StringBuffer buffer = new StringBuffer(getParent().getPName());
        buffer.append(Const.CST_SEP);
        buffer.append(name);
        return buffer.toString();
    }

    /**
     * @return
     */
    public String getStrPersistant() {
        return strPersistent;
    }
}
