package org.ow2.odis.node;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.bufferIn.BufferInManager;
import org.ow2.odis.bufferIn.PersistenceBufferInManager;
import org.ow2.odis.bufferOut.IBufferOutWatcher;
import org.ow2.odis.connection.ConnectionManager;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.engine.EngineIn;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.engine.EngineProceed;
import org.ow2.odis.exception.OdisEngineException;
import org.ow2.odis.exception.OdisNodeException;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.lifeCycle.Cause;
import org.ow2.odis.lifeCycle.engine.AbstractEngineState;
import org.ow2.odis.lifeCycle.engine.StateEngineInitialized;
import org.ow2.odis.lifeCycle.engine.StateEngineStarted;
import org.ow2.odis.lifeCycle.engine.StateEngineStopped;
import org.ow2.odis.lifeCycle.engine.StateEngineSuspended;
import org.ow2.odis.lifeCycle.engine.StateEngineUnloaded;
import org.ow2.odis.lifeCycle.node.AbstractNodeState;
import org.ow2.odis.lifeCycle.node.INodeLifeCycle;
import org.ow2.odis.lifeCycle.node.StateNodeInitialized;
import org.ow2.odis.lifeCycle.node.StateNodeLoaded;
import org.ow2.odis.lifeCycle.node.StateNodeStarted;
import org.ow2.odis.lifeCycle.node.StateNodeStopped;
import org.ow2.odis.lifeCycle.node.StateNodeSuspended;
import org.ow2.odis.lifeCycle.node.StateNodeUnloaded;
import org.ow2.odis.model.AbstractAttribute;
import org.ow2.odis.model.NodeAttribute;
import org.ow2.odis.util.TraceException;

/**
 * Node classes contains 3 engines, 1 synchronized manager. Copyright France Telecom 2005, All Rights Reserved.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class Node implements INodeLifeCycle, NodeMBean {
    /** logger of this class. */
    private static final Logger LOGGER = Monolog.initialize().getLogger(
            Node.class.getName());

    /** Engine In is responsible of getting new information for the node. */
    private EngineIn engineIn = null;

    /** Engine Out is responsible of sending message for the node. */
    private EngineOut engineOut = null;

    /** Engine Proceed is responsible of processing object for the Node. */
    private EngineProceed engineProceed = null;

    /**
     * Synchronized Manager In is responsible of storing message witch has been received by engine In.
     */
    private BufferInManager synchronizeManagerIn = null;

    /** contain all properties of this node. */
    public final NodeAttribute attribute;

    /** lifeState design the state of the Node. */
    private AbstractNodeState lifeState = new StateNodeLoaded();

    /**
     * getter.
     * @return AbstractNodeState
     */
    public AbstractNodeState getNodeState() {
        return lifeState;
    }

    /**
     * Gets the name.
     * @return node's name.
     */
    public String getName() {
        return attribute.getName();
    }

    /**
     * Gets the p name.
     * @return the p name
     */
    public String getPName() {
        return attribute.getPName();
    }

    /**
     * Gets the parent.
     * @return the parent
     */
    public AbstractAttribute getParent() {
        return attribute.getParent();
    }

    /**
     * To string.
     * @return use name to identify the node. if the node isn't already defined the object reference is used.
     */
    public String toString() {
        return attribute.getName();
    }

    /**
     * Instantiates a new node.
     * @param attribute the attribute
     */
    public Node(final NodeAttribute attribute) {
        super();
        this.attribute = attribute;
        this.attribute.setNode(this);
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "resolveComponent node "
                    + attribute.getName());
        }
        JMX.getInstance().registerMBean(attribute.getName(), this);
        engineIn = new EngineIn(this, attribute.getEngineInAttribute());
        engineOut = new EngineOut(this, attribute.getEngineOutAttribute());
        engineProceed = new EngineProceed(this, attribute
                .getEngineProceedAttribute());
        // <JDO warning="This is quick and dirty for initAndStartByString()">
        boolean shouldBePersistent = false;
        if ((attribute.getStrPersistant() != null) &&
            attribute.getStrPersistant().equalsIgnoreCase("YES")) {
            shouldBePersistent = true;
        }
        // </JDO>
        if (attribute.isPersistant() || shouldBePersistent) {
            StringBuffer path = new StringBuffer(Domain.getInstance().getPath());
            path.append(File.separator);
            path.append(getName());
            synchronizeManagerIn = new PersistenceBufferInManager(
                    engineProceed, engineIn, path.toString());
        } else {
            synchronizeManagerIn = new BufferInManager(engineProceed, engineIn);
        }
        synchronizeManagerIn.init();
    }

    /**
     * Start each node's engine.
     * @throws OdisNodeException the odis node exception
     */
    public void launchComponent() throws OdisNodeException {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "launch node \"" + attribute.getName()
                    + "\"");
        }
        AbstractEngineState state;
        List lsExceptions = new ArrayList();
        try {
            state = engineIn.getEngineState();
            state.setLifeState(engineIn, new StateEngineStarted());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineOut.getEngineState();
            state.setLifeState(engineOut, new StateEngineStarted());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineProceed.getEngineState();
            state.setLifeState(engineProceed, new StateEngineStarted());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisNodeException(errors);
        }
    }

    /**
     * Gets the engine in.
     * @return engineIn.
     */
    public EngineIn getEngineIn() {
        return engineIn;
    }

    /**
     * Gets the engine out.
     * @return engineOut.
     */
    public EngineOut getEngineOut() {
        return engineOut;
    }

    /**
     * Gets the engine proceed.
     * @return engineProceed.
     */
    public EngineProceed getEngineProceed() {
        return engineProceed;
    }

    /**
     * Gets the synchronize manager in.
     * @return synchronizeManagerIn.
     */
    public BufferInManager getSynchronizeManagerIn() {
        return synchronizeManagerIn;
    }

    /**
     * Gets the description.
     * @return description.
     */
    public String getDescription() {
        return attribute.getDescription();
    }

    /**
     * when a node is initialized the engine In, out, and proceed are initialized too.
     * @throws OdisNodeException the odis node exception
     */
    public void initComponent() throws OdisNodeException {
        AbstractEngineState state;
        List lsExceptions = new ArrayList();
        try {
            state = engineIn.getEngineState();
            state.setLifeState(engineIn, new StateEngineInitialized());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineOut.getEngineState();
            state.setLifeState(engineOut, new StateEngineInitialized());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineProceed.getEngineState();
            state.setLifeState(engineProceed, new StateEngineInitialized());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            throw new OdisNodeException("[" + getName() + "]" + errors);
        }
    }

    /**
     * when a node is suspended the engine In, out, and proceed are suspended too.
     */
    public void suspendComponent() {
        AbstractEngineState state;
        List lsExceptions = new ArrayList();
        try {
            state = engineIn.getEngineState();
            state.setLifeState(engineIn, new StateEngineSuspended());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineOut.getEngineState();
            state.setLifeState(engineOut, new StateEngineSuspended());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineProceed.getEngineState();
            state.setLifeState(engineProceed, new StateEngineSuspended());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            LOGGER.log(BasicLevel.FATAL, "unexpected exception !! " + errors);
        }
    }

    /**
     * Stops the Node by stopping the engines (In, Proceed and Out).
     */
    public void stopComponent() {
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            LOGGER.log(BasicLevel.INFO, "Stopping node " + getName());
        }
        AbstractEngineState state;
        List lsExceptions = new ArrayList();
        try {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Stopping EngineIn for node "
                        + getName());
            }
            state = engineIn.getEngineState();
            state.setLifeState(engineIn, new StateEngineStopped());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Stopping EngineProceed for node "
                        + getName());
            }
            state = engineProceed.getEngineState();
            state.setLifeState(engineProceed, new StateEngineStopped());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            if (LOGGER.isLoggable(BasicLevel.INFO)) {
                LOGGER.log(BasicLevel.INFO, "Stopping EngineOut for node "
                        + getName());
            }
            state = engineOut.getEngineState();
            state.setLifeState(engineOut, new StateEngineStopped());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            LOGGER.log(BasicLevel.FATAL, "unexpected exception !! " + errors);
        }
    }

    /**
     * Unload component.
     * @see com.francetelecom.odis.component.lifeCycle.node.INodeLifeCycle#unloadComponent()
     */
    public void unloadComponent() {
        AbstractEngineState state;
        List lsExceptions = new ArrayList();
        try {
            state = engineIn.getEngineState();
            state.setLifeState(engineIn, new StateEngineUnloaded());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineOut.getEngineState();
            state.setLifeState(engineOut, new StateEngineUnloaded());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        try {
            state = engineProceed.getEngineState();
            state.setLifeState(engineProceed, new StateEngineUnloaded());
        } catch (OdisEngineException e) {
            lsExceptions.add(e);
        }
        if (!lsExceptions.isEmpty()) {
            String errors = TraceException.formatListException(lsExceptions);
            LOGGER.log(BasicLevel.FATAL, "unexpected exception !! " + errors);
        }
        synchronizeManagerIn = null;
        JMX.getInstance().unregisterMBean(getName(), this);
        NodeManager.unregister(this);
        ConnectionManager.unregister(getName());
    }

    // -----------------------------------------------
    // implement ILifeCycle NodeMBean
    // -----------------------------------------------
    /**
     * Sets the state.
     * @param state the state
     * @return true if succeed false in other case
     */
    public boolean setState(int state) {
        boolean result = true;
        AbstractNodeState wantedState = null;
        Cause cause = new Cause(Cause.USER_REQUEST);
        switch (state) {
        case INodeLifeCycle.STATE_LOADED:
            wantedState = new StateNodeLoaded();
            break;
        case INodeLifeCycle.STATE_INITIALIZED:
            wantedState = new StateNodeInitialized();
            break;
        case INodeLifeCycle.STATE_STARTED:
            wantedState = new StateNodeStarted();
            cause.setUserLevelRequest(Cause.NODE_REQUEST);
            getEngineIn().behavior.removeUserCause(cause);
            getEngineOut().getBehavior().removeUserCause(cause);
            getEngineProceed().getBehavior().removeUserCause(cause);
            break;
        case INodeLifeCycle.STATE_SUSPENDED:
            wantedState = new StateNodeSuspended();
            cause.setUserLevelRequest(Cause.NODE_REQUEST);
            getEngineIn().behavior.addCause(cause);
            getEngineOut().getBehavior().addCause(cause);
            getEngineProceed().getBehavior().addCause(cause);
            break;
        case INodeLifeCycle.STATE_STOPPED:
            wantedState = new StateNodeStopped();
            break;
        case INodeLifeCycle.STATE_UNLOADED:
            wantedState = new StateNodeUnloaded();
            break;
        default:
            LOGGER.log(BasicLevel.ERROR, "Unknown state " + state);
            result = false;
            break;
        }
        if (wantedState != null) {
            try {
                lifeState.setLifeState(this, wantedState);
            } catch (OdisNodeException e) {
                result = false;
                throw new RuntimeException(e.getMessage());
            }
        }
        return result;
    }

    /**
     * Gets the number of pending object.
     * @return the number of object present into the buffer in of the node
     */
    public long getNumberOfPendingObject() {
        return synchronizeManagerIn.getNumberObject();
    }

    /**
     * Gets the size of pending object.
     * @return the number of object present into the buffer in of the node
     */
    public long getSizeOfPendingObject() {
        return synchronizeManagerIn.getSizeByte();
    }

    /**
     * Gets the proceed invocation.
     * @return the number of proceed invocation
     */
    public long getProceedInvocation() {
        return engineProceed.getAttribute().getInvocation();
    }

    /**
     * getter.
     * @return the number of sending message not achieved directly
     */
    public long getSendingError() {
        return engineOut.getNbrError();
    }

    /**
     * Gets the proceed error.
     * @return the proceed error
     * @see com.francetelecom.odis.component.node.NodeMBean#getProceedError()
     */
    public long getProceedError() {
        return engineProceed.getAttribute().getNbrProceedError();
    }

    /**
     * Gets the state.
     * @return the state
     * @see com.francetelecom.odis.component.node.NodeMBean#getState()
     */
    public String getState() {
        return lifeState.getStateName();
    }

    /**
     * Gets the state engine in.
     * @return the state engine in
     * @see com.francetelecom.odis.component.node.NodeMBean#getStateEngineIn()
     */
    public String getStateEngineIn() {
        return engineIn.getEngineState().getStateName();
    }

    /**
     * Gets the state engine out.
     * @return the state engine out
     * @see com.francetelecom.odis.component.node.NodeMBean#getStateEngineOut()
     */
    public String getStateEngineOut() {
        return engineOut.getEngineState().getStateName();
    }

    /**
     * Gets the state engine proceed.
     * @return the state engine proceed
     * @see com.francetelecom.odis.component.node.NodeMBean#getStateEngineProceed()
     */
    public String getStateEngineProceed() {
        return engineProceed.getEngineState().getStateName();
    }

    /**
     * Gets the decoding error.
     * @return the decoding error
     * @see com.francetelecom.odis.component.node.NodeMBean#getDecodingError()
     */
    public long getDecodingError() {
        return engineIn.getNbrErrorDecoding();
    }

    /**
     * Gets all buffer out types for this Node.
     * @return the list of all buffer out types used
     * @see com.francetelecom.odis.component.node.NodeMBean#getBufferOutType()
     */
    public String[] getBufferOutType() {
        return engineOut.getBufferOutManager().getBufferOutList();
    }

    /**
     * Checks if is able to send messages.
     * @return true if the last message cannot be sent
     */
    public boolean isAbleToSendMessage() {
        return !engineOut.isOnError();
    }

    /**
     * Sets the node state.
     * @param state the state
     */
    public void setNodeState(AbstractNodeState state) {
        this.lifeState = state;
        if (LOGGER.isLoggable(BasicLevel.INFO)) {
            StringBuffer buffer = new StringBuffer(toString());
            buffer.append(" set to ");
            buffer.append(state.getStateName());
            LOGGER.log(BasicLevel.INFO, buffer.toString());
        }
    }

    /**
     * Gets the processing interval.
     * @return the interval
     */
    public long getInterval() {
        return attribute.getEngineProceedAttribute().getInterval();
    }

    /**
     * Forwards the provided cause to the appropriate behaviors according to the kind of cause.
     * @param c the cause
     */
    public void addCause(Cause c) {
        if (c.getValue() == Cause.BUFFER_OUT_OVERFLOW) {
            engineIn.getBehavior().addCause(c);
            engineProceed.getBehavior().addCause(c);
        }
        // other cases handled elsewhere for the moment being...
    }

    /**
     * Take appropriate measures on a cause ending according to the kind of cause.
     * @param c the cause
     */
    public void causeEnds(Cause c) {
        c.setEnd();
        if (c.getValue() == Cause.BUFFER_OUT_OVERFLOW) {
            engineIn.getBehavior().checkBehaviours();
            engineProceed.getBehavior().checkBehaviours();
        }
        // other cases handled elsewhere for the moment being...
    }

    /**
     * Adds a BufferOut watcher.
     * @param bow
     */
    public void addBufferOutWatcher(IBufferOutWatcher bow) {
        engineOut.addBufferOutWatcher(bow);
    }
}
