package org.ow2.odis.node;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jdom.Element;
import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.admin.Launch;
import org.ow2.odis.domain.Domain;
import org.ow2.odis.exception.OdisException;
import org.ow2.odis.exception.OdisNodeException;
import org.ow2.odis.jmx.JMX;
import org.ow2.odis.lifeCycle.node.INodeLifeCycle;
import org.ow2.odis.model.NodeAttribute;

/**
 * The aim of this class is to manage the Node. Copyright France Telecom 2005, All Rights Reserved.
 * @author Beyler Olivier FranceTelecom R&D
 */
public final class NodeManager implements NodeManagerMBean {
    /** logger of this class. */
    private final static Logger LOGGER = Monolog.initialize().getLogger(
            NodeManager.class.getName());

    /** unique instance of this class. Implement singleton pattern. */
    static NodeManager instance = new NodeManager();

    /** The Constant LINE_SEP. */
    static final String LINE_SEP = "===========================";

    /** The Constant CST_VALUE. */
    static final String CST_VALUE = "value";

    /** The Constant WARN_DEPRECATED. */
    static final String WARN_DEPRECATED = "USED DEPECRATED PROPERTY METHODE please change your config file";

    /**
     * singleton pattern.
     * @return unique instance of NodeManager
     */
    public static NodeManager getInstance() {
        return instance;
    }

    /**
     * map beetween the name of the node and the node. Aim of this map is to be able to retrieve quickly a nome by his name.
     */
    private static Map mapNodeName = new HashMap();

    /**
     * Instantiates a new node manager.
     */
    private NodeManager() {
        super();
        JMX.getInstance().registerMBean("Node Manager", this);
    }

    /**
     * create a Node attribut.
     * @param description JDom Element
     * @return NodeAttribut
     */
    public NodeAttribute createNodeAttribut(Element description) {
        NodeAttribute nodeAttribut = new NodeAttribute(
                Domain.getInstance().attribute);
        nodeAttribut.setter(description);
        return nodeAttribut;
    }

    /**
     * launch each node.
     * @throws OdisNodeException the odis node exception
     */
    public void launchComponent() throws OdisNodeException {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        List lsExceptions = null;
        for (int i = 0; i < nodes.length; i++) {
            try {
                nodes[i].launchComponent();
            } catch (OdisException e) {
                if (lsExceptions == null) {
                    lsExceptions = new ArrayList();
                }
                lsExceptions.add(e.getMessage());
            }
        }
        if (lsExceptions != null && !lsExceptions.isEmpty()) {
            throw new OdisNodeException(lsExceptions.toString());
        }
    }

    /**
     * return the node associated to to a name.
     * @param name the name
     * @return Node
     */
    public Node getNode(final String name) {
        return (Node) mapNodeName.get(name);
    }

    /**
     * initialize the NodeManager.
     * @throws OdisNodeException can occured if the initialization failed.
     */
    public void initComponent() throws OdisNodeException {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        List lsExceptions = null;
        for (int i = 0; i < nodes.length; i++) {
            try {
                nodes[i].initComponent();
            } catch (OdisException e) {
                if (lsExceptions == null) {
                    lsExceptions = new ArrayList();
                }
                lsExceptions.add(e.getMessage());
            }
        }
        if (lsExceptions != null && !lsExceptions.isEmpty()) {
            throw new OdisNodeException(lsExceptions.toString());
        }
    }

    /**
     * initialize Nodes by a string.
     * @param config String
     * @deprecated use Domain initByString
     */
    public void initByString(String config) {
        Domain domain = Domain.getInstance();
        LOGGER.log(BasicLevel.WARN,
                "use deprecated initString on NodeManager :\n " + config);
        domain.initByString(config);
    }

    /**
     * suspend all nodes referenced by the node manager. To do this, we use the same way as a user would use if he do it by a JMX
     * invocation. It doesn't apply directly on live cycle as it usually do.
     */
    public void suspendComponent() {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        for (int i = 0; i < nodes.length; i++) {
            nodes[i].setState(INodeLifeCycle.STATE_SUSPENDED);
        }
    }

    /**
     * Stop each node. To do this, we use the same way as a user would use if he do it by a JMX invocation. It doesn't apply directly on
     * live cycle as it usually do.
     */
    public void stopComponent() {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        for (int i = 0; i < nodes.length; i++) {
            if (nodes[i].getNodeState().getState() == INodeLifeCycle.STATE_STARTED) {
                nodes[i].setState(INodeLifeCycle.STATE_SUSPENDED);
            }
            nodes[i].setState(INodeLifeCycle.STATE_STOPPED);
        }
    }

    /**
     * Unload the nodeManager and reinitialize the singleton to a new one.
     */
    public void unloadComponent() {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        for (int i = 0; i < nodes.length; i++) {
            nodes[i].unloadComponent();
        }
        JMX.getInstance().unregisterMBean("Node Manager", this);
        instance = new NodeManager();
    }

    /**
     * unregister a node to the Nodemanager. Consequence of this action is : <br>
     * this node won't be able to be accessed by nodemanager and if no other component know this node, the garbage collector will destroy
     * it.
     * @param node Node
     */
    public static void unregister(final Node node) {
        final Object test = mapNodeName.remove(node.getName());
        if (test == null && LOGGER.isLoggable(BasicLevel.WARN)) {
            final StringBuffer warnMsg = new StringBuffer("try to remove node ");
            warnMsg.append(node.getName());
            warnMsg.append(" witch is unknown by the NodeManager");
            LOGGER.log(BasicLevel.WARN, warnMsg.toString());
        }
        if (node != null) {
            Domain domain = Domain.getInstance();
            domain.attribute.lsNodeAttributs.remove(node.attribute);
        }
    }

    /**
     * register a node to the nodemanager.
     * @param node Node
     * @throws OdisNodeException if a node using the same name is already present
     */
    public static void register(final Node node) throws OdisNodeException {
        if (mapNodeName.containsKey(node.getName())) {
            final StringBuffer warnMsg = new StringBuffer("try to add node ");
            warnMsg.append(node.getName());
            warnMsg.append(" witch is already known by the NodeManager");
            LOGGER.log(BasicLevel.WARN, warnMsg.toString());
            throw new OdisNodeException(warnMsg.toString());
        }
        mapNodeName.put(node.getName(), node);
    }

    /**
     * get the list of the node's name.
     * @return string array.
     * @see com.francetelecom.odis.component.node.NodeManagerMBean#getListNodes()
     */
    public String[] getListNodes() {
        final Node[] nodes = (Node[]) mapNodeName.values().toArray(
                new Node[mapNodeName.size()]);
        final String[] result = new String[nodes.length];
        for (int index = 0; index < result.length; index++) {
            result[index] = nodes[index].getName();
        }
        return result;
    }

    /**
     * This main function should be not use.
     * @param args String name of the config file
     * @throws OdisException Exception on error
     * @deprecated Use Launch methode to beneficiate of last framework evolution.
     */
    public static void main(String[] args) throws OdisException {
        final String config = "config.xml";
        String[] newArgue = new String[2];
        newArgue[0] = "-add";
        if (args.length == 0) {
            newArgue[1] = config;
            Launch.main(newArgue);
        } else if (args.length == 1) {
            newArgue[1] = args[0];
            Launch.main(newArgue);
        } else {
            LOGGER.log(BasicLevel.ERROR, "Invalid argument number");
        }
    }
}
