package org.ow2.odis.policy;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.objectweb.util.monolog.api.BasicLevel;

/**
 * This policy will elected element with the higher priority using roundrobin.
 * @author Beyler Olivier FranceTelecom R&D
 */
public final class RoundRobinWithPriorityPolicy extends AbstractPolicy {
    /**
     * map witch associate a priority and the list of element having this priority.
     */
    private final transient Map mapPriorityList = new HashMap();

    /**
     * list witch contains all priority present into this policy
     */
    private final transient List lsPriorities = new ArrayList();

    /**
     * constructor.
     */
    public RoundRobinWithPriorityPolicy() {
        super();
    }

    /**
     * constructor by copy.
     * @param copy
     */
    public RoundRobinWithPriorityPolicy(RoundRobinWithPriorityPolicy copy) {
        this();
        duplicate(copy);
    }

    /**
     * overload the addPolicyElement.
     * @param element
     * @return true if the element wasn't already present into this policy
     * @see com.francetelecom.odis.component.policy.AbstractPolicy#addPolicyElement(com.francetelecom.odis.component.policy.IPolicyElement)
     */
    public boolean addPolicyElement(IPolicyElement element) {
        final boolean result = super.addPolicyElement(element);
        if (result) {
            List lsElement;// ensemble d'element ayant le meme niveau de
            // priorite
            lsElement = (List) mapPriorityList.get(new Integer(element
                    .getPriority()));
            final Integer priority = new Integer(element.getPriority());
            if (lsElement == null) {
                synchronized (this) {
                    lsElement = (List) mapPriorityList.get(priority);
                    if (lsElement == null) {
                        // the list of priority is unknow will have to create
                        // it.
                        lsElement = new LinkedList();
                        lsPriorities.add(priority);
                        Collections.sort(lsPriorities);
                        mapPriorityList.put(priority, lsElement);
                    }
                }
            }
            lsElement.add(element);
        }
        return result;
    }

    /**
     * find the next element and compare it with the other element witch can be elected to be sure that no one element whith a best (ie
     * inferior) priority can be elected.
     * @return the selected element
     * @see com.francetelecom.odis.component.policy.AbstractPolicy#getPolicyElement()
     */
    public IPolicyElement getPolicyElement() {
        IPolicyElement result = null;
        // recherche en essayant de trouver l'element en recherchant d'abord
        // dans les listes de plus haute priorite
        final Iterator itPriorite = lsPriorities.iterator();
        LinkedList lsElement; // ensemble des elements a tester ayant le meme
        // niveau de priorite
        Integer priority; // sert au parcours des niveaux de priorite
        IPolicyElement element; // element qui peut etre elue si il a des
        // messages
        int nbr;
        while (itPriorite.hasNext() && result == null) {
            priority = (Integer) itPriorite.next();
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "search on priority list :"
                        + priority);
            }
            lsElement = (LinkedList) mapPriorityList.get(priority);
            if (lsElement != null && !lsElement.isEmpty()) {
                final Iterator itElement = lsElement.listIterator();
                while (itElement.hasNext() && result == null) {
                    element = (IPolicyElement) itElement.next();
                    nbr = element.getPendingMessage();
                    if (nbr > 0) {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG, "find "
                                    + element.toString() + " with " + nbr
                                    + " waiting object");
                        }
                        result = element;
                    } else {
                        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                            LOGGER.log(BasicLevel.DEBUG, element.toString()
                                    + " has no waiting object");
                        }
                    }
                }
                if (result != null) {
                    // enleve l'element ou qu il se trouve dans la liste des
                    // element a traiter
                    lsElement.remove(result);
                    // le replacer en fin de la liste
                    lsElement.add(result);
                }
            } else { // cas anomal de traitement
                if (lsElement == null) {
                    LOGGER.log(BasicLevel.ERROR,
                            "list not found for priority :" + priority);
                } else {
                    LOGGER.log(BasicLevel.ERROR, "list empty for priority :"
                            + priority);
                }
            }
        }
        if (result == null) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG, "no Element has pendingMessage");
            }
        }
        return result;
    }
}
