package org.ow2.odis.port;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.IConnectionIn;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.decoder.IDecoder;
import org.ow2.odis.engine.EngineIn;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionInitialized;
import org.ow2.odis.lifeCycle.connection.StateConnectionStarted;
import org.ow2.odis.lifeCycle.connection.StateConnectionStopped;
import org.ow2.odis.lifeCycle.connection.StateConnectionUnloaded;
import org.ow2.odis.model.FromAttribute;
import org.ow2.odis.policy.IPolicyElement;

/**
 * The Class PortIn.
 * @author olivier beyler FT R&D MAPS AMS Grenoble
 */
public class PortIn implements IPolicyElement {
    // ---------------------------------------
    // attributes
    // ---------------------------------------
    /** keep a reference to the connection In. */
    private IConnectionIn connectionIn = null;

    /** Keep a reference to the engineIn. */
    private EngineIn engineIn = null;

    /** attribute witch contains properties of this component. */
    final public FromAttribute attribut;

    /** logger of this class. */
    final static Logger LOGGER = Monolog.initialize().getLogger(PortIn.class.getName());

    /**
     * constructor.
     * @param attribute the attribute
     */
    public PortIn(final FromAttribute attribute) {
        this.attribut = attribute;
    }

    /**
     * ask to the connection if there is some pending message.
     * @return the pending message
     */
    public int getPendingMessage() {
        return connectionIn.getPendingMessage();
    }

    /**
     * On Message wake up EngineIn.
     */
    public void onMessage() {
        engineIn.activeRead();
    }

    /**
     * Gets the context message.
     * @return the context message
     */
    public ContextMessage getContextMessage() {
        return connectionIn.getContextMessage();
    }

    /**
     * ack the message to the connection.
     * @param pValue the value
     */
    public void ackMessage(boolean pValue) {
        connectionIn.ackMessage(pValue);
    }

    /**
     * implement comparable interface.
     * @param arg0 the arg0
     * @return -1,0,1 depends of link's priority
     */
    public int compareTo(Object arg0) {
        int result = 0;
        final PortIn portIn = (PortIn) arg0;
        int priority = getPriority();
        if (priority > portIn.getPriority()) {
            result = 1;
        } else if (priority < portIn.getPriority()) {
            result = -1;
        }
        return result;
    }

    /**
     * Sets the engine.
     * @param engine the new engine
     */
    public void setEngine(EngineIn engine) {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "setEngine..In");
        }
        engineIn = engine;
    }

    /**
     * Gets the decoder.
     * @return the decoder
     */
    public IDecoder getDecoder() {
        return attribut.getDecoderInstance();
    }

    /**
     * Gets the priority.
     * @return the priority
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getPriority()
     */
    public int getPriority() {
        return attribut.getPriority();
    }

    /**
     * To string.
     * @return the string
     * @see java.lang.Object#toString()
     */
    public String toString() {
        StringBuffer buffer = new StringBuffer("PortIn_");
        buffer.append(attribut.getLink());
        buffer.append("(");
        buffer.append(super.hashCode());
        buffer.append(")");
        return buffer.toString();
    }

    /**
     * ask to initialize the connection.
     * @throws OdisConnectionException if the connection refused to be initialized
     */
    public void initComponent() throws OdisConnectionException {
        if (connectionIn==null){
            throw new OdisConnectionException("no connection available");
        }
        AbstractConnectionState state = connectionIn.getConnectionState();
        state.setLifeState(connectionIn, new StateConnectionInitialized());
    }

    /**
     * ask to start the connection.
     * @throws OdisConnectionException if the connectionIn refuse to to be started.
     */
    public void launchComponent() throws OdisConnectionException {
        AbstractConnectionState state = connectionIn.getConnectionState();
        state.setLifeState(connectionIn, new StateConnectionStarted());
    }

    /**
     * stop the connection associated to this portIn.
     */
    public void stopComponent() {
        AbstractConnectionState state = connectionIn.getConnectionState();
        try {
            state.setLifeState(connectionIn, new StateConnectionStopped());
        } catch (OdisConnectionException e) {
            LOGGER.log(BasicLevel.ERROR, "Unexpected exception on stop portIn", e);
        }
    }

    /**
     * ask to connection to be unloaded.
     */
    public void unloadComponent() {
        AbstractConnectionState state = connectionIn.getConnectionState();
        try {
            state.setLifeState(connectionIn, new StateConnectionUnloaded());
        } catch (OdisConnectionException e) {
            LOGGER.log(BasicLevel.ERROR, "Unexpected exception on unload portIn", e);
        }
    }

    /**
     * getter.
     * @return IConnectionIn
     */
    public IConnectionIn getConnectionIn() {
        return connectionIn;
    }

    /**
     * setter.
     * @param connectionIn IConnectionIn
     */
    public void setConnectionIn(IConnectionIn connectionIn) {
        this.connectionIn = connectionIn;
    }

    /**
     * getter.
     * @return String Id
     * @see com.francetelecom.odis.component.policy.IPolicyElement#getId()
     */
    public String getId() {
        return attribut.getLink();
    }
}
