package org.ow2.odis.port;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.bufferOut.BufferOut;
import org.ow2.odis.connection.IConnectionOut;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.connection.context.IContextualConnectionOut;
import org.ow2.odis.engine.EngineOut;
import org.ow2.odis.exception.OdisConnectionException;
import org.ow2.odis.lifeCycle.connection.AbstractConnectionState;
import org.ow2.odis.lifeCycle.connection.StateConnectionInitialized;
import org.ow2.odis.lifeCycle.connection.StateConnectionStarted;
import org.ow2.odis.lifeCycle.connection.StateConnectionStopped;
import org.ow2.odis.lifeCycle.connection.StateConnectionUnloaded;
import org.ow2.odis.model.ToAttribute;
import org.ow2.odis.policy.IPolicyElement;

/**
 * The Class PortOut.
 * 
 * @author Beyler Olivier FranceTelecom R&D
 */
public class PortOut implements IPolicyElement {
	/** Logger for this class. */
	static final Logger LOGGER = Monolog.initialize().getLogger(
			PortOut.class.getName());

	/** The associated ConnectionOut. */
	private IConnectionOut connectionOut = null;

	/** The EngineOut this PortOut belongs to. */
	private final EngineOut engineOut;

	/** The attribute containing all port properties. */
	private final ToAttribute attribut;

	/** Denotes that the connection associated with this port handles contexts. */
	boolean isContextual = false;

	/**
	 * Instantiates a new PortOut for the provided EngineOut and with the
	 * parameters provided in the attribute.<br>
	 * 
	 * @param pAttribut
	 *            the attribute containing all parameters
	 * @param pEngineOut
	 *            the EngineOut the port belongs to
	 */
	public PortOut(ToAttribute pAttribut, EngineOut pEngineOut) {
		super();
		if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
			LOGGER.log(BasicLevel.DEBUG, "create PortOut");
		}
		this.attribut = pAttribut;
		this.engineOut = pEngineOut;
	}

	/**
	 * Sends a message on the associated connection.
	 * 
	 * @param msg
	 *            the message
	 * @return boolean true if succeed false in other case
	 */
	public boolean sendMessage(byte[] msg) {
		if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
			LOGGER.log(BasicLevel.DEBUG, this.toString()
					+ " sendMessage to connection ");
		}
		// send a byte[]
		return connectionOut.send(msg);
	}

	/**
	 * Sends a contextual message on the associated connection.
	 * 
	 * @param msg
	 *            the contextual message
	 * @return true if the send operation succeeded, false if it failed or this
	 *         port is not contextual
	 */
	public boolean sendMessage(ContextMessage msg) {
		if (isContextual) {
			if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
				LOGGER.log(BasicLevel.DEBUG, this.toString()
						+ " sendMessage (with Context) to connection");
			}
			// send a ContextMessage
			return ((IContextualConnectionOut) connectionOut).send(msg);
		} else {
			LOGGER.log(BasicLevel.ERROR, "Non contextual " + this.toString()
					+ " can't send a ContextMessage");
			return false;
		}
	}

	/**
	 * Starts the associated OUT connection.
	 * 
	 * @throws OdisConnectionException
	 *             if the connection failed to start.
	 */
	public void launchComponent() throws OdisConnectionException {
		AbstractConnectionState state = connectionOut.getConnectionState();
		state.setLifeState(connectionOut, new StateConnectionStarted());
	}

	/**
	 * Gets the pending message.
	 * 
	 * @return the number of pending message contain into associated buffer
	 * @see com.francetelecom.odis.component.policy.IPolicyElement#getPendingMessage()
	 */
	public int getPendingMessage() {
		BufferOut bufferOut = null;
		int result = 0;
		bufferOut = engineOut.getBufferOutManager().getBufferOut(
				attribut.getId());
		if (bufferOut != null) {
			int count = bufferOut.getPendingMessage();
			result += count;
			if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
				LOGGER.log(BasicLevel.DEBUG, "" + bufferOut.toString()
						+ " have " + count + " pending message");
			}
		}
		if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
			LOGGER.log(BasicLevel.DEBUG, "PortOut " + " have " + result
					+ " pending message");
		}
		return result;
	}

	/**
	 * Gets a priority value for use by policy.
	 * 
	 * @return the priority
	 */
	public int getPriority() {
		return 0;
	}

	/**
	 * Gets the attribute.
	 * 
	 * @return attribute witch is used to make the relation between link and the
	 *         connection
	 * @see com.francetelecom.odis.component.policy.IPolicyElement#getId()
	 */
	public ToAttribute getAttribut() {
		return attribut;
	}

	/**
	 * Initialize the the component.
	 * 
	 * @throws OdisConnectionException
	 *             the odis connection exception
	 */
	public void initComponent() throws OdisConnectionException {
		if (connectionOut == null) {
			throw new OdisConnectionException("no connection available");
		}
		AbstractConnectionState state = connectionOut.getConnectionState();
		state.setLifeState(connectionOut, new StateConnectionInitialized());
	}

	/**
	 * stop the connection associated to this portIn.
	 */
	public void stopComponent() {
		AbstractConnectionState state = connectionOut.getConnectionState();
		try {
			state.setLifeState(connectionOut, new StateConnectionStopped());
		} catch (OdisConnectionException e) {
			LOGGER.log(BasicLevel.ERROR,
					"Unexpected exception on stop portOut", e);
		}
	}

	/**
	 * ask to connection to be unloaded.
	 */
	public void unloadComponent() {
		if (connectionOut != null) {
			AbstractConnectionState state = connectionOut.getConnectionState();
			try {
				state
						.setLifeState(connectionOut,
								new StateConnectionUnloaded());
			} catch (OdisConnectionException e) {
				LOGGER.log(BasicLevel.ERROR,
						"Unexpected exception on unload portOut", e);
			}
		}
	}

	/**
	 * To string.
	 * 
	 * @return the string
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer buffer = new StringBuffer("PortOut_");
		buffer.append(attribut.getLink());
		buffer.append("(");
		buffer.append(super.hashCode());
		buffer.append(")");
		return buffer.toString();
	}

	/**
	 * Getter.
	 * 
	 * @return String
	 * @see com.francetelecom.odis.component.policy.IPolicyElement#getId()
	 */
	public String getId() {
		return attribut.getId();
	}

	/**
	 * Sets the associated connection.
	 * 
	 * @param pConnectionOut
	 *            any OUT connection
	 */
	public void setConnectionOut(IConnectionOut pConnectionOut) {
		this.connectionOut = pConnectionOut;
	}

	/**
	 * Sets the associated connection.
	 * 
	 * @param pConnectionOut
	 *            a Context-able connection
	 */
	public void setConnectionOut(IContextualConnectionOut pConnectionOut) {
		this.connectionOut = pConnectionOut;
		isContextual = true;
	}

	/**
	 * Checks whether that port handles contextual send operations.
	 * 
	 * @return true if able
	 */
	public boolean isContextual() {
		return isContextual;
	}
}
