package org.ow2.odis.proceed;

import java.util.HashMap;
import java.util.Map;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.Context;
import org.ow2.odis.engine.IProceed;
import org.ow2.odis.node.Node;

/**
 * give the opportunity to access to the node and all of his mechanism from the proceed. <br>
 * <i>note : use singleton pattern</i>
 * @author Beyler Olivier FranceTelecom R&D
 */
public final class ProceedManager {
    /** logger of this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            ProceedManager.class.getName());

    /** unique instance of proceed Manager (implement singleton pattern). */
    private static ProceedManager instance = new ProceedManager();

    /**
     * map between proceed and node to be able to find node associate to a proceed.
     */
    private final Map mapProceedNode = new HashMap();

    /**
     * implement singleton pattern.
     * @return unique instance of ProceedManager
     */
    public static ProceedManager getInstance() {
        return instance;
    }

    /**
     * Instantiates a new proceed manager.
     */
    private ProceedManager() {
        if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
            LOGGER.log(BasicLevel.DEBUG, "Create a Proceed Manager");
        }
    }

    /**
     * destructor of singleton. Allow destruction of unique instance of proceedManager and create a new one.
     */
    public static void removeInstance() {
        instance.mapProceedNode.clear();
        instance = new ProceedManager();
    }

    /**
     * register a node by this proceed to populate mapProceedNode.
     * @param proceed unique instance of proceed attached to this node.
     * @param node unique instance of node attached to this proceed.
     */
    public void registerProceed(final IProceed proceed, final Node node) {
        if (proceed != null) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "register a new proceed into ProceedManager "
                                + proceed.getClass());
            }
            mapProceedNode.put(proceed, node);
        }
    }

    /**
     * remove the entry of this proceed from the map node. This case occurs when the node is deleted.
     * @param proceed proceed entry of the map.
     */
    public void unregisterProceed(final IProceed proceed) {
        if (proceed != null) {
            if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                LOGGER.log(BasicLevel.DEBUG,
                        "unregister a new proceed into ProceedManager "
                                + proceed.getClass());
            }
            mapProceedNode.remove(proceed);
        }
    }

    /**
     * getter.
     * @param proceed instance of proceed
     * @return unique node attached to this proceed.
     */
    public Node getNode(final IProceed proceed) {
        return (Node) mapProceedNode.get(proceed);
    }

    /**
     * Gets the current context.
     * @param proceed the proceed
     * @return the current context
     */
    public static Context getCurrentReceiveContext(final IProceed proceed) {
        Node node = getInstance().getNode(proceed);
        return node.getSynchronizeManagerIn().getCurrentContext();
    }

    /**
     * Allow user to request a new invocation of proceed by the framework.
     * @param proceed allow identify the node witch need to be activate.
     */
    public void activeProceed(final IProceed proceed) {
        Node node = (Node) mapProceedNode.get(proceed);
        node.getEngineProceed().setRunOnce(true);
        node.getEngineProceed().activeProceed();
    }
}
