package org.ow2.odis.routing;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.connection.context.ContextMessage;
import org.ow2.odis.port.PortOut;

/**
 * The Routing class handles data sending on the adequate ports, based on the destination provided by EngineOut.<br>
 * It handles ports separately depending on whether they are context-capable or not.
 * @author Beyler Olivier FranceTelecom R&D
 */
public class Routing {
    /** Logger for this class. */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            Routing.class.getName());

    /**
     * A map of ArrayLists of PortOuts indexed by simple destination (String), for standard routing.
     */
    final private Map mapRouting;

    /**
     * A map of ArrayLists of context-able PortOuts indexed by simple destination (String), for contextual routing.
     */
    final private Map mapContextualRouting;

    /** Default Constructor. */
    public Routing() {
        super();
        mapRouting = new HashMap();
        mapContextualRouting = new HashMap();
    }

    /**
     * Gets all non contextual ports associated with the specified simple destination.<br>
     * If no port list is currently associated with that destination, an empty list is created and returned.
     * @param destination a simple destination (object type or explicit routing name)
     * @return the ArrayList of ports (as a Collection) associated with the provided destination
     */
    public Collection getPortList(String destination) {
        Collection result = (Collection) mapRouting.get(destination);
        if (result == null) {
            result = new ArrayList();
            mapRouting.put(destination, result);
        }
        return result;
    }

    /**
     * Gets all contextual ports associated with the specified simple destination.<br>
     * If no port list is currently associated with that destination, an empty list is created and returned.
     * @param destination a simple destination (object type or explicit routing name)
     * @return the ArrayList of contextual ports (as a Collection) associated with the provided destination
     */
    public Collection getContextualPortList(String destination) {
        Collection result = (Collection) mapContextualRouting.get(destination);
        if (result == null) {
            result = new ArrayList();
            mapContextualRouting.put(destination, result);
        }
        return result;
    }

    // /**
    // * @param type String
    // * @return Object[]
    // */
    // public Object[] getArray(final String type) {
    // final Collection col = getPortList(type);
    // final Object[] obs = col.toArray();
    // return obs;
    // }
    // /**
    // * @return the list of usable route type.
    // */
    // public Collection getList() {
    // return mapRouting.keySet();
    // }
    /**
     * @return Collection
     */
    public Collection getListPortOut() {
        final Set set = new HashSet();
        Collection list = mapRouting.values();
        Iterator it = list.iterator();
        while (it.hasNext()) {
            ArrayList ls = (ArrayList) it.next();
            Iterator itLs = ls.iterator();
            while (itLs.hasNext()) {
                set.add(itLs.next());
            }
        }
        list = mapContextualRouting.values();
        it = list.iterator();
        while (it.hasNext()) {
            ArrayList ls = (ArrayList) it.next();
            Iterator itLs = ls.iterator();
            while (itLs.hasNext()) {
                set.add(itLs.next());
            }
        }
        return set;
    }

    /**
     * Adds a portOut associated with a simple destination.<br>
     * The port is added to the contextual or non-contextual map, according to its nature.
     * @param destination a simple destination name (object type or explicit routing name)
     * @param portOut the port to be added
     */
    public void addPortOut(final String destination, final PortOut portOut) {
        Map map = portOut.isContextual() ? mapContextualRouting : mapRouting;
        Collection portList = (Collection) map.get(destination);
        if (portList == null) {
            portList = new ArrayList();
            map.put(destination, portList);
        }
        portList.add(portOut);
    }

    // /**
    // * send msg on all portOut. When a msg is sent to a portOut, this portOut
    // is set to null.
    // * @param type the type
    // * @param msg the msg
    // * @param portOuts the port outs
    // * @return true if succeed false in other case.
    // */
    // public boolean sendMessage(final String type, final byte[] msg, final
    // Object[] portOuts) {
    // boolean result = true;
    // PortOut portOut = null;
    // boolean isSended;
    // StringBuffer buffer;
    // for (int i = 0; i < portOuts.length; i++) {
    // portOut = (PortOut) portOuts[i];
    // if (portOut != null) {
    // // just for debugging
    // if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
    // buffer = new StringBuffer("routing to ");
    // buffer.append(portOut);
    // buffer.append(" message =");
    // Byte b = null;
    // for (int index = 0; index < msg.length; index++) {
    // b = new Byte(msg[index]);
    // buffer.append(b);
    // }
    // LOGGER.log(BasicLevel.DEBUG, buffer.toString());
    // }
    // // end just for debugging
    // isSended = portOut.sendMessage(msg);
    // if (isSended) {
    // portOuts[i] = null;
    // }
    // result = result && isSended;
    // }
    // }
    // return result;
    // }
    /**
     * Tries to send the provided message, always non contextually, on all adequate OUT ports.<br>
     * If portOuts is null, the method uses all non contextual ports for that destination. If it is not, the method uses that collection.<br>
     * In both cases, when the method returns, portOuts holds the list of ports on which the sending failed.
     * @param dest the destination
     * @param msg the message to be sent
     * @param portOuts the ports that should be used
     * @return true if the data was successfully sent on all selected destinations
     */
    public boolean sendMessage(final Destination dest, final byte[] msg,
            Collection portOuts) {
        boolean result = true;
        if (portOuts == null) {
            portOuts = new ArrayList(getPortList(dest.getName()));
        }
        for (Iterator itPorts = portOuts.iterator(); itPorts.hasNext();) {
            PortOut portOut = (PortOut) itPorts.next();
            if (portOut != null) {
                // just for debugging
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    StringBuffer buffer = new StringBuffer("routing to ");
                    buffer.append(portOut);
                    buffer.append(" message =");
                    Byte b = null;
                    for (int index = 0; index < msg.length; index++) {
                        b = new Byte(msg[index]);
                        buffer.append(b);
                    }
                    LOGGER.log(BasicLevel.DEBUG, buffer.toString());
                }
                // end just for debugging
                if (portOut.sendMessage(msg)) {
                    itPorts.remove();
                } else {
                    result = false;
                }
            }
        }
        return result;
    }

    /**
     * If the provided destination is contextual, tries to contextually send the provided message.
     * @param dest the destination
     * @param msg the message to be sent
     * @return true if the data was successfully sent OR the destination is non contextual OR no contextual connection is available
     */
    public boolean sendContextualMessage(final Destination dest,
            final byte[] msg) {
        if (!dest.isContextual()) {
            return true;
        }
        boolean result = true;
        Collection portOuts = new ArrayList(getContextualPortList(dest
                .getName()));
        for (Iterator itPorts = portOuts.iterator(); itPorts.hasNext();) {
            PortOut portOut = (PortOut) itPorts.next();
            if (portOut != null) {
                // just for debugging
                if (LOGGER.isLoggable(BasicLevel.DEBUG)) {
                    StringBuffer buffer = new StringBuffer(
                            "Contextually routing to ");
                    buffer.append(portOut);
                    buffer.append(" message =");
                    Byte b = null;
                    for (int index = 0; index < msg.length; index++) {
                        b = new Byte(msg[index]);
                        buffer.append(b);
                    }
                    LOGGER.log(BasicLevel.DEBUG, buffer.toString());
                }
                // end just for debugging
                result &= portOut.sendMessage(new ContextMessage(msg, dest
                        .getContext()));
            }
        }
        return result;
    }
}
