package org.ow2.odis.util;

import java.io.UnsupportedEncodingException;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Utility class witch provide basic encoder decoder.
 * @author Vincent OLIVE,Olivier BEYLER Copyright FT/R&D (2004)
 */
public class CodingDecoding {
    /**
     * logger of this class.
     */
    static final Logger LOGGER = Monolog.initialize().getLogger(
            CodingDecoding.class.getName());

    /**
     * hidde default constructor (unused for an utility class).
     */
    private CodingDecoding() {
    };

    /**
     * constant.
     */
    public static final int CID_LENGTH = 21;

    /**
     * constant.
     */
    public static final int LONG_LENGTH = 8;

    /**
     * constant.
     */
    public static final int INT_LENGTH = 4;

    /**
     * constant.
     */
    public static final int SHORT_LENGTH = 2;

    /**
     * constant.
     */
    public static final String EMPTY_CID_CHAR = "0";

    /**
     * constant.
     */
    private static final int CST_OXFF = 0xff;

    /**
     * constant.
     */
    private static final int CST_DECAL_8 = 8;

    /** The charset used for any text encoding */
    public static final String BYTE_ENCODING = "ISO-8859-1";

    // Coding methods
    /**
     * convert a short into a byte array.
     * @param value short
     * @return byte array representing the short.
     */
    public static final byte[] shortTo2Bytes(final short value) {
        byte[] r = new byte[SHORT_LENGTH];
        short v;
        for (int i = r.length - 1; i >= 0; i--) {
            // $ANALYSIS-IGNORE,codereview.java.rules.casting.
            // RuleCastingPrimitives
            v = (short) (value >>> i * CST_DECAL_8);
            // $ANALYSIS-IGNORE,codereview.java.rules.casting.
            // RuleCastingPrimitives
            r[i] = (byte) (v & CST_OXFF);
        }
        return r;
    }

    /**
     * onvert a boolean into a byte array.
     * @param value boolean
     * @return byte array.
     */
    public static final byte[] booleanTo1Bytes(final boolean value) {
        byte[] r = new byte[1];
        if (value) {
            r[0] = 1;
        } else {
            r[0] = 0;
        }
        return r;
    }

    /**
     * Converts an integer value into a byte array.<br>
     * The conversion is done according to the internal encoding of int type.
     * @param value the integer to convert
     * @return a 4-byte array
     */
    public static final byte[] intTo4Bytes(final int value) {
        byte[] r = new byte[INT_LENGTH];
        int v;
        for (int i = r.length - 1; i >= 0; i--) {
            v = (value >>> i * CST_DECAL_8);
            // $ANALYSIS-IGNORE,codereview.java.rules.casting.
            // RuleCastingPrimitives
            r[i] = (byte) (v & CST_OXFF);
        }
        return r;
    }

    // Coding Decoding methods for long
    /**
     * convert a long into a byte array.
     * @param value long
     * @return byte array
     */
    public static final byte[] longTo8Bytes(final long value) {
        byte[] r = new byte[LONG_LENGTH];
        long v = value;
        for (int i = r.length - 1; i >= 0; i--) {
            v = (value >>> i * CST_DECAL_8);
            // $ANALYSIS-IGNORE,codereview.java.rules.casting.
            // RuleCastingPrimitives
            r[i] = (byte) (v & CST_OXFF);
        }
        return r;
    }

    /**
     * convert double into byte array.
     * @param value double
     * @return byte array
     */
    public static final byte[] doubleTo8Bytes(final double value) {
        return longTo8Bytes(Double.doubleToLongBits(value));
    }

    /**
     * extract a long from a byte array at a specified index position.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return long
     */
    public static final long getLongFrom8bytes(final byte[] bytes,
            final IntegerModif value) {
        long result = 0;
        int offset = value.intValue();
        for (int i = LONG_LENGTH - 1 + offset; i >= offset; i--) {
            result = (result << CST_DECAL_8 | (bytes[i]) & CST_OXFF);
        }
        value.setValue(offset + LONG_LENGTH);
        return result;
    }

    /**
     * extract a boolean from a byte array at a specified position.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return a boolean
     */
    public static final boolean getBooleanFrom1Bytes(final byte[] bytes,
            final IntegerModif value) {
        final int offset = value.intValue();
        value.setValue(offset + 1);
        final boolean result = bytes[offset] == 1;
        return result;
    }

    /**
     * extract a double from a byte array at a specific position.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return double
     */
    public static final double getDoubleFrom8bytes(final byte[] bytes,
            final IntegerModif value) {
        long temp = getLongFrom8bytes(bytes, value);
        return Double.longBitsToDouble(temp);
    }

    /**
     * extract a float from a byte array at a specific position.
     * @param bytes byte array.
     * @param value modif integer value representing an index into the byte array
     * @return float value
     */
    public static final float getFloatFrom4bytes(final byte[] bytes,
            final IntegerModif value) {
        int temp = getIntFrom4bytes(bytes, value);
        return Float.intBitsToFloat(temp);
    }

    /**
     * convert a float into a byte array
     * @param value float
     * @return byte array
     */
    public static final byte[] floatTo4Bytes(float value) {
        int temp = Float.floatToRawIntBits(value);
        return intTo4Bytes(temp);
    }

    /**
     * extract an int from a byte array
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return int value
     */
    public static final int getIntFrom4bytes(final byte[] bytes,
            final IntegerModif value) {
        int result = 0;
        int offset = value.intValue();
        for (int i = INT_LENGTH - 1 + offset; i >= offset; i--) {
            result = (result << CST_DECAL_8 | (bytes[i]) & CST_OXFF);
        }
        value.setValue(offset + INT_LENGTH);
        return result;
    }

    /**
     * extract a short from a bytes array at a specific position.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return short value
     */
    public static final short getShortFrom2bytes(final byte[] bytes,
            final IntegerModif value) {
        short result = 0;
        int offset = value.intValue();
        for (int i = SHORT_LENGTH - 1 + offset; i >= offset; i--) {
            // $ANALYSIS-IGNORE,codereview.java.rules.casting.
            // RuleCastingPrimitives
            result = (short) (result << CST_DECAL_8 | (bytes[i]) & CST_OXFF);
        }
        value.setValue(offset + SHORT_LENGTH);
        return result;
    }

    /**
     * convert the result of toString function of object into a byte array.
     * @param value Object
     * @return byte array
     * @throws java.io.UnsupportedEncodingException in case of invalid encoding.
     */
    public static final byte[] stringToBytes(final Object value)
            throws UnsupportedEncodingException {
        return stringToBytes(value.toString());
    }

    /**
     * Encodes a string into a byte array.<br>
     * <li>The first 4 bytes hold a representation of the string's length, or -1 if the string is null (as per {@link #intTo4Bytes(int)}).</li>
     * <li>The following bytes hold the string's characters, according to a constant charset (see {@link #BYTE_ENCODING})</li>
     * @param value the string to encode
     * @throws java.io.UnsupportedEncodingException if the encoding failed
     * @return the byte[] representation of the provided string
     */
    public static final byte[] stringToBytes(final String value)
            throws UnsupportedEncodingException {
        byte[][] v = new byte[2][];
        if (value == null) {
            v[0] = intTo4Bytes(-1);
            v[1] = null;
        } else {
            v[1] = value.getBytes(BYTE_ENCODING);
            v[0] = intTo4Bytes(v[1].length);
        }
        return getLinearByte(v);
    }

    /**
     * transform a byte array into a byte array ;-) <br>
     * in fact it will add the size of the byte array before the byte array. <br>
     * the aim of this procedure is to be able to extract the original byte array from a message.
     * @param value byte array
     * @return byte array
     */
    public static final byte[] bytesToBytes(final byte[] value) {
        byte[][] v = new byte[2][];
        if (value == null) {
            v[0] = intTo4Bytes(-1);
            v[1] = null;
        } else {
            v[1] = value;
            v[0] = intTo4Bytes(v[1].length);
        }
        return getLinearByte(v);
    }

    /**
     * extract a string from a byte array at a specific position.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @throws java.io.UnsupportedEncodingException happend if the encoding is corrupted
     * @return string value
     */
    public static final String getString(final byte[] bytes,
            final IntegerModif value) throws UnsupportedEncodingException {
        final int lenght = getIntFrom4bytes(bytes, value);
        String result = null;
        if (lenght != -1) {
            if (lenght > bytes.length) {
                throw (new UnsupportedEncodingException("bad string decoding"));
            }
            byte[] tmp = new byte[lenght];
            int offset = value.intValue();
            System.arraycopy(bytes, offset, tmp, 0, lenght);
            value.setValue(offset + lenght);
            result = new String(tmp, BYTE_ENCODING);
        }
        return result;
    }

    /**
     * This function is not secure, so it's depracated. Please use getBytesSecure instead of getBytes
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @return byte[] message
     * @deprecated use getBytesSecure instead of getBytes
     */
    public static final byte[] getBytes(final byte[] bytes,
            final IntegerModif value) {
        byte[] result = null;
        try {
            result = getBytesSecure(bytes, value);
        } catch (UnsupportedEncodingException e) {
            LOGGER.log(BasicLevel.FATAL,
                    "error on decoding getBytes check your encoder/decoder");
        }
        return result;
    }

    /**
     * Extract from the byte array the byte array encoded.
     * @param bytes byte array
     * @param value modif integer value representing an index into the byte array
     * @throws java.io.UnsupportedEncodingException <br>
     * if the extract request is over the content of the byte message that would only mean that the encoding / decoding are not correct and
     * it'need to be considered as noise
     * @return byte array
     */
    public static final byte[] getBytesSecure(final byte[] bytes,
            final IntegerModif value) throws UnsupportedEncodingException {
        final int lenght = getIntFrom4bytes(bytes, value);
        byte[] result = new byte[0];
        if (lenght != -1) {
            // 
            if (lenght > bytes.length) {
                throw (new UnsupportedEncodingException("bad byte decoding"));
            }
            result = new byte[lenght];
            final int offset = value.intValue();
            try {
                System.arraycopy(bytes, offset, result, 0, lenght);
            } catch (RuntimeException e) {
                LOGGER.log(BasicLevel.ERROR, "decode pb long" + lenght
                        + "offset" + offset, e);
            }
            value.setValue(offset + lenght);
        }
        return result;
    }

    /**
     * convert a array of byte array into a byte array. <br>
     * exemple : [[0,1,0],[1,1],[1,0,0]]=>[0,1,0,1,1,1,0,0]
     * @param value byte[ ][ ]
     * @return byte[ ]
     */
    public static final byte[] getLinearByte(final byte[][] value) {
        int resultLength = 0;
        for (int i = 0; i < value.length; i++) {
            if (value[i] != null) {
                resultLength += value[i].length;
            }
        }
        byte[] result = new byte[resultLength];
        int offset = 0;
        int len;
        for (int i = 0; i < value.length; i++) {
            if (value[i] != null) {
                len = value[i].length;
                System.arraycopy(value[i], 0, result, offset, len);
                offset += len;
            }
        }
        return result;
    }
}
