package org.ow2.odis.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.RandomAccessFile;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import org.objectweb.util.monolog.Monolog;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.odis.domain.Domain;

/**
 * @author olivier beyler - orange-labs
 */
public class StreamHelper {
    private final static Logger logger = Monolog.initialize().getLogger(
            StreamHelper.class.getName());

    /**
     * Gets the input stream.
     * @param name the name
     * @return the input stream
     * @throws Exception the exception
     */
    public static InputStream getInputStream(String name) throws Exception {
        InputStream is = null;
        try {
            is = getInputStreamFromClasspath(name);
        } catch (Exception e) {
        }
        if (is == null) {
            try {
                is = getInputStreamFromAbsolutePath(name);
            } catch (Exception e) {
                throw new Exception("Resources " + name + " is not found");
            }
        }
        return is;
    }

    /**
     * Gets the input stream from absolute path.
     * @param absName the abs name
     * @return the input stream from absolute path
     * @throws IOException Signals that an I/O exception has occurred.
     */
    private static InputStream getInputStreamFromAbsolutePath(String absName)
            throws IOException {
        return new FileInputStream(new File(absName));
    }

    /*
     * private static InputStream getInputStreamFromRelativePath(String relName) { return StreamHelper.class.getResourceAsStream(relName);
     * }
     */
    /**
     * Gets the input stream from classpath.
     * @param name the name
     * @return the input stream from classpath
     */
    private static InputStream getInputStreamFromClasspath(String name) {
        return Domain.getInstance().getFactoryInstance().getClass()
                .getClassLoader().getResourceAsStream(name);
    }

    /**
     * Gets the file.
     * @param name the name
     * @return the file
     * @throws Exception the exception
     */
    public static File getFile(String name) throws Exception {
        File file = null;
        URL url = Domain.getInstance().getFactoryInstance().getClass()
                .getClassLoader().getResource(name);
        if (url != null) {
            URI uri = url.toURI();
            file = new File(uri);
        } else {
            file = new File(name);
        }
        return file;
    }

    /**
     * Check directory. Throw an exception if the directory name does'nt correspond to an existing directory.
     * @param propertyName the property name (used in the exception message).
     * @param directory the directory
     * @throws Exception the exception
     */
    public static void checkDirectory(String propertyName, String directory)
            throws Exception {
        if (directory == null) {
            throw new Exception("Error " + propertyName + " is null");
        }
        File dir = StreamHelper.getFile(directory);
        if (!dir.exists()) {
            throw new Exception("Error directory " + directory
                    + " is not found.");
        } else if (!dir.isDirectory()) {
            throw new Exception("Error " + directory + " is not a directory.");
        }
    }

    /**
     * Check file. Throw an exception if the file name doesn't correspond to an existing file.
     * @param propertyName the property name (used in the exception message).
     * @param fileName the file name
     * @throws Exception the exception
     */
    public static void checkFile(String propertyName, String fileName)
            throws Exception {
        checkFile(propertyName, fileName, false);
    }

    /**
     * Check file. Throw an exception if the file name doesn't correspond to a file. If the file doesn't exist, it will be create if the
     * createIt parameter is set to true
     * @param propertyName the property name
     * @param fileName the file name
     * @param createIt if true, the file is created
     * @throws Exception the exception
     */
    public static void checkFile(String propertyName, String fileName,
            boolean createIt) throws Exception {
        if (fileName == null) {
            throw new Exception("Error " + propertyName + " is null");
        }
        File file = StreamHelper.getFile(fileName);
        if (createIt) {
            if (!file.exists()) {
                file.createNewFile();
            } else {
                file.delete();
                file.createNewFile();
            }
        }
        if (!file.exists()) {
            throw new Exception("Error file " + fileName + " is not found.");
        }
        if (file.isDirectory()) {
            throw new Exception("Error " + fileName + " is a directory.");
        }
    }

    /**
     * Recherche un fichier dans le classpath et le retourne sous forme d'inputStream.
     * @param fileName fichier recherche dans le classpath
     * @return InputStream
     * @throws IOException
     */
    public static InputStream getResourcesStream(String fileName)
            throws IOException {
        InputStream result = null;
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger
                    .log(BasicLevel.DEBUG, "getResourcesStream(" + fileName
                            + ')');
        }
        URL url = StreamHelper.class.getResource("/" + fileName);
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "url = " + url);
        }
        if (url != null) {
            result = url.openStream();
        } else {
            throw new FileNotFoundException("unable to find Ressource :"
                    + fileName);
        }
        return result;
    }

    /**
     * Recherche un fichier dans le classpath et le retourne sous forme de fichier.
     * @param fileName fichier recherche dans le classpath
     * @return File
     * @throws IOException
     */
    public static File getResourcesFile(String fileName) throws IOException {
        File file = null;
        URL url = StreamHelper.class.getResource("/" + fileName);
        if (url == null) {
            throw new FileNotFoundException("unable to find Ressource :"
                    + fileName);
        }
        try {
            URI uri = url.toURI();
            file = new File(uri);
        } catch (URISyntaxException e) {
            throw new IOException("URI Exception on " + fileName);
        }
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            StringBuffer buffer = new StringBuffer("=> ");
            buffer.append(file);
            buffer.append("(");
            if (file != null) {
                buffer.append(file.getAbsolutePath());
            }
            buffer.append(")");
            logger.log(BasicLevel.DEBUG, buffer.toString());
        }
        return file;
    }

    /**
     * copy le fichier in dans le fichier out.
     * @param in fichier in
     * @param out fichier out
     * @throws IOException
     */
    static public void copy(File in, File out) throws IOException {
        copy(new FileInputStream(in), out);
    }

    /**
     * copy l'inputStream dans un fichier out.
     * @param is inputStream
     * @param out File out
     * @throws IOException
     */
    static public void copy(InputStream is, File out) throws IOException {
        // System.out.println("in = " + in.getAbsolutePath() +", out = "+
        // out.getAbsolutePath());
        FileOutputStream fos = new FileOutputStream(out);
        byte[] buf = new byte[1024];
        int i = 0;
        while ((i = is.read(buf)) != -1) {
            fos.write(buf, 0, i);
        }
        is.close();
        fos.close();
    }

    /**
     * Deplace tous les fichiers, qui commence par "fileName". dans le repertoire "path" sous le nom "prefix+fileName".
     * @param fileName fichier qui commence par fileName
     * @param toPath repertoire de destination
     * @param prefix si null prefix = "it."
     * @throws Exception
     */
    static public void moveFileStartsWith(final String fileName, String toPath,
            String prefix) throws Exception {
        FilenameFilter filter = new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.startsWith(fileName) && !name.endsWith(".lck");
            }
        };
        File[] files = new File(".").listFiles(filter);
        for (int i = 0; i < files.length; i++) {
            if (prefix == null) {
                copy(files[i], new File(toPath, "it." + files[i].getName()));
            } else {
                copy(files[i], new File(toPath, prefix + files[i].getName()));
            }
            files[i].delete();
        }
    }

    /**
     * ajoute un message a la fin d'un fichier.
     * @param fileName nom du fichier
     * @param message message a jouter
     */
    static public void addToFile(String fileName, String message) {
        try {
            RandomAccessFile file = new RandomAccessFile(fileName, "rw");
            file.seek(file.length());
            file.writeUTF(message + "\n");
            file.close();
        } catch (FileNotFoundException e) {
            logger.log(BasicLevel.ERROR, "Exception :: ", e);
            throw new RuntimeException(e);
        } catch (IOException e) {
            logger.log(BasicLevel.ERROR, "Exception :: ", e);
            throw new RuntimeException(e);
        }
    }

    /**
     * Create a directory. If the directory exist, it will be recreate if the createIt parameter is set to true
     * @param directoryName the directory name
     * @param createIt if true, the file is created
     * @throws Exception the exception
     */
    public static void createDirectory(String directoryName, boolean createIt)
            throws Exception {
        File directory = getFile(directoryName);
        if (createIt) {
            if (!directory.exists()) {
                directory.mkdir();
            } else {
                directory.delete();
                directory.mkdir();
            }
        }
    }
}
